/*
 * $Id$
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation du filtre (interne) EPG.
 *
 *	Ce fichier implmente le filtre (interne) EPG, qui sert 
 *	rcuprer, dans le flux TS du multiplex courant, toutes les informations
 *	destines au guide lectronique des programmes (EPG).
 **/

#include "stdafx.h"
#include "epgfilter.h"
#include "..\..\channels.h"
#include "crc32.h"
#include "mpeg2defs.h"
#include "..\..\epg.h"

CEPGFilter::CEPGFilter(IUnknown *pUnk, HRESULT *phr) :
	CBaseRenderer(__uuidof(this), TEXT(EPG_FILTER_NAME), pUnk, phr)
{
	myprintf(TEXT("%") W2t TEXT(" construit") EOL, szFilterName);
}

CEPGFilter::~CEPGFilter()
{
	myprintf(TEXT("%") W2t TEXT(" detruit") EOL, szFilterName);
}

/**
 * \brief Extraction d'une chane de l'EPG
 *
 * Les codes spciaux (codes 0x80  0x9f) sont nettoys,
 * ainsi que les espaces de dbut, de fin et multiples
 **/
static void get_epg_string(const MDESCR::S_str & sMs, tstring & strStr)
{
	TCHAR		str[256];

	sMs.get(str);

	LPCTSTR	src = str;
	LPTSTR	dst = str;
	UINT	ch;	// non sign !

	while ((ch=*src++) != 0) {
		switch (ch) {
		case DCTL_(CRLF):								// Saut de ligne
			ch = '\n';
			break;

		case '\t':										// Tabulation
			ch = ' ';									// remplacer par espace
			break;

		case DCTL_(EMPH_ON):							// Codes d'emphase : ignorer
		case DCTL_(EMPH_OFF):
			continue;

#ifdef _UNICODE
		default:
			if (ch>=DCTL_(MIN) && ch<=DCTL_(MAX)) {		// Autre code de contrle :
				// On va supposer que c'est une erreur de la chane TV, et le traiter comme
				// de l'encodage natif Windows (Windows 1252) (ce genre d'erreur a t constat).
				// En mode UNICODE, le caractre d'origine a juste t augment de la valeur
				// 0xE000, donc en pointer l'octet de poids faible doit permettre de le retrouver.
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, LPCSTR(&src[-1]), 1, LPWSTR(&ch), 1);
			}
#endif
		}

		*dst++ = (TCHAR)ch;
	}

	// Enlever les espaces de fin :
	while (dst > str && (unsigned)dst[-1]<= (unsigned)' ')
		--dst;
	*dst = 0;
	strStr += str;
}

LPCWSTR CEPGFilter::szFilterName = WTXT(EPG_FILTER_NAME);

CUnknown * WINAPI CEPGFilter::CreateInstance(IUnknown *pUnk, HRESULT *phr) // static
{
	CEPGFilter *pNewFilter = new CEPGFilter(pUnk, phr);

	if (phr) {
		if (pNewFilter == NULL)
			*phr = E_OUTOFMEMORY;
		else
			*phr = S_OK;
	}

	return pNewFilter;
}

static bool parse_eit_entry(Chaine & cChan, int ixChaine, const SI_EIT::entry & sEitEn, BYTE nFlags)
{
	UINT64		nUtcBegT	= sEitEn.start_time();	// Heure de dbut UTC pour tests
	UINT64		nUtcEnd10	= nUtcBegT + sEitEn.duration + 10*(60+1000);	// Heure de fin + 10 minutes
	SYSTEMTIME	sUtcTime;

	GetSystemTime(&sUtcTime);

	UINT64		nUtcTime	= SI_EIT::e_time(sUtcTime)();

	// Ne pas retenir les missions dont l'heure de fin est dpasse de plus de 10 minutes
	if (nUtcTime > nUtcEnd10)
		return false;

	UINT16		nEvtId		= sEitEn.event_id();
	Emission *	pEmis		= NULL;

	// Chercher si l'mission n'est pas dj mmorise, et trouver le point d'insertion, s'il y a lieu
	CAutoLock				cLock(cChan.crit_sec());	// Protection contre les modifications concurrentes
	ChanEventList::iterator	it	= cChan.begin();		// itrateur d'exploration

	while (it != cChan.end()) {
		Emission &	cEmis	= **it;

		if (nUtcTime > cEmis.nUtcEnd10) {
			// Descripteur d'mission prim: dtruire
			it = cChan.erase(it);
			continue;
		}
		if (nEvtId == cEmis.nEvtId) {	// Vrifier si item existant
			cEmis.nFlags |= nFlags;
			return false;		// L'mission est dj mmorise
		}
		if (nUtcBegT < cEmis.nUtcBegT) {	// Vrifier si on n'a pas dpass l'heure dans la liste
			break;
		}
		++it;
	}

	// Insertion d'un nouvel lment au point d'insertion
	pEmis = &**cChan.insert(it, new Emission);
	pEmis->debut	= sEitEn.start_time;	// (format SYSTEMTIME)

	// transforme l'heure GMT du programme en local
	SystemTimeToLocalSystemTime(pEmis->debut);

	// on ajoute la dure pour trouver l'heure de fin
	pEmis->fin = pEmis->debut;
	AddTime(pEmis->fin, sEitEn.duration);

	// Ajouter le nom et le descriptif
	for (MDESCR::D_scanner sed(sEitEn.descrRange()); !sed; ++sed) {
		MDESCR::D_tag eTag = sed().tag;

		switch (eTag) {

		case MDESCR::mdt_Short_Event: {
			const MDESCR::D_4d & d4d = sed().as<MDESCR::D_4d>();

			// Extraction et nettoyage des chanes :
			get_epg_string(d4d.event_name(), pEmis->strNom);
			if (pEmis->strNom.empty())
				pEmis->strNom = TEXT("< >");
			get_epg_string(d4d.text(), pEmis->strDesc);
			break; }

		case MDESCR::mdt_Extended_Event: {
			// Attention, les "mdt_Extended_Event" peuvent tre multiples,
			// et les donnes qu'on en extrait doivent tre concatnes.
			const MDESCR::D_4e & d4e = sed().as<MDESCR::D_4e>();

			get_epg_string(d4e.text(), pEmis->strExtDesc);
			break; }

		case MDESCR::mdt_Parental_Rating: {
			for (MDESCR::D_55::sdScanner s55 = sed().as<MDESCR::D_55>().subDescrRange(); !s55; ++s55) {
				// S'il y a plusieurs limites d'ge, on prend la plus leve.
				// On ne tient pas compte du code de pays pour le moment.
				if (pEmis->nRating < s55().rating)
					pEmis->nRating = s55().rating;
			}
			break; }

		case MDESCR::mdt_Content:	// Exprimental
			for (MDESCR::D_54::sdScanner s54 = sed().as<MDESCR::D_54>().subDescrRange(); !s54; ++s54) {
				if (pEmis->nContent == 0) {
					pEmis->nContent = MAKELONG(
						MAKEWORD(s54().content_nibble_level_1(), s54().content_nibble_level_2()),
						MAKEWORD(s54().user_nibble_1(), s54().user_nibble_2()));
					break;
				}
			}
			break;

#ifdef _DEBUG
		case MDESCR::mdt_Component:
			break;

		default:
			eTag = eTag;
#endif
		}
	}

	pEmis->nUtcBegT		= nUtcBegT;
	pEmis->nUtcEnd10	= nUtcEnd10;
	pEmis->nEvtId		= nEvtId;
	pEmis->nFlags		= nFlags;
	//
	pEmis->cleChaine	= cChan.cle();
	pEmis->ixChaine		= ixChaine;

	//myprintf(TEXT("Ajout d'mission 0x%04x : %s") EOL,
	//	pEmis->nEvtId, pEmis->getName().c_str());
	return true;
}

/// \todo Vrifier souci probable en cas de rception multiple de la mme chane sur plusieurs frquences
static void parse_eit(const SI_EIT & eit, size_t size)
{
	bool		modif		= false;
	ChaineIDs	sIDs(eit.onid, eit.tsid, eit.service_id());
	int			ixChaine	= Canaux.trouve_par_Ids(sIDs);
	BYTE		nFlags		= (eit.table_id() > si_EIT_oth) ? DEF_IN_SCHEDULE : DEF_IN_ACTUAL;

	if (ixChaine < 0)
		return;

	Chaine & cChan = Canaux[ixChaine];

	for (SI_EIT::entryScanner se(eit.entryRange()); !se; ++se) {
		if (parse_eit_entry(cChan, ixChaine, se(), nFlags))
			modif = true;
	}

	// Passe par le thread principal pour mettre  jour l'affichage, sinon a plante
	if (modif)
		CEpgDialog::PostDlgMessage(WM_APP_EPG_UPDATED);

	if (modif && ixChaine == ixChaineCourante && (nFlags & DEF_IN_ACTUAL)!=0)
		PostMessage(hMainWnd, WM_APP_EPG_UPDATED, 0, 0);
}

HRESULT CEPGFilter::DoRenderSample(IMediaSample *pMediaSample)
{
	const SI_EIT *	peit;
	size_t			taille = pMediaSample->GetActualDataLength();

	pMediaSample->GetPointer((LPBYTE *)(&peit));

	if (
		(peit->table_id()>=si_EIT_act && peit->table_id()<=si_EIT_osf) &&
		taille >= peit->size()
		&& peit->check_crc32()
	) {
		parse_eit(*peit, taille);
	}

	return S_OK;
}

HRESULT CEPGFilter::CheckMediaType(const CMediaType * pmt)
{
	if ( *pmt->Type() == MEDIATYPE_MPEG2_SECTIONS) return S_OK;
	return S_FALSE;
}
