/*
 * $Id: index_parser.cpp 48 2010-11-10 13:50:05Z gingko $
 * Copyright (C) 2010 gingko - http://gingko.homeip.net/
 *
 * This file is part of the Pouchin TV Mod documentation.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/** \file
 *	\brief Utilitaire de gnration de table de contenu et de table d'index.
 *
 *	Ce fichier et le code source d'une application qui s'occupe d'analyser les fichiers de
 *	la documentation de Pouchin TV Mod, rdigs avec des tags spcialement adapts,
 *	et de gnrer les tables de contenu et d'indexation qui leur correspondent.
 *
 *	\see Voir le fichier "..\Lisez-moi.txt" pour les instructions d'utilisation.
 **/

#include <windows.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include "tinyxml.h"

using namespace std;

// ----------------------------------------------------------------------------

/**
 * Macro d'assistance  l'numration d'lments de vecteurs
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_VECTOR(parent, type, name) \
	std::vector<type>::iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de vecteurs constants
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_CONST_VECTOR(parent, type, name) \
	std::vector<type>::const_iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

// ----------------------------------------------------------------------------

/**
 * Macro d'allocation d'un tableau de \p taille lments de type \p type sur la pile,
 * en tirant parti des particularits de la fonction systme \p _alloca.
 * Permet un effet similaire  celui produit par l'extension non standard permise sur
 * les compilateurs de type GCC,  savoir de pouvoir dfinir un tableau  l'intrieur
 * d'une fonction, en utilisant un terme variable dans la taille du tableau.
 *
 * Par rapport  une allocation sur le tas (heap) avec \p malloc ou \p new, les
 * avantages sont les suivants :
 * - Traitement infiniment plus rapide (quelques instructions machine seulement)
 * - Pas besoin de \p free ou \p delete, la destruction est implicite  la fin de la
 *   fonction.
 *
 * Pour ce qui est des inconvnients :
 * - Usage strictement limit au corps d'une fonction : pas question de retourner
 *   le rsultat.
 * - Ne devrait pas tre utilis dans une boucle,  moins qu'il soit clair que le
 *   nombre d'appels sera trs rduit : les allocations successives ne seraient pas
 *   libres avant la sortie de la fonction.
 * - La taille de l'allocation doit rester limite  ce qui pourrait se dclarer en
 *   tant que variable locale (sinon, risque de dbordement de la pile)
 *
 * Exemple : \code
 *   int		taille=16;
 *   wchar_t *	buffer = STK_ALLOC(wchar_t, taille);
 * \endcode
 * Ceci est  peu prs quivalent  cette syntaxe, qui serait valide sous GCC : \code
 *   int		taille=16;
 *   wchar_t	buffer[taille];
 * \endcode
 * \param[in]	type	Type des lments du tableau
 * \param[in]	taille	Nombre d'lments  allouer
**/
#define STK_ALLOC(type, taille) reinterpret_cast<type *>(_alloca((taille)*sizeof(type)))

/**
 * Convertit une chane de caractres UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres UTF-8
 * \return     chane de caractres convertie
 **/
string ConvertUTF82Char(LPCSTR aconvertir)
{
	LPSTR _tmpout;
	LPWSTR	_tmp;
	UINT len;

	if (aconvertir) {
		UINT	i, j;								// Itrateurs
		UCHAR	c;									// Caractres  convertir
		UINT	taille = (UINT)strlen(aconvertir);	// La taille de la chane rsultante est forcment plus courte
		len = taille;

		_tmpout = STK_ALLOC(CHAR, taille+1);	// Chane temporaire
		_tmp = STK_ALLOC(WCHAR, taille+1);	// Chane temporaire
		for( i=0, j=0; i<taille; i++) {
			c = aconvertir[i];

			UINT32 nTmp = '?'; // Caractre par dfaut si conversion impossible

			if (c <= 0x7F)
				nTmp = c;
			else if (c < 0xE0)
				nTmp = ((c & 0x1F) << 6) + (aconvertir[++i] & 0x3F);
			_tmp[j++] = WCHAR(nTmp);
		}
		_tmp[j] = 0;
	}

	UINT j = WideCharToMultiByte(
			CP_ACP, 0, _tmp,
			len, _tmpout, len, NULL, NULL);
	_tmpout[j] = 0;
	return _tmpout;	// Conversion implicite en tstring
}

/**
 * Convertir un caractre en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[out] convertis   chane de caractres convertie
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     le nombre de caractres convertis
 * \note \p aconvertir est de type \p UINT32 car :
 * - TCHAR pose des problmes tant donn qu'il s'agit d'une valeur *signe*
 * - La fonction est capable de convertir des caractres cods sur plus de 16 bits
 * \note \p convertis doit pointer sur une variable de taille suffisante
 * pour tre capable de recevoir au moins 6 caractres (terminateur compris).
 **/
int ConvertChar2UTF8(UINT32 aconvertir, LPSTR convertis)
{
	int	j;		// Itrateur

	j = WideCharToMultiByte(
			CP_UTF8, 0, reinterpret_cast<LPCWSTR>(&aconvertir),
			1, convertis, 5, NULL, NULL);
	convertis[j] = 0;	// zro terminateur

	return j;
}

void CleanString(string & strTxt)
{
	size_t	nLen	= strTxt.length();
	UCHAR	nCh		= 0;

	// Nettoyer les espaces et la ponctuation  la fin
	while (nLen > 0 && (nCh = strTxt[nLen-1]) <= ' ' || nCh==0xa0 || nCh==':' || nCh=='.')
		strTxt.resize(--nLen);

	// Nettoyer les espaces au dbut
	while (nLen > 0 && ((nCh = strTxt[0]) <= ' ' || nCh==0xa0)) {
		strTxt.erase(0, 1);
		--nLen;
	}

	size_t nPos;

	// Remplacer tous les sauts de lignes par des espaces
	while ((nPos = strTxt.find("\n")) != string::npos)
		strTxt.replace(nPos, 1, " ");
	// Idem pour les tabulations
	while ((nPos = strTxt.find("\t")) != string::npos)
		strTxt.replace(nPos, 1, " ");
	// Remplacer tous les espaces doubls par des espaces simples
	while ((nPos = strTxt.find("  ")) != string::npos)
		strTxt.erase(nPos, 1);
}

// ----------------------------------------------------------------------------

struct Entity
{
	LPCSTR			pszName;
	unsigned char	szUtf8[3];
	unsigned char	nAnsi;
};

#define STD_CHAR(ch, win, uni, sgml) {sgml, {win, 0, 0}, win}
#define UTF_CHAR(ch, win, uni, sgml) {sgml, {CHAR((unsigned(win) >> 6) | 0xc0), CHAR((unsigned(win) & 0x3f) | 0x80), 0}, win}

static const Entity aMinimal[] =
{
	UTF_CHAR(' ', 0xA0, 0x00A0, "nbsp"),
	STD_CHAR('<', 0x3C, 0x003C, "lt"),
	STD_CHAR('>', 0x3E, 0x003E, "gt"),
	STD_CHAR('"', 0x22, 0x0022, "quot"),
	STD_CHAR('&', 0x26, 0x0026, "amp")
};

static const Entity aEntities[] =
{
	UTF_CHAR(' ', 0xA0, 0x00A0, "nbsp"),
	// STD_CHAR('\'',0x27, 0x003C, "apos"),	// trait par TinyXML
	// STD_CHAR('<', 0x3C, 0x003C, "lt"),	// trait par TinyXML
	// STD_CHAR('>', 0x3E, 0x003E, "gt"),	// trait par TinyXML
	// STD_CHAR('"', 0x22, 0x0022, "quot"),	// trait par TinyXML
	// STD_CHAR('&', 0x26, 0x0026, "amp"),	// trait par TinyXML
	UTF_CHAR('', 0xE9, 0x00E9, "eacute"),
	UTF_CHAR('', 0xE8, 0x00E8, "egrave"),
	UTF_CHAR('', 0xEA, 0x00EA, "ecirc"),
	UTF_CHAR('', 0xE0, 0x00E0, "agrave"),
	UTF_CHAR('', 0xE7, 0x00E7, "ccedil"),
	UTF_CHAR('', 0xF9, 0x00F9, "ugrave"),
	UTF_CHAR('', 0xE2, 0x00E2, "acirc"),
	UTF_CHAR('', 0xEE, 0x00EE, "icirc"),
	UTF_CHAR('', 0xF4, 0x00F4, "ocirc"),
	UTF_CHAR('', 0xFB, 0x00FB, "ucirc"),
	UTF_CHAR('', 0xE4, 0x00E4, "auml"),
	UTF_CHAR('', 0xEB, 0x00EB, "euml"),
	UTF_CHAR('', 0xEF, 0x00EF, "iuml"),
	UTF_CHAR('', 0xF6, 0x00F6, "ouml"),
	UTF_CHAR('', 0xFC, 0x00FC, "uuml"),
	UTF_CHAR('', 0xE1, 0x00E1, "aacute"),
	UTF_CHAR('', 0xE6, 0x00E6, "aelig"),
	UTF_CHAR('', 0xE5, 0x00E5, "aring"),
	UTF_CHAR('', 0xE3, 0x00E3, "atilde"),
	UTF_CHAR('', 0xED, 0x00ED, "iacute"),
	UTF_CHAR('', 0xEC, 0x00EC, "igrave"),
	UTF_CHAR('', 0xF1, 0x00F1, "ntilde"),
	UTF_CHAR('', 0xF3, 0x00F3, "oacute"),
	UTF_CHAR('', 0x9C, 0x0153, "oelig"),
	UTF_CHAR('', 0xF2, 0x00F2, "ograve"),
	UTF_CHAR('', 0xF5, 0x00F5, "otilde"),
	UTF_CHAR('', 0xFA, 0x00FA, "uacute"),
	UTF_CHAR('', 0xFD, 0x00FD, "yacute"),
	UTF_CHAR('', 0xFF, 0x00FF, "yuml"),
	UTF_CHAR('', 0x80, 0x20AC, "euro"),
	UTF_CHAR('', 0xC9, 0x00C9, "Eacute"),
	UTF_CHAR('', 0xC8, 0x00C8, "Egrave"),
	UTF_CHAR('', 0xCA, 0x00CA, "Ecirc"),
	UTF_CHAR('', 0xC0, 0x00C0, "Agrave"),
	UTF_CHAR('', 0xC7, 0x00C7, "Ccedil"),
	UTF_CHAR('', 0xD9, 0x00D9, "Ugrave"),
	UTF_CHAR('', 0xC2, 0x00C2, "Acirc"),
	UTF_CHAR('', 0xCE, 0x00CE, "Icirc"),
	UTF_CHAR('', 0xD4, 0x00D4, "Ocirc"),
	UTF_CHAR('', 0xDB, 0x00DB, "Ucirc"),
	UTF_CHAR('', 0xC4, 0x00C4, "Auml"),
	UTF_CHAR('', 0xCB, 0x00CB, "Euml"),
	UTF_CHAR('', 0xCF, 0x00CF, "Iuml"),
	UTF_CHAR('', 0xD6, 0x00D6, "Ouml"),
	UTF_CHAR('', 0xDC, 0x00DC, "Uuml"),
	UTF_CHAR('', 0xC1, 0x00C1, "Aacute"),
	UTF_CHAR('', 0xC6, 0x00C6, "AElig"),
	UTF_CHAR('', 0xC5, 0x00C5, "Aring"),
	UTF_CHAR('', 0xC3, 0x00C3, "Atilde"),
	UTF_CHAR('', 0xCD, 0x00CD, "Iacute"),
	UTF_CHAR('', 0xCC, 0x00CC, "Igrave"),
	UTF_CHAR('', 0xD1, 0x00D1, "Ntilde"),
	UTF_CHAR('', 0xD3, 0x00D3, "Oacute"),
	UTF_CHAR('', 0x8C, 0x0152, "OElig"),
	UTF_CHAR('', 0xD2, 0x00D2, "Ograve"),
	UTF_CHAR('', 0xD5, 0x00D5, "Otilde"),
	UTF_CHAR('', 0xDA, 0x00DA, "Uacute"),
	UTF_CHAR('', 0xDD, 0x00DD, "Yacute"),
	UTF_CHAR('', 0x9F, 0x0178, "Yuml"),
	UTF_CHAR('', 0xA9, 0x00A9, "copy"),
	UTF_CHAR('', 0xAE, 0x00AE, "reg"),
	UTF_CHAR('', 0x99, 0x2122, "trade"),
	UTF_CHAR('', 0x83, 0x0192, "fnof"),
	UTF_CHAR('', 0x85, 0x2026, "hellip"),
	UTF_CHAR('', 0x85, 0x2026, "mldr"),
	UTF_CHAR('', 0x96, 0x2013, "ndash"),
	UTF_CHAR('', 0x97, 0x2014, "mdash"),
	UTF_CHAR('', 0x86, 0x2020, "dagger"),
	UTF_CHAR('', 0x87, 0x2021, "Dagger"),
	UTF_CHAR('', 0xF0, 0x00F0, "eth"),
	UTF_CHAR('', 0xD0, 0x00D0, "ETH"),
	UTF_CHAR('', 0x88, 0x02C6, "circ"),
	UTF_CHAR('', 0x89, 0x2030, "permil"),
	UTF_CHAR('', 0x9A, 0x0161, "scaron"),
	UTF_CHAR('', 0x8A, 0x0160, "Scaron"),
	UTF_CHAR('', 0x9E, 0x017E, "zcaron"),
	UTF_CHAR('', 0x8E, 0x017D, "Zcaron"),
	UTF_CHAR('', 0x95, 0x2022, "bull"),
	UTF_CHAR('', 0x98, 0x02DC, "tilde"),
	UTF_CHAR('', 0xAB, 0x00AB, "laquo"),
	UTF_CHAR('', 0xBB, 0x00BB, "raquo"),
	UTF_CHAR('', 0x93, 0x201C, "ldquo"),
	UTF_CHAR('', 0x94, 0x201D, "rdquo"),
	UTF_CHAR('', 0x91, 0x2018, "lsquo"),
	UTF_CHAR('', 0x92, 0x2019, "rsquo"),
	UTF_CHAR('', 0x9B, 0x203A, "rsaquo"),
	UTF_CHAR('', 0x8B, 0x2039, "lsaquo"),
	UTF_CHAR('', 0x82, 0x201A, "lsquor"),
	UTF_CHAR('', 0x84, 0x201E, "ldquor"),
	UTF_CHAR('', 0xBF, 0x00BF, "iquest"),
	UTF_CHAR('', 0xA1, 0x00A1, "iexcl"),
	UTF_CHAR('', 0xA2, 0x00A2, "cent"),
	UTF_CHAR('', 0xA3, 0x00A3, "pound"),
	UTF_CHAR('', 0xA4, 0x00A4, "curren"),
	UTF_CHAR('', 0xA5, 0x00A5, "yen"),
	UTF_CHAR('', 0xA6, 0x00A6, "brvbar"),
	UTF_CHAR('', 0xA7, 0x00A7, "sect"),
	UTF_CHAR('', 0xA8, 0x00A8, "uml"),
	UTF_CHAR('', 0xA8, 0x00A8, "die"),
	UTF_CHAR('', 0xA8, 0x00A8, "Dot"),
	UTF_CHAR('', 0xAA, 0x00AA, "ordf"),
	UTF_CHAR('', 0xAC, 0x00AC, "not"),
	UTF_CHAR('', 0xAD, 0x00AD, "shy"),
	UTF_CHAR('', 0xAF, 0x00AF, "macr"),
	UTF_CHAR('', 0xB0, 0x00B0, "deg"),
	UTF_CHAR('', 0xB1, 0x00B1, "plusmn"),
	UTF_CHAR('', 0xB2, 0x00B2, "sup2"),
	UTF_CHAR('', 0xB3, 0x00B3, "sup3"),
	UTF_CHAR('', 0xB4, 0x00B4, "acute"),
	UTF_CHAR('', 0xB5, 0x00B5, "micro"),
	UTF_CHAR('', 0xB6, 0x00B6, "para"),
	UTF_CHAR('', 0xB7, 0x00B7, "middot"),
	UTF_CHAR('', 0xB8, 0x00B8, "cedil"),
	UTF_CHAR('', 0xB9, 0x00B9, "sup1"),
	UTF_CHAR('', 0xBA, 0x00BA, "ordm"),
	UTF_CHAR('', 0xBC, 0x00BC, "frac14"),
	UTF_CHAR('', 0xBD, 0x00BD, "half"),
	UTF_CHAR('', 0xBD, 0x00BD, "frac12"),
	UTF_CHAR('', 0xBE, 0x00BE, "frac34"),
	UTF_CHAR('', 0xD7, 0x00D7, "times"),
	UTF_CHAR('', 0xD8, 0x00D8, "Oslash"),
	UTF_CHAR('', 0xDE, 0x00DE, "THORN"),
	UTF_CHAR('', 0xDF, 0x00DF, "szlig"),
	UTF_CHAR('', 0xF7, 0x00F7, "divide"),
	UTF_CHAR('', 0xF8, 0x00F8, "oslash"),
	UTF_CHAR('', 0xFE, 0x00FE, "thorn"),
	STD_CHAR('!', 0x21, 0x0021, "excl"),
	STD_CHAR('#', 0x23, 0x0023, "num"),
	STD_CHAR('$', 0x24, 0x0024, "dollar"),
	STD_CHAR('%', 0x25, 0x0025, "percnt"),
	STD_CHAR('(', 0x28, 0x0028, "lpar"),
	STD_CHAR(')', 0x29, 0x0029, "rpar"),
	STD_CHAR('*', 0x2A, 0x002A, "ast"),
	STD_CHAR('+', 0x2B, 0x002B, "plus"),
	STD_CHAR(',', 0x2C, 0x002C, "comma"),
	STD_CHAR('-', 0x2D, 0x002D, "hyphen"),
	STD_CHAR('.', 0x2E, 0x002E, "period"),
	STD_CHAR('/', 0x2F, 0x002F, "sol"),
	STD_CHAR(':', 0x3A, 0x003A, "colon"),
	STD_CHAR(';', 0x3B, 0x003B, "semi"),
	STD_CHAR('=', 0x3D, 0x003D, "equals"),
	STD_CHAR('?', 0x3F, 0x003F, "quest"),
	STD_CHAR('@', 0x40, 0x0040, "commat"),
	STD_CHAR('[', 0x5B, 0x005B, "lsqb"),
	STD_CHAR('\\',0x5C, 0x005C, "bsol"),
	STD_CHAR('\\',0x5C, 0x005C, "sbsol"),
	STD_CHAR(']', 0x5D, 0x005D, "rsqb"),
	STD_CHAR('_', 0x5F, 0x005F, "lowbar")
};

static LPCSTR EntityToUtf8(LPCSTR pszEnt)
{
	for (int nInx=0; nInx < _countof(aEntities); ++nInx) {
		const Entity & sEnt = aEntities[nInx];

		if (strcmp(sEnt.pszName, pszEnt)==0)
			return reinterpret_cast<LPCSTR>(sEnt.szUtf8);
	}

	static CHAR szWrk[16];

	sprintf_s(szWrk, "&%s;", pszEnt);
	return szWrk;
}

static LPCSTR EntityToWin1252(LPCSTR pszEnt)
{
	static CHAR szWrk[16];

	for (int nInx=0; nInx < _countof(aEntities); ++nInx) {
		const Entity & sEnt = aEntities[nInx];

		if (strcmp(sEnt.pszName, pszEnt)==0) {
			szWrk[0] = sEnt.nAnsi;
			szWrk[1] = 0;
			return szWrk;
		}
	}

	sprintf_s(szWrk, "&%s;", pszEnt);
	return szWrk;
}

static LPCSTR Win1252ToEntity(CHAR nCh, const Entity * pEnt, size_t nCount)
{
	static CHAR szWrk[16];

	for (size_t nInx=0; nInx < nCount; ++nInx) {
		const Entity & sEnt = pEnt[nInx];

		if ((UCHAR)sEnt.nAnsi == (UCHAR)nCh) {
			sprintf_s(szWrk, "&%s;", sEnt.pszName);
			return szWrk;
		}
	}

	sprintf_s(szWrk, "&#x%02X;", (unsigned char)nCh);
	return szWrk;
}

static LPCSTR SgmlEncode(LPCSTR pszStr)
{
	static CHAR	szWrk[512];
	UINT		nInd = 0;
	CHAR		nCh;

	while (nCh = *pszStr++) {
		if (nCh < 0 || nCh == '"' || nCh == '<' || nCh == '>' || nCh == '&') {
			LPCSTR pszEnt = Win1252ToEntity(nCh, aEntities, _countof(aEntities));

			if (nInd <= _countof(szWrk)-1-strlen(pszEnt)) {
				strcpy_s(szWrk+nInd, _countof(szWrk)-nInd, pszEnt);
				nInd += (UINT)strlen(pszEnt);
			}
		} else {
			if (nInd <= _countof(szWrk)-1-1)
				szWrk[nInd++] = nCh;
		}
	}

	szWrk [nInd] = 0;
	return szWrk;
}

static LPCSTR HtmlEncode(LPCSTR pszStr)
{
	static CHAR	szWrk[512];
	UINT		nInd = 0;
	CHAR		nCh;

	while (nCh = *pszStr++) {
		if ((UCHAR)nCh == 0xa0 || nCh == '"' || nCh == '<' || nCh == '>' || nCh == '&') {
			LPCSTR pszEnt = Win1252ToEntity(nCh, aMinimal, _countof(aMinimal));

			if (nInd <= _countof(szWrk)-1-strlen(pszEnt)) {
				strcpy_s(szWrk+nInd, _countof(szWrk)-nInd, pszEnt);
				nInd += (UINT)strlen(pszEnt);
			}
		} else if (nCh < 0) {
			CHAR	szTmp[8];

			int nLen = ConvertChar2UTF8((UCHAR)nCh, szTmp);
			strcpy_s(szWrk+nInd, _countof(szWrk)-nInd, szTmp);
			nInd += nLen;
		} else {
			if (nInd <= _countof(szWrk)-1-1)
				szWrk[nInd++] = nCh;
		}
	}

	szWrk [nInd] = 0;
	return szWrk;
}

static void DisplayError(DWORD dwErr)
{
	CHAR	szBuffer[512];

	FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dwErr,
		0, // Langue par dfaut (pourrait tre 0x040c = franais)
		szBuffer,
		_countof(szBuffer),
		NULL);

	string strErr = szBuffer;

	CleanString(strErr);
	cout << strErr << endl;
}

static DWORD LoadFileInString(LPCSTR pszFile, string & strData)
{
	DWORD	dwErr = 0;

	strData.clear();

	HANDLE hFile = CreateFile(
						pszFile, GENERIC_READ, FILE_SHARE_READ, NULL,
						OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (hFile == INVALID_HANDLE_VALUE) {
		dwErr = GetLastError();
		return dwErr;
	}

	DWORD	nRead = 0;
	DWORD	nSize = GetFileSize(hFile, NULL);

	if (nSize > 0) {
		strData.resize(nSize + 1);
		if (ReadFile(hFile, &strData[0], nSize, &nRead, NULL) == 0)
			dwErr = GetLastError();
	}

	CloseHandle(hFile);

	if (nRead==0 || strData.empty())
		return ERROR_EMPTY;

	// Remplacer tous les CRLF par des LF
	size_t nPos = 0;

	while ((nPos = strData.find("\r\n")) != string::npos) {
		strData.erase(nPos++, 1);
	}

	return dwErr;
}

// ----------------------------------------------------------------------------

struct XmlStuff;

struct XmlHeading : public vector<XmlHeading> {
	string				strFile;
	string				strAnchor;
	string				strText;
	int					nIcon;
	int					nLevel;
	XmlHeading *		pParent;

	XmlHeading & operator = (const XmlHeading & s2)
	{
		vector<XmlHeading>::operator = (s2);
		strFile		= s2.strFile;
		strAnchor	= s2.strAnchor;
		strText		= s2.strText;
		nIcon		= s2.nIcon;
		nLevel		= s2.nLevel;
		pParent		= s2.pParent;
		return		*this;
	}

	bool operator < (const XmlHeading & s2) const
	{
		// Comparaison avec prise en charge des caractres accentus
		int nRes = CompareString(
					LOCALE_NEUTRAL, NORM_IGNORECASE,
					strText.c_str(), -1, s2.strText.c_str(), -1);

		if (nRes != 0)
			return nRes == CSTR_LESS_THAN;

		// On ne devrait pas arriver ici, parce que a voudrait dire qu'il y a eu une erreur
		// (nanmoins, dans ce cas, on utilise la comparaison basique)
		return strText < s2.strText;
	}

	bool operator == (const XmlHeading & s2) const
	{
		return strText == s2.strText;
	}

	bool operator == (LPCSTR pszStr) const
	{
		return strText == pszStr;
	}

	XmlHeading(int nLev = 0) :
		nIcon(-1),
		nLevel(nLev),
		pParent(NULL)
	{
	}

	XmlHeading & AddChild()
	{
		push_back(XmlHeading(nLevel+1));

		return back();
	}

	void OutHhwTocNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent = "") const;

	void OutHhwIdxNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent = "");

	void OutHtmlTocNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent = "") const;

	void OutHtmlIdxNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent = "") const;
};

struct TemplateParts
{
	string	strBefore;
	string	strIndent;
	string	strAfter;
	string	strNewFile;
};

struct XmlStuff
{
	vector<string>	vFiles;		//!< Liste de fichiers HTML
	vector<string>	vTables;	//!< Liste de fichiers de tables de contenu
	vector<string>	vIndex;		//!< Liste de fichiers d'index
	string			strStart;	//!< Page par dfaut
	string			strTitle;	//!< Titre du document .chm
	XmlHeading		sToc;		//!< Construction de la table
	XmlHeading		sIdx;		//!< Construction de l'index
	XmlHeading *	pCurToc;	//!< Pointeur sur l'item de table courant
	int				nBaseLevel;	//!< Niveau d'intendation du fichier courant
	string			strTarget;	//!< Attribut "target" de la table de contenu
	string			strDir;		//!< Rpertoire des fichiers HTML
	string			strDirU;	//!< Rpertoire des fichiers HTML (format UNIX)

	XmlStuff() :
		pCurToc(NULL),
		nBaseLevel(0)
	{
		pCurToc = &sToc;
	}

	int	GetLevel() const
	{
		return pCurToc->nLevel;
	}

	XmlHeading & GetCurToc()
	{
		return *pCurToc;
	}

	XmlHeading & GetCurIdx()
	{
		return sIdx.back();
	}

	void IncToc()
	{
		XmlHeading & sNewHd = pCurToc->AddChild();

		sNewHd.pParent = pCurToc;
		sNewHd.strFile = sToc.strFile;
		pCurToc = & sNewHd;
	}

	void DecToc()
	{
		if (pCurToc->pParent)
			pCurToc = pCurToc->pParent;
	}

	typedef DWORD (XmlStuff::* ParseDataProc)(LPCTSTR , string &);

	DWORD	ProcessFile(LPCSTR pszFile, ParseDataProc proc);

	DWORD	ParseHtml(LPCTSTR pszName, string & strData);

	DWORD	ParseTopFilesXml(LPCTSTR pszName, string & strData);

	void	OutHhHeader(ostream & os);

	DWORD	OutHhwPrj(const string & strBaseName, const string & strOutFile);

	DWORD	OutHhwToc(const string & strBaseName);

	DWORD	OutHhwIdx(const string & strBaseName);

	DWORD	LoadTemplate(LPCSTR pszTemplate, TemplateParts & sParts);

	DWORD	OutHtmlToc(LPCSTR pszTemplate);

	DWORD	OutHtmlIdx(LPCSTR pszTemplate);
};

void XmlHeading::OutHhwTocNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent) const
{
	if (!strText.empty() || nLevel > 0) {
		string strTitle = !strText.empty() ? SgmlEncode(strText.c_str()) : "[titre manquant]";
		string strLocal;

		if (!strAnchor.empty() && strAnchor != "titre")
			strLocal += "#" + strAnchor;

		os	<< pszIndent << "<LI> <OBJECT type=\"text/sitemap\">" << endl
			<< pszIndent << "\t<param name=\"Name\" value=\"" << strTitle << "\">" << endl
			<< pszIndent << "\t<param name=\"Local\" value=\"" << sStuff.strDirU.c_str() << strFile.c_str() << strLocal.c_str() << "\">" << endl;
		if (nIcon >= 0)
			os	<< pszIndent << "\t<param name=\"ImageNumber\" value=\"" << nIcon << "\">" << endl;

		os	<< pszIndent << "\t</OBJECT>" << endl;
	}

	if (!empty()) {
		os << pszIndent << "<UL>" << endl;

		string strIndent(pszIndent);

		strIndent += "\t";
		for (ITERATE_CONST_VECTOR(*this, XmlHeading, it)) {
			it->OutHhwTocNode(sStuff, os, strIndent.c_str());
		}
		os << pszIndent << "</UL>" << endl;
	}
}

void XmlHeading::OutHhwIdxNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent)
{
	setlocale(LC_CTYPE, "French"); 
	sort(begin(), end());

	if (!strText.empty()) {
		string strName = SgmlEncode(strText.c_str());
		string strLocal;

		if (!strAnchor.empty() && strAnchor != "titre")
			strLocal += "#" + strAnchor;

		os	<< pszIndent << "<LI> <OBJECT type=\"text/sitemap\">" << endl
			<< pszIndent << "\t<param name=\"Name\" value=\"" << strName.c_str() << "\">" << endl;
		if (!strFile.empty())
			os	<< pszIndent << "\t<param name=\"Local\" value=\"" << sStuff.strDirU.c_str() << strFile.c_str() << strLocal.c_str() << "\">" << endl;
		else
			os	<< pszIndent << "\t<param name=\"See Also\" value=\"" << strName.c_str() << "\">" << endl;
		os	<< pszIndent << "\t</OBJECT>" << endl;
	}

	if (!empty()) {
		os << pszIndent << "<UL>" << endl;

		string strIndent(pszIndent);

		strIndent += "\t";
		for (ITERATE_VECTOR(*this, XmlHeading, it)) {
			it->OutHhwIdxNode(sStuff, os, strIndent.c_str());
		}
		os << pszIndent << "</UL>" << endl;
	}
}

void XmlStuff::OutHhHeader(ostream & os)
{
	os	<< "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">" << endl
		<< "<HTML>" << endl
		<< "<HEAD>" << endl
		<< "<meta name=\"GENERATOR\" content=\"Microsoft&reg; HTML Help Workshop 4.1\">" << endl
		<< "<!-- Sitemap 1.0 -->" << endl
		<< "</HEAD><BODY>" << endl;
}

DWORD XmlStuff::OutHhwPrj(const string & strBaseName, const string & strOutFile)
{
	ofstream	os((strBaseName+ ".hhp").c_str());

	os	<< "[OPTIONS]" << endl
		<< "Auto Index=Yes" << endl
		<< "Compatibility=1.1 or later" << endl
		<< "Compiled file=" << strOutFile << endl
		<< "Contents file=" << strBaseName << ".hhc" << endl;
	if (!strStart.empty())
		os	<< "Default topic=" << strDir << strStart << endl;
	os	<< "Display compile progress=Yes" << endl
		<< "Enhanced decompilation=Yes" << endl
		<< "Error log file=" << strBaseName << ".log" << endl
		<< "Full-text search=Yes" << endl
		<< "Index file=" << strBaseName << ".hhk" << endl
		<< "Language=0x402 Bulgarian (Bulgaria)" << endl;
	if (!strTitle.empty())
		os	<< "Title=" << ConvertUTF82Char(strTitle.c_str()) << endl;
	os	<< endl << endl << "[FILES]" << endl;
	for (ITERATE_CONST_VECTOR(vFiles, string, it))
		os	<< strDir << *it << endl;
	os	<< endl << "[INFOTYPES]" << endl;
	return 0;
}

DWORD XmlStuff::OutHhwToc(const string & strBaseName)
{
	ofstream	os((strBaseName + ".hhc").c_str());

	OutHhHeader(os);
	sToc.OutHhwTocNode(*this, os);

	os	<< "</BODY></HTML>" << endl;
	return 0;
}

DWORD XmlStuff::OutHhwIdx(const string & strBaseName)
{
	ofstream	os((strBaseName + ".hhk").c_str());

	OutHhHeader(os);
	sIdx.OutHhwIdxNode(*this, os);

	os	<< "</BODY></HTML>" << endl;
	return 0;
}

void XmlHeading::OutHtmlTocNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent) const
{
	if (!strText.empty() || nLevel > 0) {
		string strTitle = !strText.empty() ? HtmlEncode(strText.c_str()) : "<b style=\"color:red\">[titre manquant]</b>";
		string strLocal;
		string strAttr = "";

		if (strAnchor == "titre") {
			strAttr = " class=\"page\"";
			os << endl;
		} else if (!strAnchor.empty())
			strLocal += "#" + strAnchor;
		if (!sStuff.strTarget.empty())
			strAttr += " target=\"" + sStuff.strTarget + "\"";

		os	<< endl << pszIndent << "<li><a href=\"" << strFile << strLocal << "\"" << strAttr << ">"
			<< strTitle << "</a>";
	}

	if (!empty()) {
		os << "<ul>";

		string strIndent(pszIndent);

		strIndent += pszIndent[0] == '\t' ? "\t" : "  ";

		for (ITERATE_CONST_VECTOR(*this, XmlHeading, it)) {
			it->OutHtmlTocNode(sStuff, os, strIndent.c_str());
		}
		os << "</ul>";
	}
	if (!strAnchor.empty())
		os << "</li>";
}

void XmlHeading::OutHtmlIdxNode(const XmlStuff & sStuff, ostream & os, LPCSTR pszIndent) const
{
	if (!strText.empty()) {
		string strLocal;
		string strAttr;
		string strLink = HtmlEncode(strText.c_str());

		if (!strAnchor.empty())
			strLocal += "#" + strAnchor;
		if (!sStuff.strTarget.empty())
			strAttr += " target=\"" + sStuff.strTarget + "\"";

		if (!strFile.empty())
			strLink = "<a href=\"" + strFile + strLocal + "\"" + strAttr + ">" + strLink + "</a>";
		else
			strLink = "<span class=\"nolink\">" + strLink + "</span>";

		os	<< endl << pszIndent << "<li>" << strLink;
	}

	if (!empty()) {
		os << "<ul>";

		string strIndent(pszIndent);

		strIndent += pszIndent[0] == '\t' ? "\t" : "  ";

		for (ITERATE_CONST_VECTOR(*this, XmlHeading, it)) {
			it->OutHtmlIdxNode(sStuff, os, strIndent.c_str());
		}
		os << "</ul>";
	}
	if (!strText.empty())
		os << "</li>";
}

DWORD XmlStuff::LoadTemplate(LPCSTR pszTemplate, TemplateParts & sParts)
{
	cout << "Traitement du template \"" << pszTemplate << "\"" << endl;

	string strData;

	DWORD dwErr = LoadFileInString(pszTemplate, strData);

	if (dwErr!=0)
		return dwErr;

	size_t nBeg = strData.find("<!-- <template ");

	if (nBeg == string::npos) {
		cout << "Impossible de trouver le point de substitution" << endl;
		return MAXWORD;
	}

	size_t nEnd = strData.find("-->", nBeg+14);

	if (nEnd == string::npos) {
		cout << "Impossible de trouver la fin du point de substitution" << endl;
		return MAXWORD;
	}

	size_t nBol = strData.rfind("\n", nBeg);

	if (nBol == string::npos)
		nBol = 0;
	else
		++nBol;

	string strComment = strData.substr(nBeg+4, (nEnd - nBeg)-4);

	sParts.strBefore	= strData.substr(0, nBeg);
	sParts.strIndent	= strData.substr(nBol, nBeg-nBol);
	sParts.strAfter		= strData.substr(nEnd+3);

	// Analyse du contenu du commentaire
	TiXmlDocument	sTiXmlDoc;

	sTiXmlDoc.Parse(strComment.c_str());
	if (sTiXmlDoc.Error()) {
		cout << "Erreur  l'analyse des paramtres de substitution" << endl;
		return MAXWORD;
	}

	TiXmlElement * pszElm = sTiXmlDoc.FirstChildElement("template");

	strTarget.clear();
	if (pszElm) {
		LPCSTR pszTarget = pszElm->Attribute("target");

		if (pszTarget)
			strTarget = pszTarget;
	}

	sParts.strNewFile	= pszTemplate;
	size_t	nExt		= sParts.strNewFile.find(".tpl");

	if (nExt == string::npos) {
		cout << "Le fichier d'entre n'a pas l'extension \".tpl\"" << endl;
		return MAXWORD;
	}
	sParts.strNewFile.replace(nExt, 4, ".html");
	return 0;
}

DWORD XmlStuff::OutHtmlToc(LPCSTR pszTemplate)
{
	TemplateParts sParts;

	DWORD	dwErr = LoadTemplate(pszTemplate, sParts);

	if (dwErr != 0)
		return dwErr;

	ofstream	os(sParts.strNewFile.c_str());

	os	<< sParts.strBefore.c_str();

	sToc.OutHtmlTocNode(*this, os, sParts.strIndent.c_str());

	os	<< sParts.strAfter.c_str();

	return 0;
}


DWORD XmlStuff::OutHtmlIdx(LPCSTR pszTemplate)
{
	TemplateParts sParts;

	DWORD	dwErr = LoadTemplate(pszTemplate, sParts);

	if (dwErr != 0)
		return dwErr;

	ofstream	os(sParts.strNewFile.c_str());

	os	<< sParts.strBefore.c_str();

	sIdx.OutHtmlIdxNode(*this, os, sParts.strIndent.c_str());

	os	<< sParts.strAfter.c_str();

	return 0;
}

DWORD XmlStuff::ProcessFile(LPCSTR pszFile, ParseDataProc proc)
{
	cout << "Traitement du fichier \"" << pszFile << "\"" << endl;

	string strData;
	DWORD	dwErr = LoadFileInString(pszFile, strData);

	if (dwErr == 0)
		dwErr = (this->*proc)(pszFile, strData);
	return dwErr;
}

// ----------------------------------------------------------------------------

class HtmlParserCb : public TiXmlVisitor
{
	XmlStuff &				sStuff;
	const TiXmlElement *	pTocOpen;
	const TiXmlElement *	pIdxOpen;
	XmlHeading *			pIdxRoot;
	string					strId;
	string					strKey;

	int isHeading (LPCSTR pszTag)
	{
		if (pszTag[0] == 'h' && pszTag[1] >= '1' && pszTag[1] <= '6' && pszTag[2]==0)
			return pszTag[1] - '0';
		return 0;
	}

	//bool isCharTag(LPCSTR pszTag) {
	//	return
	//		strcmp(pszTag, "a")		== 0 ||
	//		strcmp(pszTag, "abbr")	== 0 ||
	//		strcmp(pszTag, "b")		== 0 ||
	//		strcmp(pszTag, "big")	== 0 ||
	//		strcmp(pszTag, "blink")	== 0 ||
	//		strcmp(pszTag, "cite")	== 0 ||
	//		strcmp(pszTag, "code")	== 0 ||
	//		strcmp(pszTag, "dfn")	== 0 ||
	//		strcmp(pszTag, "em")	== 0 ||
	//		strcmp(pszTag, "font")	== 0 ||
	//		strcmp(pszTag, "i")		== 0 ||
	//		strcmp(pszTag, "kbd")	== 0 ||
	//		strcmp(pszTag, "nobr")	== 0 ||
	//		strcmp(pszTag, "q")		== 0 ||
	//		strcmp(pszTag, "s")		== 0 ||
	//		strcmp(pszTag, "samp")	== 0 ||
	//		strcmp(pszTag, "small")	== 0 ||
	//		strcmp(pszTag, "span")	== 0 ||
	//		strcmp(pszTag, "spell")	== 0 ||
	//		strcmp(pszTag, "strike")== 0 ||
	//		strcmp(pszTag, "strong")== 0 ||
	//		strcmp(pszTag, "tt")	== 0 ||
	//		strcmp(pszTag, "u")		== 0 ||
	//		strcmp(pszTag, "var")	== 0;
	//}

	/// Tag H1  H6 avec "id" : dbuter la mmorisation du contenu
	void AddTocItem(const TiXmlElement & sXmlElm, int nNewLev)
	{
		while (nNewLev <= sStuff.GetLevel())
			sStuff.DecToc();
		while (nNewLev > sStuff.GetLevel())
			sStuff.IncToc();

		XmlHeading & sCurToc = sStuff.GetCurToc();

		sCurToc.strAnchor = strId;

		LPCSTR pszIcon = sXmlElm.Attribute("icon");
		if (pszIcon)
			sCurToc.nIcon = atoi(pszIcon);
		pTocOpen = & sXmlElm;
	}

	void CompleteKeyItem(LPCSTR pszKey)
	{
		// Chercher si le titre existe
		XmlHeading::iterator	it = find(pIdxRoot->begin(), pIdxRoot->end(), pszKey);
		XmlHeading &			sCurToc = sStuff.GetCurToc();
		XmlHeading *			pNewIdx;

		if (it != pIdxRoot->end() &&it->strText.empty())
			pNewIdx = &*it;
		else
			pNewIdx = &pIdxRoot->AddChild();

		// Utiliser le dernier "id" dfini comme cible.
		pNewIdx->strAnchor = strId;
		pNewIdx->strFile = sCurToc.strFile;
		pNewIdx->strText = pszKey;
	}


	void ProcessKeyItem(const TiXmlElement & sXmlElm, string strTitle)
	{
		pIdxRoot = &sStuff.sIdx;	// Point d'ajout

		// Chercher des catgories et des sous-catgories
		for (size_t nPos; (nPos = strTitle.find("|")) != string::npos; strTitle.erase(0, nPos+1)) {
			// Catgorie prsente : chercher si elle existe
			string					strCat = strTitle.substr(0, nPos);
			XmlHeading::iterator	it = find(pIdxRoot->begin(), pIdxRoot->end(), strCat.c_str());

			if (it == pIdxRoot->end()) {
				pIdxRoot = &pIdxRoot->AddChild();

				pIdxRoot->strText = strCat;
			} else
				pIdxRoot = &*it;
		}

		// Ne pas retenir "1" comme titre
		if (strTitle=="1")
			strTitle.erase();

		// Si l'attribut "ix" a un contenu, utiliser celui-ci comme libell. Sinon,
		// on se servira du contenu du "tag".
		if (!strTitle.empty()) {
			CompleteKeyItem(strTitle.c_str());
		} else {
			pIdxOpen = & sXmlElm;
		}
	}

	/// Ouverture de tag
	virtual bool VisitEnter (const TiXmlElement & sXmlElm, const TiXmlAttribute * pXmlAttr)
	{
		LPCSTR	pszValue = sXmlElm.Value();

		if (pszValue) {
			// Extraire le tag et le convertir en minuscules
			CHAR	szWrk[8];

			strncpy_s(szWrk, pszValue, _countof(szWrk)-1);
			szWrk[_countof(szWrk)-1] = 0;
			_strlwr_s(szWrk);

			// Retenir le contenu de tout attribut "id" ou "<a name=..."
			LPCSTR	pszId = sXmlElm.Attribute("id");

			if (pszId)
				strId = pszId;
			else if (strcmp(szWrk, "a")) {
				pszId = sXmlElm.Attribute("name");
				if (pszId)
					strId = pszId;
			}

			int		nHeading;

			if (!pTocOpen && (nHeading = isHeading(szWrk)) != 0) {	// Tester si c'est un tag H1  H6

				if (pszId)
					AddTocItem(sXmlElm, nHeading + sStuff.nBaseLevel);
			}

			LPCSTR	pszIx = sXmlElm.Attribute("ix");

			if (pszIx) {
				if (!pIdxOpen)	// Tester si un index n'est pas dj ouvert
					ProcessKeyItem(sXmlElm, ConvertUTF82Char(pszIx));
				else {
					cout	<< "Erreur durant l'analyse du fichier\"" << sStuff.pCurToc->strFile << "\"" << endl
							<< " Ligne=" << sXmlElm.Row() << ", Colonne=" << sXmlElm.Column() << endl
							<< " Description=\"Attribut ix='" << pszIx << "' illegalement imbrique\"." << endl;
				}
			}
		}

		return true;
	}

	/// Fermeture de tag
	virtual bool VisitExit (const TiXmlElement & sXmlElm)
	{
		LPCSTR			pszValue = sXmlElm.Value();
		if (pTocOpen == &sXmlElm) {
			XmlHeading &	sCurToc = sStuff.GetCurToc();

			CleanString(sCurToc.strText);
			pTocOpen = NULL;
		}
		if (pIdxOpen == &sXmlElm) {
			CleanString(strKey);
			CompleteKeyItem(strKey.c_str());
			pIdxOpen = NULL;
			pIdxRoot = NULL;
			strKey.clear();
		}

		return true;
	}

	/// Contenu de tag
	virtual bool Visit (const TiXmlText & sXmlTxt)
	{	// Rcupration de fragments de titre
		if (pTocOpen) {
			LPCSTR			pszValue = sXmlTxt.Value();
			XmlHeading &	sCurToc = sStuff.GetCurToc();

			sCurToc.strText += ConvertUTF82Char(pszValue).c_str();
		}
		if (pIdxOpen) {
			LPCSTR			pszValue = sXmlTxt.Value();

			strKey += ConvertUTF82Char(pszValue);
		}
		return true;
	}

public:
	HtmlParserCb(XmlStuff & sStuf) :
		sStuff(sStuf),
		pTocOpen(NULL),
		pIdxOpen(NULL),
		pIdxRoot(NULL)
	{}
};

// ----------------------------------------------------------------------------

void CheckTag(LPCSTR pszTag, string & strData)
{
	string::size_type nInx = 0;

	while ((nInx = strData.find(pszTag, nInx)) != string::npos) {
		string::size_type nIny = strData.find(">", nInx);

		if (nIny == string::npos || nIny == 0)
			break;
		nInx = nIny+1;
		if (strData[nIny-1] != '/') {
			strData.insert(nIny, "/");
			++nInx;
		}
	}
}

void PreprocessData(string & strData)
{
	CheckTag("<meta", strData);
	CheckTag("<link", strData);
	CheckTag("<img", strData);
	CheckTag("<br", strData);
	CheckTag("<hr", strData);
	CheckTag("<wbr", strData);
	CheckTag("<frame ", strData);

	string::size_type nInx = 0;

	while ((nInx = strData.find("&", nInx)) != string::npos) {
		string::size_type	nIny = strData.find(";", nInx);

		if (nIny == string::npos)
			break;

		CHAR				szEntity[10];
		string::size_type	nLen = nIny-nInx-1;

		if (nLen > _countof(szEntity)-1) {
			nInx++;
			continue;
		}

		strData._Copy_s(szEntity, _countof(szEntity), nLen, nInx+1);
		szEntity[nLen] = 0;

		LPCSTR pszUtf = EntityToUtf8(szEntity);

		strData.replace(nInx, nLen+2, pszUtf);
		nInx += strlen(pszUtf);
	}
}

DWORD XmlStuff::ParseHtml(LPCTSTR pszName, string & strData)
{
	PreprocessData(strData);

	TiXmlDocument	sTiXmlDoc;

	sTiXmlDoc.SetCondenseWhiteSpace(false);
	sTiXmlDoc.Parse(strData.c_str());
	if (sTiXmlDoc.Error()) {
		CHAR	szName[MAX_PATH];

		sprintf_s(szName, "%s.xml", pszName);

		cout	<< "Erreur durant l'analyse du fichier\"" << pszName << "\"" << endl
				<< " Ligne=" << sTiXmlDoc.ErrorRow() << ", Colonne=" << sTiXmlDoc.ErrorCol() << endl
				<< " Description=\"" << sTiXmlDoc.ErrorDesc() << "\"." << endl
				<< " Gnration du fichier \"" << szName << "\" pour diagnostic." << endl;

		TiXmlDeclaration	sXmlDecl("1.0", "UTF-8", "yes");

		sTiXmlDoc.InsertBeforeChild(sTiXmlDoc.FirstChild(), sXmlDecl);
		sTiXmlDoc.SaveFile(szName);
		return ERROR_INVALID_DATA;
	}

	HtmlParserCb	sParserCb(*this);

	sTiXmlDoc.Accept(&sParserCb);
	return 0;
}

class TopXmlParserCb : public TiXmlVisitor
{
public:
	XmlStuff &	sStuff;
	int			nLev;
	DWORD		dwErr;

	virtual bool 	VisitEnter (const TiXmlElement & sXmlElm, const TiXmlAttribute * pXmlAttr)
	{
		if (dwErr != 0)
			return false;

		LPCSTR	pszValue	= sXmlElm.Value();
		LPCSTR	pszName		= sXmlElm.Attribute("name");
		LPCSTR	pszAttr;

		if (_stricmp(pszValue, "files")==0) {
			pszAttr = sXmlElm.Attribute("title");
			if (pszAttr)
				sStuff.strTitle = pszAttr;
		} else if (_stricmp(pszValue, "file")==0) {
			if (pszName) {
				sStuff.sToc.strFile = pszName;
				sStuff.nBaseLevel = nLev++;
				dwErr = sStuff.ProcessFile((sStuff.strDir + pszName).c_str(), &XmlStuff::ParseHtml);
				if (dwErr != 0) {
					DisplayError(dwErr);
					return false;
				}
				sStuff.vFiles.push_back(pszName);
				pszAttr = sXmlElm.Attribute("default");
				if (pszAttr)
					sStuff.strStart = pszName;
			}
		} else if (_stricmp(pszValue, "table")==0) {
			if (pszName)
				sStuff.vTables.push_back(pszName);
		} else if (_stricmp(pszValue, "index")==0) {
			if (pszName)
				sStuff.vIndex.push_back(pszName);
		}
		return dwErr == 0;
	}

	virtual bool 	VisitExit (const TiXmlElement & sXmlElm)
	{
		LPCSTR	pszValue = sXmlElm.Value();

		if (_stricmp(pszValue, "file")==0)
			sStuff.nBaseLevel = --nLev;
		return dwErr == 0;
	}

	TopXmlParserCb(XmlStuff & sStuf) :
		sStuff(sStuf),
		nLev(0),
		dwErr(0)
	{}
};

DWORD XmlStuff::ParseTopFilesXml(LPCTSTR pszName, string & strData)
{
	TiXmlDocument	sTiXmlDoc(pszName);

	sTiXmlDoc.Parse(strData.c_str());
	if (sTiXmlDoc.Error()) {
		cout	<< "Erreur durant l'analyse du fichier\"" << pszName << "\"" << endl
				<< " ligne=" << sTiXmlDoc.ErrorRow() << ", colonne=" << sTiXmlDoc.ErrorCol() << endl
				<< " description=\"" << sTiXmlDoc.ErrorDesc() << "\"" << endl;
	}

	TopXmlParserCb	sParserCb(*this);

	sTiXmlDoc.Accept(&sParserCb);
	return sParserCb.dwErr;
}

int main(int argc, LPCSTR * argv)
{
	UINT	nCp		= GetConsoleOutputCP();
	DWORD	dwErr	= 0;

	SetConsoleOutputCP(1252);
	nCp = GetConsoleOutputCP();
	if (argc < 3 || argc >4) {
		cout	<< "Usage  : index_parser <fichier liste> <base nom Hhw> [<fichier sortie>]" << endl
				<< "    o :" << endl
				<< "  <fichier liste> est un fichier XML contenant la liste des fichiers  traiter ;" << endl
				<< "  <base nom Hhw> est le nom de base des fichiers destins  HTML Help Workshop ;" << endl << endl
				<< "  <fichier sortie> est chemin optionnel complet du fichier .CHM de sortie  compiler" << endl
				<< "    (par dfaut dtermin  partir de <base nom Hhw>)." << endl << endl
				<< "Logiciel d'analyse d'une srie de fichiers HTML en vue de produire des tables" << endl
				<< " de contenu et des fichiers d'index, soit en HTML, soit sous forme de fichiers" << endl
				<< " de projet pour gnrer des documents .chm  l'aide de HTML Help Workshop." << endl;
		return -1;
	}

	cout << "Index_parser - compiled on " __TIMESTAMP__  << endl << endl;

	string		strXml = argv[1];		// 1er argument : nom du fichier XML de description

	// Pas de ".\" au dbut
	if (strncmp(strXml.c_str(), ".\\", 2)==0)
		strXml.erase(0, 2);

	string		strHhw = argv[2];		// 2nd argument : nom de base des fichiers Hhw
	string		strOut = argc==4 ? argv[3] : strHhw + ".chm";
	XmlStuff	sStuff;

	size_t	nPos = strXml.rfind("\\");

	if (nPos != string::npos)
		sStuff.strDir = strXml.substr(0, nPos+1);
	strXml.erase(0, nPos+1);
	sStuff.strDirU = sStuff.strDir;

	while ((nPos = sStuff.strDirU.find("\\")) != string::npos)
		sStuff.strDirU.replace(nPos, 1, "/");

	// Traitement du fichier XML d'entre et des fichiers HTML qui y sont inclus
	dwErr = sStuff.ProcessFile((sStuff.strDir + strXml).c_str(), &XmlStuff::ParseTopFilesXml);

	// Sortie du fichier projet Hhw (.hhp)
	if (dwErr == 0)
		dwErr = sStuff.OutHhwPrj(strHhw, strOut);
	// Sortie du fichier de table de contenu Hhw (.hhc)
	if (dwErr == 0)
		dwErr = sStuff.OutHhwToc(strHhw);
	// Sortie du fichier d'index Hhw (.hhk)
	if (dwErr == 0)
		dwErr = sStuff.OutHhwIdx(strHhw);

	if (dwErr == 0) {
		// Sortie des tables de contenu HTML
		for (ITERATE_CONST_VECTOR(sStuff.vTables, string, it)) {
			dwErr = sStuff.OutHtmlToc((sStuff.strDir + *it).c_str());
			if (dwErr != 0)
				break;
		}
	}

	if (dwErr == 0) {
		// Sortie des tables d'index HTML
		for (ITERATE_CONST_VECTOR(sStuff.vIndex, string, it)) {
			dwErr = sStuff.OutHtmlIdx((sStuff.strDir + *it).c_str());
			if (dwErr != 0)
				break;
		}
	}

	if (dwErr != 0)
		DisplayError(dwErr);
#ifdef _DEBUG
	Sleep(2000);
#endif
	return dwErr;
}
