/*
 * $Id: capture.cpp 264 2010-11-01 22:49:12Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion des enregistrements (TS, PS, multiplex)
 *
 *	Ce fichier implmente l'enregistrement en mode TS, PS et multiplex,  partir
 *	des chantillons fournis par l'interface "callback" du filtre "Sample Grabber".
 *
 *	Il assure donc, en particulier, la conversion des flux TS (Transport Stream)
 *	en enregistrements PS (Program Stream).
 **/

#include "stdafx.h"
#include "console.h"
#include "capture.h"

// Taille maximale des paquests PS mis :
#define CHUNK_SIZE 0x800

// ----------------------------------------------------------------------------
//	Implmentations de la classe "CCapture"
// ----------------------------------------------------------------------------

/**
 * \brief Constructeur de base
 * \param[in]	hFil	"Handle" de fichier ouvert en criture pour enregistrer le
 *						rsultat de la capture
 **/
CCapture::CCapture(HANDLE hFil) :
	hFile(hFil)
{
#ifdef TRACE_CCAPTURE_CONSTR
	myprintf(TEXT("## Construction CCapture, dbut d'enregistrement (0x%08x)") EOL, hFile);
#endif
}

//UINT64 CCapture::pos_fichier(void)
//{
//	LARGE_INTEGER filepos;
//
//	filepos.QuadPart = 0;
//	if (SetFilePointerEx(hFile, filepos, &filepos, FILE_CURRENT)!=0)
//		return UINT64(filepos.QuadPart);
//	return 0;
//}

/**
 * \brief Sortie d'un bloc de donnes vers le fichier
 * \param[in]	pBuf	Pointeur sur le premier octet du bloc
 * \param[in]	sizBuf	Longueur du bloc
 * \return		0 si succs, code d'erreur systme si chec
 **/
DWORD CCapture::ecrit_fichier(LPCVOID pBuf, DWORD sizBuf) const
{
	DWORD written;

	if (WriteFile(hFile, pBuf, sizBuf, &written, NULL) == 0)
		return GetLastError();
	return 0;
}

/**
 * \brief (Re)lecture d'un bloc de donnes depuis le fichier
 * \param[in]	pBuf	Pointeur sur le premier octet du tampon de rception
 * \param[in]	nToRead	Longueur  lire
 * \return		0 si succs, code d'erreur systme si chec
 **/
DWORD CCapture::lit_fichier(LPVOID pBuf, DWORD nToRead) const
{
	DWORD read;

	if (ReadFile(hFile, pBuf, nToRead, &read, NULL) == 0)
		return GetLastError();
	if(read != nToRead)
		return ERROR_HANDLE_EOF;
	return 0;
}

UINT64 CCapture::seek(UINT64 nPos, DWORD dwMoveMethod) const throw (...)
{
	LARGE_INTEGER	sPos = {0};
	LARGE_INTEGER	sRes = {0};

	sPos.QuadPart = nPos;
	if (!SetFilePointerEx(hFile, sPos, &sRes, dwMoveMethod))
		throw GetLastError();
	return sRes.QuadPart;
}

/* Destructeur gnral */

CCapture::~CCapture()
{
#ifdef TRACE_CCAPTURE_CONSTR
	myprintf(TEXT("## Destruction CCapture, fin d'enregistrement (0x%08x)") EOL, hFile);
#endif
	CloseHandle(hFile);
}

// ----------------------------------------------------------------------------
//	Implmentations de la classe "CCapture_TS"
// ----------------------------------------------------------------------------

CCapture_TS::CCapture_TS(HANDLE hFil, const ChainePMT & lst_pistes) :
	CCapture(hFil),
	pistes(lst_pistes)
{
}

DWORD CCapture_TS::traite_paquet(const TS_packet & tsp)	// virtual
{
	// une chane en TS
	DWORD nErr = 0;
	UINT16 pid = tsp.hdr.pid();

	if (pid == p_pid_PAT || pistes.cherche_pid(pid))
		nErr = ecrit_fichier(tsp.bytes, TS_SIZE);
	return nErr;
}

// ----------------------------------------------------------------------------
//	Classe "PS_chunk", dfinitions et implmentations
// ----------------------------------------------------------------------------

// Limites du champ "program mux rate".
// Le programme essaie d'estimer la valeur relle, mais si le rsultat sort de ces limites,
// la limite correspondante sera utilise en remplacement.
#define MIN_PROGRAM_MUX_RATE (0x1000*50)
#define MAX_PROGRAM_MUX_RATE (0x3000*50)

class PS_chunk
{
	const PS_hdr_wd &	src_hdr;	// Rfrence sur l'en-tte source
	CCapture_PS &		cCapture;
	PS_buffer_type		eBType;
public:
	UINT				ixBuf;
	PS_hdr_wd *			p_curPES;	// Pointeur sur l'en-tte du PES en cours de construction
	UINT8				buffer[CHUNK_SIZE];

	PS_chunk(const PS_hdr_wd & hdr, PS_buffer & cBuff, PS_buffer_type eBTyp);

	// Ajout d'un bloc "Pack Header" :
	// (retourne la taille du bloc ajout, ou bien 0 si chec)
	UINT	addPackHeader();

	// Ajout d'un bloc "System Header" :
	// (retourne la taille du bloc ajout, ou bien 0 si chec)
	UINT	addSystemHeader();

	// Ajout d'un bloc PES de base, driv du bloc source 'p_src_hdr' :
	// (retourne la taille du bloc ajout, ou bien 0 si chec)
	UINT 	addPES_Header();

	// Ajout d'un bloc PES d'extension, driv du bloc source 'p_src_hdr' :
	// (retourne la taille du bloc ajout, ou bien 0 si chec)
	UINT 	addPES_extHeader();

	// Complter le bloc avec un PES de padding, en laissant nanmoins
	// ventuellement un espace de taille gale  'guard'  la fin.
	// Si l'ajustement requis est trs faible, alors l'ajustement se fera,  la place,
	// en ajoutant des octets de "stuffing" dans l'en-tte.
	// Retourne la quantit de donnes *ajoutes* au buffer :
	UINT	add_Padding(size_t guard=0);

	// Ajout de donnes dans le bloc. Retourne 'true' si la totalit des donnes ont pu tre ajoutes,
	// 'false' sinon ; dans ce dernier cas, rien n'est modifi.
	bool addData(const UINT8 * ptr, UINT32 size);

	// Ajout de donnes dans le bloc. Retourne 'true' si la totalit des donnes ont pu tre ajoutes,
	// 'false' sinon ; dans ce dernier cas, 'ptr' et 'size' sont ajusts sur les donnes restantes.
	bool addPES_data(const UINT8 * & ptr, UINT32 & size);

	// Ajustement de l'index du buffer pour que les donnes subsquentes s'inscrivent
	//  partir de l'adresse 'ptr' :
	PUINT8	setBufferIndex(PUINT8 ptr) {
				ixBuf = UINT16(ptr-&buffer[0]); return ptr;}

	// Construction d'un buffer contenant autant de donnes que possible parmi celles restant disponibles.
	// S'il ne reste pas assez de donnes disponibles pour remplir le buffer, des donnes de remplissage
	// sont ajoutes.
	// La fonction retourne le nombre d'octets *ajouts* (en-ttes, remplissage, etc) par rapport au flot
	// d'origine.
	UINT	build(bool start, const UINT8 * & ptr, UINT32 & size);

	// Vidange du tampon dans le fichier :
	void write_to_file() throw(...);
};

PS_chunk::PS_chunk(const PS_hdr_wd & hdr, PS_buffer & cBuff, PS_buffer_type eBTyp) :
	src_hdr(hdr),
	cCapture(cBuff.cCapture),
	eBType(eBTyp),
	ixBuf(0),
	p_curPES(NULL)
{
}

// Construction d'un buffer contenant autant de donnes que possible parmi celles restant disponibles.
// S'il ne reste pas assez de donnes disponibles pour remplir le buffer, des donnes de remplissage
// sont ajoutes.
// La fonction retourne le nombre d'octets *ajouts* (en-ttes, remplissage, etc) par rapport au flot
// d'origine.
UINT PS_chunk::build(bool start, const UINT8 * & ptr, UINT32 & size)
{
	UINT32 added_bytes = 0;

	if (start) {
		added_bytes += addPackHeader();
		if (!cCapture.sys_header_set) {
			added_bytes += addSystemHeader();
			cCapture.sys_header_set = true;
		}
		//
		UINT32 hdsize = addPES_Header();
		// Le premier en-tte existait dans le tampon source, on le saute donc en lecture
		if (hdsize > size)
			hdsize = size;	// prcaution en cas de corruption de donnes
		ptr += hdsize;
		size -= hdsize;
	} else {
		// Les en-ttes subsquents sont ajouts, donc on n'enlve pas de donnes de la source
		added_bytes += addPES_extHeader();
	}

	addPES_data(ptr, size);
	if (size==0)
		added_bytes += (UINT32)add_Padding();
	return added_bytes;
}

void PS_chunk::write_to_file() throw(...)
{
	DWORD nRes = cCapture.ecrit_fichier(buffer, sizeof(buffer));

	if (nRes != 0)
		throw nRes;
}

// Ajout de donnes dans le bloc. Retourne 'true' si la totalit des donnes ont pu tre ajoutes,
// 'false' sinon ; dans ce dernier cas, rien n'est modifi.
bool PS_chunk::addData(const UINT8 * ptr, UINT32 size)
{
	if (ixBuf+size <= _countof(buffer)) {
		ASSERT(ptr != NULL);
		ASSERT(ixBuf < _countof(buffer));
		ASSERT(size <= _countof(buffer) - ixBuf);
		memcpy(&buffer[ixBuf], ptr, size);
		ixBuf	= ixBuf + UINT16(size);
		return true;
	}
	return false;
}

// Ajout de donnes dans le bloc. Retourne 'true' si la totalit des donnes ont pu tre ajoutes,
// 'false' sinon ; dans ce dernier cas, 'ptr' et 'size' sont ajusts sur les donnes restantes.
bool PS_chunk::addPES_data(const UINT8 * & ptr, UINT32 & size)
{
	UINT32	adjustedSize	= UINT32(size);
	bool	fit				= ixBuf+size <= _countof(buffer);

	if (!fit)
		adjustedSize = sizeof(buffer) - ixBuf;

	addData(ptr, adjustedSize);
	ptr		+= adjustedSize;
	size	-= adjustedSize;
	if (p_curPES)
		p_curPES->set_data_size(&buffer[ixBuf]);

	return fit;
}

// Ajout d'un bloc "Pack header" :
UINT PS_chunk::addPackHeader()
{
	if (sizeof(PS_chunk)-ixBuf >= sizeof(PS_PackHeader_v2)) {
		PS_PackHeader_v2 & pkh = *new(&buffer[ixBuf])PS_PackHeader_v2();	// "Placement" new

		pkh.scr			= cCapture.get_dst_pcr();	// transformation par oprateur surcharg
		pkh.mux_rate	= cCapture.byte_mux_rate/50;
		pkh.length		= 0xf8;
		setBufferIndex(pkh.end_of_header());
		return pkh.packet_size();
	}

	return 0;
}

// Ajout d'un bloc "System Header" :
UINT PS_chunk::addSystemHeader()
{
	if (sizeof(PS_chunk)-ixBuf >= sizeof(PS_SystemHeader)+2*sizeof(PS_SystemHeader::s_entry)) {
		PS_SystemHeader &	shd = *new (&buffer[ixBuf])PS_SystemHeader();	// "Placement" new

		shd.rate_bound						= MAX_PROGRAM_MUX_RATE;
		shd.audio_bound_and_flags			= 1<<2;		// SH_FIXED_FLAG=0, SH_CSPS_FLAG=0
		shd.video_bound_and_flags			= 1|0x20;	// SH_SYSTEM_AUDIO_LOCK_FLAG=0, SH_SYSTEM_VIDEO_LOCK_FLAG =0, 0x20=marker bit
		shd.packet_rate_restriction_flag 	= 0;		// SH_PACKET_RATE_RESTRICTION_FLAG=0

		shd[0].stream_id = sid_SB_AllAudio;
		shd[0].bound_buf.set_size(0, 0x20);	// 0x001000
		shd[1].stream_id = sid_SB_AllVideo;
		shd[1].bound_buf.set_size(1, 0xe8);	// 0x03a000
		shd.set_data_size(setBufferIndex(PUINT8(shd[1].next())));
		return shd.packet_size();
	}

	return 0;
}

// Ajout d'un bloc PES de base, driv du bloc source 'p_src_hdr' :
// (retourne la taille du bloc ajout, ou bien 0 si chec)
UINT PS_chunk::addPES_Header()
{
	if (sizeof(PS_chunk)-ixBuf >= sizeof(PS_hdr_wd)) {
		PS_hdr_wd &	dst_hdr	= *new (&buffer[ixBuf]) PS_hdr_wd(src_hdr);	// "Placement" new

		p_curPES = &dst_hdr;

#ifdef _DEBUG
		PS_data_unwrapper_full uw_ph(dst_hdr);
#else
		PS_data_unwrapper_base uw_ph(dst_hdr);
#endif

		// Ajustement des champs de timing pour les ramener  une base commenant au dbut du
		// fichier :
		if (uw_ph.p_pts) {
			*uw_ph.p_pts = cCapture.adjust_pcr(*uw_ph.p_pts);
			// myprintf(TEXT("PS pts = %10.6f (%") A2t TEXT(")") EOL, UINT64(*uw_ph.p_pts)/27000000.0, pbt2str(eBType));
		}
		if (uw_ph.p_dts) {
			*uw_ph.p_dts = cCapture.adjust_pcr(*uw_ph.p_dts);
			// myprintf(TEXT("PS dts = %10.6f (%") A2t TEXT(")") EOL, UINT64(*uw_ph.p_dts)/27000000.0, pbt2str(eBType));
		}

		setBufferIndex(dst_hdr.end_of_header());
		return dst_hdr.header_size();
	}

	return 0;
}

// Ajout d'un bloc PES d'extension, driv du bloc source 'p_src_hdr' :
UINT32 PS_chunk::addPES_extHeader()
{
	if (sizeof(PS_chunk)-ixBuf >= sizeof(PS_hdr_wd)) {
		PS_hdr_wd	&	dst_hdr	= *new (&buffer[ixBuf])PS_hdr_wd(src_hdr);	// "Placement" new

		p_curPES			=	&dst_hdr;
		dst_hdr.info		&=	~DATA_ALIGNMENT_INDICATOR;	// reset data aligment indicator
		dst_hdr.flags		=	0;	// aucun champ ajout
		dst_hdr.data_length	=	0;	// longueur de champs ajouts nulle
		setBufferIndex(dst_hdr.end_of_header());
		return dst_hdr.header_size();
	}

	return 0;
}

// Complter le bloc avec un PES de padding, en laissant nanmoins
// ventuellement un espace de taille gale  'guard'  la fin.
// Retourne la quantit de donnes *ajoutes* au buffer.
UINT PS_chunk::add_Padding(size_t guard)
{
	UINT16 remain = UINT16(sizeof(buffer)-guard-ixBuf);

	if (remain>0) {
		if (remain >= 0x10) {
			PS_hdr_wl &	hdr	= *new (&buffer[ixBuf])PS_hdr_wl(sid_Padding_Stream);	// "Placement" new

			remain -= sizeof(PS_hdr_wl);
			hdr.hdr_length = remain;
			memset(hdr.base_of_data_size(), 0xff, remain);
			cCapture.byte_count += remain;
			return hdr.packet_size();
		} else {
			if (p_curPES) {
				PUINT8 base_ptr = p_curPES->base_of_data();

				memmove(base_ptr+remain, base_ptr, &buffer[ixBuf]-base_ptr);
				memset(base_ptr, 0xff, remain);
				ixBuf = ixBuf + UINT16(remain);
				p_curPES->data_length = p_curPES->data_length + UINT8(remain);
				p_curPES->set_data_size(&buffer[ixBuf]);
				cCapture.byte_count += p_curPES->packet_size();
				return remain;
			}
		}
	}

	return 0;
}

// ----------------------------------------------------------------------------
//	Implmentations de la classe "PS_buffer"
// ----------------------------------------------------------------------------

PS_buffer::PS_buffer(CCapture_PS & cCapt, UINT16 used_pid, UINT8 maxChnk, PS_buffer_type eBTyp) :
	cCapture(cCapt),
	pid(used_pid),
	valid(false),
	continuity_counter(0),
	eBType(eBTyp)
{
	reserve(maxChnk*sizeof(PS_chunk));
}

DWORD PS_buffer::traite_paquet(const TS_packet & tsp)
{
	try {
		const TS_unwrapper	uw_th(tsp.hdr);
		bool				new_PES_starts_here	= tsp.hdr.payload_unit_start_indicator();
		UINT8				new_ccounter		= tsp.hdr.continuity_counter();
		UINT8				diff_ccounter		= (new_ccounter - continuity_counter) & 0x0f;

		// Vrification de la continuit :
		if (diff_ccounter == 0) {
			// myprintf(TEXT("paquet dupliqu dans %") A2t EOL, pbt2str(eBType));
			return 0;	// ignorer
		}

		if (diff_ccounter != 1) {
			// myprintf(TEXT("discontinuit dans %") A2t EOL, pbt2str(eBType));

			// envoie le paquet (incomplet) qui tait en cours de construction
			flush();

			// Marque la suite comme invalide, sauf si un nouveau paquet commence justement ici :
			valid = false;
		}

		if (new_PES_starts_here) {
			if (cCapture.eState >= cs_stopping) {
				flush();
				valid = false;
			} else
				valid = true;
		}

		if (valid && uw_th.p_payload) {
			INT32 ptr_size = INT32(&tsp.bytes[TS_SIZE]-uw_th.p_payload);

			if (ptr_size < 0)	// si ngatif, alors le paquet doit tre corrompu ; en consquence
				return 0;			// on l'ignore, et le test de continuit sera faux au prochain passage.

			// Paquet valide et "Payload" prsent
			cCapture.byte_count += ptr_size;
			put_data(uw_th.p_payload, ptr_size, new_PES_starts_here);
		}

		// Mmoriser maintenant le compteur de continuit
		continuity_counter = new_ccounter;
	} catch (DWORD nErr) {
		return nErr;
	}

	return 0;
}

/// Traitement d'un paquet PS reconstitu
void PS_buffer::send(PCUINT8 ptr, UINT32 ptr_size) throw(...)
{
	const PS_hdr_wd &	hdr			= *reinterpret_cast<const PS_hdr_wd *>(ptr);
	bool				start		= true;
	UINT32				added_bytes = 0;

	// Ignorer si on n'a pas un en-tte PS conforme au dbut
	if (!hdr.isPS())
		return;

	// Redcoupage du paquet PS issu du flux TS en paquets PS plus petits
	while (ptr_size>0) {
		PS_chunk chunk(hdr, *this, eBType);

		added_bytes += chunk.build(start, ptr, ptr_size);
		chunk.write_to_file();
		start = false;
	}

	cCapture.byte_count += added_bytes;
}

CCapture_PS::CCapture_PS(HANDLE hFil, UINT16 video, UINT16 audio) :
	CCapture(hFil),
	first_pcr(0),
	curr_pcr(0),
	last_pcr(0),
	curr_pcr_second(UINT32(-1)),
	pck_pcr_count(0),
	pcr_rate(7500),	// estimation
	byte_count(0),
	byte_mux_rate((MIN_PROGRAM_MUX_RATE+MAX_PROGRAM_MUX_RATE)/2),
	eState(cs_idle),
	sys_header_set(false)
{
	// Cration des buffers de traitement
	pVideoBuf.reset(new Video_PS_buffer(*this, video));
	pAudioBuf.reset(new Audio_PS_buffer(*this, audio));
	if (fichier_valide()) {
		eState = cs_started;
	} else
		eState = cs_error;
}

DWORD CCapture_PS::traite_paquet(const TS_packet & tsp)	// virtual
{
	// une chane en PS
	UINT16 pid = tsp.hdr.pid();

	if (eState==cs_idle || (pid != pVideoBuf->pid && pid != pAudioBuf->pid))
		return 0;

	const TS_unwrapper	uw_th(tsp.hdr);	// Dballage de l'en-tte

	if (uw_th.p_afield) {
		// L'en-tte contient un champ "adaptation field" : dballage de ce champ :
		TS_AF_unwrapper_base	uw_af(*uw_th.p_afield);
		if (uw_af.p_pcr) {

			// Ce champ "adaptation field" contient un chantillon
			// de l'horloge de rfrence du programme
			UINT64 new_pcr		= *uw_af.p_pcr;
			UINT32 diff_time	= UINT32(new_pcr-last_pcr);

			if (eState>=cs_first_pcr_set) {
				if (diff_time>0 && pck_pcr_count>=32) {
					// Estimation du temps moyen qui spare deux paquets TS du mme canal :
					pcr_rate_helper.submit(diff_time/pck_pcr_count);
					pcr_rate = pcr_rate_helper.mean();
					// myprintf(TEXT("pcr rate=%u (packet count=%u)") EOL, pcr_rate, pck_pcr_count);
					pck_pcr_count = 0;
				}
			} else {
				first_pcr = new_pcr;
				eState = cs_first_pcr_set;
				byte_count = 0;
				curr_pcr_second = UINT32(-1);
			}
			if (new_pcr > last_pcr)
				last_pcr = new_pcr;
			//myprintf(TEXT("TS pcr = %10.6f (ajust = %10.6f)") EOL,
			//	new_pcr/27000000.0, (new_pcr-first_pcr)/27000000.0);
		}
	}

	if (eState < cs_first_pcr_set)
		return 0;

	UINT64 new_pcr = last_pcr+(pcr_rate*(pck_pcr_count++));

	if (new_pcr > curr_pcr)
		curr_pcr = new_pcr;

	UINT32	second			= UINT32(curr_pcr/27000000);

	// Evaluations statistiques :
	if (second != curr_pcr_second) {
		// myprintf(TEXT("second=%u") EOL, second);
		// Section appele  chaque nouvelle seconde du PCR
		curr_pcr_second = second;
		switch (eState) {

		case cs_first_pcr_set:
			if (pVideoBuf->valid && pAudioBuf->valid) {
				eState = cs_running;
				byte_count = 0;
			}
			break;

		case cs_running:
			if (byte_count > 0) {
				// valuation du "multiplex rate" :
				mux_rate_helper.submit(byte_count);

				UINT32 mux_rate = mux_rate_helper.mean();

				byte_mux_rate = max(MIN_PROGRAM_MUX_RATE, min(MAX_PROGRAM_MUX_RATE, mux_rate));
				// myprintf(TEXT("mux rate=0x%05x, avg=0x%05x") EOL, byte_count/50, byte_mux_rate/50);
				byte_count = 0;
			}
		}
	}

	if (pid == pVideoBuf->pid)
		return pVideoBuf->traite_paquet(tsp);
	else if (pid == pAudioBuf->pid)
		return pAudioBuf->traite_paquet(tsp);
	return 0;
}

CCapture_PS::~CCapture_PS()
{
	eState = cs_stopping;
	Sleep(200);	// Laisser les paquets en cours se terminer

	PS_hdr prog_end(sid_Program_End);

	ecrit_fichier(&prog_end, sizeof(prog_end));
}

// ----------------------------------------------------------------------------
//	Implmentations de la classe "CCapture_Stream"
// ----------------------------------------------------------------------------

CCapture_Stream::CCapture_Stream(HANDLE hFil) :
	CCapture(hFil)
{
}

DWORD CCapture_Stream::traite_paquet(const TS_packet & tsp)	// virtual
{
	// full TS
	if (tsp.hdr.pid() != p_pid_NULL)
		return ecrit_fichier(tsp.bytes, TS_SIZE);
	return 0;
}

// ----------------------------------------------------------------------------
