/*
 * $Id: console.cpp 277 2011-03-13 10:24:41Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion de la console de diagnostic
 *
 *	Ce fichier implmente l'affichage de la console de diagnostic qui est
 *	affiche (uniquement) dans les compilations en mode \p _DEBUG ou \p _DIAGNOSTIC.
 **/

#include "stdafx.h"
#include "console.h"
#include "strsafe.h"

#define ANSI_POSITIONABLE	0
#define USE_CONSOLE_HANDLER	0

static HANDLE OpenPlacementFile(bool bWrite)
{
	TCHAR	szPlFile[MAX_PATH];

	if (GetTempPath(_countof(szPlFile), szPlFile) == 0)
		return NULL;
	StringCchCat(szPlFile, _countof(szPlFile), _T("__{da1ea02c-2f69-49b2-bbf4-988cc9053f67}__.bin"));

	return CreateFile(szPlFile, bWrite ? GENERIC_WRITE : GENERIC_READ, FILE_SHARE_READ, NULL,
		bWrite ? CREATE_ALWAYS : OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
}

static void SavePlacement(HWND hWnd)
{
	if (!hWnd)
		return;

	WINDOWPLACEMENT sPlmnt = {sizeof(sPlmnt)};

	if (GetWindowPlacement(hWnd, &sPlmnt)) {
		HANDLE hFile = OpenPlacementFile(true);

		if (hFile != INVALID_HANDLE_VALUE) {
			DWORD		dwWritten;
			static CHAR	szInfo[] = "[Pouchin TV Mod diagnostic console placement file]";

			WriteFile(hFile, &sPlmnt, sizeof(sPlmnt), &dwWritten, NULL);
			WriteFile(hFile, szInfo, sizeof(szInfo), &dwWritten, NULL);

			CloseHandle(hFile);
		}
	}
}

static bool LoadPlacement(HWND hWnd)
{
	bool bResult = false;

	if (hWnd) {
		HANDLE hFile = OpenPlacementFile(false);

		if (hFile != INVALID_HANDLE_VALUE) {
			WINDOWPLACEMENT sPlmnt = {sizeof(sPlmnt)};
			DWORD			dwRead;

			if (ReadFile(hFile, &sPlmnt, sizeof(sPlmnt), &dwRead, NULL) > 0)
				bResult = SetWindowPlacement(hWnd, &sPlmnt) != 0;
			CloseHandle(hFile);
		}
	}
	return bResult;
}

class CConsoleBase
{
	HANDLE		hConsMutex;
public:

	CConsoleBase() :
		hConsMutex(CreateMutex(NULL, FALSE, NULL))
	{
	}

	/// Ouverture du fichier journal
	virtual void OpenLog() {}

	/// Fermeture du fichier journal
	virtual void CloseLog() {}

	/**
	 * Sortie d'un message
	 **/
	virtual void Output(LPTSTR pszMsg) = 0;

	/**
	 * Affichage d'un message d'erreur non format dans la console
	 *
	 * L'affichage du message est prcd d'un "time stamp" en dbut de ligne,
	 * indiquant l'heure exacte du message avec les secondes et les millisecondes.
	 * Le message est dcoup selon les retours chariot (CR) qu'il contient, et
	 * les lignes produites sont affiches une par une, la premire avec le
	 * "time stamp", et les suivantes avec des espaces pour aligner verticalement
	 * les lignes affiches.
	 * \param[in]	msg		Message  afficher
	 **/
	void OutputMsg(LPCTSTR msg);

	/**
	 * Affichage d'un message d'erreur format dans la console
	 *
	 * Le message format est ensuite affich avec la fontion \p OutputMsg ci-dessus.
	 * \param[in]	fmt		Chane de formatage
	 * \param[in]	argptr	Arguments style \p vprintf (voir aussi note)
	 * \retval		\p true si un message a t valu
	 * \retval		\p false si un code "%?" a provoqu la lecture d'une valeur fausse
	 * \note		Si la chane de formatage dbute par "%?", alors le premier paramtre
	 *				suivant est un boolen qui conditionne si l'affichage a lieu ou pas.
	 *				Le "%?" n'est bien sr pas affich.
	 **/
	bool OutputVF(LPCTSTR fmt, va_list & argptr);

	~CConsoleBase() {
		if (hConsMutex)
			CloseHandle(hConsMutex);
	}
};

/**
 * Affichage d'un message d'erreur non format dans la console
 *
 * L'affichage du message est prcd d'un "time stamp" en dbut de ligne,
 * indiquant l'heure exacte du message avec les secondes et les millisecondes.
 * Le message est dcoup selon les retours chariot (CR) qu'il contient, et
 * les lignes produites sont affiches une par une, la premire avec le
 * "time stamp", et les suivantes avec des espaces pour aligner verticalement
 * les lignes affiches.
 * \param[in]	msg		Message  afficher
 **/
void CConsoleBase::OutputMsg(LPCTSTR msg)
{
	if (hConsMutex) {
		// Utilisation d'un mutex parce que les messages de ce type, quand on se met  les
		// utiliser, peuvent tre si frquents qu'ils ont beaucoup de chance d'entrer en
		// conflit avec des messages d'un autre thread.

		// 1/2 seconde. Si a dpasse, on y va quand mme, on ne va
		// tout de mme pas bloquer l'application pour un message de debugging !
		WaitForSingleObject(hConsMutex, 500);
	}

	// Ouverture du fichier journal
	OpenLog();

	static bool		bBegOfLine	= true;	//!< Si \p true, on est en dbut de ligne
	bool			bTimeSet	= false;
	static TCHAR	szCLS[]		= CLS;
	LPCTSTR			pszStr		= _tcsstr(msg, szCLS);
	LPCTSTR			pszNl;

	if (pszStr) {
		Output(szCLS);	// Effacement d'cran
		pszStr += _countof(szCLS)-1;
	} else
		pszStr = msg;

	for (; pszStr && *pszStr; pszStr = pszNl) {
		// Reprage des retours chariot, en vue de reformater la ligne s'il y en a en milieu de ligne
		TCHAR		szOutBuf[256];

		pszNl = _tcschr(pszStr, TCHAR('\n'));

		if (!bBegOfLine) {// Continuation d'un message prcdent
			strcpy_T(szOutBuf, pszStr);
		} else {
			if (!bTimeSet) {			// Premier retour chariot
				SYSTEMTIME	time;

				GetLocalTime(&time);
				_stprintf_s(szOutBuf, COLR(LCYAN, TEXT("%02u:%02u:%02u.%03u>")) TEXT(" "),
					time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);
				bTimeSet = true;
			} else
				_tcscpy_s(szOutBuf, TEXT("              "));
				Output(szOutBuf);

			bBegOfLine = false;
		}

		if (pszNl)
			bBegOfLine = true;	// Mmoriser qu'on vient de passer  la ligne

		// Longueur de la chane jusqu'au prochain retour chariot inclus
		int		nLSize = min<int>(
							_countof(szOutBuf)-1,
							pszNl ? (int)((++pszNl) - pszStr) : (int)_tcslen(pszStr));

		_tcsncpy_s(szOutBuf, pszStr, nLSize);
		szOutBuf[nLSize] = 0;	// Terminateur

		// Sortie
		Output(szOutBuf);
	}
	// Fermeture du fichier journal
	CloseLog();

	if (hConsMutex)
		ReleaseMutex(hConsMutex);
}

/**
 * Affichage d'un message d'erreur format dans la console
 *
 * Le message format est ensuite affich avec la fontion \p OutputMsg ci-dessus.
 * \param[in]	fmt		Chane de formatage
 * \param[in]	argptr	Arguments style \p vprintf (voir aussi note)
 * \retval		\p true si un message a t valu
 * \retval		\p false si un code "%?" a provoqu la lecture d'une valeur fausse
 * \note		Si la chane de formatage dbute par "%?", alors le premier paramtre
 *				suivant est un boolen qui conditionne si l'affichage a lieu ou pas.
 *				Le "%?" n'est bien sr pas affich.
 **/
bool CConsoleBase::OutputVF(LPCTSTR fmt, va_list & argptr)
{
	// Affichage conditionnel : si la chane de formatage dbute par "%?",
	// alors le premier paramtre suivant est un boolen qui conditionne
	// si l'affichage a lieu ou pas. Le "%?" n'est bien sr pas affich.
	if (fmt[0]==TCHAR('%') && fmt[1]==TCHAR('?')) {
		if (!va_arg(argptr, bool))
			return false;
		fmt += 2;
	}

	TCHAR		szFmtBuf[512];

	_vstprintf_s(szFmtBuf, fmt, argptr);

	OutputMsg(szFmtBuf);
	return true;
}

class CDebugOutput : public CConsoleBase
{
public:
	/// Sortie d'un message
	virtual void Output(LPTSTR pszMsg) {
					OutputDebugString(pszMsg); }

};

class AnsiConsole : public CConsoleBase
{
	HANDLE					hStdOut;
	static HWND				hWndConsole;
	static HWINEVENTHOOK	hEventHook;
	TempPathName			sLogFileName;
	FILE *					f;
	errno_t					nFileRes;
	BYTE					anAttrs[8];

#if USE_CONSOLE_HANDLER
	struct EnumWindowsInfo
	{
		DWORD	dwCurProcessId;
		HWND	hWndConsole;

		EnumWindowsInfo() :
			dwCurProcessId(GetCurrentProcessId()),
			hWndConsole(GetConsoleWindow())
		{
		}
	};

	/**	\brief Fonction "callback" pour \p EnumWindowsProc
	 *
	 *	Envoie un message \p WM_CLOSE  toutes les fentres qui ont t cres par le processus
	 *	courant,  l'exception de celle de la console de diagnostic.
	 **/
	static BOOL CALLBACK EnumWindowsProc(HWND hWnd, LPARAM lParam);

	/**
	 * Interception d'un clic sur la case de fermeture de la console, et transformation en
	 * fermeture de la fentre principale (ceci permet de s'assurer que les mises  jour faites
	 *  la fermeture de l'application sont effectues proprement).
	 **/
	static BOOL WINAPI HandlerRoutine(DWORD dwCtrlType);
#endif

	/// Fonction "callback" recevent les vnements de dplacement ou redimensionnement de la console
	static void CALLBACK ConsoleWinEventProc(HWINEVENTHOOK hHook, DWORD nEvent, HWND hWnd, LONG, LONG, DWORD, DWORD);

	/// Sortie brute d'un message vers la console
	void RawOutput(LPCTSTR pszMsg);

#if ANSI_POSITIONABLE
	**
	 * Affichage d'un message positionn aprs sauvegarde de l'tat de la console,
	 * puis restauration de l'tat antrieur.
	 * \param[in]		X		Position horizontale du message. Non modifi si < 0.
	 * \param[in]		Y		Position verticale du message. Non modifi si < 0.
	 * \param[in,out]	pszMsg	Message  afficher.
	 * \note Des caractres  l'intrieur du message peuvent tre modifis par la fonction.
	 **/
	void PosAnsiOutput(int X, int Y, LPTSTR pszMsg);
#endif

	/// Effacement complet
	void cls();

	/**
	 * Sortie d'un message format ANSI vers la console
	 * Les codes ANSI sont, au passage, limins de la chane de caractres
	 * Pour les codes ANSI, voir http://en.wikipedia.org/wiki/ANSI_escape_code
	 * Seules les couleurs et l'intensit sont couramment prises en charge
	 * \param[in,out]	pszMsg	Message  afficher.
	 * \note Des caractres  l'intrieur du message peuvent tre modifis par la fonction.
	 **/
	void AnsiOutput(LPTSTR pszMsg);

	/// Ouverture du fichier journal
	virtual void OpenLog();

	/// Fermeture du fichier journal
	virtual void CloseLog();

	/// Sortie d'un message
	virtual void Output(LPTSTR pszMsg) {
					AnsiOutput(pszMsg); }

public:
	AnsiConsole(LPCTSTR pszLogFileName) :
		hStdOut(NULL),
		sLogFileName(pszLogFileName),
		f(NULL),
		nFileRes(0)
	{
		memset(anAttrs, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE, sizeof(anAttrs));
	}

	void Init(LPCTSTR pszTitle, bool alloc);
};

#if USE_CONSOLE_HANDLER
/// Fonction appele pour chacune des fentres numres
BOOL CALLBACK AnsiConsole::EnumWindowsProc(HWND hWnd, LPARAM lParam)	// static
{
	EnumWindowsInfo	sEWInf			= *reinterpret_cast<EnumWindowsInfo *>(lParam);
	DWORD			dwWinProcessId	= NULL;

	GetWindowThreadProcessId(hWnd, &dwWinProcessId);

	if (dwWinProcessId == sEWInf.dwCurProcessId && hWnd != sEWInf.hWndConsole) {
		LONG_PTR	nStyle = GetWindowLongPtr(hWnd, GWL_STYLE);

		// Ignorer les fentres "child" et "pop up"
		if ((nStyle & (WS_CHILD|WS_POPUP)) == 0)
			PostMessage(hWnd, WM_CLOSE, 0, NULL);
	}
	return TRUE;
}
#endif

#if USE_CONSOLE_HANDLER
/**
 * Interception d'un clic sur la case de fermeture de la console, et transformation en
 * fermeture de la fentre principale (ceci permet de s'assurer que les mises  jour faites
 *  la fermeture de l'application sont effectues proprement).
 **/
BOOL WINAPI AnsiConsole::HandlerRoutine(DWORD dwCtrlType)	// static
{
	EnumWindowsInfo	sEWInf;

	EnumWindows(EnumWindowsProc, reinterpret_cast<LPARAM>(&sEWInf));
	return TRUE;	//	TRUE : la fermeture est ignore (elle sera faite, normalement,
					//			par le message WM_CLOSE envoy  la fentre principale)
}
#endif

/// Sortie d'un message vers la console
void AnsiConsole::RawOutput(LPCTSTR pszMsg)
{
	UINT	nLen = strlen_T(pszMsg);

	if (nLen > 0) {
#ifdef _UNICODE
		LPCTSTR	pszOut = pszMsg;
#else
		LPTSTR	pszOut = STK_ALLOC(TCHAR, nLen+1);

		CharToOem(pszMsg, pszOut);
#endif
		DWORD	cCharsWritten;

		WriteConsole(hStdOut, pszOut, nLen, &cCharsWritten, NULL);

		// criture dans le fichier journal
		if (f && nFileRes==0)
			_fputts(pszMsg, f);
	}
}

#if ANSI_POSITIONABLE
**
 * Affichage d'un message positionn aprs sauvegarde de l'tat de la console,
 * puis restauration de l'tat antrieur.
 * \param[in]		X		Position horizontale du message. Non modifi si < 0.
 * \param[in]		Y		Dcalage vertical (ngatif) du message. Non modifi si < 0.
 * \param[in,out]	pszMsg	Message  afficher.
 * \note Des caractres  l'intrieur du message peuvent tre modifis par la fonction.
 **/
void AnsiConsole::PosAnsiOutput(int X, int Y, LPTSTR pszMsg)
{
	CONSOLE_SCREEN_BUFFER_INFO	sCsbi;

	if (GetConsoleScreenBufferInfo(hStdOut, &sCsbi)) {
		COORD	dwCursorPosition	= sCsbi.dwCursorPosition;

		if (X >= 0)
			dwCursorPosition.X = SHORT(X);
		if (Y >= 0)
			dwCursorPosition.Y = dwCursorPosition.Y - min(dwCursorPosition.Y, SHORT(Y));
		SetConsoleCursorPosition(hStdOut, dwCursorPosition);
		AnsiOutput(pszMsg);
		SetConsoleCursorPosition(hStdOut, sCsbi.dwCursorPosition);
		SetConsoleTextAttribute(hStdOut, sCsbi.wAttributes);
	}
}
#endif

/// Effacement complet
void AnsiConsole::cls()
{
	CONSOLE_SCREEN_BUFFER_INFO  csbi;

	// Obtenir le nombre de cellules de caractres dans le tampon courant
	if (!GetConsoleScreenBufferInfo(hStdOut, &csbi))
		return;

	DWORD						dwConSize	= csbi.dwSize.X * csbi.dwSize.Y;
	DWORD						nWritten;
	COORD						sCoord		= {0, 0};	// position d'origine du curseur

	// Remplir la totalit de l'cran avec des blancs
	if (!FillConsoleOutputCharacter(hStdOut, (TCHAR) ' ', dwConSize, sCoord, &nWritten ))
		return;

	// Rinitialiser les attributs de texte
	memset(anAttrs, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE, sizeof(anAttrs));

	// Remplir le tampon des attributs ainsi
	if (!FillConsoleOutputAttribute(hStdOut, anAttrs[0], dwConSize, sCoord, &nWritten))
		return;

	// Mettre le curseur  la position d'origine
	SetConsoleCursorPosition(hStdOut, sCoord);
}

/**
 * Sortie d'un message format ANSI vers la console
 * Pour les codes ANSI, voir http://en.wikipedia.org/wiki/ANSI_escape_code
 * Seules les couleurs et l'intensit sont couramment prises en charge
 * \param[in,out]	pszMsg	Message  afficher.
 * \note Des caractres  l'intrieur du message peuvent tre modifis par la fonction.
 **/
void AnsiConsole::AnsiOutput(LPTSTR pszMsg)
{
	struct AnsiDescr
	{
		UINT16	nCode;
		BYTE	nMask;
		BYTE	nVal;
	};

	#define FOREGROUND_MASK (FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE)
	#define BACKGROUND_MASK (BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE)

	static const AnsiDescr aAnsiTbl[] = {
		{ 0,	0xff,					FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE},
		{ 1,	FOREGROUND_INTENSITY,	FOREGROUND_INTENSITY},
		{30,	FOREGROUND_MASK,		0},
		{31,	FOREGROUND_MASK,		FOREGROUND_RED},
		{32,	FOREGROUND_MASK,		FOREGROUND_GREEN},
		{33,	FOREGROUND_MASK,		FOREGROUND_RED|FOREGROUND_GREEN},
		{34,	FOREGROUND_MASK,		FOREGROUND_BLUE},
		{35,	FOREGROUND_MASK,		FOREGROUND_RED|FOREGROUND_BLUE},
		{36,	FOREGROUND_MASK,		FOREGROUND_GREEN|FOREGROUND_BLUE},
		{37,	FOREGROUND_MASK,		FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE},
		{40,	BACKGROUND_MASK,		0},
		{41,	BACKGROUND_MASK,		BACKGROUND_RED},
		{42,	BACKGROUND_MASK,		BACKGROUND_GREEN},
		{43,	BACKGROUND_MASK,		BACKGROUND_RED|BACKGROUND_GREEN},
		{44,	BACKGROUND_MASK,		BACKGROUND_BLUE},
		{45,	BACKGROUND_MASK,		BACKGROUND_RED|BACKGROUND_BLUE},
		{46,	BACKGROUND_MASK,		BACKGROUND_GREEN|BACKGROUND_BLUE},
		{47,	BACKGROUND_MASK,		BACKGROUND_RED|BACKGROUND_GREEN|BACKGROUND_BLUE}
	};

	LPTSTR		pszPtr;

	// Recherche d'une chane "ESC [", reprsentant un dbut de squence ANSI
	while ((pszPtr = _tcsstr(pszMsg, CSI))!=NULL) {
		TCHAR	szAnsiBuf[64];
		int		nSize;
		TCHAR	nCmd;

		if (_stscanf_s(pszPtr, TEXT("\x1b%63[^@-Za-z]%c%n"), &szAnsiBuf, _countof(szAnsiBuf), &nCmd, sizeof(nCmd), &nSize)==2) {
			// Sortie de la partie du message qui prcde la squence ANSI
			if (pszPtr > pszMsg) {
				*pszPtr = 0;
				RawOutput(pszMsg);
				*pszPtr = TCHAR('\x1b');
			}

			pszMsg = pszPtr+nSize;

			LPTSTR		pszSeq = szAnsiBuf+1;	// Initialisation analyse des paramtres

			switch (nCmd) {

#if ANSI_POSITIONABLE
			case '@': {	// CSI x [;y] @ 	Spcifique : affichage d'un message positionn
				int nY = -1;
				int nX = _tcstoul(pszSeq, &pszSeq, 10);

				if (*pszSeq == TCHAR(';'))
					nY = _tcstoul(pszSeq+1, &pszSeq, 10);
				pszPtr = _tcschr(pszMsg, TCHAR('\r'));
				if (pszPtr)
					*pszPtr = 0;
				PosAnsiOutput(nX, nY, pszMsg);
				if (pszPtr)
					pszMsg = pszPtr+1;
				else
					pszMsg += strlen_T(pszMsg);
				break; }
#endif

			case 'm':	// CSI n [;k] 'm' - SGR - Select Graphic Rendition
				while (*pszSeq) {
					if (pszSeq[0] == TCHAR('<')) {
						// Non standard : Empiler l'attribut courant
						memmove(anAttrs+1, anAttrs, sizeof(anAttrs)-1);
						++pszSeq;
					} else if (pszSeq[0] == TCHAR('>')) {
						// Non standard : dpiler l'attribut courant
						memmove(anAttrs, anAttrs+1, sizeof(anAttrs)-1);
						++pszSeq;
					} else {
						UINT	nAnsiCode = _tcstoul(pszSeq, &pszSeq, 10);

						for (int nInx = 0; nInx<_countof(aAnsiTbl); ++nInx) {
							AnsiDescr	sAnsiDescr = aAnsiTbl[nInx];

							if (sAnsiDescr.nCode == nAnsiCode) {
									anAttrs[0] &= ~sAnsiDescr.nMask;
									anAttrs[0] |= sAnsiDescr.nVal;
								break;
							}
						}
					}

					if (*pszSeq != TCHAR(';'))
						break;
					++pszSeq;
				}
				SetConsoleTextAttribute(hStdOut, anAttrs[0]);
				break;

			case 'J':	// CSI '2' 'J' - Erase Display
				if (pszSeq[0] == TCHAR('2') && pszSeq[1] == 0)
					cls();
			}
		} else
			pszMsg = pszPtr+1;
	}

	RawOutput(pszMsg);
}

/// Ouverture du fichier journal
void AnsiConsole::OpenLog()
{
	nFileRes = _tfopen_s(&f, sLogFileName, TEXT("a"));
}

/// Fermeture du fichier journal
void AnsiConsole::CloseLog()
{
	if (nFileRes==0 && f)
		fclose(f);
}

HWND AnsiConsole::hWndConsole = NULL;
HWINEVENTHOOK AnsiConsole::hEventHook = NULL;

/// Fonction "callback" recevent les vnements de dplacement ou redimensionnement de la console
VOID CALLBACK AnsiConsole::ConsoleWinEventProc(HWINEVENTHOOK hHook, DWORD nEvent, HWND hWnd, LONG, LONG, DWORD, DWORD)
{
	if (hHook == hEventHook && hWnd == hWndConsole && nEvent == EVENT_SYSTEM_MOVESIZEEND)
		SavePlacement(hWndConsole);
}

void AnsiConsole::Init(LPCTSTR pszTitle, bool alloc)
{
	if (alloc) {
		AllocConsole();
		SetConsoleTitle(pszTitle);
	}
	hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
	hWndConsole = GetConsoleWindow();
#if USE_CONSOLE_HANDLER
	SetConsoleCtrlHandler(HandlerRoutine, TRUE);
#else
	HMENU	hMenuConsole	= GetSystemMenu(hWndConsole, FALSE);

	DeleteMenu(hMenuConsole, SC_CLOSE, MF_BYCOMMAND);  // Dsactivation de la case de fermeture
#endif


	LoadPlacement(hWndConsole);

	// Dfinition d'un "hook" dclench si la console est dplace ou redimensionne, en vue de sauvegarder sa position
	hEventHook = SetWinEventHook(EVENT_SYSTEM_MOVESIZEEND, EVENT_SYSTEM_MOVESIZEEND,
					NULL, ConsoleWinEventProc, 0, 0, WINEVENT_OUTOFCONTEXT);
}

/// Pointeur sur la console ouverte
auto_ptr<CConsoleBase>	pConsole;

/**
 * Initialisation de la console de debugging.
 * \param[in]	pszTitle	Titre de la console
 * \param[in]	eMode		Mode d'ouverture
 * \param[in]	bTimeStamps	Mettre \p true si un horodatage est souhait
 * \param[in]	pszLogFile	Nom de fichier journal optionnel (mettre NULL si non dsir)
 **/
void CONSOLE_DECLSPEC startConsole(LPCTSTR pszTitle, StartMode eMode, bool bTimeStamps, LPCTSTR pszLogFile)
{	// Note : bTimeStamps  implmenter (toujours actif pour le moment)
	AnsiConsole * pNewConsole;

	switch (eMode) {

	case sm_Debug:
		pConsole.reset(new CDebugOutput());
		break;

	case sm_Attach:
	case sm_Console:
	case sm_ConsoleFb:
		pNewConsole = new AnsiConsole(pszLogFile);

		if (pNewConsole) {
			pNewConsole->Init(pszTitle, eMode != sm_Attach);
			pConsole.reset(pNewConsole);
			//if (pConsolePos && pConsolePos->verifie())
			//	pConsolePos->RestoreWindowPos(hWndConsole, SW_SHOW);
		}
	}
}

/**
 * Affichage d'un message d'erreur non format dans la console
 *
 * L'affichage du message est prcd d'un "time stamp" en dbut de ligne,
 * indiquant l'heure exacte du message avec les secondes et les millisecondes.
 * Le message est dcoup selon les retours chariot (CR) qu'il contient, et
 * les lignes produites sont affiches une par une, la premire avec le
 * "time stamp", et les suivantes avec des espaces pour aligner verticalement
 * les lignes affiches.
 * \param[in]	msg		Message  afficher
 **/
void CONSOLE_DECLSPEC OutputMsg(LPCTSTR msg)
{
	if (pConsole.get())
		pConsole->OutputMsg(msg);
}


/**
 * Affichage d'un message d'erreur format dans la console
 *
 * \param[in]	fmt		Chane de formatage
 * \param[in]	argptr	Arguments style \p vprintf (voir aussi note)
 * \retval		\p true si un message a t valu
 * \retval		\p false si un code "%?" a provoqu la lecture d'une valeur fausse
 * \note		Si la chane de formatage dbute par "%?", alors le premier paramtre
 *				suivant est un boolen qui conditionne si l'affichage a lieu ou pas.
 *				Le "%?" n'est bien sr pas affich.
 **/
bool CONSOLE_DECLSPEC OutputVF(LPCTSTR fmt, va_list & argptr)
{
	if (pConsole.get())
		return pConsole->OutputVF(fmt, argptr);
	return false;
}
