/*
 * $Id: internet.h 260 2010-08-04 20:42:33Z gingko $
 * Copyright (C) 2007 lolo_32 <pouchintv-dev@baysse.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions pour les connexions vers Internet
 *
 *	Ce fichier dfinit les fonctionnalits ncessaires  la communication
 *	via Internet.
 *
 *	Il est utilis en particulier pour la rcupration des mises  jour de
 *	l'application par Internet.
 **/

#pragma once

#include <Wininet.h>

// Liste des codes de status chez Microsoft:
//    http://msdn2.microsoft.com/en-us/library/aa384325.aspx

class ConnexionInternet {
public:

	/// Constantes pour les rsultats des requetes
	static const enum HTTP_ERR {
		HTTP_ERR_OK           = 200,	///< Tout s'est bien pass
		HTTP_ERR_OK_PARTIAL   = 206,	///< Fichier partiel
		HTTP_ERR_DROIT        = 403,	///< Problme de droits d'accs
		HTTP_ERR_INEXISTANT   = 404,	///< Fichier inexistant
		HTTP_ERR_INTERNE      = 500,	///< Erreur interne du serveur
		HTTP_ERR_TIMEOUT      = 900,	///< La connexion s'est acheve, car elle a pris trop de temps
		HTTP_ERR_CNINVALID,				///< Le CN du certificat ne correspont pas au nom du serveur
										///< (ex. www.ptvm.fr != www.ptv.fr)
	};
	/// Constantes pour la mthode  utiliser pour la requte
	static const enum HTTP_METHODE {
		HTTP_METHODE_GET,				///< utilise la mthode GET
		HTTP_METHODE_POST				///< utilise la mthode POST
	};

	/**
	 * Initialise les valeurs par dfaut, l'User-Agent, la mthode  GET et le type de fichiers accept
	 *
	 * \param[in]	pszUagt	"user agent"  indiquer lors de la connexion
	 * \param[in]	pszUrl	url  interroger
	**/
	ConnexionInternet(LPCTSTR pszUagt, LPCTSTR pszUrl = TEXT(""));

	~ConnexionInternet();

	/**
	 * Dfinit la mthode  utiliser pour la connexion
	 * \param[in]  methode      Mthode  utiliser pour la connexion.
	 *                          Certaines permettent d'ajouter des paramtres
	 **/
	void SetMethode(HTTP_METHODE methode) {this->m_methode = methode;};

	/** Indique l'url  utiliser pour passer la requte
	 * \param[in]  serveur      Serveur revevant la connexion (ex. www.pouchintv.fr)
	 * \param[in]  url          Url demande (ex. /forum/)
	 * \retval     0            en cas d'erreur
	 **/
	int SetUrl(LPCTSTR serveur, LPCTSTR url);

	/** Indique l'url  utiliser pour passer la requte
	 * \param[in]  url          Url complte demande (ex. http://www.pouchintv.fr/forum/)
	 * \retval     0            en cas d'erreur
	 **/
	int SetUrl(LPCTSTR url);

	/** Remplace l'User-Agent utilis pour la requte
	 * \param[in]  useragent    User-Agent devant tre utilis
	 **/
	void SetUserAgent(LPCTSTR useragent);

	/**
	 * Demande  tablir une connexion scurise (en HTTPS)
	 * \param[in]  chiffre  Si true, la connexion sera tablie en HTTPS
	 **/
	void SetChiffree(bool chiffre);

	/** Envoie la connexion
	 * \param[in]  donnees      chaine de caractres  envoyer avec la requte
	 * \warning                 il faut obligatoirement passer la chaine en CHAR, 
	 *                          car si passe en WCHAR, les 0 sont aussi transmis
	 * \param[in]  donnees_len  longueur de donnees
	 * \param[in]  debut        indique la position de lecture facultative (position pour la rcupration d'un morceau)
	 * \retval     0            en cas d'erreur
	 * \return                  une valeur de HTTP_ERR indiquant le status de la requte
	 * \note                    si donnees != NULL, la requte sera envoy en POST
	 **/
	int EnvoieRequete(LPCSTR donnees = 0, DWORD donnees_len = 0, DWORD debut = 0);

	/** Envoie la connexion
	 * \param[in]  debut        indique la position de lecture facultative (position pour la rcupration d'un morceau)
	 * \retval     0            en cas d'erreur
	 * \return                  une valeur de HTTP_ERR indiquant le status de la requte
	 **/
	int EnvoieRequete(DWORD debut) {this->EnvoieRequete(0, 0, debut);};

	/** Retourne le status de la connexion
	 * \return le code status de la connexion
	 **/
	int GetStatus() {return this->m_dwStatus;};

	/** Retourne la taille du fichier
	 * \return la taille du fichier
	 **/
	DWORD GetTaille() {return this->m_dwTaille;};

	/** Retourne le type de contenue
	 * \param[out]    content  rcupre le type de fichiers (ex. "text/html", "application/x-msdos-program")
	 * \param[in,out] longueur la taille de content maximale, et retourne le nombre d'octets copis/necessaires
	 **/
	void GetContentType(LPTSTR content, UINT *longueur);

	/** Retourne le nombre d'octets dj rcuprs
	 * \return le nombre d'octets rcuprs
	 **/
	DWORD GetRestant() {return this->m_dwTaille - this->m_dwRecupere;};

	/** Retourne la date/heure de modification du fichier
	 * \return la date de modification du fichier
	 **/
	SYSTEMTIME GetModifTime() {return this->m_date;};

	/** Lit le contenu de la rponse  la requte
	 * \param[out]    contenu   tableau d'octets du fichiers lu
	 * \param[in,out] longueur  longueur de la variable contenue
	 * \return                  le nombre d'octets lus
	 * \warning                 aucun caractre null ne sera ajout  la fin de contenu
	 **/
	DWORD RecupereFichier(LPBYTE contenu, DWORD *longueur);

private:
	/// Handle de la session
	HINTERNET	m_hSession;
	/// Handle de la connexion
	HINTERNET	m_hConnexion;
	/// Handle de la requte
	HINTERNET	m_hRequete;

	/// Connexion chiffre ou pas
	bool		m_bChiffre;
	/// Mthode de connexion (GET/POST ou autre)
	HTTP_METHODE m_methode;
	/// User Agent utilis
	LPTSTR		m_pszUserAgent;
	// Site de connexion
	LPTSTR		m_pszServeur;
	/// URL demande
	LPTSTR		m_pszUrl;

	/// Status de la requete
	DWORD		m_dwStatus;
	/// Taille du fichier  tlcharger
	DWORD		m_dwTaille;
	/// Date et heure de modification du fichier
	SYSTEMTIME	m_date;
	/// Content-Type retourn par le serveur
	LPTSTR		m_pszContentType;
	/// Octets rcuprs
	DWORD		m_dwRecupere;

	/// Initialise la connexion
	int InitConnexion();

	/**
	 * Copie les informations internes
	 * \param[in]  destination  variable destinataire. Si celle-ci est diffrente de NULL, elle sera efface
	 * \param[in]  source       chane de caractres  mmoriser
	 **/
	void SetInfo(LPTSTR *destination, LPCTSTR source);

};