/*
 * $Id: memfifo.cpp 257 2010-08-03 07:35:33Z gingko $
 * Copyright (C) 2009 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion d'une pile "Fifo" rapide (implmentation)
 *
 *	Ce fichier implmente les mthodes ncessaires  la cration d'une pile
 *	"Fifo" (First In - First Out) utilisant un minimum de ressources, qui
 *	est utilise comme tampon de transit pour les flux destins aux
 *	enregistrements programms.
 **/

#include "stdafx.h"
#include "memfifo.h"

/// Constructeur
CMemFifoBlock::CMemFifoBlock(size_t nSize, const BYTE * pPtr) :
	pNext(NULL),
	nDataSize(nSize)
{
	if (nSize) {
		if (pPtr)
			CopyMemory(this+1, pPtr, nSize);
		else
			ZeroMemory(this+1, nSize);
	}
}

/**
 * Ajout d'un chantillon en fin de pile FiFo
 * \param[in]	nSize	Taille de l'chantillon
 * \param[in]	pPtr	Pointeur sur les donnes de l'chantillon
 **/
void CMemFifo::Put(size_t nSize, const BYTE * pPtr)
{
	CMemFifoBlock * pNewBlock = CMemFifoBlock::new_block(nSize, pPtr);

	{
		CAutoLock cLock(&cCritSec);

		if (pFirst == NULL)
			pFirst = pNewBlock;
		if (pLast)
			pLast->pNext = pNewBlock;
		pLast = pNewBlock;
	}

	SetEvent(hEvent);
}

/**
 * Extraire un chantillon de la pile FiFo.
 * Retourne NULL si aucun chantillon n'est disponible.
 * \note L'chantillon retourn DOIT tre dtruit par \p delete (sans [])
 * aprs usage.
 **/
const CMemFifoBlock * CMemFifo::Get()
{
	CAutoLock		cLock(&cCritSec);
	CMemFifoBlock *	pRetBlock = pFirst;

	if (pFirst)
		pFirst = pFirst->pNext;
	if (pFirst==NULL)
		pLast = NULL;

	return pRetBlock;
}

/**
 * Extraire un chantillon de la pile FiFo.
 * Si aucun chantillon n'est disponible, reste en attente jusqu' ce qu'un
 * chantillon soit ajout (par un autre thread).
 * Retourne NULL si en cas d'erreur.
 * \note L'chantillon retourn DOIT tre dtruit par \p delete (sans [])
 * aprs usage.
 **/
const CMemFifoBlock * CMemFifo::WaitGet()
{
	while (pFirst==NULL) {
		if (WaitForSingleObject(hEvent, INFINITE)!=WAIT_OBJECT_0)
			return NULL;
	}

	return Get();
}

/// Destruction de tous les blocs allous
void CMemFifo::Free()
{
	const CMemFifoBlock * pRetBlock;

	while ((pRetBlock = Get())!=NULL)
		delete pRetBlock;
}

CMemFifo::~CMemFifo()
{
	// Dtruire tous les blocs allous :
	const CMemFifoBlock * pRetBlock;

	while ((pRetBlock = Get())!=NULL)
		delete pRetBlock;
}
