/*
 * $Id: memfifo.h 257 2010-08-03 07:35:33Z gingko $
 * Copyright (C) 2009 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion d'une pile "Fifo" rapide (dfinitions)
 *
 *	Ce fichier dfinit les structures ncessaires  la cration d'une pile
 *	"Fifo" (First In - First Out) utilisant un minimum de ressources, qui
 *	est utilise comme tampon de transit pour les flux destins aux
 *	enregistrements programms.
 **/

#pragma once

#include "wxdebug.h"	// (BaseClasses)
#include "wxutil.h"		// (BaseClasses)

/**
 * Classe d'allocation d'un chantillon en mmoire.
 * Afin de limiter le nombre de requtes d'allocation mmoire, les donnes
 * stockes sont situes dans l'espace mmoire se trouvant immdiatement aprs
 * les donnes membres de la classe. Pour cette raison, les instances de cette
 * classe ne doivent tre crs qu' l'aide de la fonction membre statique
 * \p new_block.
 **/
class CMemFifoBlock
{
	friend class CMemFifo;

	CMemFifoBlock * pNext;
	size_t			nDataSize;

	/**
	 * Oprateur \p new surcharg, utilisant la syntaxe "placement new".
	 * \param[in]	nSize		Taille de la structure (implicite au langage C++)
	 * \param[in]	nExtraSize	Taille ajoute aprs la structure.
	 **/
	void *			operator new(size_t nSize, size_t nExtraSize) {
						return new BYTE[nSize + nExtraSize]; }
public:
	/// Constructeur
	CMemFifoBlock(size_t nSize, const BYTE * pPtr=NULL);

	/// Oprateur \p delete surcharg, devant tre utilis pour dtruire
	/// les chantillons aprs usage.
	void			operator delete (void * pPtr) {
						delete [] reinterpret_cast<BYTE *>(pPtr); }
	/// Oprateur \p delete surcharg n2 (version probablement jamais utilise,
	/// mais requise par le langage).
	void			operator delete (void * pPtr, size_t) {
						delete [] reinterpret_cast<BYTE *>(pPtr); }

	/** Allocation d'un bloc pour un chantillon
	 * \param[in]	nSize	Taille de l'chantillon (devient \p nExtraSize
	 *						pour l'oprateur \p new surcharg)
	 * \param[in]	pPtr	Pointeur sur les donnes de l'chantillon,  copier
	 *						dans l'espace allou. Si NULL, l'espace allou est
	 *						initialis avec des zros.
	 * \return Pointeur sur l'object allou (structure + donnes)
	 **/
	static CMemFifoBlock *
					new_block(size_t nSize, const BYTE * pPtr=NULL) {
						return new (nSize) CMemFifoBlock(nSize, pPtr);}

	/// Obtenir la taille des donnes (seules)
	size_t			size() const {
						return nDataSize;}
	/// Obtenir un pointeur sur les donnes (\p const)
	const BYTE *	ptr() const {
						return reinterpret_cast<const BYTE *>(this+1); }
	/// Obtenir un pointeur sur les donnes
	BYTE *			ptr() {
						return reinterpret_cast<BYTE *>(this+1); }
};

/// Classe de gestion d'une pile "FiFo" destine aux chantillons reus du tuner.
/// Le chargement et le dchargement de la pile doivent normalement se faire dans
/// des "threads" diffrents.
class CMemFifo
{
	CMemFifoBlock *		pFirst;		//!< Pointeur sur le premier chantillon
	CMemFifoBlock *		pLast;		//!< Pointeur sur le dernier chantillon
	HANDLE				hEvent;		//!< Handle d'objet "Event" pour signaler au thread de
									//!< traitement que des donnes ont t ajoutes
	CCritSec			cCritSec;	//!< Objet "critical section" pour protger les ajouts
									//!< et les extractions
public:
	/// Constructeur
	CMemFifo() :
		pFirst(NULL),
		pLast(NULL),
		hEvent(CreateEvent(NULL, FALSE, FALSE, NULL))
	{
	}

	/**
	 * Ajout d'un chantillon en fin de pile FiFo
	 * \param[in]	nSize	Taille de l'chantillon
	 * \param[in]	pPtr	Pointeur sur les donnes de l'chantillon
	 **/
	void					Put(size_t nSize, const BYTE * pPtr);

	/**
	 * Extraire un chantillon de la pile FiFo.
	 * Retourne NULL si aucun chantillon n'est disponible.
	 * \note L'chantillon retourn DOIT tre dtruit par \p delete (sans [])
	 * aprs usage.
	 **/
	const CMemFifoBlock *	Get();

	/**
	 * Extraire un chantillon de la pile FiFo.
	 * Si aucun chantillon n'est disponible, reste en attente jusqu' ce qu'un
	 * chantillon soit ajout (par un autre thread).
	 * Retourne NULL si en cas d'erreur.
	 * \note L'chantillon retourn DOIT tre dtruit par \p delete (sans [])
	 * aprs usage.
	 **/
	const CMemFifoBlock *	WaitGet();

	/// Destruction de tous les blocs allous
	void Free();

	/// Destructeur
	~CMemFifo();
};


