/*
 * $Id$
 * Copyright (C) 2010 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion d'intervalles mmoire et de flux d'octets et de bits
 *
 *	Ce fichier dfinit les classes et mthodes dfinies pour les structures de
 *	rfrencement d'intervalles d'adresses mmoire, ainsi que pour la lecture de
 *  flux d'octets et de bits.
 **/

#pragma once

#include "utils.h"

// ====================================================================================

enum StreamErrors
{
	serr_NoError,
	serr_OutOfData,
	serr_BufferOverflow,
	serr_TooManyBitsRequestedAtOnce,
	serr_RewindTooLarge
};

// ====================================================================================

/// Dfinition d'un intervalle mmoire par adresse et longueur
template <typename T>
	struct PtrRng
{
	const T *	ptr;	//!< Pointeur sur l'adresse de dbut
	UINT		siz;	//!< Longueur de l'espace point

	/// Constructeur par dfaut
	PtrRng() :
		ptr(NULL),
		siz(0)
	{
	}

	/// Construction  partir d'un pointeur et d'une longueur
	PtrRng(const T * pt, UINT sz) :
		ptr(pt),
		siz(sz)
	{
	}

	/// Construction  partir de deux pointeurs
	PtrRng(const T * ptr1, const T * ptr2) :
		ptr(ptr1),
		siz((UINT)(ptr2 - ptr1))
	{
	}

	/// Constructeur copie
	PtrRng(const PtrRng<T> & sM) :
		ptr(sM.ptr),
		siz(sM.siz)
	{
	}

	/// Retourner l'adresse qui suit celle du dernier octet
	PCUINT8	end() const {
		return ptr+siz; }

	/// limination de "sz" octets par le dbut
	PtrRng & operator += (UINT sz) {
		if (sz > siz)
			sz = siz;
		ptr += sz;
		siz -= sz;
		return *this; }

	/// Recul de "sz" octets par le dbut
	PtrRng & operator -= (UINT sz) {
		ptr -= sz;
		siz += sz;
		return *this; }

	/// Redfinir l'adresse de dpart dans l'intervalle existant, en ajustant la longueur
	PtrRng & operator >>= (const T * new_ptr) {
		new_ptr = min<const T *>(max<const T *>(new_ptr, ptr), end());
		siz -= UINT(new_ptr - ptr);
		ptr = new_ptr;
		return *this; }

	/// limination de "sz" objets "T" par le dbut
	PtrRng operator + (UINT sz) const {
		return PtrRng<T>(*this) += sz; }

	/// Recul de "sz" objets "T" par le dbut
	PtrRng operator - (UINT sz) const {
		return PtrRng<T>(*this) -= sz; }

	/// Retourne le premier objet "T" et l'limine
	T get() {
		return siz == 0 ? T() : (--siz, *ptr++); }

	/// Retourne \p true si l'intervalle courant est entirement contenu dans l'intervalle \p sM
	bool isIn(const PtrRng<T> & sM) const {
		return ptr != NULL && sM.ptr != NULL && ptr >= sM.ptr && end() <= sM.end(); }

	/// Retourner le dbut de l'espace mmoire en tant que rfrence sur un objet de type dfini
	template <typename T>
		const T & as() const {
			return *reinterpret_cast<const T *>(ptr); }
};

// ====================================================================================

typedef PtrRng<UINT8> MemRng;

// ====================================================================================

/// Abstraction pour la lecture d'un flux d'octets
class ByteStream
{
protected:
	virtual UINT8	peek_byte() = 0;
	virtual UINT8	next_byte() = 0;
	virtual void	unread_bytes(UINT count) = 0;
	virtual bool	end_of_stream() const = 0;
public:
	virtual bool	empty() {
						return end_of_stream();}
	void			fill_bytes_table(PUINT8 tbl, UINT siz) {
						while (siz > 0) {*tbl++ = next_byte(); --siz;}}
	void			skip_bytes(UINT count) {
						while (count > 0) {next_byte(); --count;	}}
};

// ====================================================================================

class BitStream : virtual public ByteStream
{
	void load_byte_in_cache();
public:

	UINT	bit_cache;	//!< Cache pour les prochains bits  lire
	UINT8	bit_cnt;	//!< Nombre de bits restant  lire dans le cache

	BitStream() :
		bit_cache(0),
		bit_cnt(0)
	{
	}

	// Alignement  l'octet, avec dchargement des bits prchargs si ncessaire ; aprs appel  cette
	// fonction, il est garanti que le pointeur 'pcur' pointe le prochain octet  lire :
	void	align();

	/**
	 * \brief Lecture des \p cnt bits suivants dans le flot
	 * \param[in]	cnt	Nombre de bits  lire
	 * \return		Bits lus, sous forme d'entier non sign
	 * \note	La valeur de \p cnt est valide pour le nombre de bits contenus dans un entier moins 7,
	 *			soit 25 en compilation 32 bits, et 57 en compilation 64 bits
	 *			Cette limite est porte  32 ou 64 juste aprs l'appel de la fonction membre \p align.
	 **/
	UINT	getbits(UINT8 cnt);

	UINT8	getbit();

	bool	getflag() {
				return getbit() != 0; }

	UINT32	peekbits(UINT8 count);

	UINT8	peekbit() {
				UINT8 res = getbit(); ++bit_cnt; return res; }

	bool	peekflag() {
				return peekbit() != 0; }

	void	rewind_bits(UINT8 count);

	UINT8	get8bits() {
				return (UINT8)getbits(8); }

	UINT16	get16bits() {
				return (UINT16)getbits(16); }

	UINT32	get32bits() {
		#ifdef _WIN64
				return (UINT32)getbits(32);
		#else
				UINT32 res = getbits(16) << 16; return res |getbits(16);
		#endif
			}

	UINT64	get64bits() {
				UINT64 res = UINT64(get32bits()) << 32; return res | get32bits(); }

	/// Remplissage d'une table de valeurs scalaires avec un nombre dfinis de valeurs d'une taille
	/// identique compris entre issues d'un flot d'octets
	template <typename T>
		void	fill_table(T * tbl, UINT siz, UINT8 cnt) {
					while (siz > 0) {*tbl++ = T(getbits(cnt)); --siz;} }

	// Lecture d'aprs une table VLC (Variable Length Code).
	// Le premier lment de cette table est un mot de 16 bits de la forme '0bbbdddddddddddd', avec
	//   'bbb' = nombre de bits  lire, et 'dddddddddddd' est un offset positif vers une sous-table
	//   de longueur 2^bbb, dans laquelle on lit la valeur retourne par le nombre de bits lus.
	// Chaque lment de cette sous-table est, soit du mme type (et on continue  parcourir la table),
	//   soit du type '1vvvvvvvvvvvvvvv', et alors on retourne la valeur 'vvvvvvvvvvvvvvv'.
	UINT16	get_vlc_bits(PCUINT16 pvlc);

	/// Lecture d'une valeur non signe code en codage exponentiel-Golomb
	/// \see http://fr.wikipedia.org/wiki/Code_exponentiel-Golomb
	/// \see ISO/IEC 14496-10 section 9.1
	UINT	get_eg();

	/// Lecture d'une valeur signe code en codage exponentiel-Golomb
	/// \see ISO/IEC 14496-10 section 9.1.1
	INT		get_eg_s();

	bool	empty() {
				return !bit_cnt && __super::empty();}
};

// ====================================================================================

/// Classe de lecture d'un flux de bits placs en mmoire dans des octets adjacents
class BitRng : public BitStream, private MemRng
{
	UINT8		nbit;	/// Numro du prochain bit  lire (0 = bit de poids fort, 7 = bit de
						/// poids faible, 8 = lecture de l'octet suivant ncessaire)

	virtual bool	end_of_stream() const {
						return siz == 0; }
protected:
	virtual UINT8	peek_byte() {
						return *ptr; }
	virtual UINT8	next_byte() {
						return get(); }
	virtual void	unread_bytes(UINT count) {
						(*this) -= 1; }
public:
	BitRng(MemRng sM = MemRng()) :
		MemRng(sM),
		nbit(0)
	{
	}

	BitRng(const BitRng & sB) :
		MemRng(sB),
		nbit(sB.nbit)
	{
	}

/*
	/// Lecture d'un bit
	UINT getbit();

	bool getflag() {
			return getbit() != 0; }

	/// Lecture de N bits
	UINT getbits(UINT8 cnt);

	/// Lecture d'une valeur non signe code en codage exponentiel-Golomb
	/// \see http://fr.wikipedia.org/wiki/Code_exponentiel-Golomb
	/// \see ISO/IEC 14496-10 section 9.1
	UINT get_eg();

	/// Lecture d'une valeur signe code en codage exponentiel-Golomb
	/// \see ISO/IEC 14496-10 section 9.1.1
	INT get_eg_s();

	/// Ralignement  l'octet suivant
	void align();
*/
};

// ====================================================================================

/// Classe de lecture d'un flux de bits placs en mmoire dans des octets adjacents,
/// et dans lesquels peuvent exister des octets 0x03 de prvention d'mulation (H264)
/// \see ISO/IEC 14496-10 section 7.4.1, "emulation_prevention_three_byte"
class BitRng_ep3 : public BitRng
{
	UINT8	zerocount;	//!< comptage des zros

	virtual UINT8 next_byte();
public:
	BitRng_ep3(MemRng sM = MemRng()) :
		BitRng(sM),
		zerocount(0)
	{
	}

	BitRng_ep3(const BitRng_ep3 & sB) :
		BitRng(sB),
		zerocount(sB.zerocount)
	{
	}
};

// ====================================================================================
