/*
 * $Id$
 * Copyright (C) 2011 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Utilitaires pour calcul de tailles, rectangles et points
 *
 *	Ce fichier dfinit des fonctionnalits permettant de faciliter les calculs
 *  portant sur les rectangles, les tailles et les points.
 **/

#pragma once

#include "stdafx.h"
#include <windows.h>

// ----------------------------------------------------------------------------

/// Largeur d'un rectangle (RECT)
inline LONG RWdt(const RECT & r)
{
	return r.right - r.left;
}

/// Hauteur d'un rectangle (RECT)
inline LONG RHgt(const RECT & r)
{
	return r.bottom - r.top;
}

/// Dimensions d'un rectangle (RECT)
inline SIZE RSiz(const RECT & sRect)
{
	SIZE sSiz = {RWdt(sRect), RHgt(sRect)};

	return sSiz;
}

/// Fonction retournant un \p POINT reprsentant l'origine d'un rectangle \p RECT
inline POINT ROrg(const RECT & sRect)
{
	POINT sPt = {sRect.left, sRect.top};

	return sPt;
}

/// Fonction retournant un \p POINT reprsentant le centre d'un rectangle \p RECT
inline POINT RCtr(const RECT & r)
{
	POINT sPt = {(r.right + r.left)/2, (r.bottom + r.top)/2};

	return sPt;
}

/// Fonction retournant un \p POINT reprsentant le point infrieur droit d'un rectangle \p RECT
inline POINT REnd(const RECT & sRect)
{
	POINT sPt = {sRect.right, sRect.bottom};

	return sPt;
}

/// Fonction construisant et retournant un point \p POINT
inline POINT MkPOINT(LONG x, LONG y)
{
	POINT sPnt = {x, y};

	return sPnt;
}

/// Comparaison de \p POINT pour galit
inline bool operator == (const POINT & sPt1, const POINT & sPt2)
{
	return sPt1.x ==sPt2.x && sPt1.y == sPt2.y;
}

/// Comparaison de \p POINT pour diffrence
inline bool operator != (const POINT & sPt1, const POINT & sPt2)
{
	return ! (sPt1 == sPt2);
}

/// Fonction construisant et retournant une taille \p SIZE.
inline SIZE MkSIZE(LONG cx, LONG cy)
{
	SIZE sSiz = {cx, cy};

	return sSiz;
}

/// Comparaison de tailles \p SIZE pour galit
inline bool operator == (const SIZE & sSz1, const SIZE & sSz2)
{
	return sSz1.cx ==sSz2.cx && sSz1.cy == sSz2.cy;
}

/// Comparaison de tailles \p SIZE pour diffrence
inline bool operator != (const SIZE & sSz1, const SIZE & sSz2)
{
	return ! (sSz1 == sSz2);
}

/// Multiplication d'une taille \p SIZE par une valeur
inline SIZE operator * (const SIZE & sSz, LONG nMul)
{
	SIZE sSiz = {sSz.cx*nMul, sSz.cy*nMul};

	return sSiz;
}

/// Division d'une taille \p SIZE par une valeur
inline SIZE operator / (const SIZE & sSz, LONG nDiv)
{
	SIZE sSiz = {sSz.cx/nDiv, sSz.cy/nDiv};

	return sSiz;
}

/// Simplification de ratio d'aspect
inline SIZE Simplify(SIZE sz)
{
	LONG nPgcd = PGCD(sz.cx, sz.cy);

	return sz / nPgcd;
}

/// Fonction construisant et retournant un rectangle \p RECT.
inline RECT MkRECT(LONG left, LONG top, LONG right, LONG bottom)
{
	RECT sRct = {left, top, right, bottom};

	return sRct;
}

/// Fonction construisant et retournant un rectangle \p RECT d'origine 0 et de taille \p SIZE.
inline RECT MkRECT(SIZE sSz)
{
	RECT sRct = {0, 0, sSz.cx, sSz.cy};

	return sRct;
}

/// Comparaison de rectangles \p RECT pour galit
inline bool operator == (const RECT & sRct1, const RECT & sRct2)
{
	return EqualRect(&sRct1, &sRct2) == TRUE;
}

/// Comparaison de rectangles \p RECT pour diffrence
inline bool operator != (const RECT & sRct1, const RECT & sRct2)
{
	return EqualRect(&sRct1, &sRct2) != TRUE;
}

/// Dcalage positif d'un rectangle \p RECT par un offset \p POINT.
inline RECT operator + (RECT sRct, POINT sPt)
{
	if (!OffsetRect(&sRct, sPt.x, sPt.y))
		return RECT();
	return sRct;
}

/// Dcalage positif d'un rectangle \p RECT par un offset \p POINT.
inline RECT & operator += (RECT & sRct, POINT sPt)
{
	if (!OffsetRect(&sRct, sPt.x, sPt.y))
		sRct = RECT();
	return sRct;
}

/// Dcalage ngatif d'un rectangle \p RECT par un offset \p POINT.
inline RECT operator - (RECT sRct, POINT sPt)
{
	if (!OffsetRect(&sRct, -sPt.x, -sPt.y))
		return RECT();
	return sRct;
}

/// Dcalage ngatif d'un rectangle \p RECT par un offset \p POINT.
inline RECT & operator -= (RECT & sRct, POINT sPt)
{
	if (!OffsetRect(&sRct, -sPt.x, -sPt.y))
		sRct = RECT();
	return sRct;
}

/// Diminution de taille d'un rectangle par une valeur \p nVal.
inline RECT operator << (RECT sRct, LONG nVal)
{
	if (!InflateRect(&sRct, -nVal, -nVal))
		return RECT();
	return sRct;
}

/// Augmentation de taille d'un rectangle par une valeur \p nVal.
inline RECT & operator <<= (RECT & sRct, LONG nVal)
{
	if (!InflateRect(&sRct, -nVal, -nVal))
		sRct = RECT();
	return sRct;
}

/// Diminution de taille d'un rectangle par une taille \p SIZE.
inline RECT operator << (RECT sRct, SIZE sSz)
{
	if (!InflateRect(&sRct, -sSz.cx, -sSz.cy))
		return RECT();
	return sRct;
}

/// Diminution de taille d'un rectangle par une taille \p SIZE.
inline RECT & operator <<= (RECT & sRct, SIZE sSz)
{
	if (!InflateRect(&sRct, -sSz.cx, -sSz.cy))
		sRct = RECT();
	return sRct;
}

/// Augmentation de taille d'un rectangle par une valeur \p nVal.
inline RECT operator >> (RECT sRct, LONG nVal)
{
	if (!InflateRect(&sRct, nVal, nVal))
		return RECT();
	return sRct;
}

/// Augmentation de taille d'un rectangle par une valeur \p nVal.
inline RECT & operator >>= (RECT & sRct, LONG nVal)
{
	if (!InflateRect(&sRct, nVal, nVal))
		sRct = RECT();
	return sRct;
}

/// Augmentation de taille d'un rectangle par une taille \p SIZE.
inline RECT operator >> (RECT sRct, SIZE sSz)
{
	if (!InflateRect(&sRct, sSz.cx, sSz.cy))
		return RECT();
	return sRct;
}

/// Augmentation de taille d'un rectangle par une taille \p SIZE.
inline RECT & operator >>= (RECT & sRct, SIZE sSz)
{
	if (!InflateRect(&sRct, sSz.cx, sSz.cy))
		sRct = RECT();
	return sRct;
}

/// Union de deux rectangles
inline RECT operator | (const RECT & sRct1, const RECT & sRct2)
{
	RECT sRct;

	if (!UnionRect(&sRct, &sRct1, &sRct2))
		return RECT();
	return sRct;
}

/// Union d'un rectangle  un autre
inline RECT & operator |= (RECT & sRct1, const RECT & sRct2)
{
	if (!UnionRect(&sRct1, &sRct1, &sRct2))
		sRct1 = RECT();
	return sRct1;
}

/// Intersection de deux rectangles
inline RECT operator & (const RECT & sRct1, const RECT & sRct2)
{
	RECT sRct;

	if (!IntersectRect(&sRct, &sRct1, &sRct2))
		return RECT();
	return sRct;
}

/// Intersection d'un rectangle avec un autre
inline RECT & operator &= (RECT & sRct1, const RECT & sRct2)
{
	if (!IntersectRect(&sRct1, &sRct1, &sRct2))
		sRct1 = RECT();
	return sRct1;
}

/// Transformation d'un rectangle en le plus grand rectangle possible,
/// centr ce rectangle, qui ait un ratio d'aspect gal  \p sAr.
inline void MaxArRect(RECT & sRect, SIZE sAr)
{
	SIZE	sSiz	= RSiz(sRect);	// Dimensions du rectangle initial
	LONG	nCYxX	= sSiz.cy * sAr.cx;
	LONG	nCXxY	= sSiz.cx * sAr.cy;
	LONG	nDiff;

	if (nCYxX > nCXxY) {
		// trop haut de combien de pixels ?
		nDiff			=	(sSiz.cy - (nCXxY / sAr.cx)) / 2;
		sRect.top		+=	nDiff;
		sRect.bottom	-=	nDiff;
	} else if (nCYxX < nCXxY) {
		// trop large de combien de pixels ?
		nDiff			= (sSiz.cx - (nCYxX / sAr.cy)) / 2;
		sRect.left		+=	nDiff;
		sRect.right		-=	nDiff;
	}
}

/// Transformation d'un rectangle en le plus petit rectangle possible,
/// entourant ce rectangle, qui ait un ratio d'aspect gal  \p sAr.
inline void MinArRect(RECT & sRect, SIZE sAr)
{
	SIZE	sSiz	= RSiz(sRect);	// Dimensions du rectangle initial
	LONG	nCYxX	= sSiz.cy * sAr.cx;
	LONG	nCXxY	= sSiz.cx * sAr.cy;
	LONG	nDiff;

	if (nCYxX > nCXxY) {
		// trop troit de combien de pixels ?
		nDiff			= ((nCYxX / sAr.cy) - sSiz.cx) / 2;
		sRect.left		-=	nDiff;
		sRect.right		+=	nDiff;
	} else if (nCYxX < nCXxY) {
		// trop aplati de combien de pixels ?
		nDiff			=	((nCXxY / sAr.cx - sSiz.cy)) / 2;
		sRect.top		-=	nDiff;
		sRect.bottom	+=	nDiff;
	}
}

/**
 * \brief Transposition d'un diffrentiel de rectangle d'un rfrentiel vers un autre
 * \param[in]	r1	Rectangle source numro 1
 * \param[in]	r2	Rectangle source numro 2
 * \param[in]	rR	Rectangle pour lequel la transposition doit tre calcule
 * \return	Un rectangle qui a, par rapport au rectangle \p rR, la mme relation que celle
 *			qu'a le rectangle \p r2 par rapport au rectangle \p r1.
 * \note	Les rectangles \p r1 et \p r2 doivent tre du mme type, mais le rectangle \p rR peut avoir
 *			ses champs dfinis dans un type diffrent (par rapport \p int vs. \p float ou bien
 *			\p unsigned vs \p double. La seule contrainte est qu'ils doivent tre tous dfinis en
 *			tant que structure avec 4 champs nomms \p top, \p left, \p bottom et \p right.
 *			Le type de retour est le mme que celui du rectangle \p rR.
 *			Si un type entier est utilis, attention au fait 
 **/
#if TYPE_INFERENCE_AVAILABLE	// ->  partir de Visual Studio 2010

template <typename T, typename U>	// Version pour Visual Studio 2010
	U TransposeRect(const T & r1, const T & r2, const U & rR)
{
	// Variables intermdiaires avec infrence de types
	auto	nR1_w = r1.right   - r1.left;
	auto	nR1_h = r1.bottom  - r1.top;
	auto	nRr_w = rR.right   - rR.left;
	auto	nRr_h = rR.bottom  - rR.top;

	U rRes = {
			rR.left + static_cast<decltype(rR.left) >((r2.left   - r1.left) * nRr_w / nR1_w),
			rR.top  + static_cast<decltype(rR.top)  >((r2.top    - r1.top)  * nRr_h / nR1_h),
			rR.left + static_cast<decltype(rR.left) >((r2.right  - r1.left) * nRr_w / nR1_w),
			rR.top  + static_cast<decltype(rR.right)>((r2.bottom - r1.top)  * nRr_h / nR1_h)
		};

	return rRes;
}

#else	// #if _MSC_VER < 1600

#pragma warning (push)
#pragma warning (disable : 4244)	// C4244 = 'conversion' : conversion de 'type1' en 'type2', perte possible de donnes

template <typename T, typename U>	// Version pour Visual Studio 2005 ou 2008
	U TransposeRect(T r1, const T & r2, U rR)
{
	// Palliatif en attendant le C++0x qui permet les infrences de types
	#define width	right	// Utilisation de champs 'right' en tant que 'width'
	#define height	bottom	// Utilisation de champs 'bottom' en tant que 'height'

	r1.width  = r1.right   - r1.left;
	r1.height = r1.bottom  - r1.top;
	rR.width  = rR.right   - rR.left;
	rR.height = rR.bottom  - rR.top;

	U rRes = {
			rR.left + (r2.left   - r1.left) * rR.width  / r1.width,
			rR.top  + (r2.top    - r1.top)  * rR.height / r1.height,
			rR.left + (r2.right  - r1.left) * rR.width  / r1.width,
			rR.top  + (r2.bottom - r1.top)  * rR.height / r1.height
		};

	return rRes;
	#undef width
	#undef height
}
#pragma warning (pop)

#endif	// #if _MSC_VER
