/*
 * $Id: utils.h 276 2011-03-06 10:17:01Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Fonctions utilitaires gnriques (dfinitions)
 *
 *	Ce module contient des dfinitions et structures utilitaires simples, et non
 *	obligatoirement lies au programme principal.
 *
 *	Les objets dfinis ici ne doivent faire rfrence qu' des bibliothques externes
 *	standard du C, du C++ ou de Windows.
 **/

#pragma once

#include <string>
#include <sstream>
#include <vector>
#include <algorithm>
#include <malloc.h>

using namespace std;

// ----------------------------------------------------------------------------

typedef const UINT8 * PCUINT8;		//!< Pointeur "const" sur UINT8
typedef const UINT16 * PCUINT16;	//!< Pointeur "const" sur UINT16

/**
 * Classe modle permettant d'accder rapidement aux membres d'une structure
 * quelconque se trouvant au milieu d'un champ d'octets bruts (par exemple un
 * tampon / buffer),  partir d'un pointeur quelconque sur le dbut de cette
 * structure.
 *
 * Dans l'exemple suivant, on copie dans \p x le champ \p right d'un rectangle
 * de type \p RECT dont on sait qu'il existe  l'adresse pointe par \p ptr :
 * \code LONG x = buf_cast<RECT>(ptr).right; \endcode
 * \note
 * Il est prfrable de s'assurer de n'accder ainsi qu' des lments
 * correctement aligns, selon le type de microprocesseur utilis (les lments
 * ne doivent pas "croiser" une limite de 32 ou 64 bits).
 **/
template <class T>
	inline const T & buf_cast(PCUINT8 ptr)
{
	return *reinterpret_cast<const T *>(ptr);
}

/** @defgroup fmtMacros Macros de formatage style PRINTF
 *  Ces macros permettent d'insrer dans une chane de formatage de style \p printf
 *  des arguments structurs courants.
 *
 *  Elles fonctionnent par paires, composes d'une macro  placer au milieu de la
 *  chane de format et d'une autre macro  placer dans la liste d'arguments.
 *  Cette seconde macro prend l'objet  formater (directement, pas un pointeur)
 *  en paramtre.
 *
 *  Ces macros sont a priori davantage ddies au debugging, bien que ce ne soit
 *  pas l une limitation d'usage absolue.
 *  @{
 **/

/**
 * Macro s'insrant au milieu d'une chane de formatage style
 * \p printf pour produire le contenu d'un objet \p GUID ou \p CLSID. Une macro
 * \p GUIDPARM doit correspondre, en position, dans les arguments associs.
 */
#define GUIDFMT TEXT("{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}")

/// Identique au prcdent, mais peut se voir remplacer, en debugging, par une
/// fonction qui extrait le nom en clair correspondant au CLSID
#define GUIDFMT_N GUIDFMT

/**
 * Macro s'insrant dans les arguments d'une fonction de style \p printf
 * pour produire le contenu d'un objet \p GUID ou \p CLSID. Doit correspondre en position
 *  une macro \p GUIDFMT insre dans la chane de formatage.
 * \param[in] x Objet GUID ou CLSID dont on veut produire la sortie.
 */
#define GUIDPARM(x) (x).Data1, (x).Data2, (x).Data3, \
	(x).Data4[0],(x).Data4[1],(x).Data4[2],(x).Data4[3], \
	(x).Data4[4],(x).Data4[5],(x).Data4[6],(x).Data4[7]

/// Identique au prcdent, mais peut se voir remplacer, en debugging, par une
/// fonction qui extrait le nom en clair correspondant au CLSID
#define GUIDPARM_N GUIDPARM

/**
 * Macro s'insrant au milieu d'une chane de formatage style
 * \p printf pour produire le contenu d'une structure RECT (rectangle). Une macro
 * \p RECTPARM doit correspondre, en position, dans les arguments associs.
 * \param[in] c Le caractre qui spare les lments du rectangle dans la chane  gnrer.
*/
#define RECTFMT(c) TEXT("%i")TEXT(c)TEXT("%i")TEXT(c)TEXT("%i")TEXT(c)TEXT("%i")

/**
 * Macro s'insrant dans les arguments d'une fonction de style \p printf
 * pour produire le contenu d'une structure \p RECT (rectangle). Doit correspondre
 *  en position une macro \p RECTFMT insre dans la chane de formatage.
 * \param[in] r Objet \p RECT dont on veut produire la sortie.
 */
#define RECTPARM(r) (r).top,(r).left,(r).bottom,(r).right

/** @} */ // fin de fmtMacros

/// Fonction (modle) retournant la valeur \p Val limite par les bornes \p nInf et \p nSup
template <class T>
	inline T limit(T nInf, T nVal, T nSup)
{
	return min(max(nInf, nVal), nSup);
}

/// Calcul du plus grand commun diviseur
template <typename T>
	T PGCD(T nVal1, T nVal2)
{
	while (nVal2 != 0) {
		T	nTmp = nVal1 % nVal2;

		nVal1 = nVal2;
		nVal2 = nTmp;
	}
	return nVal1;
}

// ----------------------------------------------------------------------------

/// Macro pour amliorer la lisibilit des appels de fonctions membres par pointeur
#define CALL_MEMBER_FN(obj, funcPtr)  ((obj).*(funcPtr))

// ----------------------------------------------------------------------------

/** @defgroup A2tmacros Macros de conversion @p CHAR / @p WCHAR / @p TCHAR
 *  Ces macros s'insrent dans les chanes de format de style \p printf, permettant de choisir
 *  la lettre de formatage approprie, selon la configuration, pour la conversion correcte
 *  des arguments de type \p TCHAR vers une chane de format \p CHAR / \p WCHAR, ou inversement.
 *  \note
 *  Ne pas oublier d'inclure le caractre \p "%" (suivi de son ventuel modificateur)  la
 *  fin de la chane qui prcde.
 *  @{
 **/

/*! \def T2a
 * Exemple : \code printf("Il y a %" T2a "au menu.", TEXT("du poulet")); \endcode
 **/

/*! \def T2w
 * Exemple : \code wprintf(L"Il y a %" T2w L"au menu.", TEXT("du boeuf")); \endcode
 **/

/*! \def A2t
 * Exemple : \code _tprintf(TEXT("Il y a %") A2t TEXT("au menu."), "de l'agneau"); \endcode
 **/

/*! \def W2t
 * Exemple : \code _tprintf(TEXT("Il y a %") W2t TEXT("au menu."), L"du poisson"); \endcode
 **/

#ifdef  _UNICODE
	#define T2a "S"		//!< Conversion TCHAR vers CHAR
	#define T2w L"s"	//!< Conversion TCHAR vers WCHAR
	#define A2t L"S"	//!< Conversion CHAR vers TCHAR
	#define W2t L"s"	//!< Conversion WCHAR vers TCHAR
#else
	#define T2a "s"		//!< Conversion TCHAR vers CHAR
	#define T2w L"S"	//!< Conversion TCHAR vers WCHAR
	#define A2t "s"		//!< Conversion CHAR vers TCHAR
	#define W2t "S"		//!< Conversion WCHAR vers TCHAR
#endif
// (ces macros taient dfinies initialement uniquement avec des lettres
//  majuscules, mais en raison d'un conflit avec certaines macros de la
//  bibliothque ATL, leurs 3me caractres sont maintenant en minuscules).

/** @} */ // fin de A2tmacros

/// Conversion explicite d'une macro CHAR en macro WCHAR (base)
#define __WTXT(str) L##str

/// Conversion explicite d'une macro CHAR en macro WCHAR (appel effectif, requis)
#define WTXT(str) __WTXT(str)

// ----------------------------------------------------------------------------

typedef	basic_string<TCHAR>			tstring;		//!< Alias "TCHAR" pour \p string ou \p wstring
typedef basic_stringstream<TCHAR>	tstringstream;	//!< Alias "TCHAR" pour \p stringstream ou \p wstringstream

/**
 * Gnration d'une chane formate (format \p printf) avec rsultat dans une chane \p tstring
 **/
tstring tstr_printf(LPCTSTR pszFmt, ...);

// ----------------------------------------------------------------------------

/** @defgroup strcpy_T Copies de chanes entre @p CHAR / @p WCHAR / @p TCHAR
 *  Ces fonctions assurent la copie depuis des chanes de caractres de type \p TCHAR
 *  depuis ou vers des chanes de caractres \p CHAR ou \p WCHAR.
 *  Elles ont toutes le mme nom car elles devraient pouvoir tre discrimines par
 *  leurs arguments en C++, pargnant ainsi au programmeur d'avoir  faire le choix
 *	lui-mme.
 *
 *  \note Essayer quand mme d'viter les copies inutiles entre types diffrents,
 *  car on peut prsumer que les fonctions \p MultiByteToWideChar et
 *  \p WideCharToMultiByte que cela appelle indirectement consomment relativement
 *  plus de ressources que celles impliques au cours d'une copie simple.
 *  @{
 **/

/**
 * Version CHAR -> CHAR (alias de \p strcpy_s).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (CHAR)
 **/
inline errno_t strcpy_T(LPSTR pDst, UINT nSizDst, LPCSTR pSrc) {
	return strcpy_s(pDst, nSizDst, pSrc);}

/**
 * Version WCHAR -> WCHAR (alias de \p wcscpy_s).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
inline errno_t strcpy_T(LPWSTR pDst, UINT nSizDst, LPCWSTR pSrc) {
	return wcscpy_s(pDst, nSizDst, pSrc);}

/**
 * Version CHAR -> WCHAR (utilise \p MultiByteToWideChar).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
errno_t strcpy_T(LPWSTR pDst, UINT nSizDst, LPCSTR pSrc);

/**
 * Version WCHAR -> CHAR (utilise \p WideCharToMultiByte).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (CHAR)
 **/
errno_t strcpy_T(LPSTR pDst, UINT nSizDst, LPCWSTR pSrc);

/**
 * Version template, utilisable uniquement si la taille est drivable de la destination.
 * Cette version,  l'instar de celles de Microsoft sur les fonctions scurises de
 * traitement de chanes, permettent de se dispenser de fournir un dcompte de la taille
 * du tampon de destination si celle-ci peut en tre drive (c'est--dire, si la
 * destination est un tableau rel et non pas un pointeur).
 *
 * \param[in]  pSrc		Pointeur sur la chane source
 * \param[out] szDst	Tampon (tableau) de destination recevant la valeur copie
 **/
template <class TD, class TS, size_t S>
	inline errno_t strcpy_T(TD (&szDst)[S], const TS * pSrc) {
		return strcpy_T(szDst, S, pSrc); }

/**
 * Copie vers chane STL, version CHAR -> std::wstring.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  strDst	Chane STL de destination (std::wstring)
 **/
errno_t strcpy_T(std::wstring & strDst, LPCSTR pSrc);

/**
 * Copie vers chane STL, version WCHAR -> std::string.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  strDst	Chane STL de destination (std::string)
 **/
errno_t strcpy_T(std::string & strDst, LPCWSTR pSrc);

/**
 * Copie vers chane STL, version CHAR -> std::string.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  strDst	Chane STL de destination (std::string)
 **/
inline errno_t strcpy_T(std::string & strDst, LPCSTR pSrc)
{
	strDst = pSrc;
	return 0;
}

/**
 * Copie vers chane STL, version WCHAR -> std::wstring.
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  strDst	Chane STL de destination (std::wstring)
 **/
inline errno_t strcpy_T(std::wstring & strDst, LPCWSTR pSrc)
{
	strDst = pSrc;
	return 0;
}

/** @} */ // fin de strcpy_T

// ----------------------------------------------------------------------------

/** @defgroup strcat_T concatnation de chanes entre @p CHAR / @p WCHAR / @p TCHAR
 *  Ces fonctions assurent la concatnation depuis des chanes de caractres de type
 *  \p TCHAR depuis ou vers des chanes de caractres \p CHAR ou \p WCHAR.
 *  Elles ont toutes le mme nom car elles devraient pouvoir tre discrimines par
 *  leurs arguments en C++, pargnant ainsi au programmeur d'avoir  faire le choix
 *	lui-mme.
 *
 *  \note Essayer quand mme d'viter les copies inutiles entre types diffrents,
 *  car on peut prsumer que les fonctions \p MultiByteToWideChar et
 *  \p WideCharToMultiByte que cela appelle indirectement consomment relativement
 *  plus de ressources que celles impliques au cours d'une copie simple.
 *  @{
 **/

/**
 * Version CHAR -> CHAR (alias de \p strcat_s).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (CHAR)
 **/
inline errno_t strcat_T(LPSTR pDst, UINT nSizDst, LPCSTR pSrc) {
	return strcat_s(pDst, nSizDst, pSrc);}

/**
 * Version WCHAR -> WCHAR (alias de \p wcscat_s).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
inline errno_t strcat_T(LPWSTR pDst, UINT nSizDst, LPCWSTR pSrc) {
	return wcscat_s(pDst, nSizDst, pSrc);}

/**
 * Version CHAR -> WCHAR (utilise \p MultiByteToWideChar).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (CHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (WCHAR)
 **/
errno_t strcat_T(LPWSTR pDst, UINT nSizDst, LPCSTR pSrc);

/**
 * Version WCHAR -> CHAR (utilise \p WideCharToMultiByte).
 *
 * \param[in]  pSrc		Pointeur sur la chane source (WCHAR)
 * \param[in]  nSizDst	Taille du tampon de destination
 * \param[in]  pDst		Pointeur sur le tampon de destination (CHAR)
 **/
errno_t strcat_T(LPSTR pDst, UINT nSizDst, LPCWSTR pSrc);

/**
 * Version template, utilisable uniquement si la taille est drivable de la destination.
 * Cette version,  l'instar de celles de Microsoft sur les fonctions scurises de
 * traitement de chanes, permettent de se dispenser de fournir un dcompte de la taille
 * du tampon de destination si celle-ci peut en tre drive (c'est--dire, si la
 * destination est un tableau rel et non pas un pointeur).
 *
 * \param[in]  pSrc		Pointeur sur la chane source
 * \param[out] szDst	Tampon (tableau) de destination recevant la valeur copie
 **/
template <class TD, class TS, size_t S>
	inline errno_t strcat_T(TD (&szDst)[S], const TS * pSrc) {
		return strcat_T(szDst, S, pSrc); }

/** @} */ // fin de strcat_T

// ----------------------------------------------------------------------------

/** @defgroup strlen_T Prendre la longueur d'une chane de caractres @p CHAR / @p WCHAR / @p TCHAR
 *  Ces fonctions comptent le nombre de caractres d'une chane, quelque soit le type
 *  de caractre inclus dans la chane.
 *  Elles ont toutes le mme nom car elles devraient pouvoir tre discrimines par
 *  leur arguments en C++, pargnant ainsi au programmeur d'avoir  faire le choix
 *	lui-mme, et aussi elles facilitent leur usage dans un "template".
 *
 *  Accessoirement, elles retournent aussi un type \p UINT (plutt que \p size_t), vitant ainsi
 *  des cast de type ou des warnings agaants.
 *  @{
 **/

/**
 * Longueur d'une chane de caractres, version \p CHAR.
 *
 * \param[in]  pszStr	Pointeur sur la chane
 **/
inline UINT strlen_T(LPCSTR pszStr) {
	return (UINT)strlen(pszStr);}

/**
 * Longueur d'une chane de caractres, version \p WCHAR.
 *
 * \param[in]  pszStr	Pointeur sur la chane
 **/
inline UINT strlen_T(LPCWSTR pszStr) {
	return (UINT)wcslen(pszStr);}

/** @} */ // fin de strlen_T

// ----------------------------------------------------------------------------

/**
 *	Conversion d'une chane de caractres WCHAR vers une chane std::string
 *  \param[in]	pszStr	Chane  convertir
 **/
std::string WChar2Char(LPCWSTR pszStr);

/**
 *	Conversion d'une chane de caractres CHAR vers une chane std::wstring
 *  \param[in]	pszStr	Chane  convertir
 **/
std::wstring Char2WChar(LPCSTR pszStr);


/*! \def WCHAR2TCHAR
 * Conversion " la vole" d'une chane WCHAR vers une chane TCHAR.
 * Macro  utiliser de prfrence lorsque la conversion doit se faire
 * au milieu d'une expression. L'allocation mmoire intermdiaire est
 * automatiquement libre aprs valuation de l'expression.
 **/
#ifdef _UNICODE
	#define WCHAR2TCHAR(str) str
#else
	#define WCHAR2TCHAR(str) WChar2Char(str).c_str()
#endif

/*! \def CHAR2TCHAR
 * Conversion " la vole" d'une chane CHAR vers une chane TCHAR.
 * Macro  utiliser de prfrence lorsque la conversion doit se faire
 * au milieu d'une expression. L'allocation mmoire intermdiaire est
 * automatiquement libre aprs valuation de l'expression.
 **/
#ifdef _UNICODE
	#define CHAR2TCHAR(str) Char2WChar(str).c_str()
#else
	#define CHAR2TCHAR(str) str
#endif

/**
 * Macro d'allocation d'un tableau de \p taille lments de type \p type sur la pile,
 * en tirant parti des particularits de la fonction systme \p _alloca.
 * Permet un effet similaire  celui produit par l'extension non standard permise sur
 * les compilateurs de type GCC,  savoir de pouvoir dfinir un tableau  l'intrieur
 * d'une fonction, en utilisant un terme variable dans la taille du tableau.
 *
 * Par rapport  une allocation sur le tas (heap) avec \p malloc ou \p new, les
 * avantages sont les suivants :
 * - Traitement infiniment plus rapide (quelques instructions machine seulement)
 * - Pas besoin de \p free ou \p delete, la destruction est implicite  la fin de la
 *   fonction.
 *
 * Pour ce qui est des inconvnients :
 * - Usage strictement limit au corps d'une fonction : pas question de retourner
 *   le rsultat.
 * - Ne devrait pas tre utilis dans une boucle,  moins qu'il soit clair que le
 *   nombre d'appels sera trs rduit : les allocations successives ne seraient pas
 *   libres avant la sortie de la fonction.
 * - La taille de l'allocation doit rester limite  ce qui pourrait se dclarer en
 *   tant que variable locale (sinon, risque de dbordement de la pile)
 *
 * Exemple : \code
 *   int		taille=16;
 *   wchar_t *	buffer = STK_ALLOC(wchar_t, taille);
 * \endcode
 * Ceci est  peu prs quivalent  cette syntaxe, qui serait valide sous GCC : \code
 *   int		taille=16;
 *   wchar_t	buffer[taille];
 * \endcode
 * \param[in]	type	Type des lments du tableau
 * \param[in]	taille	Nombre d'lments  allouer
**/
#define STK_ALLOC(type, taille) reinterpret_cast<type *>(_alloca((taille)*sizeof(type)))

/**
 * Macro  insrer dans le corps d'une fonction C++, et qui permet de gnrer
 * une copie alloue sur la pile d'une chane de caractres, cette copie
 * pouvant faire l'objet d'une transformation \p CHAR / \p WCHAR / \p TCHAR
 * ventuelle au passage.
 * \param[in]	type	nom d'un type de donnes qui dfinit un caractre
 * \param[in]	dest	est le nom du pointeur sur la chane destination (copie)
 * \param[in]	source	est le nom de la chane  copier
 * \param[in]	taille	est le nom d'une variable de type \p UINT qui sera cre
 *						et qui recevra la taille de la chane + 1.
 * \note	Cette macro ne peut **pas** devenir une fonction "inline", en raison du
 * comportement trs particulier de la fonction \p _alloca incluse.
 **/
#define COPY_STR_ON_STACK(type, dest, source, taille) \
	UINT taille = strlen_T(source)+1; \
	type * dest = STK_ALLOC(type, taille); \
	strcpy_T(dest, taille, source)

// ----------------------------------------------------------------------------

/**
 * Macro d'assistance  l'encapsulation de fonctionnalits "printf"
 *
 * Aprs excution du code correspondant, la chane formate est disponible
 * dans la variable dont le nom est choisi en premier argument.
 * \param[in]	buffer	Nom de la chane tampon  utiliser
 * \param[in]	size	Taille de la chane tampon  utiliser
 * \param[in]	fmt		Chane de formatage
 * \param[in]	prevarg	Argument prddant la liste d'arguments variables
 * \note Une tentative d'utiliser un template pour ce faire a t abandonne
 * pour cause d'impossibilit de se conformer au standard C++ (cela fonctionnait
 * en compilation 32 bits mais pas en compilation 64 bits).
 **/
#define FMT_BUF(buffer, size, fmt, prevarg) \
	TCHAR buffer[size]; { \
	va_list	args; \
	va_start(args, prevarg); \
	_vstprintf_s(buffer, fmt, args); \
	va_end(args); }

// ----------------------------------------------------------------------------

/**
 * Nettoyage des espaces en dbut et fin d'une chane de caractres
 *
 * \param[in,out]	pszStr		Adresse du tampon contenant la chane  nettoyer
 * \param[in]		nBufLen		Taille du tampon
 * \param[in]		pszDefault	Chane  substituer si le rsultat est vide
 **/
void clean_string(LPTSTR pszStr, UINT32 nBufLen, LPCTSTR pszDefault=NULL);

/**
 * Assignation d'une chane sur une autre avec test modification
 *
 * \param[in]		pszSrc		Chane source
 * \param[in,out]	pszDst		Chane destination
 * \param[in]		nSize		Taille du tampon de cette chane
 * \retval			\p true si la chane destination a t modifie
 * \retval			\p false si la nouvelle valeur tait identique  l'ancienne
 **/
bool copy_check(LPTSTR pszDst, UINT nSize, LPCTSTR pszSrc);

// ----------------------------------------------------------------------------
/**
 * Structure pour la sauvegarde de la position des fentres
 **/
struct WindowPos {
	long X;
	long Y;
	long L;
	long H;

	/// Constructeur
	WindowPos() :
		X(CW_USEDEFAULT),Y(CW_USEDEFAULT),
		L(CW_USEDEFAULT),H(CW_USEDEFAULT)
	{}

	// Vrifie si cette valeur est visible sur un des moniteurs existants
	bool verifie() const;

	/// Restauration d'une position de fentre
	BOOL RestoreWindowPos(HWND hWnd, UINT nShowCmd) const;
};

/**
 * Conversion des coordonnes d'un rectangle contenant des coordonnes relatives  l'cran en un rectangle
 * contenant des coordonnes relatives au rectangle client d'une fentre.
 * \param[in]		hWnd	"Handle" d'une fentre dont le rectangle client sera utilis pour la conversion
 * \param[in,out]	sRect	Rectangle dont les coordonnes seront converties
 **/
BOOL ScreenToClient(HWND hWnd, RECT & sRect);

/**
 * Obtenir le rectangle de position d'une fentre en coordonnes relatives  celles de sa fentre parente
 * \param[in]	hWnd	Fentre dont les coordonnes sont  rcuprer
 * \param[out]	sRect	Rectangle recevant les coordonnes rsultantes
 **/
inline BOOL GetChildWindowRect(HWND hWnd, RECT & sRect)
{
	return GetWindowRect(hWnd, &sRect) && ScreenToClient(GetParent(hWnd), sRect);
}

// ----------------------------------------------------------------------------
// GDI

/// Gestion automatique de Device Context ncessitant d'tre libr.
class AutoDC
{
	HWND	hWnd;
	HDC		hDC;
public:
	AutoDC(HWND hWn) :
		hWnd(hWn),
		hDC(GetDC(hWn))
	{
	}

	operator HDC () const {
		return hDC; }

	~AutoDC() {
		ReleaseDC(hWnd, hDC); }
};

/// Gestion automatique de Handle devant tre dtruite en sortie.
template <typename T, int NULVAL = NULL>
	class AutoHandle
{
	T		hHdl;
public:
	AutoHandle(T hHd) :
		hHdl(hHd)
	{
	}

	AutoHandle(HGDIOBJ hObj) :
		hHdl(static_cast<T>(hObj))
	{
	}

	operator T () const {
		return hHdl; }

	~AutoHandle() {
		if (hHdl == NULVAL)
			DeleteObject((HGDIOBJ)hHdl); }
};

/// Slection d'un objet dans un Device Context avec dslection automatique
///  la sortie du contexte.
template <typename T>
	class AutoSelect
{
	HDC	hSavDC;
	T	hObj;
public:
	AutoSelect(HDC hDC, T hOb) :
		hSavDC(hDC),
		hObj((T)SelectObject(hDC, (HGDIOBJ)hOb))
	{
	}

	operator HDC () const {
		return hSavDC; }

	~AutoSelect() {
		SelectObject(hSavDC, (HGDIOBJ)hObj); }
};

// ----------------------------------------------------------------------------

/**
 * lment de table d'association chane <--> valeur binaire.
 * Cette structure est destine  faire partie d'une table dont le dernier lment
 * doit avoir ses deux valeurs  NULL et 0 respectivement.
 **/
struct AssocElement
{
	LPCTSTR	str;
	DWORD	bin;
};

/**
 * \brief Trouver la valeur binaire associe  une chane de caractres fournie dans une table
 *
 * Si la chane ne se trouve pas dans la table, alors la valeur \p dwDefault est retourne.
 *
 * \param[in]  assocTable	Tableau associatif d'o l'on doit dduire la valeur numrique
 *							correspondant  la chane de caractres passe
 * \param[in]	pszStr		Chane de caractres recherche
 * \param[in]	dwDefault	Valeur retourne si on n'a rien trouv
 * \return					La valeur numrique correspondante
 **/
DWORD str2dword(const AssocElement * assocTable, LPCTSTR pszStr, DWORD dwDefault = DWORD(-1));

/**
 * \brief Trouver la chane de caractres associe  une valeur binaire fournie dans une table
 *
 * Si la valeur binaire ne se trouve pas dans la table, alors la chane \p pszDefault est retourne.
 *
 * \param[in]	assocTable	Tableau associatif d'o l'on doit extraire la valeur
 * \param[in]	dwValue		Valeur devant tre convertie en chane de caractres
 * \param[in]	pszDefault	Chane retourne si aucune correspondance trouve
 * \return					Chane de caractres correspondant  \p dwValue, ou \p pszDefault si pas trouv
 **/
LPCTSTR dword2str(const AssocElement * assocTable, DWORD dwValue, LPCTSTR pszDefault = TEXT("??"));

// ----------------------------------------------------------------------------

/**
 * Obtenir une chane de caractres "Oui" ou "Non" selon la valeur d'un boolan.
 *
 * \param[in]	bValue	valeur boolenne  convertir
 * \return	Chane de caractres "Oui" ou "Non"
 **/
LPCTSTR boolToYesNo(bool bValue);

/// Liste des jours du mois (longs, minuscules)
extern LPCSTR pszMon_lower;

/// Liste des jours de la semaine (longs, minuscules)
extern LPCSTR pszDoW_lower;

/// Liste des jours de la semaine (longs, avec majuscules initiales)
extern LPCSTR pszDoW_ucfirst;

/// Liste des jours de la semaine (courts, avec majuscules initiales)
extern LPCSTR pszDoW_short_ucf;

/**
 * Constante matrialisant un temps "infini" lors des calculs de diffrence de temps :
 * (reprsentant en fait un peu moins de 25 jours, mais dans notre contexte a ne
 * devrait pas poser de problmes)
 **/
#define DUREE_INFINIE long((unsigned long)-1 >> 1)	// = 0x7fffffff, si 'long' est sur 32 bits

/**
 * Fonction qui retourne la diffrence time1 - time2, convertie en millisecondes.
 *
 * \param[in]  time1  temps devant tre compar
 * \param[in]  time2  temps soustrait au prcdent
 * \return           diffrence entre les 2 temps en millisecondes
 **/
long DiffTime(SYSTEMTIME time1, SYSTEMTIME time2);

/**
 * Fonction qui ajoute 'add' millisecondes au temps 'time.
 *
 * \param[in,out] time  date de dpart/finale
 * \param[in]     add   millisecondes  ajouter
 **/
void AddTime(SYSTEMTIME & time, long add);

/**
 * Conversion FILETIME vers UINT64 ...
 **/
inline UINT64 FILETIME_to_UINT64(const FILETIME & ft)
{
	return (UINT64(ft.dwHighDateTime)<<32) | ft.dwLowDateTime;
}

/**
 * Conversion heure (sans date) en chane de caractres
 *
 * \param[in]  st        structure SYSTEMTIME contenant les donnes  intgrer
 * \param[out] str       chane de caractres rsultant
 * \param[in]  str_size  longeur max de la chane de caractres
 **/
void TimeToStr(SYSTEMTIME st, LPTSTR str, size_t str_size);

/**
 * Conversion heure et date en chane de caractres
 *
 * \param[in]  st        structure SYSTEMTIME contenant les donnes  intgrer
 * \param[out] str       chane de caractres rsultant
 * \param[in]  str_size  longeur max de la chane de caractres
 **/
void DateTimeToStr(SYSTEMTIME st, LPTSTR str, size_t str_size);

/**
 * Conversion temps GMT en temps local
 *
 * \param[in,out] st  structure SYSTEMTIME  convertir
 **/
void SystemTimeToLocalSystemTime(SYSTEMTIME & st);

/**
 * \brief Reformatage de la date et de l'heure de compilation
 *
 * \param[out]	pszBuf		Tampon destin  recevoir le rsultat
 * \param[in]	nBufSiz		Taille du tampon
 * \param[in]	pszDtSep	Chane de sparation entre date et heure
 * \param[in]	pszRawTime	Date et heure au format natif du compilateur, spars par une virgule
 * \note	La chane contenant l'heure et la date source ne doit surtout pas tre assigne  une
 *			variable globale, car cela compromettrait la fonctionnalit "Edit & Continue" de Visual Studio.
 **/
void ReformatCompileTime(LPTSTR pszBuf, UINT nBufSiz, LPCTSTR pszDtSep = TEXT(", "), LPCSTR pszRawTime = __DATE__ ", " __TIME__);

// ----------------------------------------------------------------------------

/**
 * Conversion d'une valeur d'un type numr vers une chane de caractres
 * slectionne dans une chane fournissant une suite de rsultats possibles
 *
 * \param[in] nVal			Valeur  associer
 * \param[in] pszList		Chane contenant une suite d'items spars par des zros binaires ;
 *							un item dbutant par \p 0x01 ("un" binaire) est remplac par l'item par
 *							dfaut ; un item vide (donc deux zros successifs) DOIT terminer.
 * \param[in] pszDefault	Valeur par dfaut
 * \retval					Une copie de l'item de rang \p nVal (base 0), ou bien \p pszDefault
 *							si \p nVal excde le nombre d'items ou si l'item dbute par \p 0x01.
 * \note \p pszList doit tre une chane constante, ou bien au moins statique.
 * \note Cette fonction permet de ne pas avoir de tableau externe pour effectuer la conversion.
 * \note Aucun item ne peut tre vide, puisqu'il serait alors d'office terminateur.
 * \note tant donne la longueur possible de la chane \p pszList, on n'utilise
 * pas de caractres \p TCHAR.
 **/
LPCSTR EnumToString(int nVal, LPCSTR pszList, LPCSTR pszDefault="??");

/**
 * Fonction inverse de \p EnumToString : recherche d'une chane de caractres dans une liste.
 * La liste de chanes est fournie sous la forme d'une seule chane avec des \p \\0 
 * la fin de chaque item. Un double \p \\0 termine donc implicitement.
 * La comparaison n'est pas sensible  la casse.
 * \param[in] pszList	Chane contenant une suite d'items spars par des zros binaires ;
 *						un item vide (donc deux zros successifs) DOIT terminer.
 * \param[in] pszStr	Chane recherche.
 * \return	Index de la chane trouve, ou bien -1 si pas trouv.
 **/
int StringToEnum(LPCSTR pszList, LPCSTR pszStr);

#define		FF_RECURSIVE	0x01	//!< Recherche dans les sous-rpertoires
#define		FF_STOPATFIRST	0x02	//!< Arrt au premier rsultat trouv
#define		FF_DIRECTORY	0x04	//!< Recherche d'un rpertoire

// ----------------------------------------------------------------------------

/**
 * Recherche d'un fichier correspondant au nom pass en paramtre
 *
 * \param[in]	pszDir	Rpertoire initial pour la recherche
 * \param[in]	pszFile	Fichier, ou motif de recherche
 * \param[in]	dwFlags	Options de recherche (FF_RECURSIVE, FF_DIRECTORY)
 * \return	Nom du premier fichier trouv avec le chemin, ou une chane vide si pas trouv
 **/
tstring FindFile(LPCTSTR pszDir, LPCTSTR pszFile, DWORD dwFlags=FF_RECURSIVE);

/**
 * Recherche de tous les fichiers correspondant au motif pass en paramtre
 *
 * \param[in]	pszDir	Rpertoire initial pour la recherche
 * \param[in]	pszFile	Fichier, ou motif de recherche
 * \param[out]	vResult	Vecteur contenant les fichiers rpondant au nom pass en paramtre
 * \param[in]	dwFlags	Options de recherche (FF_RECURSIVE, FF_STOPATFIRST, FF_DIRECTORY)
 * \return	Le nombre de fichiers trouvs correspondant
 **/
UINT FindFile(LPCTSTR pszDir, LPCTSTR pszFile, std::vector<tstring> & vResult, DWORD dwFlags);

// ----------------------------------------------------------------------------

/**
 * Convertit une chane de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres UTF-8
 * \return     chane de caractres convertie
 **/
tstring ConvertUTF82Char(LPCSTR aconvertir);

/**
 * Convertir une chaine de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     chane de caractres convertie
 **/
std::string ConvertChar2UTF8(LPCWSTR aconvertir, bool for_sgml=false);

/**
 * Convertir une chaine de caractres en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     chane de caractres convertie
 **/
std::string ConvertChar2UTF8(LPCSTR aconvertir, bool for_sgml=false);

/**
 * Convertir un caractre en UTF-8
 *
 * \param[in]  aconvertir  chane de caractres contenant ventuellement des caractres devant tre encods en UTF-8
 * \param[out] convertis   chane de caractres convertie
 * \param[in]  for_sgml    si \p true, les caractres <, > et & sont convertis en entits SGML
 * \return     le nombre de caractres convertis
 * \note \p aconvertir est de type \p UINT32 car :
 * - TCHAR pose des problmes tant donn qu'il s'agit d'une valeur *signe*
 * - La fonction est capable de convertir des caractres cods sur plus de 16 bits
 * \note \p convertis doit pointer sur une variable de taille suffisante
 * pour tre capable de recevoir au moins 6 caractres (terminateur compris).
 **/
int ConvertChar2UTF8(UINT32 aconvertir, LPSTR convertis, bool for_sgml=false);

// ----------------------------------------------------------------------------

/**
 * Rcupration du menu parent ainsi que de la position d'un item de menu donn dans une arborescence de menus
 * \param[in]	hMenu	Handle de la racine de l'arborescence de menus contenant l'item dans un sous-menu quelconque
 * \param[in]	uIDItem	Identifiant de l'item de menu recherch
 * \param[out]	pPos	Si != NULL, la variable pointe reoit la position (base 0) de l'item de menu
 * \return		Handle du menu parent, ou bien NULL si l'item n'existe pas
 **/
HMENU GetMenuParentAndPos(HMENU hMenu, UINT uIDItem, INT * pPos = NULL);

/**
 * Cration d'un sous-menu en remplacement d'un autre, ou bien d'un item
 * \param[in]		hMenu	Handle du menu contenant l'item et/ou le sous-menu
 * \param[in, out]	nPos	Position du sous-menu, ou -1 si on rfrence un item.
 *							En sortie, reoit la position trouve si elle n'tait pas dfinie.
 * \param[in]		uIDItem	Item de menu rfrenc si la position est gale = -1
 * \return			Handle du nouveau sous-menu
 **/
HMENU CreatePopupMenuFromItem(HMENU hMenu, INT & nPos, UINT uIDItem);

/**
 * Ajout d'un item de menu avec icne optionnelle
 * \param[in]	hMenu	Handle du menu contenant l'item et/ou le sous-menu
 * \param[in]	uIDItem	Item de menu rfrenc si la position est gale = -1
 * \param[in]	pszName	Texte de l'item de menu
 * \param[in]	fState	tat
 * \param[in]	hImage	Si != NULL, icne associe
 * \return		Handle du nouveau sous-menu
 **/
bool AppendMenuItemExt(HMENU hMenu, UINT uIDItem, LPCTSTR pszName, UINT fState = MFS_ENABLED, HBITMAP hImage = NULL);

/**
 * Modification de l'tat d'un item de menu
 * \param[in]	hMenu		Menu dans lequel se trouve l'item  modifier.
 * \param[in]	uItem		Identificateur de l'item  modifier.
 * \param[in]	fStateMask	Masque indiquant l'tat  modifier.
 * \param[in]	fState		Nouvel tat.
 **/
BOOL SetMenuItemState(HMENU hMenu, UINT uItem, UINT fStateMask, UINT fState);

// ----------------------------------------------------------------------------

/**
 * Macro d'assistance  l'numration d'interface
 * \param[in]	parent	Nom de l'objet dont les lments doivent tre numrs
 * \param[in]	type	Type des objets numrs
 * \param[in]	name	Nom de l'objet recevant les instances d'objets numrs
 **/
#define ENUM_ITF(parent, type, name) \
	CComPtr<type> name; \
	(parent)->Next(1, &name, NULL)==S_OK; \
	name.Release()

/**
 * Macro d'assistance  l'numration d'lments de listes
 * \param[in]	cont	Type de contenant
 * \param[in]	parent	Nom de la liste dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE(cont, parent, type, name) \
	std::cont<type>::iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de listes constantes
 * \param[in]	cont	Type de contenant
 * \param[in]	parent	Nom de la liste dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_CONST(cont, parent, type, name) \
	std::cont<type>::const_iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration inverse d'lments de vecteurs
 * \param[in]	cont	Type de contenant
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define REV_ITERATE(cont, parent, type, name) \
	std::cont<type>::reverse_iterator name = (parent).rbegin(); \
	name != (parent).rend(); \
	name++

/**
 * Macro d'assistance  l'numration inverse d'lments de vecteurs constants
 * \param[in]	cont	Type de contenant
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define REV_ITERATE_CONST(cont, parent, type, name) \
	std::cont<type>::const_reverse_iterator name = (parent).rbegin(); \
	name != (parent).rend(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de vecteurs
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_VECTOR(parent, type, name) \
	std::vector<type>::iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de vecteurs constants
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_CONST_VECTOR(parent, type, name) \
	std::vector<type>::const_iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration inverse d'lments de vecteurs
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define REV_ITERATE_VECTOR(parent, type, name) \
	std::vector<type>::reverse_iterator name = (parent).rbegin(); \
	name != (parent).rend(); \
	name++

/**
 * Macro d'assistance  l'numration inverse d'lments de vecteurs constants
 * \param[in]	parent	Nom du vecteur dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define REV_ITERATE_CONST_VECTOR(parent, type, name) \
	std::vector<type>::const_reverse_iterator name = (parent).rbegin(); \
	name != (parent).rend(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de listes
 * \param[in]	parent	Nom de la liste dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_LIST(parent, type, name) \
	std::list<type>::iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/**
 * Macro d'assistance  l'numration d'lments de listes constantes
 * \param[in]	parent	Nom de la liste dont les lments doivent tre numrs
 * \param[in]	type	Type des lments numrs
 * \param[in]	name	Nom du pointeur sur les lments numrs
 **/
#define ITERATE_CONST_LIST(parent, type, name) \
	std::list<type>::const_iterator name = (parent).begin(); \
	name != (parent).end(); \
	name++

/// Recherche de la prsence d'un lment dans une liste ou un vecteur
template <class T, class U>
	bool is_inside(T & cont, const U & val)
{
	return std::find(cont.begin(), cont.end(), val) != cont.end();
}

// ----------------------------------------------------------------------------

typedef	TCHAR SZPathName[MAX_PATH];

/// Structure d'assistance  la cration de noms de fichiers situs dans le rpertoire %TEMP%
struct TempPathName
{
	SZPathName	szPath;

	/// Constructeur
	template <class T>
		TempPathName(const T * pszFileName)
	{
		GetTempPath(_countof(szPath), szPath);
		strcat_T(szPath, pszFileName);
	}

	/// Constructeur pour nom de fichier automatique
	TempPathName()
	{
		TCHAR szTmp[MAX_PATH];

		GetTempPath(_countof(szTmp), szTmp);
		GetTempFileName(szTmp, TEXT("ptv"), 0, szPath);
	}

	/// Rcupration du rsultat
	operator LPCTSTR() {
		return szPath;}
};

/**
 * Nettoyage des espaces et caractres spciaux (tels CR/LF)  la fin d'une chane de caractres
 * \param[in,out]	pszStr	Chane de caractres  nettoyer, et rsultat
 * \return	Pointeur sur cette mme chane
 **/
LPTSTR RTrimStr(LPTSTR pszStr);

/// Conversion d'un entier en chane de caractres, avec sparateur des milliers
tstring thousandsSepFmt(UINT64 nVal);

/**
 * Conversion d'un code d'erreur systme (retourn par \p GetLastError)
 * en une chane de caractres affichable.
 *
 * \param[in]	dwErr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 **/
tstring SystemErrorToString(DWORD dwErr = GetLastError());

// ----------------------------------------------------------------------------

class ByteVector : public std::vector<BYTE>
{
public:
	LPBYTE get() {
		return &front(); }
	const BYTE * get() const {
		return &front(); }
	operator LPCSTR() const {
		return reinterpret_cast<LPCSTR>(&front()); }

	// Ajout d'un zro terminateur (pour exploitation en tant que chane de type C)
	void AddStrZ() {
		size_type nSize = size();

		resize(nSize+1);
		get()[nSize] = 0; }

	/**
	 * Enregistrement du contenu du vecteur dans un fichier
	 *
	 * \param[in]	pszFileName	Nom du fichier  sauvegarder
	 * \param[in]	pSysTime	Date de modification  appliquer au fichier si != NULL
	 * \retval		0 si tout s'est bien droul
	 * \retval		Code d'erreur systme en cas d'erreur
	 **/
	DWORD SaveAsFile(LPCTSTR pszFileName, const SYSTEMTIME * pSysTime = NULL) const;
};

// ----------------------------------------------------------------------------

/// Classe permettant de rcuprer les fichiers inclus dans les resources de PTVM
class RawResource : public ByteVector
{
public:
	/**
	 * Extrait une resource du fichier. Charge galement la taille alloue.
	 * \param[in]	hInstance	Instance de l'application.
	 * \param[in]	idRessource	Identifiant de la resource.
	 * \param[in]	lpType		Type de la resource.
	 * \retval		\p true en cas de succs, \p false en cas d'chec.
	 **/
	bool Extraction(HINSTANCE hInstance, UINT idRessource, LPCTSTR lpType=RT_RCDATA);
};

// ----------------------------------------------------------------------------

/// Classe de gestion d'un mutex
class CSyncMutex
{
	HANDLE	hMutex;		//!< "Handle du mutex"
public:
	/**
	 * \brief Constructeur
	 * \param[in]	pszName	Nom optionnel du mutex
	 **/
	CSyncMutex(LPCTSTR pszName = NULL) :
		hMutex(CreateMutex(NULL, FALSE, pszName))
	{
	}

	/// Test si mutex dj existant (valide seulement immdiatement aprs la construction)
	bool	alreadyExists() const {
				return hMutex == NULL && (GetLastError()==ERROR_ALREADY_EXISTS); }

	/**
	 * \brief Verrouillage de l'accs aux ressources protges par le mutex
	 * \param[in]	dwMilliseconds	Temps d'attente (mS) au bout duquel le verrouillage est
	 *								considr comme impossible
	 * \return		\p true ou \p false selon succs ou chec (aprs expiration de temps d'attente)
	 **/
	bool	Lock(DWORD dwMilliseconds = 1000L) {
				return WaitForSingleObject(hMutex, dwMilliseconds)==WAIT_OBJECT_0; }

	/// Libration de l'accs aux ressources protges par le mutex
	void	Unlock() {
				ReleaseMutex(hMutex); }

	/// Destructeur
	~CSyncMutex() {
		CloseHandle(hMutex); hMutex = NULL; }
};

// ----------------------------------------------------------------------------

/// Classe de gestion d'un thread de travail
class WorkingThread
{
	HANDLE	hThread;
	DWORD	dwThreadId;
	bool	bDeleteOnExit;	//!< \p true si doit tre dtruit par \p delete

	virtual DWORD		ThreadProc() = 0;

	/**
	 * Point d'entre (statique) du thread
	 * \param[in]	lpParameter	Pointeur \p this de la classe associe
	 **/
	static DWORD WINAPI	ThreadProc_static(LPVOID lpParameter);

public:
	/**
	 * Constructeur
	 * \param[in]	bDelOnExit si \p true, l'objet sera dtruit par \p delete aprs sortie du thread
	 **/
	WorkingThread(bool bDelOnExit) :
		hThread(NULL),
		dwThreadId(0),
		bDeleteOnExit(bDelOnExit)
	{
	}

	/**
	 * Dmarrage du thread
	 * \return \p true si le thread a dmarr correctement, \p false sinon
	 **/
	bool	StartThread();

	/**
	 * Envoi d'un message dans la file d'attente de messages du thread
	 **/
	BOOL	PostMessageInThread(UINT nMsg, WPARAM wParam=0, LPARAM lParam=0) {
				return ::PostThreadMessage(dwThreadId, nMsg, wParam, lParam);}

	/**
	 * Attente de la terminaison du thread
	 * \param[in]	dwMilliseconds	Temps d'attente maximal
	 * \return \p true si le thread s'est termin correctement, \p false sinon
	 **/
	bool	WaitThreadExit(DWORD dwMilliseconds);

	/// Obtenir l'identifiant du thread
	DWORD	getThreadId() const {
				return dwThreadId; }

	/// Obtenir la handle du thread
	HANDLE	getThreadHdl() const {
				return hThread; }

	/// Test si le thread est actif
	bool	isThreadRunning() const {
				return hThread != NULL && WaitForSingleObject(hThread, 0) != WAIT_OBJECT_0; }

	/// Destructeur
	virtual ~WorkingThread();
};

// ----------------------------------------------------------------------------

/// Type des options de ligne de commande
enum CmdOptionType
{
	cot_BOOL,				//!< Boolen (\p bool, mis  true sur seule prsence de l'option)
	cot_INT,				//!< Entier (\p int)
	cot_STR					//!< Chane (\p tstring)
};

struct OneCmdOption
{
	LPCTSTR			pszStr;	//!< Nom de l'option
	CmdOptionType	eType;	//!< Type de l'option
	INT				nOffs;	//!< Offset dans la structure
};

/** \brief Analyse des options de ligne de commande
 *	\param[in]	pOptTable	Table des options reconnues
 *	\param[in]	nSize		Taille de la table
 *	\param[out]	pDest		Pointeur sur la structure recevant les valeurs
 *	\param[out]	pOthers		Si non \p NULL, tous les lments prsents sur la ligne de commande qui ne
 *							sont pas spcifis dans la table sont ajouts  ce vecteur.
 **/
void ProcessCmdLine(OneCmdOption * pOptTable, int nSize, void * pDest, vector<tstring> * pOthers = NULL);

// ----------------------------------------------------------------------------

/// Macro utilitaire pour faciliter l'numration de cls de registre dans une boucle "for"
#define ENUM_REG_KEYS(creg, str) \
	DWORD creg##_inx_=0; \
	creg.EnumKey(creg##_inx_, str); \
	++creg##_inx_

/// Classe "CRegKey" adapte  l'usage des types \p tstring
/// (exploration de la base de registre)
class CRegKey_tstr
{
	HKEY	hKey;
public:
	/// Constructeur
	CRegKey_tstr() :
		hKey(NULL)
	{
	}

	/// Ouverture d'une cl, dfinie par une cl racine et un chemin
	bool	Open(HKEY hParKey, LPCTSTR pszSubKey, REGSAM samDesired = KEY_READ);

	/// Lecture d'une valeur chane
	bool	QueryStringValue(LPCTSTR pszName, tstring & strRes) const;

	/// criture d'une valeur chane
	bool	SetStringValue(LPCTSTR pszName, LPCTSTR pszValue);

	/// Lecture de la valeur d'index \p nInx prsente sous une cl
	bool	EnumKey(DWORD nInx, tstring & strRes) const;

	bool	DeleteValue(LPCTSTR pszName)  {
				return RegDeleteValue(hKey, pszName) == ERROR_SUCCESS; }

	operator HKEY() {
				return hKey;}

	/// Destructeur
	~CRegKey_tstr() {
		if (hKey!=NULL) RegCloseKey(hKey);}
};

