/*
 * $Id: filters.h 270 2010-12-02 20:57:32Z gingko $
 * Copyright (C) 2009 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions pour l'assistance  l'insertion des filtres
 *
 *	Ce fichier dfinit des classes et fonctions facilitant la manipulation
 *	des filtres dans le graphe.
 **/

#pragma once

// ====================================================================================
// Structures, dfinitions et mthodes concernant les filtres
// ====================================================================================

/** \defgroup Interfaces Gestion d'interfaces en C++
	Macros permettant de grer la notion d'interfaces, en dpit
    du fait que le langage C++ ne fournit pas (encore) cette possibilit.
	Voir http://www.codeguru.com/cpp/cpp/cpp_mfc/oop/article.php/c9989
 * @{ */

#define _Interface_ class

/// Dclaration d'une interface de base
#define DeclareInterface(name) _Interface_ name { \
          public: \
          virtual ~name() {}

/// Dclaration d'une interface base sur une autre interface
#define DeclareBasedInterface(name, base) class name : \
        public base { \
           public: \
           virtual ~name() {}

/// Fin de dclaration d'interface
#define EndInterface };

/// Dfinition d'un "mot cl" plus spcifique
#define implements public

/** @} */

// ====================================================================================

typedef CUnknown * (WINAPI * CreateFilterProc)(IUnknown *, HRESULT *);

class IPTvM_GraphBase;
class CMediaRenderer;

/// Donnes relatives aux filtres de sortie
class FilterData
{
public:
	CComPtr<IBaseFilter>	pFilter;	//!< Interface du filtre
	LPCTSTR					pszName;	//!< Nom du filtre, pour affichage
	bool					bLocal;		//!< \p true si filtre local,  \p false si filtre externe
	union {
		const IID *			pclsid;		//!< CLSID du filtre si externe
		LPFNNewCOMObject	procCrea;	//!< Fonction de cration du filtre si interne
	};

	/// Constructeur pour filtre dj instanci
	FilterData(LPCTSTR pszNam=NULL, IBaseFilter * pFiltr=NULL) :
		pFilter(pFiltr),
		pszName(pszNam),
		bLocal(false),
		pclsid(NULL)
	{
	}

	/// Constructeur pour filtre externe (COM)
	FilterData(LPCTSTR pszNam, REFCLSID clsid) :
		pszName(pszNam),
		bLocal(false),
		pclsid(&clsid)
	{
	}

	/// Constructeur pour filtre interne
	FilterData(LPCTSTR pszNam, LPFNNewCOMObject proc) :
		pszName(pszNam),
		bLocal(proc!=NULL),
		procCrea(proc)
	{
	}

	/// Constructeur copie
	FilterData(const FilterData & sFltFrom) :
		pFilter(sFltFrom.pFilter),
		pszName(sFltFrom.pszName),
		bLocal(sFltFrom.bLocal),
		pclsid(sFltFrom.pclsid)
	{
	}

	/**
	 * \brief Cration du filtre
	 **/
	virtual HRESULT			Create();

	/// Cration et ajout du filtre.
	HRESULT					get_and_add(IPTvM_GraphBase & iGraph);

	/**
	 * \brief Appel du dialogue de proprits du filtre
	 **/
	VOID					create_property_dialog();

	/**
	 * \brief Retourne \p true si le filtre a t dclar comme "Aucun"
	 **/
	bool					isNull() const {
								return IsNullName(pszName); }

	/**
	 * \brief Vrification de l'existence d'un dialogue de proprits sur ce filtre
	 *
	 * \return	\p true si un dialogue de proprits existe
	 * \note	La mthode cherchera  crer une instance du filtre s'il n'y en a pas encore
	 **/
	bool					check_property_dialog();

	/**
	 * \brief Obtenir un pointeur sur le graphe dans lequel est insr un filtre
	 * \return Pointeur sur le graphe, ou bien NULL si pas de filtre ou filtre non insr
	 **/
	CComPtr<IFilterGraph>	graph_from_filter();

	void					Release() {
								pFilter.Release(); }

	virtual ~FilterData() {
				}
};

typedef shared_ptr<FilterData>	FilterDataPtr;

class DemuxOutPin
{
public:
	CComPtr<IPin>			pPin;		//!< Broche de sortie du dmultiplexeur
	CComPtr<IMPEG2PIDMap>	pMap;		//!< Interface de la broche de sortie du dmultiplexeur
	const LPCWSTR			pszName;	//!< Nom de la broche
	MEDIA_SAMPLE_CONTENT	eMsc;		//!< Type de contenu

	/// Constructeur
	DemuxOutPin(LPCWSTR pszName, MEDIA_SAMPLE_CONTENT eMs=MEDIA_ELEMENTARY_STREAM) :
		pszName(pszName),
		eMsc(eMs)
	{
	}

	/// Cration d'une broche de sortie du dmultiplexeur
	HRESULT					CreatePin(IBaseFilter * pDemuxFilter, const AM_MEDIA_TYPE & sAm, bool bFindExisting);

	/**
	 * Branchement d'un flux lmentaire (identifi par PID) sur la connexion \p pMap
	 * (en sortie du dmultiplexeur), en dbranchant tous les autres ventuels
	 * \param[in]	pid		N de PID  y connecter, ou 0 si aucun
	 * \param[in]	bKeep	Ne dbranche PAS les autres PIDs si \p true
	 * \return	Code d'erreur DirectShow
	 * \note	Utiliser 0 comme paramtre \p pid pour tout dbrancher
	 **/
	HRESULT					Map(ULONG pid, bool bKeep = false) const;

	/// Dbranchement de tous les flux lmentaires de cette broche
	HRESULT					Unmap() const {
								return Map(0); }

	/// Fonction de slection du rendu appropri, s'il y a lieu
	virtual HRESULT			SelectRenderer(IPTvM_GraphBase & iGraph, bool bNoRendering) {
								return S_OK; }

	/// Obtenir un pointeur sur le filtre de rendu
	virtual CMediaRenderer* GetRenderer() {
								return NULL; }
};

/// Donnes relatives aux filtres connects  une broche de sortie du dmultiplexeur
class DecoderData : public FilterData
{
public:
	DemuxOutPin &			cDemuxOut;	//!< Donnes d'une sortie du dmultiplexeur
	const AM_MEDIA_TYPE *	pAmt;		//!< Descripteur du mdia
	const LPCSTR			pszType;	//!< Type de codec, pour affichage

	/// Constructeur pour filtre externe (COM)
	DecoderData(DemuxOutPin & cDmxOut, const AM_MEDIA_TYPE * pAm, LPCSTR pszTyp,
				LPCTSTR pszNam, REFCLSID clsid) :
		FilterData(pszNam, clsid),
		cDemuxOut(cDmxOut),
		pAmt(pAm),
		pszType(pszTyp)
	{
	}

	/// Constructeur pour filtre interne
	DecoderData(DemuxOutPin & cDmxOut, const AM_MEDIA_TYPE * pAm, LPCSTR pszTyp,
				LPCTSTR pszNam, LPFNNewCOMObject proc=NULL) :
		FilterData(pszNam, proc),
		cDemuxOut(cDmxOut),
		pAmt(pAm),
		pszType(pszTyp)
	{
	}

	HRESULT	render(IPTvM_GraphBase & iGraph, FilterData * pFilterOut=NULL);
};

#define UNDEF_CODEC_PTR reinterpret_cast<DecoderData *>(-1)	// pour indiquer codec indfini

class MediaOutPin : public DemuxOutPin
{
public:
	DecoderData *				pCurCodec;	//!< Codec prcdemment associ
	shared_ptr<CMediaRenderer>	pRenderer;	//!< Rendu de mdia associ

	/// Constructeur
	MediaOutPin(LPCWSTR pszName) :
		DemuxOutPin(pszName),
		pCurCodec(UNDEF_CODEC_PTR)
	{
	}

	/// Obtenir un pointeur sur le filtre de rendu
	virtual CMediaRenderer *	GetRenderer() {
									return &*pRenderer; }

	/// Tester si l'application d'un codec et d'un rendu ncessitent un changement du graphe
	bool						NeedGraphChange(const DecoderData * pNewCodec, RdrTypes eNewVmrMode) const;
};

class VideoOutPin : public MediaOutPin
{
public:
	/// Constructeur
	VideoOutPin(LPCWSTR pszName) :
		MediaOutPin(pszName)
	{
	}

	/**
	 * Slection, avec cration ventuelle, d'un rendu vido
	 * \param[in]	iGraph			Graphe dans lequel slectionner le rendu vido
	 * \param[in]	bNoRendering	Si \p true, remplacer par un non rendu vido (croix bleue)
	 * \return		Code d'erreur DirectShow
	 **/
	virtual HRESULT	SelectRenderer(IPTvM_GraphBase & iGraph, bool bNoRendering);

	virtual ~VideoOutPin();
};

class AudioOutPin : public MediaOutPin
{
public:
	/// Constructeur
	AudioOutPin(LPCWSTR pszName) :
		MediaOutPin(pszName)
	{
	}

	/**
	 * Slection, avec cration ventuelle, d'un rendu audio
	 * \param[in]	iGraph			Graphe dans lequel slectionner le rendu audio
	 * \param[in]	bNoRendering	Si \p true, remplacer par un rendu audio NULL
	 * \return		Code d'erreur DirectShow
	 **/
	virtual HRESULT	SelectRenderer(IPTvM_GraphBase & iGraph, bool bNoRendering);
};

class MediaDecoderData : public DecoderData
{
	/**
	 * \brief Cration du filtre
	 **/
	virtual HRESULT			Create();
public:
	/// Constructeur
	MediaDecoderData(DemuxOutPin & cDmxOut, const AM_MEDIA_TYPE * pAm, LPCSTR pszTyp, LPCTSTR pszNam) :
		DecoderData(cDmxOut, pAm, pszTyp, pszNam)
	{
	}

	HRESULT render(IPTvM_GraphBase & iGraph);
};

/// Donnes d'accs au "grabber"
struct GrabberData
{
	CComPtr<ISampleGrabber>		pInterface;
	CComPtr<ISampleGrabberCB>	pCallback;

	void Release() {
		pCallback.Release();
		pInterface.Release(); }

	HRESULT SetCallback() {
		return pInterface->SetCallback(pCallback, 0);
	}

	HRESULT ResetCallback() {
		return pInterface->SetCallback(NULL, 0);
	}
};

// ====================================================================================
// Assistance  la recherche de filtres
// ====================================================================================

/// Classe d'assistance  l'interprtation des rsultats d'une recherche de filtres
class MonikerBindings
{
	/// Rcupration \p IPropertyBag et nom du filtre
	HRESULT					BindProperties();

public:
	CComPtr<IMoniker>		pMoniker;	//!< Pointeur vers l'interface "Moniker"
	CComPtr<IBindCtx>		pBindCtx;	//!< Pointeur vers l'interface "Bind context"
	CComPtr<IPropertyBag>	pPropBag;	//!< Pointeur vers l'interface "Property bag"
	CComPtr<IBaseFilter>	pFilter;	//!< Pointeur vers l'interface du filtre, si requis
	tstring					strName;	//!< Nom du filtre	

	/**
	 * Constructeur  partir d'une numration.
	 * \param[in]	pMonikr	Pointeur vers l'interface "Moniker" numre
	 * \param[in]	phr		Pointeur sur variable o mettre le code retour si non NULL
	 **/
	MonikerBindings(IMoniker * pMonikr, HRESULT * phr = NULL);

	/**
	 * Constructeur  partir d'un "device path".
	 * \param[in]	pszDevicePath	Chane de caractres "device path"
	 * \param[in]	phr				Pointeur sur variable o mettre le code retour si non NULL
	 **/
	MonikerBindings(LPCTSTR pszDevicePath, HRESULT * phr = NULL);

	/**
	 * Charger l'interface du filtre dans \p pFilter (non ralis  la construction car pas
	 * toujours requis et parfois gnant).
	 * \return	Code d'erreur DirectShow
	 **/
	HRESULT BindToFilter();

	/**
	 * Rcuprer le "device path" dans une variable \p tstring. Le "device path" contient  lui seul
	 * toutes les informations ncessaires pour identifier un filtre existant de faon unique.
	 * \param[out]	strDevicePath	"Device path" rcupr
	 * \return	Code d'erreur DirectShow
	 **/
	HRESULT GetDevicePath(tstring & strDevicePath);

	/// Test validit des donnes
	bool isOk() const {
		return pPropBag && !strName.empty();}
};

// ====================================================================================
// Recherche de broche dans un filtre
// ====================================================================================

/// Paramtre possible de FilterData::cherche_pin
#define NOTCONNECTED	((AM_MEDIA_TYPE *)-1)

/**
 * Vrification de la compatibilit d'une broche avec un type de mdia donn.
 * \param[in]	pPin	Interface de la broche  tester
 * \param[in]	am		Descripteur de type de mdia
 * \retval \p true en cas de succs
 **/
bool check_pin_for_media(CComPtr<IPin> & pPin, const AM_MEDIA_TYPE & am);

/**
 * Recherche de la premire broche d'un filtre, pour une direction donne.
 * \param[in]	pFilter	Interface du filtre dans lequel la recherche s'effectue
 * \param[in]	dir		Direction de la broche recherche (entre ou sortie)
 * \param[out]	result	Interface de la broche trouve
 * \param[in]	pmt		Si != \p NULL, alors la broche doit tre non connecte, et si la valeur
 *						est diffrente de NOTCONNECTED, elle doit en plus accepter le mdia spcifi
 * \retval \p S_OK	en cas de succs
 **/
HRESULT cherche_pin(IBaseFilter * pFilter, PIN_DIRECTION dir,
					CComPtr<IPin> & result, const AM_MEDIA_TYPE * pmt=NULL);

