/*
 * $Id: grabber.cpp 265 2010-11-10 15:26:43Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation de l'interface "callback" du "Sample Grabber"
 *
 *	Ce fichier implmente l'interface "callback" qui se connecte au filtre
 *	\p DirectShow "Sample Grabber", utilis pour rcuprer les chantillons
 *	du flux TS destins aux enregistrements programms.
 **/

#include "stdafx.h"

#include "grabber.h"
#include "record.h"
#include "memfifo.h"

/// Constructeur de base
CSampleGrabber::CSampleGrabber() :
	WorkingThread(false),
	CUnknown(TEXT("Sample grabber"), NULL),
	taille_paquet_buf(0)
{
	myprintf(TEXT("## Construction CSampleGrabber") EOL);
}

STDMETHODIMP CSampleGrabber::SampleCB(double SampleTime, IMediaSample *pSample)	// virtual
{

	long	l = pSample->GetActualDataLength();

	// (note : pSample->IsDiscontinuity ne semble pas donner d'informations significatives sur les
	// tuners tests, mais qui sait d'autres tuners donneront peut-tre des rsultats meilleurs)
	myprintf(CMyP TEXT("SampleCB : Discontinuit") EOL, pSample->IsDiscontinuity()==S_OK);

	// myprintf(TEXT("Echantillon SampleCB, t=%f, taille %u (%u paquets, frag=%u)") EOL, SampleTime, l, l/TS_SIZE, l%TS_SIZE);

	if (l>0) {
		BYTE *	p;

		pSample->GetPointer(&p);
		cMemFifo.Put(l, p);
	}
	return S_OK;
}

/**
 * Traitement d'un chantillon reu du tuner
 * param[in]	nSize	Taille de l'chantillon
 * param[in]	pPtr	Pointeur sur les donnes de l'chantillon
 **/
void CSampleGrabber::traite_sample(size_t nSize, const BYTE * pPtr)
{
	UINT32 i = 0;

	// Rcupration du fragment de paquet ventuellement reu  la fin de l'appel prcdent :
	if (taille_paquet_buf) {
		i = TS_SIZE - taille_paquet_buf;

		memcpy(paquet_buf.bytes+taille_paquet_buf, pPtr, i);
		traite_paquet(paquet_buf);
		taille_paquet_buf = 0;
	}

	// Traitement de tous les paquets entiers reus ici :
	while (i + TS_SIZE <= nSize){

		if (reinterpret_cast<const TS_packet *>(pPtr+i)->isTS(i+TS_SIZE >= nSize)) {
			traite_paquet(*reinterpret_cast<const TS_packet *>(pPtr+i));
			i += TS_SIZE;
		} else {
			// cherche prochaine synchro
			BYTE * res = (BYTE *)memchr(&pPtr[i+1], TS_SYNC, nSize-i-1);

			if (res == NULL) {
				// rien trouv, plus rien a faire
				return;
			}

			// myprintf(TEXT("trouve index @ %i") EOL, i);

			i = (long)(res - pPtr);
		}
	}

	// Report du fragment rsiduel  la fin pour l'appel suivant :
	if (i != nSize) {
		taille_paquet_buf = BYTE(nSize - i);
		memcpy(paquet_buf.bytes, pPtr+i, taille_paquet_buf);
	}
}

/**
 * Traitement d'un paquet TS reu du tuner (issu du dcoupage de l'chantillon, et mme
 * ventuellement d'une reconstitution de celui-ci).
 *
 * param[in]	p		Rfrence sur le paquet reu (qui a une longueur fixe)
 **/
void CSampleGrabber::traite_paquet(const TS_packet & p)	// virtual
{
	if (cMutex.Lock()) {
		for (int nInx=0; nInx < enregistrements_actuels.count(); ++nInx) {
			Enregistrement &	sRec	= enregistrements_actuels[nInx];
			Mpeg_TS_InpPtr		pCapt	= sRec.pCapture;

			if (pCapt) {
				DWORD			nErr	= pCapt->traite_paquet(p);

				if (nErr != 0 && sRec.nErrCode==0) {
					sRec.nErrCode = nErr;
					myprintf(ERRX(TEXT("## Erreur criture fichier, code=0x%08x")), nErr);
					PostMessage(hMainWnd, WM_COMMAND, IDM_STOP_RECORD_BASE+nInx, epr_fileerr);
				}
			}
		}
		cMutex.Unlock();
	}
}

DWORD CSampleGrabber::ThreadProc()
{
	myprintf(TEXT("## Entre dans CSampleGrabber::GrabThread") EOL);

	// viter que l'ordinateur passe en mode veille par inactivit pendant qu'un enregistrement est en cours
	EXECUTION_STATE	dwExecState = SetThreadExecutionState(ES_CONTINUOUS|ES_SYSTEM_REQUIRED|ES_AWAYMODE_REQUIRED);

	const CMemFifoBlock * pBlock;

	while ((pBlock = cMemFifo.WaitGet())!=NULL) {
		const BYTE *	pPtr	= pBlock->ptr();
		size_t			nSize	= pBlock->size();

		if (nSize)
			traite_sample(nSize, pPtr);

		delete	pBlock;
		if (nSize==0)
			break;
	}

	// Rtablir la gestion du mode veille
	SetThreadExecutionState(dwExecState);

	myprintf(TEXT("## Sortie de CSampleGrabber::GrabThread") EOL);
	return 0;
}

/// Arrt du "thread" de capture
bool CSampleGrabber::StopThread()
{
	if (isThreadRunning()) {
		// Envoyer un chantillon vide pour terminer le "Thread"
		cMemFifo.Put(NULL, 0);

		// Attendre une seconde
		if (!WaitThreadExit(1000)) {
			myprintf(TEXT("Le thread du grabber ne s'est pas termin correctement") EOL);
			return false;
		}
	}
	cMemFifo.Free();	// Libration des chantillons mmoire rsiduels
	return true;
}

/// Destructeur gnral
CSampleGrabber::~CSampleGrabber()
{
	StopThread();	// S'assurer de l'arrt du "thread"

	myprintf(TEXT("## Destruction CSampleGrabber") EOL);
}
