/*
 * $Id: graph.cpp 276 2011-03-06 10:17:01Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation des fonctionnalits ncessaires  la contruction du graphe
 *
 *	Ce fichier implmente toutes les classes et fonctions ncessaires  la construction
 *	et  la gestion des graphes \p DirectShow grs par Pouchin TV Mod, dans toutes leurs
 *	variations.
 **/

#include "stdafx.h"

#include "Filters\Misc\network.h"
#include "Filters\Misc\epgfilter.h"

#include "search.h"
#include "record.h"

#include "grabber.h"

#include <Mferror.h>

#include <dxerr.h>
#pragma comment (lib, "dxerr.lib")

// ====================================================================================
// GUIDs
// ====================================================================================

/// CLSID pour E-AC3
// A7FB87AF-2D02-42FB-A4D4-05CD93843BDD
CLSID MEDIASUBTYPE_DOLBY_DDPLUS =
	{0xA7FB87AF, 0x2D02, 0x42FB, {0xA4, 0xD4, 0x05, 0xCD, 0x93, 0x84, 0x3B, 0xDD}};

/// CLSID pour TIF
// FC772AB0-0C7F-11D3-8FF2-00A0C9224CF4
static CLSID CLSID_BDA_MPEG2_TIF =
	{0xFC772AB0, 0x0C7F, 0x11D3, {0x8F, 0xF2, 0x00, 0xA0, 0xC9, 0x22, 0x4C, 0xF4}};


// ====================================================================================
// Descripteurs de mdias
// ====================================================================================

#define DIMH_MPG2 720
#define DIMV_MPG2 576
#define	AR_H_MPG2 4
#define AR_V_MPG2 3

/// Format, Vido MPEG2
static const MPEG2VIDEOINFO sMp2v_fmt = {
	{								// hdr
		{0,0,DIMH_MPG2,DIMV_MPG2},	// rcSource
		{0,0,0,0},					// rcTarget
		0,							// dwBitRate
		0,							// dwBitErrorRate
		0,							// AvgTimePerFrame
		0,							// dwInterlaceFlags
		0,							// dwCopyProtectFlags
		AR_H_MPG2,					// dwPictAspectRatioX
		AR_V_MPG2,					// dwPictAspectRatioY
		{0},						// dwControlFlag & dwReserved1
		0,							// dwReserved2
		{							// bmiHeader
			sizeof(BITMAPINFOHEADER),// biSize
			DIMH_MPG2,				// biWidth
			DIMV_MPG2				// biHeight
		}
		// le reste  zro (implicite)
	}
};

/// Media type, Vido MPEG2
static const AM_MEDIA_TYPE sMp2v_am = {
	MEDIATYPE_Video,				// majortype
	MEDIASUBTYPE_MPEG2_VIDEO,		// subtype
	FALSE,							// bFixedSizeSamples
	TRUE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_MPEG2Video,				// formattype
	NULL,							// pUnk
	sizeof(sMp2v_fmt),				// cbFormat
	(LPBYTE)&sMp2v_fmt				// pbFormat
};

#define FCC_AVC1 MAKEFOURCC('A', 'V', 'C', '1')	// Possible aussi ...
#define FCC_h264 MAKEFOURCC('h', '2', '6', '4')
#define DIMH_H264 720
#define DIMV_H264 576
#define	AR_H_H264 4
#define AR_V_H264 3

/// Format, Vido H264
static const VIDEOINFOHEADER2 sH264_fmt = {
	{0,0,DIMH_H264,DIMV_H264},		// rcSource
	{0,0,0,0},						// rcTarget
	0,								// dwBitRate,
	0,								// dwBitErrorRate
	0,								// AvgTimePerFrame
	0,								// dwInterlaceFlags
	0,								// dwCopyProtectFlags
	AR_H_H264,						// dwPictAspectRatioX
	AR_V_H264,						// dwPictAspectRatioY
	{0},							// dwControlFlag & dwReserved1
	0,								// dwReserved2
	{								// bmiHeader
		sizeof(BITMAPINFOHEADER),	// biSize
		DIMH_H264,					// biWidth
		DIMV_H264,					// biHeight
		0,							// biPlanes
		0,							// biBitCount
		FCC_h264					// biCompression ( FCC_AVC1 ? )
	}
	// le reste  zro (implicite)
};

/// Media type, Vido H264
static const AM_MEDIA_TYPE sH264_am = {
	MEDIATYPE_Video,				// majortype
	MEDIASUBTYPE_H264,				// subtype
	FALSE,							// bFixedSizeSamples
	TRUE,							// bTemporalCompression
	1,								// lSampleSize
	FORMAT_VideoInfo2,				// formattype
	NULL,							// pUnk
	sizeof(sH264_fmt),				// cbFormat
	(LPBYTE)&sH264_fmt				// pbFormat
};

/// Format, Audio (commun)
static const WAVEFORMATEX sAudioFmt = {
	WAVE_FORMAT_PCM,				// wFormatTag
	2,								// nChannels
	48000,							// nSamplesPerSec
	4*48000,						// nAvgBytesPerSec
	4,								// nBlockAlign
	16,								// wBitsPerSample
	0								// cbSize
};

/// Media type, Audio MPEG2
static const AM_MEDIA_TYPE sMp2a_am = {
	MEDIATYPE_Audio,				// majortype
	MEDIASUBTYPE_MPEG2_AUDIO,		// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_WaveFormatEx,			// formattype
	NULL,							// pUnk
	sizeof(sAudioFmt),				// cbFormat
	(LPBYTE)&sAudioFmt				// pbFormat
};

/// Media type, Audio AC3
static const AM_MEDIA_TYPE sAc3__am = {
	MEDIATYPE_Audio,				// majortype
	MEDIASUBTYPE_DOLBY_AC3,			// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_WaveFormatEx,			// formattype
	NULL,							// pUnk
	sizeof(sAudioFmt),				// cbFormat
	(LPBYTE)&sAudioFmt,				// pbFormat
};

/// Media type, Audio E-AC3
static const AM_MEDIA_TYPE sEac3_am = {
	MEDIATYPE_Audio,				// majortype
	MEDIASUBTYPE_DOLBY_DDPLUS,		// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_WaveFormatEx,			// formattype
	NULL,							// pUnk
	sizeof(sAudioFmt),				// cbFormat
	(LPBYTE)&sAudioFmt,				// pbFormat
};

/// Media type, PSI
static const AM_MEDIA_TYPE sPsi__am = {
	MEDIATYPE_MPEG2_SECTIONS,		// majortype
	MEDIASUBTYPE_MPEG2DATA,			// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_None,					// formattype
	NULL,							// pUnk
	0,								// cbFormat
	NULL							// pbFormat
};

/// Media type, TIF
static const AM_MEDIA_TYPE sTif__am = {
	MEDIATYPE_MPEG2_SECTIONS,		// majortype
	MEDIASUBTYPE_DVB_SI,			// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_None,					// formattype
	NULL,							// pUnk
	0,								// cbFormat
	NULL							// pbFormat
};

/// Media type, EPG
static const AM_MEDIA_TYPE sEpg__am = {
	MEDIATYPE_MPEG2_SECTIONS,		// majortype
	MEDIASUBTYPE_DVB_SI,			// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_None,					// formattype
	NULL,							// pUnk
	0,								// cbFormat
	NULL,							// pbFormat
};

/// Media type, PMT
static const AM_MEDIA_TYPE sPmt__am = {
	MEDIATYPE_MPEG2_SECTIONS,		// majortype
	MEDIASUBTYPE_DVB_SI,			// subtype
	TRUE,							// bFixedSizeSamples
	FALSE,							// bTemporalCompression
	0,								// lSampleSize
	FORMAT_None,					// formattype
	NULL,							// pUnk
	0,								// cbFormat
	NULL							// pbFormat
};

// ====================================================================================
/// Pointeur d'accs global au graphe
// ====================================================================================

IPTvM_FullGraph *		pGraph = NULL;

// ====================================================================================
// Exportation du graphe
// ====================================================================================

#if EXPORT_GRAPH

class CExportToRot
{
	DWORD pdwRegister;
public:

	/// Constructeur
	CExportToRot() :
		pdwRegister(0)
	{
	}

	HRESULT Add(IGraphBuilder * pGrph)
	{
		CComPtr<IMoniker>				pMoniker;
		CComPtr<IRunningObjectTable>	pROT;
		OLECHAR							wsz[256];

		if (FAILED(GetRunningObjectTable(0, &pROT)))
			return E_FAIL;

		swprintf_s(wsz, L"FilterGraph %08x pid %08x",	// TCHAR impossible ici
			(DWORD_PTR)pGrph, GetCurrentProcessId());

		HRESULT hr = CreateItemMoniker(L"!", wsz, &pMoniker);

		if (SUCCEEDED(hr))
			hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pGrph,
								pMoniker, &pdwRegister);

		return hr;
	}

	void Remove()
	{
		CComPtr<IRunningObjectTable> pROT;

		if (SUCCEEDED(GetRunningObjectTable(0, &pROT)))
			pROT->Revoke(pdwRegister);
	}

	/// Destructeur
	~CExportToRot()
	{
		if (pdwRegister)
			Remove();
	}
};

#endif // #if EXPORT_GRAPH

// ====================================================================================
/// "Tuning space" et "tune requests"
// ====================================================================================

class CDVBTTuningSpace
{
	friend class CDVBTTuneRequest;

	CComPtr<ITuningSpace>	pTuningSpace;
public:
	CDVBTTuningSpace(HRESULT * phr = NULL);

	HRESULT	GetNetworkType(CLSID & clsid) const {
		return pTuningSpace ? pTuningSpace->get__NetworkType(&clsid) : E_NOINTERFACE;}
};

CDVBTTuningSpace::CDVBTTuningSpace(HRESULT * phr)
{
	HRESULT hr = pTuningSpace.CoCreateInstance(__uuidof(DVBTuningSpace));

	if (SUCCEEDED(hr)) {
		CComQIPtr<IDVBTuningSpace2> pDVBTuningSpace(pTuningSpace);

		if (pDVBTuningSpace) {
			hr = pDVBTuningSpace->put_SystemType(DVB_Terrestrial);
			if (SUCCEEDED(hr))
				hr = pDVBTuningSpace->put__NetworkType(DVB_TERRESTRIAL_TV_NETWORK_TYPE);
		} else
			hr = E_NOINTERFACE;
	}

	if (phr)
		*phr = hr;
}

class CDVBTTuneRequest
{
	CDVBTTuningSpace &		cTuningSpace;
	CComPtr<ITuneRequest>	pTuneRequest;
public:
	CDVBTTuneRequest(CDVBTTuningSpace & cTunSpace, long frequency, long bandwidth, long ONID, long TSID, long SID, HRESULT * phr = NULL);

	operator ITuneRequest * () const {
							return pTuneRequest;}
};

CDVBTTuneRequest::CDVBTTuneRequest(CDVBTTuningSpace & cTunSpace, long frequency, long bandwidth, long ONID, long TSID, long SID, HRESULT * phr) :
	cTuningSpace(cTunSpace)
{
	HRESULT hr = S_OK;

	if (cTuningSpace.pTuningSpace) {
		//Get an interface to the TuningSpace
		CComQIPtr<IDVBTuningSpace2> pDVBTuningSpace(cTuningSpace.pTuningSpace);

		if (pDVBTuningSpace) {
			//Get new TuneRequest from tuning space
			hr = pDVBTuningSpace->CreateTuneRequest(&pTuneRequest);
			if (!check_hr(hr, "CreateTuneRequest")) {
				//Get interface to the TuneRequest
				CComQIPtr<IDVBTuneRequest> pDVBTuneRequest(pTuneRequest);

				if (pDVBTuneRequest) {
					CComPtr<IDVBTLocator> pDVBTLocator;

					hr = pDVBTLocator.CoCreateInstance(__uuidof(DVBTLocator));
					if (!check_hr(hr, "pDVBTLocator.CoCreateInstance")) {
						hr = pDVBTLocator->put_CarrierFrequency(frequency);
						if (!check_hr(hr, "put_CarrierFrequency")) {
							hr = pDVBTLocator->put_Bandwidth(bandwidth);
							if (!check_hr(hr, "put_Bandwidth")) {
								// Bind the locator to the tune request.
								hr = pDVBTuneRequest->put_Locator(pDVBTLocator);
								if (!check_hr(hr, "put_Locator")) {
									hr = pDVBTuneRequest->put_ONID(ONID);
									check_hr(hr, "put_ONID");
								}
								if (SUCCEEDED(hr)) {
									hr = pDVBTuneRequest->put_TSID(TSID);
									check_hr(hr, "put_TSID");
								}
								if (SUCCEEDED(hr)) {
									hr = pDVBTuneRequest->put_SID(SID);
									check_hr(hr, "put_SID");
								}
							}
						}
					}
				} else
					hr =E_NOINTERFACE;
			}
		} else
			hr = E_NOINTERFACE;
	} else
		hr = E_INVALIDARG;

	if (phr)
		*phr = hr;
}

// ====================================================================================
/// Classe de base de construction du graphe
// ====================================================================================

class CPTvM_GraphBase : public virtual IPTvM_GraphBase
{
	bool			bRunning;	//!< \p true si le graphe est en route
protected:
	CComPtr<IFilterGraph2>
					pFGrph;		//!< Interface d'accs au graphe

  #if LOG_DSHOW
	HANDLE			logFile;

	HRESULT			AddLog(LPCTSTR pszLogBaseFileName);
  #endif // #if LOG_DSHOW

  #if EXPORT_GRAPH
	CExportToRot	cExport;

	/// Exportation du graphe
	HRESULT			AddExport();
  #endif // #if EXPORT_GRAPH

	CComPtr<IMediaEventEx>
					pEvents;

	/**
	 * Traitement d'un vnement du graphe
	 * \param[in] hWnd Handle de la fentre principale
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	HandleGraphEvent(HWND hWnd) {
						return S_OK;}

public:
	/// Constructeur
	CPTvM_GraphBase();

	/**
	 * Programmation d'une notification des vnements du graphe
	 * \param[in] hWnd Handle de la fentre principale
	 * \param[in] uMsg Identifiant du message  transmettre
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	AddEventNotifier(HWND hWnd, UINT uMsg);

	/**
	 * Ajout d'un filtre au graphe
	 *
	 * \param[in]	pFilter	Filtre	 ajouter
	 * \param[in]	pszName	Nom de ce filtre
	 **/
	virtual HRESULT	AddFilterToGraph(IBaseFilter * pFilter, LPCTSTR pszName);

	/**
	 * Cration et ajout d'une instance de filtre dfini par son "clsid" (le paramtre "nom"
	 * n'est ici qu'informatif,  l'usage des messages d'erreur et de l'exportation du graphe).
	 * \param[in,out]	sFData	Donnes d'aprs lesquelles crer le filtre
	**/
	virtual HRESULT	create_and_add_filter(FilterData & sFData) {
						return sFData.get_and_add(*this); }

	/**
	 * Connexion d'une broche de sortie  la premire broche d'entre disponible d'un filtre.
	 * \param[in] pOutPin	Broche de sortie  connecter
	 * \param[in] pFilter	Filtre prsentant la broche d'entre  connecter
	 * \return Code d'erreur DirectShow
	 **/
	HRESULT			connect_pin_to_filter(IPin * pOutPin, IBaseFilter * pFilter);

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	HRESULT			connect_pin_to_filter_via_filter(IPin * pOutPin, CComPtr<IBaseFilter> & pFilter, FilterData sVia);
#endif

	/**
	 * Connexion de la premire broche de sortie d'un filtre  la premire broche
	 * d'entre d'un autre.
	 * \param[in] source	Filtre prsentant la broche de sortie  connecter
	 * \param[in] dest		Filtre prsentant la broche d'entre  connecter
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	connect_filters(IBaseFilter * source, IBaseFilter * dest);

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/**
	 * Dconnexion d'une broche de la broche correspondante du filtre adjacent.
	 * \param[in]	pPin	Broche  dconnecter
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	disconnect_pin(IPin * pPin);
#endif

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/**
	 * Dconnexion d'une broche d'un filtre de la broche correspondante du filtre adjacent.
	 * \param[in]	pFilter	Filtre prsentant la broche  dconnecter
	 * \param[in]	dir		Direction de la broche  dconnecter (entre ou sortie)
	 * \return Code d'erreur DirectShow
	 **/
	 virtual HRESULT	disconnect_filter(IBaseFilter * pFilter, PIN_DIRECTION dir);
#endif

	/**
	 * \brief Retrait d'un filtre du graphe.
	 * \param[in]	pFilter	Filtre  retirer
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	remove_filter(IBaseFilter * pFilter);

	/**
	 * \brief	Mise en route du graphe
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT	graph_Run();

	/**
	 * Mise en pause du graphe
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	graph_Pause();

	/**
	 * Arrt du graphe
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT	graph_Stop();

	/**
	 * Interrogation de l'tat du graphe
	 * \return Valeurs de l'numration FILTER_STATE (strmif.h)
	 * \retval 0 (State_Stopped) Le graphe est arrt
	 * \retval 1 (State_Paused) Le graphe est en mode "pause"
	 * \retval 2 (State_Running) Le graphe est en marche
	 * \retval -1 Le graphe ne s'est pas initialis correctement
	 **/
	virtual OAFilterState	graph_GetState();

	/**
	 * Vrifie si le graphe est en route
	 * \return \p true (en route), ou \p false (arrte)
	 * \note	Se base sur une variable interne  la classe, alors que
				\p graph_GetState interroge le graphe lui-mme.
	 **/
	virtual bool			graph_isRunning() const {
								return bRunning;}

	/// Construction du graphe
	virtual HRESULT			Build();

  #ifdef _DIAGNOSTIC
	virtual tstring			get_graph_state();
  #endif

	/// Destructeur
	virtual ~CPTvM_GraphBase();
};

/// Constructeur
CPTvM_GraphBase::CPTvM_GraphBase() :
	bRunning(false)
  #if LOG_DSHOW
	, logFile(INVALID_HANDLE_VALUE)
  #endif // #if LOG_DSHOW
{
}

#if LOG_DSHOW
HRESULT	CPTvM_GraphBase::AddLog(LPCTSTR pszLogBaseFileName)
{
	TempPathName sLogFileName(pszLogBaseFileName);

	logFile = CreateFile(sLogFileName, GENERIC_WRITE, FILE_SHARE_READ,
				NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (logFile == INVALID_HANDLE_VALUE) {
		myprintf(ERRX(TEXT("chec CPTvM_GraphBase::AddLog : %s")), SystemErrorToString().c_str());
		return E_FAIL;
	}
	return pFGrph->SetLogFile((DWORD_PTR)logFile);
}
#endif // #if LOG_DSHOW

#if EXPORT_GRAPH
HRESULT	CPTvM_GraphBase::AddExport()
{
	HRESULT hr = cExport.Add(pFGrph);
	if (FAILED(hr))
		return erreur(TEXT("chec de l'exportation du graphe"), hr);
	return S_OK;
}
#endif // #if EXPORT_GRAPH

/**
 * Programmation d'une notification des vnements du graphe
 * \param[in] hWnd Handle de la fentre principale
 * \param[in] uMsg Identifiant du message  transmettre
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::AddEventNotifier(HWND hWnd, UINT uMsg)
{
	HRESULT hr = pFGrph.QueryInterface(&pEvents);

	if (SUCCEEDED(hr))
		hr = pEvents->SetNotifyWindow((OAHWND)hWnd, uMsg, reinterpret_cast<LONG_PTR>(dynamic_cast<IPTvM_GraphBase *>(this)));
	return hr;
}

/**
 * Ajout d'un filtre au graphe (version CHAR 8 bits)
 *
 * \param[in]	pFilter	Filtre	 ajouter
 * \param[in]	pszName	Nom de ce filtre
 **/
HRESULT CPTvM_GraphBase::AddFilterToGraph(IBaseFilter * pFilter, LPCTSTR pszName)
{
  #ifdef _UNICODE
	#define pszTmp pszName
  #else
	COPY_STR_ON_STACK(WCHAR, pszTmp, pszName, nLen);
  #endif

	// myprintf(COLR(HYELLOW, TEXT("AddFilterToGraph")) TEXT("(0x%x, ") COLR(HWHITE, TEXT("%") W2t) TEXT(")") EOL, pFilter, pszTmp);
	return pFGrph->AddFilter(pFilter, pszTmp);
  #undef pszTmp
}

/**
 * Connexion d'une broche de sortie  la premire broche d'entre disponible d'un filtre.
 * \param[in] pOutPin	Broche de sortie  connecter
 * \param[in] pFilter	Filtre prsentant la broche d'entre  connecter
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::connect_pin_to_filter(IPin * pOutPin, IBaseFilter * pFilter)
{
	CComPtr<IPin>	pInPin;
	HRESULT			hr = cherche_pin(pFilter, PINDIR_INPUT, pInPin, NOTCONNECTED);

	if (!check_hr(hr, "connect_pin_to_filter/cherche_pin(IN)")) {
		hr = pFGrph->ConnectDirect(pOutPin, pInPin, NULL);
		check_hr(hr, "connect_pin_to_filter/ConnectDirect");
	}
	return hr;
}

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
HRESULT CPTvM_GraphBase::connect_pin_to_filter_via_filter(IPin * pOutPin, CComPtr<IBaseFilter> & pFilter, FilterData sVia)
{
	HRESULT		hr = create_and_add_filter(sVia);

	if (FAILED(hr))
		return hr;
	if (FAILED(hr = connect_pin_to_filter(pOutPin, sVia.pFilter)))
		return hr;
	return connect_filters(sVia.pFilter, pFilter);
}
#endif	// #if BUILD_UNUSED_CODE

/**
 * Connexion de la premire broche de sortie non connecte d'un filtre 
 * la premire broche d'entre d'un autre.
 * \param[in] source	Filtre prsentant la broche de sortie  connecter
 * \param[in] dest		Filtre prsentant la broche d'entre  connecter
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::connect_filters(IBaseFilter * source, IBaseFilter * dest)
{
	CComPtr<IPin>	pOutPin;
	HRESULT			hr = cherche_pin(source, PINDIR_OUTPUT, pOutPin, NOTCONNECTED);

	if (!check_hr(hr, "connect_filters/cherche_pin(OUT)"))
		hr = connect_pin_to_filter(pOutPin, dest);
	return hr;
}

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
/**
 * Dconnexion d'une broche de la broche correspondante du filtre adjacent.
 * \param[in]	pPin	Broche  dconnecter
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::disconnect_pin(IPin * pPin)
{
	IPin *	pOtherPin;
	HRESULT	hr = pPin->ConnectedTo(&pOtherPin);

	if (hr == VFW_E_NOT_CONNECTED)
		return S_OK;	// Dj dconnect, rien  faire
	if (FAILED(hr))
		return hr;
	if (FAILED(hr = pFGrph->Disconnect(pOtherPin)))
		return hr;
	return pFGrph->Disconnect(pPin);
}
#endif	// #if BUILD_UNUSED_CODE

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
/**
 * Dconnexion d'une broche d'un filtre de la broche correspondante du filtre adjacent.
 * \param[in]	pFilter	Filtre prsentant la broche  dconnecter
 * \param[in]	dir		Direction de la broche  dconnecter (entre ou sortie)
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::disconnect_filter(IBaseFilter * pFilter, PIN_DIRECTION dir)
{
	CComPtr<IPin>	pPin;
	HRESULT			hr;

	if (FAILED(hr = cherche_pin(pFilter, dir, pPin)))
		return hr;
	return disconnect_pin(pPin);
}
#endif	// #if BUILD_UNUSED_CODE

/**
 * \brief Retrait d'un filtre du graphe.
 * \param[in]	pFilter	Filtre  retirer
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::remove_filter(IBaseFilter * pFilter)
{
	HRESULT hr = S_OK;

	if (pFilter) {
		// myprintf(COLR(HCYAN, TEXT("RemoveFilter")) TEXT("(0x%x)") EOL, pFilter);
		hr = pFGrph->RemoveFilter(pFilter);
		check_hr(hr, "RemoveFilter");
	}
	return hr;
}

#ifdef _DIAGNOSTIC
tstring CPTvM_GraphBase::get_graph_state()
{
	OAFilterState			state	= graph_GetState();
	CComPtr<IEnumFilters>	pEnumfilters;
	tstring					strState;

	pFGrph->EnumFilters(&pEnumfilters);

	for (ENUM_ITF(pEnumfilters, IBaseFilter, pFilter)) {
		HRESULT				hr;
		FILTER_INFO			sInfo;
		CComHeapPtr<WCHAR>	pszVendorInfo;
		FILTER_STATE		eState;
		// CComPtr<IReferenceClock>
		//					pRefClock;
		CLSID				clsidFlt;
		tstring				strFlt;

		hr = pFilter->QueryFilterInfo(&sInfo);
		hr = pFilter->QueryVendorInfo(&pszVendorInfo);
		hr = pFilter->GetState(2000, &eState);
		// hr = pFilter->GetSyncSource(&pRefClock);
		hr = pFilter->GetClassID(&clsidFlt);

		strFlt = tstr_printf(EOL TEXT("- %") W2t, sInfo.achName);
		if (pszVendorInfo)
			strFlt += tstr_printf(TEXT(" [%") W2t TEXT("]"), pszVendorInfo.m_pData);
		strFlt += tstr_printf(TEXT(", clsid=") GUIDFMT_N, GUIDPARM_N(clsidFlt));
		/*
		if (pRefClock) {
			REFERENCE_TIME			nTime;

			hr = pRefClock->GetTime(&nTime);
			strState += tstr_printf(TEXT(", %I64u"), nTime);
		} */
		strFlt += tstr_printf(TEXT(", tat=%") A2t,
					EnumToString(eState, "Arrt\0Pause\0Marche\0", "tat inconnu"));
		strState.insert(0, strFlt);	// Insertion au dbut (pour inverser l'ordre)
	}

	strState.insert(0,
		tstr_printf(
			TEXT("%") A2t EOL TEXT("\n**     :"),
			EnumToString(state,
				" \0"
				"  \0"
				" \0",
				" ")));
	return strState;
}
#endif

/// Construction du graphe
HRESULT CPTvM_GraphBase::Build()
{
	HRESULT hr = pFGrph.CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER);

	if (FAILED(hr) )
		return erreur(TEXT("   DirectShow !"), hr);

	// on loggue
  #if LOG_DSHOW
	AddLog(TEXT("mmzin_graph.log"));	/// \todo rsoudre collision de noms si graphes multiples
  #endif // #if LOG_DSHOW

  #if EXPORT_GRAPH
	if (FAILED(hr = AddExport()))
		return hr;
  #endif // #if EXPORT_GRAPH

	return hr;
}

/// Destructeur
CPTvM_GraphBase::~CPTvM_GraphBase() {
	if (pEvents)
		pEvents->SetNotifyWindow(NULL, 0, 0);
  #if LOG_DSHOW
	if (logFile != INVALID_HANDLE_VALUE)
		CloseHandle(logFile);
  #endif // #if LOG_DSHOW
}

// ====================================================================================
/// Graphe Tuner
// ====================================================================================

class CPTvM_TunerGraph : public virtual IPTvM_TunerGraph, public CPTvM_GraphBase
{
	friend class SignalStatistics;

	/**
	 * Recherche d'une interface spcifique du filtre tuner, via son interface gnrique \p IBDA_Topology.
	 * \param[in]	pTuner	Tuner dans lequel rechercher l'interface
	 * \param[in]	iid		IID de l'interface recherche
	 * \param[out]	pUnk	Noeud de contrle de l'interface trouve
	 * \return Code d'erreur DirectShow
	 **/
	static HRESULT Search_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner, REFIID iid, CComPtr<IUnknown> & pUnk);

#ifdef _DIAGNOSTIC
	static HRESULT View_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner);
#endif

protected:

	/**
	 * Recherche d'une interface spcifique du filtre tuner, via son interface gnrique \p IBDA_Topology.
	 * \param[in]	pTuner	Tuner dans lequel rechercher l'interface
	 * \param[in]	pItf	Pointeur de l'interface trouve (le type du pointeur dtermine le type d'interface)
	 * \return		Code d'erreur DirectShow
	 **/
	template <class ITF>
		static HRESULT Search_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner, CComPtr<ITF> & pItf)
	{
		CComPtr<IUnknown>	pUnk;
		HRESULT				hr = Search_IBDA_Topology(pTuner, __uuidof(ITF), pUnk);

		if (SUCCEEDED(hr)) {
			hr = pUnk.QueryInterface(&pItf);
		}
		return !pItf ? E_NOINTERFACE : hr;
	}

	/// Mmorisation du dernier filtre ajout en bout de graphe
	/// (mmorisation simple : ne ncessite pas de faire un "AddRef").
	IBaseFilter *					pTailFilter;

	CComPtr<IBaseFilter>			pNetworkTuner;

	CComPtr<IBDA_SignalStatistics>	pStats;

	CComPtr<IBDA_DeviceControl>		pBDAControl;
	CComPtr<IBDA_FrequencyFilter>	pBDAFreq;

	CPTvM_TunerGraph() :
		pTailFilter(NULL)
	{
	}

	virtual HRESULT					AddProvider() = 0;

	/**
	 * Ajout d'un filtre derrire un autre filtre du graphe.
	 * \param[in]	pOutFilter	Filtre auquel le connecter
	 * \param[in]	pFilter		Filtre  ajouter
	 * \param[in]	pszName		Nom qui lui sera attribu
	 * \return		Code d'erreur DirectShow
	 **/
	HRESULT							AppendFilter(IBaseFilter * pOutFilter, IBaseFilter * pFilter, LPCTSTR pszName);

	/**
	 * Ajout d'un filtre  la fin du tronc principal du graphe.
	 * \param[in]	pFilter		Filtre  ajouter
	 * \param[in]	pszName		Nom qui lui sera attribu
	 * \return		Code d'erreur DirectShow
	 **/
	virtual HRESULT					PushFilter(IBaseFilter * pFilter, LPCTSTR pszName);

	/**
	 * Retrait du dernier filtre du tronc principal du graphe, avec restauration de l'tat prcdent.
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT					PopFilter();

	/// Rcupration des interfaces du filtre tuner
	virtual HRESULT					QueryTunerInterfaces();

	HRESULT							AddTunerAndReceiver();

	/// Construction du graphe
	HRESULT							Build();

	/**
	 * Syntonisation d'une frquence dans le tuner (bas niveau).
	 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
	 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
	 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
	 **/
	HRESULT							LowLevelTune(LONG nFreq, LONG nBandwidth, tstring * pstrErr);

	virtual ~CPTvM_TunerGraph() {
				}
};

// ====================================================================================
/// Graphe TV complet
// ====================================================================================

class CPTvM_FullGraph : public IPTvM_FullGraph, public virtual CPTvM_TunerGraph
{
protected:
	GrabberData						sGrabber;
	FilterData						cDemux;

	VideoOutPin						cVideoOut;
  #if SINGLE_AUDIO_RENDERER
	AudioOutPin						cAudioOut;
  #else
	AudioOutPin						cMp2aOut;
	AudioOutPin						cAc3_Out;
	AudioOutPin						cEac3Out;
  #endif
	DemuxOutPin						cPmtOut;
	DemuxOutPin						cPsi_Out;

	MediaDecoderData				cMp2vCodec;
	MediaDecoderData				cH264Codec;
	MediaDecoderData				cMp2aCodec;
	MediaDecoderData				cAc3_Codec;
	MediaDecoderData				cEac3Codec;

  #if USE_PMTFILTER
	DecoderData						cPMT_Filter;
  #endif

	DecoderData						cPSI_Filter;

	VideoStreamType					eInitialVideoType;	//!< Type de flux vido au lancement du graphe
#if SINGLE_AUDIO_RENDERER == 0
	AudioOutPin *					pCurrAudioOut;		//!< sortie audio couramment utilise
#endif
	bool							bSuspended;			//!< \p true si sorties mdias suspendues
	EXECUTION_STATE					dwExecState;

	HRESULT							AddGrabber();
	HRESULT							AddDemux();

	virtual HRESULT					AddDataFilters();
	HRESULT							AddDecoders();

	/**
	 * Traitement d'un vnement du graphe
	 * \param[in] hWnd Handle de la fentre principale
	 * \return Code d'erreur DirectShow
	 **/
	virtual HRESULT					HandleGraphEvent(HWND hWnd);

	/**
	 * Changement de volume sonore
	 * \param[in]	volume	Nouveau volume sonore, valeur de 0  100
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT					set_volume(long volume);

	HRESULT							switch_codec(MediaOutPin & cMediaOut, DecoderData * pNewCodec);

	/**
	 * Slection d'une (nouvelle ou pas) paire de codecs vido et audio
	 * \param[in]	pNewVideoCodec	Pointeur sur la description du nouveau codec vido
	 * \param[in]	pNewAudioCodec	Pointeur sur la description du nouveau codec audio
	 * \return Code d'erreur DirectShow
	 **/
	HRESULT							select_codecs(MediaDecoderData * pNewVideoCodec, MediaDecoderData * pNewAudioCodec);

	/**
	 * Branchement de tous les mdias de sortie d'une chane.
	 *
	 * \param[in]	sMData	Informations sur les mdias  connecter
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT					branche_medias(const MappingData & sMData);

	/**
	 * Branchement des pistes sur le dmultiplexeur.
	 * \param[in]	sMData	Informations sur les mdias  connecter
	 * \return	Code d'erreur DirectShow
	 **/
	virtual HRESULT					MapPIDs(const MappingData & sMData);

	/**
	 * \brief Vrification de l'existence d'un dialogue de proprits pour filtre
	 *
	 * \param[in]	eFltType	Type du filtre pour lequel l'existence du dialogue doit tre vrifie
	 * \return	\p true si un dialogue de proprits existe pour ce filtre
	 * \note	La mthode cherchera  crer une instance du filtre s'il n'y en a pas encore
	 **/
	 virtual bool					check_property_dialog(FilterType eFltType);

	 /**
	 * \brief Appel du dialogue de proprits d'un filtre
	 * \param[in]	eFltType	Type du filtre dont le dialogue de proprits doit tre appel
	 **/
	virtual void					create_property_dialog(FilterType eFltType);

	/// Obtenir l'interface du grabber
	virtual GrabberData &			GetGrabberInterface() {
										return sGrabber; }

	/**
	 * Obtenir l'interface \p IMpeg2Data (pour la recherche de chanes)
	 **/
	virtual CComPtr<IMpeg2Data>		GetMpeg2Data();

#if USE_PMTFILTER
	/**
	 * Obtenir l'interface \p IPMTMonitor (pour la surveillance des changements de PMT)
	 **/
	virtual CComPtr<IPMTMonitor>	GetPMTMonitor();
#endif
public:
	/// Constructeur
	CPTvM_FullGraph();

	/// Construction du graphe
	virtual HRESULT					Build();

	/// Destructeur
	virtual	~CPTvM_FullGraph();
};

/// Constructeur
CPTvM_FullGraph::CPTvM_FullGraph() :
	cDemux(TEXT("MPEG2 Demultiplexer"), CLSID_MPEG2Demultiplexer),
	cVideoOut(L"video"),
  #if SINGLE_AUDIO_RENDERER
	cAudioOut(L"audio"),
  #else
	cMp2aOut(L"mp2a"),
	cAc3_Out(L"ac3"),
	cEac3Out(L"eac3"),
  #endif
	cPmtOut(L"pmt", MEDIA_MPEG2_PSI),
	cPsi_Out(L"data", MEDIA_MPEG2_PSI),
	cMp2vCodec(cVideoOut, &sMp2v_am, " MPEG2", filtreMPEG2),
	cH264Codec(cVideoOut, &sH264_am, " H264",  filtreH264),
  #if SINGLE_AUDIO_RENDERER
	cMp2aCodec(cAudioOut, &sMp2a_am, " MPEG2", filtreAudio),
	cAc3_Codec(cAudioOut, &sAc3__am, " AC3",   filtreAc3),
	cEac3Codec(cAudioOut, &sEac3_am, " E-AC3", filtreEac3),
  #else
	cMp2aCodec(cMp2aOut,  &sMp2a_am, " MPEG2", filtreAudio),
	cAc3_Codec(cAc3_Out,  &sAc3__am, " AC3",   filtreAc3),
	cEac3Codec(cEac3Out,  &sEac3_am, " E-AC3", filtreEac3),
  #endif
  #if USE_PMTFILTER
	cPMT_Filter(cPmtOut,  &sPmt__am, "PMT", TEXT(PMT_FILTER_NAME), CPMTFilter::CreateInstance),
  #endif
	cPSI_Filter(cPsi_Out, &sPsi__am, "PSI", TEXT("MPEG-2 "), _uuidof(Mpeg2Data)),
  #if SINGLE_AUDIO_RENDERER == 0
	pCurrAudioOut(NULL),
  #endif
	bSuspended(true),
	dwExecState(0)
{
}

// ====================================================================================
/// Graphe Tuner, version hrite de Pouchin TV
// ====================================================================================

class CLegacyTunerGraph : public virtual CPTvM_TunerGraph
{
	FilterData		cNetworkProvider;

	/**
	 * Syntonisation d'une frquence dans le tuner.
	 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
	 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
	 * \param[in]	pChIDs		Autres informations optionnelles concernant la chane
	 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
	 **/
	virtual HRESULT	change_frequence(LONG nFreq, LONG nBandwidth, const ChaineIDs * pChIDs, tstring * pstrErr) {
						return LowLevelTune(nFreq, nBandwidth, pstrErr); }

	virtual HRESULT	AddProvider();

	/**
	 * Identification du type de graphe.
	 * \retval	false	Normal
	 * \retval	true	Alternatif
	 **/
	virtual bool	isLegacyGraph() const {
						return true; }

public:
	/// Constructeur
	CLegacyTunerGraph() :
		cNetworkProvider(TEXT(NETWORK_FILTER_NAME), CNetworkProvider::CreateInstance)
	{
	}

	/// Destructeur
	virtual ~CLegacyTunerGraph() {
				}
};

HRESULT	CLegacyTunerGraph::AddProvider()
{
	HRESULT hr = create_and_add_filter(cNetworkProvider);

	pTailFilter = cNetworkProvider.pFilter;
	return hr;
}

/// Ajout informations si \p pstr non \p NULL
static void addStr(tstring * pstr, LPCTSTR pszFmt, ...)
{
	if (!pstr)
		return;

	if (!pstr->empty())
		*pstr += TEXT("; ");
	FMT_BUF(szBuf, 96, pszFmt, pszFmt);
	*pstr += szBuf;
}

class CLegacyFullGraph : public CLegacyTunerGraph, public CPTvM_FullGraph
{
public:
	/// Constructeur
	CLegacyFullGraph() :
		CPTvM_FullGraph()
	{
	}

	/// Destructeur
	virtual ~CLegacyFullGraph() {
				}
};

// ====================================================================================
/// Graphe Tuner, version "Microsoft DVB-T Network Provider"
// ====================================================================================

class CMicrosoftTunerGraph : public virtual CPTvM_TunerGraph
{
	FilterData			cNetworkProvider;
	CDVBTTuningSpace	cTuningSpace;

	virtual HRESULT		AddProvider();

	virtual HRESULT		ApplyTuneRequest(ITuneRequest * pTuneRequest);

	/**
	 * Syntonisation d'une frquence dans le tuner.
	 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
	 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
	 * \param[in]	pChIDs		Autres informations optionnelles concernant la chane
	 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
	 **/
	virtual HRESULT		change_frequence(LONG nFreq, LONG nBandwidth, const ChaineIDs * pChIDs=NULL, tstring * pstrErr=NULL);

	/**
	 * Identification du type de graphe.
	 * \retval	false	Normal
	 * \retval	true	Alternatif
	 **/
	virtual bool		isLegacyGraph() const {
							return false; }
public:
	/// Constructeur
	CMicrosoftTunerGraph() :
		cNetworkProvider(TEXT("Microsoft DVB-T Network Provider"), CLSID_DVBTNetworkProvider)
	{
	}

	/// Destructeur
	virtual ~CMicrosoftTunerGraph() {
				}
};

HRESULT CMicrosoftTunerGraph::AddProvider()
{
	HRESULT hr = create_and_add_filter(cNetworkProvider);

	pTailFilter = cNetworkProvider.pFilter;

	if (SUCCEEDED(hr)) {
		// Ncessaire  certains tuners
		HRESULT hr2 = change_frequence(-1, -1);

		myprintf(CMyP ERRX(TEXT("Erreur : slection frquence impossible, hr=0x%x")), FAILED(hr2), hr2);
	}
	return hr;
}

HRESULT CMicrosoftTunerGraph::ApplyTuneRequest(ITuneRequest * pTuneRequest)
{
	CComQIPtr<ITuner> pTuner(cNetworkProvider.pFilter);
	if (!pTuner)
		return E_NOINTERFACE;

	HRESULT hr = pTuner->put_TuneRequest(pTuneRequest);

	check_hr(hr, "put_TuneRequest");
	return hr;
}

/**
 * Syntonisation d'une frquence dans le tuner.
 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
 * \param[in]	pChIDs		Autres informations optionnelles concernant la chane
 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
 **/
HRESULT CMicrosoftTunerGraph::change_frequence(LONG nFreq, LONG nBandwidth, const ChaineIDs * pChIDs, tstring * pstrErr)
{
	HRESULT				hr;

	if (nFreq >= 0)
		nFreq += sTunProps.offset_tuner;

	CDVBTTuneRequest	cTuneReqZ(cTuningSpace, nFreq, nBandwidth,
							-1, -1, -1, &hr);

	if (SUCCEEDED(hr))
		hr =  ApplyTuneRequest(cTuneReqZ);
	if (FAILED(hr) || !pChIDs)
		return hr;

	// Slection en deux fois : le cas a t constat o la slection de chane chouait
	// quand les PIDs de deux chanes diffrentes sur deux multiplex diffrents taient
	// les mmes, et ceci semble y pallier.
	CDVBTTuneRequest	cTuneReqN(cTuningSpace, nFreq, nBandwidth,
							pChIDs->ONID, pChIDs->TSID, pChIDs->SID, &hr);

	if (SUCCEEDED(hr))
		hr =  ApplyTuneRequest(cTuneReqN);
	return hr;
}

class CMicrosoftFullGraph : public CMicrosoftTunerGraph, public CPTvM_FullGraph
{
	virtual HRESULT	AddDataFilters();
public:
	/// Constructeur
	CMicrosoftFullGraph() :
		CPTvM_FullGraph()
	{
	}

	/// Destructeur
	virtual ~CMicrosoftFullGraph() {
				}
};

HRESULT	CMicrosoftFullGraph::AddDataFilters()
{
	// ####### TIF :
	DemuxOutPin		cTifOut(L"tif", MEDIA_MPEG2_PSI);
	HRESULT			hr = cTifOut.CreatePin(cDemux.pFilter, sTif__am, true);

	if (FAILED(hr))
		return hr;

	DecoderData		sTIF_Filter(cTifOut, &sTif__am, "TIF", TEXT("BDA MPEG2 Transport Information Filter"), CLSID_BDA_MPEG2_TIF);

	if (FAILED(hr = sTIF_Filter.render(*this)))
		return hr;

//#if USE_CONSOLE
//	liste_pins(cDemux);
//#endif
	// Contrles communs :
	return __super::AddDataFilters();
}

// ====================================================================================
/// Destruction du graphe
// ====================================================================================

void clean_dshow(void)
{
	if (!pGraph)
		return;		// Dj dtruit

	// Arrt de tous les enregistrements en cours
	myprintf(TEXT("Arrt des enregistrements") EOL);
	enregistrements_actuels.stop_all();

	pGraph->graph_Stop();

	myprintf(TEXT("Destruction du graphe") EOL);

	// viter les accs au graphe pendant la destruction
	IPTvM_FullGraph * pTmpGraph = pGraph;

	pGraph = NULL;
	delete pTmpGraph;		// #### Destruction effective
  #ifdef _DEBUG
	Sleep(200);
	DbgDumpObjectRegister();
	DbgTerminate();
  #endif
}

// ====================================================================================
// Topologie du filtre tuner
// ====================================================================================

/**
 * Recherche d'une interface spcifique du filtre tuner, via son interface gnrique \p IBDA_Topology.
 * \param[in]	pTuner	Tuner dans lequel rechercher l'interface
 * \param[in]	iid		IID de l'interface recherche
 * \param[out]	pUnk	Noeud de contrle de l'interface trouve
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_TunerGraph::Search_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner, REFIID iid, CComPtr<IUnknown> & pUnk)
{
	CComQIPtr<IBDA_Topology> pTop(pTuner);

	if (!pTop)
		return erreur(TEXT("  IBDA_Topology   "), E_NOINTERFACE);

	ULONG	NodeTypes;
	ULONG	NodeType[32];

	HRESULT	hr = pTop->GetNodeTypes(&NodeTypes, _countof(NodeType), NodeType);

	if (FAILED(hr))
		return hr;

	for (ULONG i = 0; i < NodeTypes; i++) {
		ULONG	nInterfaces;
		GUID	aInterface[32];

		hr = pTop->GetNodeInterfaces(NodeType[i], &nInterfaces, _countof(aInterface), aInterface);

		if (FAILED(hr))
			continue;

		for (ULONG j = 0; j < nInterfaces; j++) {
			if (aInterface[j] == iid)
				return pTop->GetControlNode(0, 1, NodeType[i], &pUnk);
		}
	}

	return hr;
}

#ifdef _DIAGNOSTIC

HRESULT CPTvM_TunerGraph::View_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner)
{
	CComQIPtr<IBDA_Topology> pTop(pTuner);

	if (!pTop)
		return erreur(TEXT("  IBDA_Topology   "), E_NOINTERFACE);

	ULONG	NodeTypes;
	ULONG	NodeType[32];

	HRESULT	hr = pTop->GetNodeTypes(&NodeTypes, _countof(NodeType), NodeType);

	if (FAILED(hr))
		return hr;

	myprintf(COLR(LYELLOW, TEXT("Topologie BDA (NodeTypes count = %u) :")) EOL, NodeTypes);
	for (ULONG i = 0; i < NodeTypes; i++) {
		myprintf(COLR(LYELLOW, TEXT("-- NodeTypes[%u] = %u")) EOL, i, NodeType[i]);
		ULONG	nInterfaces;
		GUID	aInterface[32];

		hr = pTop->GetNodeInterfaces(NodeType[i], &nInterfaces, _countof(aInterface), aInterface);

		if (FAILED(hr))
			continue;

		myprintf(COLR(LYELLOW, TEXT("   Interfaces count = %u")) EOL, nInterfaces);
		for (ULONG j = 0; j < nInterfaces; j++) {
			myprintf(COLR(LYELLOW, TEXT("   -- Interface = ") GUIDFMT_N) EOL,
				GUIDPARM_N(aInterface[j]));
		}
	}

	return hr;
}
#endif

/**
 * Recherche d'une interface spcifique du filtre tuner, via son interface gnrique \p IBDA_Topology.
 * \param[in]	pTuner	Tuner dans lequel rechercher l'interface
 * \param[in]	pItf	Pointeur de l'interface trouve (le type du pointeur dtermine le type d'interface)
 * \return		Code d'erreur DirectShow
 **/
template <class ITF>
	static HRESULT Search_IBDA_Topology(const CComPtr<IBaseFilter> & pTuner, CComPtr<ITF> & pItf)
{
	CComPtr<IUnknown>	pUnk;
	HRESULT				hr = Search_IBDA_Topology(pTuner, __uuidof(ITF), pUnk);

	if (SUCCEEDED(hr)) {
		hr = pUnk.QueryInterface(&pItf);
	}
	return !pItf ? E_NOINTERFACE : hr;
}

// ====================================================================================
// Configuration du dmultiplexeur
// ====================================================================================

HRESULT CPTvM_FullGraph::AddDataFilters()
{
	HRESULT hr;

	// ####### PSI :
	if (FAILED(hr = cPsi_Out.CreatePin(cDemux.pFilter, sPsi__am, true)))
		return hr;
	hr = cPSI_Filter.render(*this);
	if (FAILED(hr))
		return hr;

	// ####### EPG :
	DemuxOutPin		cEpgOut(L"epg", MEDIA_MPEG2_PSI);

	if (FAILED(hr = cEpgOut.CreatePin(cDemux.pFilter, sEpg__am, false)))
		return hr;

	DecoderData		cEPG_Filter(cEpgOut, &sEpg__am, "EPG", TEXT(EPG_FILTER_NAME), CEPGFilter::CreateInstance);

	if (SUCCEEDED(hr = cEPG_Filter.render(*this)))
		hr = cEpgOut.Map(p_pid_EIT);

#if USE_PMTFILTER
	// ####### PMT :
	if (FAILED(hr = cPmtOut.CreatePin(cDemux.pFilter, sPmt__am, false)))
		return hr;
	if (SUCCEEDED(hr))
		hr = cPMT_Filter.render(*this);
#endif
	return hr;
}

HRESULT CPTvM_FullGraph::AddDecoders()
{
	HRESULT hr;

	// Cration ou recherche des broches de sortie "mdia" du dmultiplexeur :
	// ####### Vido
	if (FAILED(hr = cVideoOut.CreatePin(cDemux.pFilter, sMp2v_am, USE_EXISTING_DEMUX_PINS)))
		return hr;
#if SINGLE_AUDIO_RENDERER
	// ####### Audio
	if (FAILED(hr = cAudioOut.CreatePin(cDemux.pFilter, sMp2a_am, USE_EXISTING_DEMUX_PINS)))
		return hr;
#else
	// ####### Audio MPEG2
	if (FAILED(hr = cMp2aOut.CreatePin(cDemux.pFilter, sMp2a_am, USE_EXISTING_DEMUX_PINS)))
		return hr;
	// ####### Audio AC3
	if (FAILED(hr = cAc3_Out.CreatePin(cDemux.pFilter, sAc3__am, USE_EXISTING_DEMUX_PINS)))
		return hr;
	// ####### Audio E-AC3
	if (FAILED(hr = cEac3Out.CreatePin(cDemux.pFilter, sEac3_am, USE_EXISTING_DEMUX_PINS)))
		return hr;
#endif

#if SINGLE_AUDIO_RENDERER == 0
	// Rendu des sorties audio :
	// ####### Audio MPEG2 :
	if (FAILED(hr = cMp2aCodec.render(*this)))
		return hr;

	// ####### Audio AC3 :
	if (FAILED(hr = cAc3_Codec.render(*this)))
		return hr;

	// ####### Audio E-AC3 :
	if (FAILED(hr = cEac3Codec.render(*this)))
		return hr;
#endif
	return hr;
}

// ====================================================================================
// Mise en route, arrt, tat du graphe
// ====================================================================================

/**
 * \brief	Mise en route du graphe
 * \return	Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::graph_Run()
{
	/**
	 * \todo Voir s'il ne serait pas intressant de dmarrer et arrter le graphe dans un
	 * "thread" distinct afin d'viter que l'application "gle" pendant cette opration.
	 **/
	if (bRunning)
		return S_OK;

	CComQIPtr<IMediaControl> pControl(pFGrph);

	if (!pControl)
		return E_NOINTERFACE;

	myprintf(BOLD(TEXT("Dmarrage du graphe")) EOL);
	HRESULT hr = pControl->Run();

	check_hr(hr, "    ");
	if (hr==S_FALSE) {	// Dmarrage en instance : on attend que tout se stabilise
		for (UINT nInx=0; nInx<10; nInx++) {	// 10 * 200 mS = 2 secondes
			Sleep(200);
			if (graph_GetState()==State_Running) {
				bRunning = true;
				hr = S_OK;
				break;
			}
		}
	}

	SetTimer(hMainWnd, TIMER_ENABLE_OSD, DUREE_ENABLE_OSD, NULL);
	myprintf(BOLD(TEXT("Graphe dmarr")) EOL);
	return hr;
}

/**
 * Mise en pause du graphe
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::graph_Pause()
{
	CComQIPtr<IMediaControl> pControl(pFGrph);

	if (!pControl)
		return E_NOINTERFACE;
	return pControl->Pause();
}

/**
 * Arrt du graphe
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_GraphBase::graph_Stop()
{
	if (!bRunning)
		return S_OK;

	if (pOSD)
		pOSD->Disable();

	CComQIPtr<IMediaControl> pControl(pFGrph);

	if (!pControl)
		return E_NOINTERFACE;

	myprintf(BOLD(TEXT("Arrt du graphe")) EOL);

#ifdef _DEBUG
	// DumpGraph(pFGrph, 3);
#endif

	HRESULT hr = pControl->Stop();

	if (SUCCEEDED(hr))
		bRunning = false;
	return hr;
}

/**
 * Interrogation de l'tat du graphe
 * \return Valeurs de l'numration FILTER_STATE (strmif.h)
 * \retval 0 (State_Stopped) Le graphe est arrt
 * \retval 1 (State_Paused) Le graphe est en mode "pause"
 * \retval 2 (State_Running) Le graphe est en marche
 * \retval -1 Le graphe ne s'est pas initialis correctement
 **/
OAFilterState CPTvM_GraphBase::graph_GetState()
{
	if (pFGrph) {
		CComQIPtr<IMediaControl> pControl(pFGrph);
		OAFilterState state;

		if (pControl && SUCCEEDED(pControl->GetState(1500, &state)))
			return state;
	}

	return -1;
}

// ====================================================================================
// Gestion des vnements du graphe
// ====================================================================================

/**
 * Traitement d'un vnement du graphe
 * \param[in] hWnd Handle de la fentre principale
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_FullGraph::HandleGraphEvent(HWND hWnd)
{
	// Disregard if we don't have an IMediaEventEx pointer.
	if (!pEvents)
		return E_FAIL;

	// Get all the events
	long		evCode;
#if USE_AUTO_RETUNE
	bool		bSendTuneCmd = false;
#endif
	bool		bVideoValid	 = false;
	LONG_PTR	param1, param2;
#if USE_CONSOLE
	int			nEvCount = 0;	// Comptage des vnements traits en mme temps
#endif

	while (SUCCEEDED(pEvents->GetEvent(&evCode, &param1, &param2, 0))) {
#if USE_CONSOLE
		myprintf(
			TEXT("Graph 0x%08x Event (%u) 0x%02x=") COLR(HWHITE, TEXT("EC_%") A2t) TEXT(", %08x, %08x") EOL,
			(IFilterGraph2 *)pFGrph, ++nEvCount, evCode,
			EnumToString(evCode,
				"BASE\0"						// 0x00
				"COMPLETE\0"					// 0x01
				"USERABORT\0"					// 0x02
				"ERRORABORT\0"					// 0x03
				"TIME\0"						// 0x04
				"REPAINT\0"						// 0x05
				"STREAM_ERROR_STOPPED\0"		// 0x06
				"STREAM_ERROR_STILLPLAYING\0"	// 0x07
				"ERROR_STILLPLAYING\0"			// 0x08
				"PALETTE_CHANGED\0"				// 0x09
				"VIDEO_SIZE_CHANGED\0"			// 0x0a
				"QUALITY_CHANGE\0"				// 0x0b
				"SHUTTING_DOWN\0"				// 0x0c
				"CLOCK_CHANGED\0"				// 0x0d
				"PAUSED\0"						// 0x0e
				"??\0"							// 0x0f
				"OPENING_FILE\0"				// 0x10
				"BUFFERING_DATA\0"				// 0x11
				"FULLSCREEN_LOST\0"				// 0x12
				"ACTIVATE\0"					// 0x13
				"NEED_RESTART\0"				// 0x14
				"WINDOW_DESTROYED\0"			// 0x15
				"DISPLAY_CHANGED\0"				// 0x16
				"STARVATION\0"					// 0x17
				"OLE_EVENT\0"					// 0x18
				"NOTIFY_WINDOW\0"				// 0x19
				"STREAM_CONTROL_STOPPED\0"		// 0x1a
				"STREAM_CONTROL_STARTED\0"		// 0x1b
				"END_OF_SEGMENT\0"				// 0x1c
				"SEGMENT_STARTED\0"				// 0x1d
				"LENGTH_CHANGED\0"				// 0x1e
				"DEVICE_LOST\0"					// 0x1f
				"SAMPLE_NEEDED\0"				// 0x20
				"PROCESSING_LATENCY\0"			// 0x21
				"SAMPLE_LATENCY\0"				// 0x22
				"SCRUB_TIME\0"					// 0x23
				"STEP_COMPLETE\0"				// 0x24
				"EC_SKIP_FRAMES\0"				// 0x25
				"??\0??\0??\0??\0??\0"			// 0x26  0x2a
				"??\0??\0??\0??\0??\0"			// 0x2b  0x2f
				"TIMECODE_AVAILABLE\0"			// 0x30
				"EXTDEVICE_MODE_CHANGE\0"		// 0x31
				"STATE_CHANGE\0"				// 0x32
				"??\0??\0??\0??\0??\0??\0??\0"	// 0x33  0x39
				"??\0??\0??\0??\0??\0??\0"		// 0x3a  0x3f
				"PLEASE_REOPEN\0"				// 0x40
				"STATUS\0"						// 0x41
				"MARKER_HIT\0"					// 0x42
				"LOADSTATUS\0"					// 0x43
				"FILE_CLOSED\0"					// 0x44
				"ERRORABORTEX\0"				// 0x45
				"EOS_SOON\0"					// 0x46
				"CONTENTPROPERTY_CHANGED\0"		// 0x47
				"BANDWIDTHCHANGE\0"				// 0x48
				"VIDEOFRAMEREADY\0"				// 0x49
				"??\0??\0??\0??\0??\0??\0"		// 0x4a  0x4f
				"GRAPH_CHANGED\0"				// 0x50
				"CLOCK_UNSET\0"					// 0x51
				"??\0"							// 0x52
				"VMR_RENDERDEVICE_SET\0"		// 0x53
				"VMR_SURFACE_FLIPPED\0"			// 0x54
				"VMR_RECONNECTION_FAILED\0"		// 0x55
				"PREPROCESS_COMPLETE\0"			// 0x56
				"CODECAPI_EVENT\0"				// 0x57
				),
				param1, param2);
#endif
		switch (evCode) {

#if USE_CONSOLE
		case EC_ERRORABORT:					// 0x03
			myprintf(ERRX(TEXT("EC_ERRORABORT, raison=%08x: %s")), HRESULT(param1),
				DShowErrorToString(HRESULT(param1)).c_str());
			break;
#endif

		case EC_PAUSED:						// 0x0e
		/* Note : lien vers une discussion en relation avec ce problme d'vnement EC_PAUSED :
		http://www.eggheadcafe.com/conversation.aspx?messageid=29757074&threadid=29757074 */
#if USE_CONSOLE
			myprintf(ERRX(TEXT("EC_PAUSED, raison=0x%08x: %s")), HRESULT(param1),
				DShowErrorToString(HRESULT(param1)).c_str());
#endif
#if USE_AUTO_RETUNE
			bSendTuneCmd = sTunProps.auto_resyntonisation;
#endif
			break;

		case EC_STREAM_ERROR_STILLPLAYING:	// 0x07

#if USE_CONSOLE
			myprintf(ERRX(TEXT("EC_STREAM_ERROR_STILLPLAYING, code=0x%08x: %s, minor=0x%08x")),
				HRESULT(param1),
				DShowErrorToString(HRESULT(param1)).c_str(), DWORD(param2));
#endif
#if USE_AUTO_RETUNE
			if (sTunProps.auto_resyntonisation) {
				// Auto-resyntonisation seulement si 3 vnements de ce type surviennent
				//  intervalle de moins de 5 secondes
				static BYTE		nCount = 0;	// nombre de rptitions de l'vnement
				static DWORD	dwLastEvt = 0;
				DWORD			dwCurrEvt = GetTickCount();

				if (dwCurrEvt-dwLastEvt < 5000) {
					++nCount;
					if (nCount>=2) {
						bSendTuneCmd = true;
						nCount = 0;
					}
				} else
					nCount = 0;
				dwLastEvt = dwCurrEvt;
			}
#endif	// USE_AUTO_RETUNE
			break;

		case EC_DEVICE_LOST:
			if (param2)	// "device is available again"
				PostMessage(hWnd, WM_COMMAND, IDM_APP_RESTART, NULL);
			break;

	//	case EC_VMR_RENDERDEVICE_SET:		// 0x53
	//		myprintf(BOLD(TEXT("Nouveau rendu VMR = %u")) EOL, param1);
	//		break;

		case EC_VIDEO_SIZE_CHANGED:			// 0x0a
			bVideoValid = true;
			myprintf(BOLD(TEXT("Nouvelle taille vido = %u/%u")) EOL, LOWORD(param1), HIWORD(param1));
		}
		pEvents->FreeEventParams(evCode, param1, param2);
	}
#if USE_AUTO_RETUNE
	if (bSendTuneCmd)
		PostMessage(hWnd, WM_COMMAND, IDM_CHAN_TUNE, 0);
#endif
	if (bVideoValid)	// ( poster APRES WM_COMMAND + IDM_CHAN_TUNE)
		PostMessage(hWnd, WM_APP_VIDEO_UPDATED, 0, NULL);
	return S_OK;
}

// ====================================================================================
// Construction du graphe
// ====================================================================================

/**
 * Ajout d'un filtre derrire un autre filtre du graphe.
 * \param[in]	pOutFilter	Filtre auquel le connecter
 * \param[in]	pFilter		Filtre  ajouter
 * \param[in]	pszName		Nom qui lui sera attribu
 * \return		Code d'erreur DirectShow
 **/
HRESULT CPTvM_TunerGraph::AppendFilter(IBaseFilter * pOutFilter, IBaseFilter * pFilter, LPCTSTR pszName)
{
	HRESULT hr = AddFilterToGraph(pFilter, pszName);

	if (check_hr(hr, "AppendFilter/AddFilterToGraph"))
		return hr;

	hr = connect_filters(pOutFilter, pFilter);
	if (check_hr(hr, "AppendFilter/connect_filters"))
		remove_filter(pFilter);

	return hr;
}

/**
 * Ajout d'un filtre  la fin du tronc principal du graphe.
 * \param[in]	pFilter		Filtre  ajouter
 * \param[in]	pszName		Nom qui lui sera attribu
 * \return		Code d'erreur DirectShow
 **/
HRESULT CPTvM_TunerGraph::PushFilter(IBaseFilter * pFilter, LPCTSTR pszName)
{
	HRESULT hr = AppendFilter(pTailFilter, pFilter, pszName);

	if (SUCCEEDED(hr))
		pTailFilter = pFilter;

	return hr;
}

/**
 * Retrait du dernier filtre du tronc principal du graphe, avec restauration de l'tat prcdent.
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_TunerGraph::PopFilter()
{
	PIN_INFO		pin_info = {NULL};
	HRESULT			hr;

	{	// Trouver la broche d'entre du dernier filtre
		CComPtr<IPin>	pInPin;

		hr = cherche_pin(pTailFilter, PINDIR_INPUT, pInPin);
		if (check_hr(hr, "PopFilter/cherche_pin(IN)"))
			return hr;

		CComPtr<IPin>	pOutPin;

		// Trouver la broche de sortie du filtre auquel il est connect
		hr = pInPin->ConnectedTo(&pOutPin);
		if (check_hr(hr, "PopFilter/ConnectedTo"))
			return hr;

		// Trouver  quel fitre appartient cette broche de sortie
		hr = pOutPin->QueryPinInfo(&pin_info);
		check_hr(hr, "PopFilter/QueryPinInfo");
		if (pin_info.pFilter)
			pin_info.pFilter->Release();	// liminer la rfrence en suspens (le pointeur reste valide)
		// Ignorer le code retour : si pas de filtre, pTailFilter sera NULL

	} // -> sortie de bloc, pour s'assurer que les broches n'ont plus de rfrences en cours

	// Retirer le filtre
	if (FAILED(hr = remove_filter(pTailFilter)))
		return hr;

	// Le filtre prcdent devient le dernier de la file
	pTailFilter = pin_info.pFilter;
	return hr;
}

HRESULT CPTvM_TunerGraph::AddTunerAndReceiver()
{
	HRESULT			hr = E_FAIL;

	MonikerBindings	cMnkbTuner(sTunProps.szPathTuner, &hr);

	if (SUCCEEDED(hr))
		cMnkbTuner.BindToFilter();

	// Ajout du tuner
	if (SUCCEEDED(hr)) {
		myprintf(TEXT("Ajout tuner \"%s\" (0x%x)") EOL, cMnkbTuner.strName.c_str(), &*cMnkbTuner.pFilter);
		hr = PushFilter(cMnkbTuner.pFilter, cMnkbTuner.strName.c_str());
	}
	if (FAILED(hr))
		return erreur(TEXT("  DVB-T   .      "), hr);

	pNetworkTuner = cMnkbTuner.pFilter;

	// Ajout du filtre dmodulateur (optionnel et sans doute assez rare)
	if (!IsNullName(sTunProps.szPathDemod)) {
		MonikerBindings	cMnkbDemod(sTunProps.szPathDemod, &hr);

		if (SUCCEEDED(hr))
			cMnkbDemod.BindToFilter();
		if (SUCCEEDED(hr)) {
			myprintf(TEXT("Ajout dmodulateur \"%s\" (0x%x)") EOL, cMnkbDemod.strName.c_str(), &*cMnkbDemod.pFilter);
			hr = PushFilter(cMnkbDemod.pFilter, cMnkbDemod.strName.c_str());
		}
		if (FAILED(hr))
			return erreur(TEXT("  DVB-T   .      "), hr);
	}

	// Ajout du filtre de capture (optionnel, mais le plus souvent prsent)
	if (!IsNullName(sTunProps.szPathCapture)) {
		MonikerBindings	cMnkbCapture(sTunProps.szPathCapture, &hr);

		if (SUCCEEDED(hr))
			cMnkbCapture.BindToFilter();
		if (SUCCEEDED(hr)) {
			myprintf(TEXT("Ajout capture \"%s\" (0x%x)") EOL, cMnkbCapture.strName.c_str(), &*cMnkbCapture.pFilter);
			hr = PushFilter(cMnkbCapture.pFilter, cMnkbCapture.strName.c_str());
		}
		if (FAILED(hr))
			return erreur(TEXT("    DVB-T   .      "), hr);
	}

	return QueryTunerInterfaces();
}

/// Rcupration des interfaces du filtre tuner
HRESULT CPTvM_TunerGraph::QueryTunerInterfaces()
{
	HRESULT hr;

 #ifdef _DIAGNOSTIC
	View_IBDA_Topology(pNetworkTuner);
 #endif

	hr = pNetworkTuner.QueryInterface(&pBDAControl);
	if (FAILED(hr))
		return erreur(TEXT("  IBDA_DeviceControl   "), hr);
	hr = Search_IBDA_Topology(pNetworkTuner, pBDAFreq);
	if (FAILED(hr))
		return erreur(TEXT("       "), hr);
	hr = Search_IBDA_Topology(pNetworkTuner, pStats);
	if (FAILED(hr))
		return erreur(TEXT("       "), hr);

	return hr;
}

/// Construction du graphe
HRESULT CPTvM_TunerGraph::Build()
{
	// Construction du graphe de base :
	HRESULT		hr = __super::Build();

	if (FAILED(hr))
		return hr;
	hr = AddProvider();		// Network provider
	return hr;
}

HRESULT CPTvM_FullGraph::AddGrabber()
{
	FilterData	sGrabberFilter(TEXT("Grabber"), CLSID_SampleGrabber);
	HRESULT		hr = sGrabberFilter.Create();

	if (FAILED(hr))
		return erreur(TEXT(" \"%s\"     "), hr, sGrabberFilter.pszName);
	hr = PushFilter(sGrabberFilter.pFilter, sGrabberFilter.pszName);
	if (FAILED(hr))
		return erreur(TEXT("       grabber"), hr);

	hr = sGrabberFilter.pFilter.QueryInterface(&sGrabber.pInterface);
	if (FAILED(hr))
		return erreur(TEXT("  grabber     "), hr);

	sGrabber.pCallback = new CSampleGrabber();
	if (sGrabber.pCallback==NULL)
		return erreur(TEXT("   grabber     "), E_FAIL);
	return hr;
}

HRESULT	CPTvM_FullGraph::AddDemux()
{
	HRESULT		hr = cDemux.Create();

	if (FAILED(hr))
		return hr;
	hr = PushFilter(cDemux.pFilter, cDemux.pszName);
	if (FAILED(hr))
		return erreur(TEXT("Grabber       "), hr);
	return hr;
}

/// Construction du graphe
HRESULT CPTvM_FullGraph::Build()
{
	// Construction du graphe de base :
	HRESULT 				hr = __super::Build();

	if (FAILED(hr))
		return hr;

	if (FAILED(hr = AddTunerAndReceiver()))	// Tuner et rcepteur
		return hr;
	if (FAILED(hr = AddGrabber()))			// Grabber
		return hr;
	if (FAILED(hr = AddDemux()))			// Dmultiplexeur
		return hr;

	// connections
	if (FAILED(hr = AddDataFilters()))		// Filtres de contrle
		return erreur(TEXT("       "), hr);
	if (FAILED(hr = AddDecoders()))			// Codecs
		return erreur(TEXT("      "), hr);
	if (FAILED(hr))
		return erreur(TEXT("    "), hr);

	if (use_osd && !pOSD) {
		myprintf(TEXT("Initialisation de l'OSD") EOL);
		pOSD.reset(new COsdAllObjects());
	}

	AddEventNotifier(hMainWnd, WM_APP_GRAPHNOTIFY);

	return hr;
}

/// Destructeur
CPTvM_FullGraph::~CPTvM_FullGraph()
{
	MSG	msg;

	// Dtruire tous les messages de notification du graphe qui seraient encore en attente
	// (et qui pourraient causer des plantages si traits aprs destruction du graphe)
	while (PeekMessage(&msg, hMainWnd, WM_APP_GRAPHNOTIFY, WM_APP_GRAPHNOTIFY, PM_REMOVE));

	// Destruction des objets OSD
	pOSD.reset();
}

/// Construction d'un graphe partiel (pour essais tuners), selon le type de graphe requis
IPTvM_TunerGraph * new_TunerGraph(bool bLegacy)
{
	if (bLegacy)
		return new CLegacyTunerGraph();
	else
		return new CMicrosoftTunerGraph();
}

/// Construction du graphe complet, selon le type de graphe requis
IPTvM_FullGraph * new_FullGraph(bool bLegacy)
{
	if (bLegacy)
		return new CLegacyFullGraph();
	else
		return new CMicrosoftFullGraph();
}

/**
 * Construction et mise en route du graphe DirectShow
 * \return Code d'erreur DirectShow
 **/
HRESULT build_graph()
{
	if (pGraph)
		return E_FAIL;	// Dj existant

  #ifdef _DEBUG
	DbgInitialise(hAppInstance);
  #endif

	pGraph = new_FullGraph(autre_graphe);
	if (!pGraph)
		return E_FAIL;

	// Construction
	return pGraph->Build();
}

/**
 * Changement de volume sonore
 * \param[in]	volume	Nouveau volume sonore, valeur de 0  100
 * \return	Code d'erreur DirectShow
 **/
HRESULT set_volume(long volume)
{
	HRESULT hr = E_FAIL;

	if (pGraph)
		hr = pGraph->set_volume(volume);
	return hr;
}

/**
 * Branchement de tous les mdias de sortie d'une chane.
 *
 * \param[in]	sMData	Informations sur les mdias  connecter
 * \return	Code d'erreur DirectShow
 **/
HRESULT branche_medias(const MappingData & sMData)
{
	if (!pGraph)
		return E_FAIL;
	return pGraph->branche_medias(sMData);
}

/**
 * Dbranchement de TOUS les mdias
 **/
void debranche_medias()
{
#if USE_PMTFILTER
	if (pGraph) {
		CComPtr<IPMTMonitor> pPmtMonitor(pGraph->GetPMTMonitor());

		if (pPmtMonitor)
			pPmtMonitor->StopPmtMonitoring();
	}
#endif
	branche_medias(MappingData());
}

// ====================================================================================
// Contrle de la rception
// ====================================================================================

/**
 * Syntonisation d'une frquence dans le tuner (bas niveau).
 * \param[in]	nFreq		Frquence  syntoniser, exprime en kilohertz
 * \param[in]	nBandwidth	Largeur de bande, exprime en megahertz
 * \param[in]	pstrErr		Pointeur sur une variable recevant un dtail des erreurs si non NULL
 **/
HRESULT CPTvM_TunerGraph::LowLevelTune(LONG nFreq, LONG nBandwidth, tstring * pstrErr)
{
	if (!pBDAControl || !pBDAFreq)
		return E_FAIL;

	HRESULT hr	= pBDAControl->StartChanges();

	if (hr != S_OK) {
		addStr(pstrErr, TEXT("StartChanges: hr=0x%08x"), hr);
		return hr;
	}

	hr = pBDAFreq->put_Bandwidth(nBandwidth);
	if (hr != S_OK)
		addStr(pstrErr, TEXT("put_Bandwidth(%ul): hr=0x%08x"), nBandwidth, hr);

	nFreq += sTunProps.offset_tuner;	// Correction de frquence

	/// \todo Vrifier
	HRESULT hr2	= pBDAFreq->put_FrequencyMultiplier(1000);

	if (hr2 != S_OK) {
		addStr(pstrErr, TEXT("put_FrequencyMultiplier(%ul): hr=0x%08x"), 1000, hr2);
		if (hr == S_OK)
			hr = hr2;
	}

	hr2	= pBDAFreq->put_Frequency(nFreq);
	if (hr2 != S_OK) {
		addStr(pstrErr, TEXT("put_Frequency(%ul): hr=0x%08x"), nFreq, hr2);
		if (hr == S_OK)
			hr = hr2;
	}

	hr2 = pBDAControl->CheckChanges();
	if (hr2 != S_OK) {
		addStr(pstrErr, TEXT("CheckChanges(): hr=0x%08x"), hr2);
		if (hr == S_OK)
			hr = hr2;
	}

	hr2 = pBDAControl->CommitChanges();
	if (hr2 != S_OK) {
		addStr(pstrErr, TEXT("CommitChanges(): hr=0x%08x"), hr2);
		if (hr == S_OK)
			hr = hr2;
	}
	return hr;
}

/**
 * Branchement des pistes sur le dmultiplexeur.
 * \param[in]	sMData	Informations sur les mdias  connecter
 * \return	Code d'erreur DirectShow
 **/
HRESULT CPTvM_FullGraph::MapPIDs(const MappingData & sMData)
{
  #if PROTECT_UNMAP_PID
	/// Exprimental (il y a problme de deadlock dans UnmapPID, et le Mpeg2 Demultiplexer a l'air d'encadrer cette fonction avec de tels appels)
	bool			bNeedCommit		= false;
	HRESULT			hr2;

	if (graph_isRunning()) {
		hr2 = pBDAControl->StartChanges();	/// Exprimental
		check_hr(hr2, "CPTvM_FullGraph::MapPIDs / StartChanges");
		if (SUCCEEDED(hr2))
			bNeedCommit = true;
	}
  #endif

	HRESULT			hr				= cVideoOut.Map(sMData.video_pid);

	if (FAILED(hr))
		return hr;
  #if SINGLE_AUDIO_RENDERER
	if (FAILED(hr = cAudioOut.Map(sMData.audio_pid)))
		return hr;
  #else
	AudioOutPin *	pNewAudioOut	= NULL;

	if (sMData.audio_pid != 0) {
		switch (sMData.eAudioType) {

		case tst_MPEG2:
			pNewAudioOut = &cMp2aOut;
			break;

		case tst_EAC3:
		case tst_EAC3_hc:
			// Si le filtre E-AC3 n'existe pas, on se rabat sur la broche AC3
			// (car certains anciens codecs peuvent dcoder l'E-AC3 dans ce mode).
			// Ce "mode de compatibilit" devrait pouvoir tre supprim d'ici quelque temps.
			pNewAudioOut = !cEac3Codec.isNull() ? &cEac3Out : &cAc3_Out;
			break;

		case tst_AC3:
		case tst_AC3_hc:
			pNewAudioOut = &cAc3_Out;
			break;

		default:
			myprintf(TEXT("Type de piste audio non identifi") EOL);
		}
	}

	if (FAILED(hr = cMp2aOut.Map(pNewAudioOut == &cMp2aOut ? sMData.audio_pid : 0)))
		return hr;
	if (FAILED(hr = cAc3_Out.Map(pNewAudioOut == &cAc3_Out ? sMData.audio_pid : 0)))
		return hr;
	if (FAILED(hr = cEac3Out.Map(pNewAudioOut == &cEac3Out ? sMData.audio_pid : 0)))
		return hr;
  #endif // #if SINGLE_AUDIO_RENDERER
  #if USE_PMTFILTER
	// pmt
	if (FAILED(hr = cPmtOut.Map(sMData.pmt_pid)))
		return hr;
  #endif
  #if PROTECT_UNMAP_PID
	if (bNeedCommit) {
		hr2 = pBDAControl->CommitChanges();	/// Exprimental
		check_hr(hr2, "CPTvM_FullGraph::MapPIDs / CommitChanges");
	}
  #endif
	return hr;
}

HRESULT CPTvM_FullGraph::switch_codec(MediaOutPin & cMediaOut, DecoderData * pNewCodec)
{
	if (pNewCodec == cMediaOut.pCurCodec)
		return S_OK;	// Codec inchang

	HRESULT hr = cMediaOut.SelectRenderer(*this, pNewCodec == NULL);

	// Slection du rendu appropri au nouveau codec
	if (FAILED(hr))
		return hr;
	if (cMediaOut.pCurCodec != pNewCodec && cMediaOut.pCurCodec != NULL && cMediaOut.pCurCodec != UNDEF_CODEC_PTR) {
		if (FAILED(hr = remove_filter(cMediaOut.pCurCodec->pFilter)))
			return hr;
	}
	if (pNewCodec) {
		// Slection d'un nouveau type de mdia pour la broche de sortie
		CComQIPtr<IMpeg2Demultiplexer>	pMpeg2Demux(cDemux.pFilter);

		if (pMpeg2Demux) {
			// Note : utilisation de "const_cast" ici car on ne s'attend pas normalement  ce que
			// "SetOutputPinMediaType" modifie les objets points.
			hr = pMpeg2Demux->SetOutputPinMediaType(
					const_cast<LPWSTR>(cMediaOut.pszName),
					const_cast<AM_MEDIA_TYPE*>(pNewCodec->pAmt));
			if (FAILED(hr))
				return hr;
		}
	}
	if (cMediaOut.pCurCodec != pNewCodec) {
		if (pNewCodec != NULL) {
			hr = pNewCodec->render(*this, &*cMediaOut.pRenderer);
			if (hr != S_OK) {
				// Insertion ou connexion du codec impossible : Slectionner un rendu vido NULL
				hr = cMediaOut.SelectRenderer(*this, true);
				check_hr(hr, "invalid codec / SelectRenderer");
				hr = connect_pin_to_filter(pNewCodec->cDemuxOut.pPin, cMediaOut.pRenderer->pFilter);
				check_hr(hr, "invalid codec / connect_pin_to_filter");
			}
		} else {
			// Aucun codec dfini : Connecter le rendu vido NULL directement  la sortie du dmultiplexeur
			hr = connect_pin_to_filter(cMediaOut.pPin, cMediaOut.pRenderer->pFilter);
			check_hr(hr, "null codec / connect_pin_to_filter");
		}
		cMediaOut.pCurCodec = pNewCodec;
	}
	return S_OK;
}

/**
 * Slection d'une (nouvelle ou pas) paire de codecs vido et audio
 * \param[in]	pNewVideoCodec	Pointeur sur la description du nouveau codec vido
 * \param[in]	pNewAudioCodec	Pointeur sur la description du nouveau codec audio
 * \return Code d'erreur DirectShow
 **/
HRESULT CPTvM_FullGraph::select_codecs(MediaDecoderData * pNewVideoCodec, MediaDecoderData * pNewAudioCodec)
{
	HRESULT				hr = S_OK;

	if (pNewVideoCodec && pNewVideoCodec->isNull())
		pNewVideoCodec = NULL;
#if SINGLE_AUDIO_RENDERER
	if (pNewAudioCodec && pNewAudioCodec->isNull())
		pNewAudioCodec = NULL;
#else
	// (pNewAudioCodec inutilis)
#endif

	bool				bVideoOutChanged	= cVideoOut.NeedGraphChange(pNewVideoCodec, vmr_mode);
#if SINGLE_AUDIO_RENDERER
	bool				bAudioOutChanged	= cAudioOut.NeedGraphChange(pNewAudioCodec, rdt_DirectSound);
#endif

#if SINGLE_AUDIO_RENDERER
	if (!bVideoOutChanged && !bAudioOutChanged)
		return S_OK;
#else
	if (!bVideoOutChanged)
		return S_OK;
#endif

	// Suspension du graphe requise
	GraphSuspender		cGSusp(*this);

	// Slection du rendu appropri au nouveau codec vido
	if (FAILED(hr = switch_codec(cVideoOut, pNewVideoCodec)))
		return hr;

#if SINGLE_AUDIO_RENDERER
	// Slection du rendu appropri au nouveau codec audio
	if (FAILED(hr = switch_codec(cAudioOut, pNewAudioCodec)))
		return hr;

#endif	// #if SINGLE_AUDIO_RENDERER
	cGSusp.Ok();
	return hr;
}

/**
 * Branchement de tous les mdias de sortie d'une chane.
 *
 * \param[in]	sMData	Informations sur les mdias  connecter
 * \return	Code d'erreur DirectShow
 **/
HRESULT CPTvM_FullGraph::branche_medias(const MappingData & sMData)
{
	HRESULT	hr	= E_FAIL;	// valeur par dfaut

	// Passage en mode suspendu si tous les PIDS sont gaux  zro
	bool bNewSuspend = sMData.noMediaPids();

	// viter que l'ordinateur passe en mode veille par "inactivit" pendant que la vido
	// y est affiche
	if (bNewSuspend != bSuspended) {
		if (bNewSuspend)	// Restauration de l'tat par dfaut
			SetThreadExecutionState(dwExecState);
		else				// Empcher le mode veille et sauvegarder l'tat par dfaut
			dwExecState = SetThreadExecutionState(ES_CONTINUOUS|ES_DISPLAY_REQUIRED);
		bSuspended = bNewSuspend;
	}

	if (!bSuspended)  {
		MediaDecoderData *	pNewVideoCodec	= NULL;

		// Dtermination du codec vido
		switch (sMData.eVideoType) {

		case vst_MPEG2:
		case vst_MPEG2_encrypted:
			pNewVideoCodec = &cMp2vCodec;
			break;

		case vst_H264:
		case vst_H264_encrypted:
			pNewVideoCodec = &cH264Codec;
		}

  #if SINGLE_AUDIO_RENDERER
		// Dtermination du codec audio
		MediaDecoderData *	pNewAudioCodec	= NULL;

		switch (sMData.eAudioType) {

		case tst_MPEG2:
			pNewAudioCodec = &cMp2aCodec;
			break;

		case tst_EAC3:
		case tst_EAC3_hc:
			// Si le filtre E-AC3 n'existe pas, on se rabat sur le filtre AC3
			// (car certains anciens codecs peuvent dcoder l'E-AC3 dans ce mode).
			// Ce "mode de compatibilit" devrait pouvoir tre supprim d'ici un ou deux ans.
			pNewAudioCodec = !cEac3Codec.isNull() ? &cEac3Codec : &cAc3_Codec;
			break;

		case tst_AC3:
		case tst_AC3_hc:
			pNewAudioCodec = &cAc3_Codec;
		}
  #endif	// #if SINGLE_AUDIO_RENDERER

  #if SINGLE_AUDIO_RENDERER
		// Mise  jour de la configuration de codecs si ncessaire
		if (FAILED(hr = select_codecs(pNewVideoCodec, pNewAudioCodec)))
			return hr;
  #else
		// Mise  jour de la configuration du codec vido si ncessaire
		if (FAILED(hr = select_codecs(pNewVideoCodec, NULL)))
			return hr;
  #endif	// #if SINGLE_AUDIO_RENDERER
	}

	if (FAILED(hr = MapPIDs(sMData)))
		return hr;
	return hr;
}

/**
 * Changement de volume sonore
 * \param[in]	volume	Nouveau volume sonore, valeur de 0  100
 * \return	Code d'erreur DirectShow
 **/
HRESULT CPTvM_FullGraph::set_volume(long volume)
{
	CComQIPtr<IBasicAudio> pBas(pFGrph);

	if (!pBas)
		return E_NOINTERFACE;

	int nDbVolume = -8000;

	if (volume > 0)
		nDbVolume = (int)((log10(1.0*volume)-2)*5000);
	return pBas->put_Volume(nDbVolume);
}

/**
 * Obtenir l'interface \p IMpeg2Data (pour la recherche de chanes)
 **/
CComPtr<IMpeg2Data> CPTvM_FullGraph::GetMpeg2Data()
{
	// query MPEG-2 Sections and Tables Filter
	CComQIPtr<IMpeg2Data> pMpeg2Data(cPSI_Filter.pFilter);

	if (!pMpeg2Data) {
		erreur(TEXT("    %") A2t TEXT("   "), E_NOINTERFACE, cPSI_Filter.pszType);
		affiche_erreurs(NULL, E_NOINTERFACE);
	}

	return pMpeg2Data;
}

#if USE_PMTFILTER
/**
 * Obtenir l'interface \p IPMTMonitor (pour la surveillance des changements de PMT)
 **/
CComPtr<IPMTMonitor> CPTvM_FullGraph::GetPMTMonitor()
{
	CComQIPtr<IPMTMonitor>	pPMTMonitor(cPMT_Filter.pFilter);

	if (!pPMTMonitor) {
		erreur(TEXT("L'interface de surveillance %") A2t TEXT(" n'a pas t trouve"), E_NOINTERFACE, cPMT_Filter.pszType);
		affiche_erreurs(NULL, E_NOINTERFACE);
	}

	return pPMTMonitor;
}
#endif

// ====================================================================================
// Proprits des filtres
// ====================================================================================

/**
 * \brief Appel du dialogue de proprits d'un filtre
 * \param[in]	eFltType	Type du filtre dont le dialogue de proprits doit tre appel
 **/
void CPTvM_FullGraph::create_property_dialog(FilterType eFltType)
{
	switch (eFltType) {
	case ft_Video_Mp2v: cMp2vCodec.create_property_dialog();	break;
	case ft_Video_H264: cH264Codec.create_property_dialog();	break;
	case ft_Audio_Mp2a: cMp2aCodec.create_property_dialog();	break;

	case ft_Audio_Eac3:
		if (!cEac3Codec.isNull()) {
			cEac3Codec.create_property_dialog();
			break;
		}
		// si le filtre E-AC3 n'est pas dfini, on utilise le filtre AC3
		// ... (passage  travers)
	case ft_Audio_Ac3:
		cAc3_Codec.create_property_dialog();
		break;

	case ft_AudioRenderer:
#if SINGLE_AUDIO_RENDERER
		if (cAudioOut.pRenderer)
			cAudioOut.pRenderer->create_property_dialog();
#else
		if (pCurrAudioOut && pCurrAudioOut->pRenderer)
			pCurrAudioOut->pRenderer->create_property_dialog();
#endif
		break;

	case ft_VideoRenderer:
		if (cVideoOut.pRenderer)
			cVideoOut.pRenderer->create_property_dialog();
	}
}

 /**
 * \brief Vrification de l'existence d'un dialogue de proprits pour filtre
 *
 * \param[in]	eFltType	Type du filtre pour lequel l'existence du dialogue doit tre vrifie
 * \return	\p true si un dialogue de proprits existe pour ce filtre
 * \note	La mthode cherchera  crer une instance du filtre s'il n'y en a pas encore
 **/
bool CPTvM_FullGraph::check_property_dialog(FilterType eFltType)
{
	switch (eFltType) {
	case ft_Video_Mp2v:	return cMp2vCodec.check_property_dialog();
	case ft_Video_H264: return cH264Codec.check_property_dialog();
	case ft_Audio_Mp2a: return cMp2aCodec.check_property_dialog();
	case ft_Audio_Ac3:	return cAc3_Codec.check_property_dialog();
	case ft_Audio_Eac3:	return cEac3Codec.check_property_dialog();

	case ft_AudioRenderer:
#if SINGLE_AUDIO_RENDERER
		if (cAudioOut.pRenderer)
			return cAudioOut.pRenderer->check_property_dialog();
#else
		if (pCurrAudioOut && pCurrAudioOut->pRenderer)
			return pCurrAudioOut->pRenderer->check_property_dialog();
#endif
		return false;

	case ft_VideoRenderer:
		if (cVideoOut.pRenderer)
			return cVideoOut.pRenderer->check_property_dialog();
	}
	return false;
}

// ====================================================================================
// Utilitaire d'arrt temporaire du graphe
// ====================================================================================

/// Constructeur
GraphSuspender::GraphSuspender(IPTvM_GraphBase & iGrph) :
	iGraph(iGrph),
	bRunning(iGrph.graph_isRunning()),
	bCanRestart(false),
	hr(S_OK)
{
	if (bRunning) {
		hr = iGraph.graph_Stop();
		check_hr(hr, "arrt du graphe");
	}
}

/// Destructeur
GraphSuspender::~GraphSuspender()
{
	if (bRunning && bCanRestart && hr == S_OK) {
		hr = iGraph.graph_Run();
		check_hr(hr, "redmarrage du graphe");
	}
}

// ====================================================================================
// Interface de qualit du signal
// ====================================================================================

/**
 * Tableau associatif dcrivant l'tat d'un lment ternaire de statistiques du signal
 **/
AssocElement SignalStatistics::aSSBool[] =
{					// Stat_BOOL
	{TEXT(""),	sb_NO},
	{TEXT(""),	sb_YES},
	{TEXT("n/a"),	sb_UNK},
	{NULL,0}		// -> terminateur
};

/// Chargement depuis l'interface de qualit du signal
void SignalStatistics::getStats()
{
	BOOLEAN	bBool;
	HRESULT	hr=S_OK;

	nErrStep = 0;

	if (!pGraph)
		return;
	IBDA_SignalStatistics * pStats = dynamic_cast<CPTvM_FullGraph *>(pGraph)->pStats;

	if (!pStats)	// pStats parfois NULL transitoirement si fermeture du programme
		return;		// avec fentre de statistiques ouverte

	if (FAILED(hr = pStats->get_SignalPresent(&bBool))) {
		if (hr!=E_PROP_ID_UNSUPPORTED) {
			myprintf(FMT_ERRX("get_SignalPresent"), hr);
			if (nErrStep==0) {
				hr_1st = hr;
				nErrStep |= 0x01;
			}
			// Erreur dans la fonction, on prsume signal prsent et on passe  la suite
			ePresent = sb_UNK;
		}
	} else
		ePresent = Stat_BOOL(bBool);

	if (FAILED(hr = pStats->get_SignalLocked(&bBool))) {
		if (hr!=E_PROP_ID_UNSUPPORTED) {
			myprintf(FMT_ERRX("get_SignalLocked"), hr);
			if (nErrStep==0) {
				hr_1st = hr;
				nErrStep |= 0x02;
			}
			// Erreur dans la fonction, on prsume signal verrouill et on passe  la suite
			eLocked = sb_UNK;
		}
	} else
		eLocked = Stat_BOOL(bBool);

	if (FAILED(hr = pStats->get_SignalStrength(&nStrength))) {
		if (hr!=E_PROP_ID_UNSUPPORTED) {
			myprintf(FMT_ERRX("get_SignalStrength"), hr);
			if (nErrStep==0) {
				hr_1st = hr;
				nErrStep |= 0x04;
			}
			nStrength = STATS_UNKNOWN;
		}
	}

	if (isPresenceOk() && isLockedOk()) {
		if (FAILED(hr = pStats->get_SignalQuality(&nQuality))) {
			if (hr!=E_PROP_ID_UNSUPPORTED) {
				myprintf(FMT_ERRX("get_SignalQuality"), hr);
				if (nErrStep==0) {
					hr_1st = hr;
					nErrStep |= 0x08;
				}
			}
			nQuality = STATS_UNKNOWN;
		} else
			nQuality = max(0, min(nQuality, 100));
	} else
		nQuality = 0;
}

/// Chargement de la force et de la qualit du signal dans des barres de niveau.
/// Retourne la valeur ajuste (pour affichage) de la force du signal.
void SignalStatistics::SetLevelBars(HWND hSBar, HWND hQBar) const
{
	LONG nAdjStrength = 0;

	if (nStrength!=STATS_UNKNOWN) {
		nAdjStrength = min(abs(nStrength - sTunProps.force_signal_zero), sTunProps.force_signal_max);

		// EnableWindow(hSBar, true); // -> Inutile : cela ne change pas l'aspect du contrle
		SendMessage(hSBar, PBM_SETPOS, nAdjStrength, 0);
	} /*else
		EnableWindow(hSBar, false);*/

	if (nQuality!=STATS_UNKNOWN) {
		// EnableWindow(hQBar, true);
		SendMessage(hQBar, PBM_SETPOS, min(max(0, nQuality), 100), 0);
	} /*else
		EnableWindow(hQBar, false);*/
}

void SignalStatistics::getStrengthStr(bool bPct, LPTSTR pszBuf, UINT32 nBufLen, LPCTSTR pszPctSign) const
{
	if (nStrength==STATS_UNKNOWN) {
		_tcscpy_s(pszBuf, nBufLen, TEXT("n/a"));
	} else {
		if (bPct) {
			LONG	nAdjStrength = min(abs(nStrength - sTunProps.force_signal_zero), sTunProps.force_signal_max);

			_stprintf_s(pszBuf, nBufLen, TEXT("%i%s"),
				sTunProps.force_signal_max ? nAdjStrength*100/sTunProps.force_signal_max : 0,	// viter risque de division par zro
				pszPctSign);
		} else {
			_stprintf_s(pszBuf, nBufLen, TEXT("%i"), nStrength);
		}
	}
}

void SignalStatistics::getQualityStr(LPTSTR pszBuf, UINT32 nBufLen, LPCTSTR pszPctSign) const
{
	if (nQuality==STATS_UNKNOWN) {
		_tcscpy_s(pszBuf, nBufLen, TEXT("n/a"));
	} else {
		_stprintf_s(pszBuf, nBufLen, TEXT("%i%s"), nQuality, pszPctSign);
	}
}

// ====================================================================================
// Chargement des tables SI
// ====================================================================================

void Mpeg2SectionsHandler::Init(PID pid, TID tid, DWORD dwTimeout, FOURCC tblName)
{
	tableName = tblName;
	pSectionList.Release();
	nextSection = 0;
	lastSection = 0;
	length = 0;
	complete = false;
	//
	memset(&filter, 0, sizeof(filter));
	filter.bVersionNumber = 1;
	filter.wFilterSize = MPEG2_FILTER_VERSION_1_SIZE;
	filter.fSpecifySectionNumber = TRUE;
	//
	_pid = pid;
	_tid = tid;
	_dwTimeout = dwTimeout;
}

bool Mpeg2SectionsHandler::GetNextSection()
{
	HRESULT hr;

	// Libration de la section prcdemment charge :
	pSectionList.Release();
	length = 0;

	if (nextSection > lastSection) {
		complete = true;
		return false;
	}
	filter.SectionNumber = nextSection++;
	hr = pMpeg2Data->GetSection(_pid, _tid, &filter, _dwTimeout, &pSectionList);
	myprintf(CMyP TEXT("Pas eu la %") A2t TEXT(", hr=0x%08x") EOL, FAILED(hr), (LPCSTR)&tableName, hr);
	if (FAILED(hr))
		return false;

	hr = pSectionList->GetSectionData(0, &length, &data);
	myprintf(CMyP TEXT("Pas eu les donnes de la %") A2t TEXT(", hr=0x%08x") EOL, FAILED(hr), (LPCSTR)&tableName, hr);
	if (FAILED(hr))
		return false;

	lastSection = ((const SI_hdr *)data)->last_section_number;
	myprintf(TEXT("Taille %") A2t TEXT(" : %u (dernire section = %u)") EOL, (LPCSTR)&tableName, length, lastSection);
	return true;
}

// ====================================================================================
// Conversion de code d'erreur
// ====================================================================================

/**
 * Conversion d'un code d'erreur DirectShow en une chane de caractres affichable.
 *
 * \param[in]	hr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 * \todo		Voir s'il ne serait pas mieux d'utiliser DXGetErrorString et DXGetErrorDescription
 *				http://msdn.microsoft.com/en-us/library/microsoft.directx_sdk.errorhandling.dxgeterrorstring(VS.85).aspx
 *				http://msdn.microsoft.com/en-us/library/microsoft.directx_sdk.errorhandling.dxgeterrordescription(VS.85).aspx
 **/
tstring DShowErrorToString(HRESULT hr)
{
	// Note : contrairement  ce qui est dit dans la documentation de Microsoft,
	// il existe des messages plus longs que MAX_ERROR_TEXT_LEN ...
	TCHAR	szBuffer[MAX_ERROR_TEXT_LEN*2];

	if (AMGetErrorText(hr, szBuffer, _countof(szBuffer))==0) {
		UINT nFacility = HIWORD(hr) & 0xfff;

		if (nFacility == FACILITY_MF || nFacility == FACILITY_MF_WIN32)
			_stprintf_s(szBuffer, TEXT("  Media Foundation,  %u"), LOWORD(hr));
		else
			strcpy_T(szBuffer, TEXT("[   ]"));
	}
	return RTrimStr(szBuffer);	// Conversion implicite en tstring
};

/**
 * Conversion d'un code d'erreur DirectX en une chane de caractres affichable.
 *
 * \param[in]	hr	Code d'erreur  convertir
 * \return		Une chane de caractres, format \p tstring (= \p string ou \p wstring),
 *				contenant le message d'erreur correspondant
 **/
tstring DXErrorToString(HRESULT hr)
{
	// Note : contrairement  ce qui est dit dans la documentation de Microsoft,
	// il existe des messages plus longs que MAX_ERROR_TEXT_LEN ...
	TCHAR	szBuffer[MAX_ERROR_TEXT_LEN*2];

	strcpy_T(szBuffer, DXGetErrorString(hr));
	strcat_T(szBuffer, TEXT(" - "));
	strcat_T(szBuffer, DXGetErrorDescription(hr));

	return RTrimStr(szBuffer);	// Conversion implicite en tstring
};
