/*
 * $Id$
 * Copyright (C) 2011 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Gestion de la surface superpose  l'image pour les besoins de l'OSD.
 *
 *	Dfinitions des interfaces ncessaires  construire une surface
 *  de rendu de texte et autres objets, selon des mthodes variables
 *  (la mthode de base tant la bibliothque GDI, les alternatives
 *  possibles tant les diverses versions de Direct3D).
 **/

#pragma once

#include "osd_objects.h"

// ====================================================================================
// Dfinitions
// ====================================================================================

/**
 * Couleur de fond de la fentre, remplace par l'image
 * \note L'utilisation de RGB(1,1,1) plutt que RGB(0,0,0) (= noir parfait) vise  viter que les
 * caractres (noirs) des textes des dialogues superposs  la vido deviennent transparents dans
 * certains mode de cette vido. Prcdemment, RGB(0,0,1) tait utilis, mais pour une raison
 * inconnue, cette valeur semble poser des problmes en EVR.
 **/
#define colorKey			RGB(1,1,1)

// ====================================================================================

/// numration des types de couches utilisables pour l'OSD
enum LayerTypes
{
	lt_GDI,		//!< GDI (basique), directement dans la fentre
	lt_BMP,		//!< GDI via bitmap
	lt_D3D9		//!< Direct3D9
};

// ====================================================================================

/**
 * Calcul du rectangle client de la fentre (hors barre d'tat)
 **/
RECT CalcCliRect();

// ====================================================================================
// Interface d'allocation des polices de caractres
// ====================================================================================

/// Classe d'assistance  la gestion de la police OSD
class COsdFontAllocator
{
protected:
	LONG	nFntHgt;	//!< Hauteur de la police utilise par l'OSD

	/// Constructeur
	COsdFontAllocator() :
		nFntHgt(0)
	{
	}

	/// Rinitialisation de la police de caractres
	virtual void Reset() = 0;

public:
	/// Cration de l'instance correspondant au type de couche OSD qu'on utilise
	static COsdFontAllocator *	CreateInstance(LayerTypes eType);

	/// Obtenir la taille (hauteur) de police couramment dfinie
	LONG Size() const {
			return nFntHgt; }

	/**
	 * \brief Rinitialisation de la police de caractres en vue de changement de taille
	 * \param[in]	nNewHgt	Taille  utiliser lorsque la police sera recre
	 **/
	void SetSize(LONG nNewHgt)
	{
		if (nNewHgt != nFntHgt) {
			Reset();
			nFntHgt = nNewHgt;
		}
	}

	/// Destructeur
	virtual ~COsdFontAllocator() {
				}
};

// ====================================================================================
// Interface de gestion du plan de dessin de l'OSD
// ====================================================================================

/// Gestion du plan de dessin de l'OSD, classe de base abstraite
class COsdLayer
{
protected:
	SIZE				m_sSize;		//!< Taille attendue de l'objet
	bool				m_bNotEmpty;	//!< \p true si au moins un objet est affich dans l'OSD
public:

	COsdLayer() :
		m_sSize(SIZE()),
		m_bNotEmpty(false)
	{
	}

	/**
	 * Cration d'une instance destine  dessiner directement dans le rectangle client
	 * de la fentre.
	 * \param[in]	hDC	"Device context",  utiliser seulement si on dispose de la valeur
	 *					fournie par \p WM_PAINT; sinon, mettre NULL.
	 * \return		Pointeur sur l'objet cr.
	 **/
	static COsdLayer *	CreateWinInstance(HDC hDC);

	/**
	 * Cration d'une instance dont le type est dfini par \p eType.
	 * \param[in]	eType	Type de couche OSD  crer.
	 * \return		Pointeur sur l'objet cr.
	 **/
	static COsdLayer *	CreateInstance(LayerTypes eType);

	/// Affichage d'un objet OSD
	/// \param[in]	cOsdObj	Rfrence sur l'objet  afficher
	HRESULT				DrawOsdObject(COsdObjectPtr pObj);

	virtual HRESULT		DrawOsdText(COsdObjectText & cOsdTxt, const RECT & sRct) = 0;

	/// Retourne \p true si la couche OSD n'est pas vierge
	bool				notEmpty() const {
							return m_bNotEmpty; }

	/// Vrifie (en retournant \p true) que la classe a t correctement initialise
	virtual bool		layer_ok() const = 0;

	/// Vrifie (en retournant \p true) qu'une surface utilisable existe
	virtual bool		layer_set() const = 0;

	/// Cration d'un allocateur de polices de caractres dans le type qui correspond
	///  la mthode de gestion instancie
	virtual COsdFontAllocator *
						CreateFontAllocator() = 0;

	/// Initialisation des prrequis au fonctionnement de l'interface
	virtual HRESULT		Initialize() = 0;

	/// Cration d'une surface de dimensions dfinies (remplacement de l'ventuelle surface prcdente)
	virtual HRESULT		Create() = 0;

	/// Destruction de la surface
	virtual void		Reset() = 0;

	/// Effacement de la surface existante
	virtual HRESULT		Erase() = 0;

	/// Obtenir la taille courante de la surface
	virtual SIZE		GetSize() const = 0;

	/// Dfinir une nouvelle taille, rinitialiser si diffrente de la taille actuelle
	void				SetSize(SIZE sNewSize) {
							if (sNewSize != GetSize()) Reset();
							m_sSize = sNewSize;	}

	/// Prsentation de la surface dans un rendu vido VMR9
	virtual HRESULT		Present(IVMRMixerBitmap9 * pBmp) = 0;

	/// Prsentation de la surface dans un rendu vido EVR
	virtual HRESULT		Present(IMFVideoMixerBitmap * pBmp) = 0;

	/// Destructeur
	virtual ~COsdLayer() {
				}
};

typedef shared_ptr<class COsdLayer>	COsdLayerPtr;

// ====================================================================================
