/*
 * $Id: parse.cpp 264 2010-11-01 22:49:12Z gingko $
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation de l'exploration des tables SI des flux Mpeg2 TS
 *
 *	Ce fichier implmente les fonctions d'exploration des tables SI flux Mpeg2 TS.
 **/

#include "stdafx.h"
#include "parse.h"

// ====================================================================================

/**
 * Recherche dans la table si on a des informations
 * concernant la chane identifie par \p sCIds.
 * \param[in]	sCIds	Identifiants de la chane recherche
 * \return		Pointeur sur le descripteur existant, ou bien \p NULL si pas trouv
 **/
const ChaineIdsNIT * CInfoNIT::Cherche(const ChaineIDs & sCIds) const
{
	for (ITERATE_CONST_VECTOR(*this, ChaineIdsNIT, it)) {
		if (it->EquIds(sCIds))
			return &*it;
	}

	return NULL;
}

// ====================================================================================
// Exploration des tables SI
// ====================================================================================

/**
 * Analyse d'une table PAT (Program Association Table) reue dans le flux TS
 * \param[in]	pat			Table  analyser
 * \param[in]	nNoCanal	Numro du canal dans lequel se trouve la table
 **/
void parse_pat(const SI_PAT & pat, BYTE nNoCanal)
{
	ChaineIDs	sCIds;

	sCIds.TSID = pat.tsid(true);
	sCIds.canal_no = nNoCanal;

	for (SI_PAT::entryScanner se(pat.entryRange(true)); !se; ++se) {

		sCIds.SID = se().program_number();

		if (sCIds.SID != 0) {
			int		ixChaine	= Canaux.trouve_par_Ids(sCIds);
			UINT16	pid			= se().pid();

			myprintf(TEXT("Service : 0x%04x, PMT PID : 0x%04x") EOL, sCIds.SID, pid);
			if (ixChaine>=0)
				Canaux[ixChaine].pmt_pid = pid;
		}
	}
}

void parse_pmt(const SI_PMT & pmt, bool bMicrosoft, ChainePMT & lpmt_e)
{
	// Paramtre "bMicrosoft" ncessaire parce que Microsoft permute les octets de certains champs,
	// dans **certaines** fonctions DirectShow ..., et que "parse_pmt" est utilise parfois via
	// ces fonctions, et parfois de faon plus directe

	bool	bEncrypted = false;

	lpmt_e.pcr_pid = pmt.PCR_PID();
	lpmt_e.nbr = lpmt_e.nbr_tot = 0;

	for (MDESCR::D_scanner sd(pmt.descrRange()); !sd; ++sd) {
		if (sd().tag == MDESCR::mdt_Conditional_Access) {
			bEncrypted = true;
			break;
		}
	}
	for (SI_PMT::entryScanner se(pmt.entryRange(bMicrosoft)); !se; ++se) {
		UINT16				pid			= se().elementary_PID();
		StreamType			stream_type	= se().stream_type;
		MDESCR::D_scanner	sd(se().descrRange());	// Prparation de l'exploration des descripteurs
		une_piste			pmt_piste;

		switch (stream_type) {
		case st_MPEG_1_video: //ISO_IEC_11172_2_VIDEO, MPEG-1 video streams
		case st_MPEG_2_video: //ISO_IEC_13818_2_VIDEO, MPEG-2 video streams
#if TEST_RNT_RADIO
			if (pid == TEST_RNT_RADIO)
				continue;
#endif
			lpmt_e.video_pid=pid;
			lpmt_e.video_type = vst_MPEG2;

			// Recherche de descripteur de cryptage (y'a pas encore a en France en MPEG2
			// mais il faut tout prvoir et y'a pas que la France) :
			for (; !sd; ++sd) {
				if (sd().tag == MDESCR::mdt_Conditional_Access) {
					bEncrypted = true;
					break;
				}
			}
			break;
		case st_MPEG_1_audio: //ISO_IEC_11172_3_AUDIO, MPEG-1 audio streams
		case st_MPEG_2_audio: //ISO_IEC_13818_3_AUDIO, MPEG-2 audio streams
			pmt_piste.pid	= pid;
			pmt_piste.type	= tst_MPEG2;
			// Recherche de descripteurs de langue :
			for (; !sd; ++sd) {
				if (sd().tag == MDESCR::mdt_ISO_639_Language && sd().length > 0)
					sd().as<MDESCR::D_0a>()().language_code.get(pmt_piste.lang);
			}
			break;
		case st_PES_private:
			pmt_piste.type	= tst_Unknown;	// type par dfaut
			pmt_piste.pid	= pid;

			// Recherche de descripteurs audio et autres :
			for (; !sd; ++sd) {
				MDESCR::D_tag	eTag	= sd().tag;
				UINT8			length	= sd().length;

				switch (eTag) {

				case MDESCR::mdt_AC3_Audio:
					if (sd().as<MDESCR::D_6a>().number_of_channels() >= MDESCR::anc_Multi_MoreThan_2_0)
						pmt_piste.type = tst_AC3_hc;
					else
						pmt_piste.type = tst_AC3;
					break;

				case MDESCR::mdt_Enhanced_AC3_Audio:
					if (sd().as<MDESCR::D_7a>().number_of_channels() >= MDESCR::anc_Multi_MoreThan_2_0)
						pmt_piste.type = tst_EAC3_hc;
					else
						pmt_piste.type = tst_EAC3;
					break;

				case MDESCR::mdt_Teletext:
					pmt_piste.type = tst_Teletext;
					break;

				case MDESCR::mdt_Subtitling:
					pmt_piste.type = tst_Subtitling;
					break;

				case MDESCR::mdt_ISO_639_Language:
					if (length > 0)
						sd().as<MDESCR::D_0a>()().language_code.get(pmt_piste.lang);
					break;

				case MDESCR::mdt_Conditional_Access:
					bEncrypted = true;
				}
			}

			break;
		case st_H264_video: // H264
			lpmt_e.video_pid=pid;
			lpmt_e.video_type = vst_H264;

			// Recherche de descripteur de cryptage :
			for (; !sd; ++sd) {
				if (sd().tag == MDESCR::mdt_Conditional_Access) {
					bEncrypted = true;
					break;
				}
			}
			break;
		}

		// Ajout des pistes audio et spciales :
		if (pmt_piste.pid!=0)
			lpmt_e.ajoute(pmt_piste);
	}

	// Cryptage dtect : ajuster le type vido dfini
	if (bEncrypted) {
		if (lpmt_e.video_type==vst_MPEG2)
			lpmt_e.video_type = vst_MPEG2_encrypted;
		else if (lpmt_e.video_type==vst_H264)
			lpmt_e.video_type = vst_H264_encrypted;
	}
}

/**
 * Analyse d'une table SDT reue dans le flux TS
 * \param[in]	sdt			Table  analyser
 * \param[in]	nNoCanal	Numro du canal dans lequel se trouve la table
 * \param[in]	khz			Frquence du canal dans lequel se trouve la table
 **/
void parse_sdt(const SI_SDT & sdt, BYTE nNoCanal, INT32 khz)
{
	UINT16	tsid = sdt.tsid(true);	// transport_stream_id
	UINT16	onid = sdt.onid();		// original_network_id

	myprintf(TEXT("Transport stream id : %i") EOL, tsid);
	myprintf(TEXT("Onid : %i") EOL, onid);
	myprintf(TEXT("chanes : ") EOL);

	for (SI_SDT::entryScanner se(sdt.entryRange(true)); !se; ++se) {
		for (MDESCR::D_scanner sd(se().descrRange()); !sd; sd++) {
			if (sd().tag == MDESCR::mdt_Service) {
				const MDESCR::D_48 &	d48	= sd().as<MDESCR::D_48>();
				Chaine					canal;

				// Charger les identifiants de la chane :
				canal.ONID		= onid;
				canal.TSID		= tsid;
				canal.SID		= se().service_id();
				canal.canal_no	= nNoCanal;
				d48.service_provider_name().get(canal.groupe);
				d48.service_name().get(canal.nom);
				canal.khz		= khz;

				myprintf(TEXT("Groupe : %") A2t TEXT(", Chaine : %") A2t TEXT(", SID : %i") EOL,
					canal.groupe, canal.nom, se().service_id());

				// Tester si la chane est dj dans la liste :
				if (Canaux.trouve_par_Ids(canal)<0) {
					// Elle n'est pas dans la liste : on l'ajoute.
					Canaux.push_back(canal);
				} else
					myprintf(TEXT("   (%") A2t TEXT(" dj dans la table des chanes : ignor)") EOL, canal.nom);

				break;
			}
		}
	}
}

/**
 * Analyse d'une table NIT reue dans le flux TS
 * \param[in]	nit			Table  analyser
 * \param[in]	cInfoNIT	Donnes rcupres
 **/
void parse_nit(const SI_NIT & nit, CInfoNIT & cInfoNIT)
{
	for (MDESCR::D_scanner sd(nit.descrRange()); !sd; ++sd) {
		TCHAR nom[256];

		nom[0] = 0;
		if (sd().tag == MDESCR::mdt_Network_Name) {
			sd().as<MDESCR::D_40>().get_network_name(nom);
			if (*nom) {
				myprintf(TEXT("parse_nit: Nom rseau : %s") EOL, nom);
				cInfoNIT.strNetworkName = nom;
				break;
			}
		}
	}

	for (SI_NIT::entries::entryScanner se = nit.get_entries().entryRange(); !se; ++se) {
		ChaineIdsNIT chIDs;

		chIDs.TSID = se().transport_stream_id();
		chIDs.ONID = se().original_network_id();

		// myprintf(TEXT("parse_nit: TSID=%i, ONID=%i, taille=%i") EOL, chIDs.TSID, chIDs.ONID, se().dloop_length());

		for (MDESCR::D_scanner sed(se().descrRange()); !sed; ++sed) {
			// myprintf(TEXT("parse_nit: type : %02X, taille : %i") EOL, sed().tag, sed().length);

			if (sed().tag == MDESCR::mdt_Logical_Channel) {
				for (MDESCR::D_83::sdScanner seds(sed().as<MDESCR::D_83>().subDescrRange()); !seds; ++seds) {
					chIDs.SID = seds().service_id();
					chIDs.nNumeroNit = seds().logical_channel_number();
					//myprintf(TEXT("parse_nit: ONID=%i, TSID=%i, SID=%i, numero=%i") EOL,
					// 	chIDs.ONID, chIDs.TSID, chIDs.SID, chIDs.nNumeroNit);

					// On ne recherche pas les ventuels doublons : cette table est trs provisoire,
					// vrifier l'existence d'un doublon risquerait d'tre plus pnalisant en temps
					// d'excution que le balayage (occasionnel puisque la premire occurrence
					// interrompra la recherche) d'une table plus longue, et trier la table ne
					// permettrait pas de faire ressortir les ajouts  la fin.
					cInfoNIT.push_back(chIDs);
				}
			}
		}
	}
}

// ====================================================================================
