/*
 * $Id: rendering.cpp 277 2011-03-13 10:24:41Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentations pour le rendu vido et pour l'OSD.
 *
 *	Implmentations des fonctionnalits ncessaires au rendu vido (\p VMR7,
 *	\p VMR9, \p EVR...) ainsi que pour l'affichage sur cran (OSD).
 **/

#include "stdafx.h"
#include "rendering.h"
#include "search.h"
#include "settings.h"

#if USE_VMR_DEINTERLACE
	#include <dxva.h>
#endif

// ====================================================================================
// Macros  usage local
// ====================================================================================

/// Intervalle minimal entre deux actualisations de la position de la vido
#define DUREE_VIDEO_MOVE	100

#define VMR_INSTANCE		0x118218

/// Accs  un membre de structure dfini par un offset extrait par \p offsetof
#define memberof(s, t, of) *reinterpret_cast<t *>((reinterpret_cast<BYTE *>(&s))+of)

// ====================================================================================
// Fonctions locales
// ====================================================================================

/**
 * Application d'un facteur de zoom  un rectangle vido
 * \param[in]	sRect	Rectangle sur lequel appliquer le zoom.
 **/
static void ApplyZoom(RECT & sRect)
{
	if (zoom_ratio!=100) {
		SIZE sSiz	= RSiz(sRect);	// Dimensions du rectangle initial
		SIZE sNSz	= {(sSiz.cx * zoom_ratio)/100, (sSiz.cy * zoom_ratio)/100};	// Nouvelle taille

		sRect.left		-=	(sNSz.cx - sSiz.cx) / 2;
		sRect.top		-=	(sNSz.cy - sSiz.cy) / 2;
		sRect.right		=	sRect.left + sNSz.cx;
		sRect.bottom	=	sRect.top + sNSz.cy;
	}
}

/**
 * Positionnement de la vido dans un rectangle de forme et de dimensions quelconque
 * (prsum tre un rectangle client de fentre d'affichage).
 * Les valeurs des variables globales \p #video_pan_mode et \p #zoom_ratio sont prises en
 * compte dans ce positionnement.
 *
 * \param[in]	sSrcRect	Rectangle client de la fenetre d'affichage (normalement avec origine
 *							zro, mais les cas o il en irait autrement sont pris en compte).
 * \param[in]	sAr			Ratio d'aspect  utiliser
 * \param[in]	pVRect		Pointeur optionnel (si non NULL) sur un rectangle destin  recevoir
 *							l'intersection entre le rectangle client et la vido positionne
 * \return					Rectangle de la vido positionne.
 *
 * \note Du fait des dplacements produits par les variables \p #video_pan_mode et
 * \p #zoom_ratio, la taille de la vido peut excder celle de la fentre. Ceci induit, selon
 * les valeurs fournies en entre, la possibilit d'avoir des valeurs ngatives en rsultat,
 * particulirement sur le bords gauche et le bord suprieur.
 **/
RECT PlaceVideoInRect(const RECT & sSrcRect, SIZE sAr, LPRECT pVRect)
{
	RECT	sRect	= sSrcRect;

	if (!IsRectEmpty(&sRect)) {
		// Correction panoramique selon la variable "video_pan_mode" :
		if (video_pan_mode != pm_Stretch) {	// (si la vido doit tre tire, on ne change rien)
			SIZE	sSiz	= RSiz(sRect);	// Dimensions du rectangle initial
			LONG	nCYxX	= sSiz.cy * sAr.cx;
			LONG	nCXxY	= sSiz.cx * sAr.cy;
			LONG	nDiff;

			if (video_pan_mode==pm_AllWidth || video_pan_mode==pm_Normal && nCYxX > nCXxY) {
				// trop haut de combien de pixels ?
				nDiff			=	(sSiz.cy - (nCXxY / sAr.cx)) / 2;
				sRect.top		+=	nDiff;
				sRect.bottom	-=	nDiff;
			} else if (video_pan_mode==pm_AllHeight || video_pan_mode==pm_Normal && nCYxX < nCXxY) {
				// trop large de combien de pixels ?
				nDiff			= (sSiz.cx - (nCYxX / sAr.cy)) / 2;
				sRect.left		+=	nDiff;
				sRect.right		-=	nDiff;
			}
		}

		// Correction selon le facteur de zoom "zoom_ratio"
		ApplyZoom(sRect);
	}

	if (pVRect)
		IntersectRect(pVRect, &sSrcRect, &sRect);
	return sRect;
}

#if USE_VMR_DEINTERLACE

template <typename VD, typename IDC>
	HRESULT GetDeinterlaceModes(IDC * pDeint, VD & desc, vector<GUID> & vGuids)
{
	DWORD	num;
	HRESULT	hr = pDeint->GetNumberOfDeinterlaceModes(&desc, &num, NULL);

	if (check_hr(hr, "GetNumberOfDeinterlaceModes (1)"))
		return hr;

	myprintf(TEXT("Nombre de modes de dsentrelacement : %i") EOL, num);
	vGuids.resize(num);

	hr = pDeint->GetNumberOfDeinterlaceModes(&desc, &num, &vGuids[0]);
	check_hr(hr, "GetNumberOfDeinterlaceModes (2)");
	return hr;
}

#endif	// #if USE_VMR_DEINTERLACE

// ====================================================================================
// Objets  usage local
// ====================================================================================

struct InputMediaType : public AM_MEDIA_TYPE
{
	/// Constructeur
	InputMediaType(IPin * pPin) {
		ZeroMemory(this, sizeof(AM_MEDIA_TYPE));
		// myprintf(COLR(HBLUE, TEXT("InputMediaType(0x%08x), thread ID=0x%x")) EOL, pPin, GetCurrentThreadId());
		HRESULT hr = pPin ? pPin->ConnectionMediaType(this) : E_FAIL;
		check_hr(hr, "ConnectionMediaType"); }

	bool isValid() const {
			return pbFormat != NULL; }

#if USE_VMR_DEINTERLACE
	const VIDEOINFOHEADER * asVideoInfo() const {
			return formattype == FORMAT_VideoInfo ? reinterpret_cast<const VIDEOINFOHEADER *>(pbFormat) : NULL; }
#endif

	const VIDEOINFOHEADER2 * asVideoInfo2() const {
			return formattype == FORMAT_VideoInfo2 ? reinterpret_cast<const VIDEOINFOHEADER2 *>(pbFormat) : NULL; }

#if USE_VMR_DEINTERLACE
	static void ConvertFrequency(DWORD nTim, DWORD & dwNumerator, DWORD & dwDenominator)
	{
		switch (nTim) {

		case 166833: dwNumerator = 60000;		dwDenominator = 1001;	break;
		case 333667: dwNumerator = 30000;		dwDenominator = 1001;	break;
		case 417188: dwNumerator = 24000;		dwDenominator = 1001;	break;
		case 200000: dwNumerator = 50;			dwDenominator = 1;		break;
		case 400000: dwNumerator = 25;			dwDenominator = 1;		break;
		case 416667: dwNumerator = 24;			dwDenominator = 1;		break;
		default:	 dwNumerator = 10000000;	dwDenominator = nTim;
		}
	}
#endif

#if USE_VMR_DEINTERLACE
	/// Dfinition d'un descripteur vido (VMR7)
	void FillVideoDesc(VMRVideoDesc & sDesc) const
	{
		ZeroMemory(&sDesc, sizeof(sDesc));
		sDesc.dwSize		= sizeof(sDesc);

		const VIDEOINFOHEADER *		pVih;
		const VIDEOINFOHEADER2 *	pVih2;

		if ((pVih2 = asVideoInfo2()) != NULL) {
			sDesc.dwSampleWidth			= pVih2->bmiHeader.biWidth;
			sDesc.dwSampleHeight		= abs(pVih2->bmiHeader.biHeight);
			if (pVih2->dwInterlaceFlags & AMINTERLACE_IsInterlaced)
				sDesc.SingleFieldPerSample	= TRUE;
			sDesc.dwFourCC			= pVih2->bmiHeader.biCompression;
			ConvertFrequency((DWORD)pVih2->AvgTimePerFrame, sDesc.InputSampleFreq.dwNumerator, sDesc.InputSampleFreq.dwDenominator);
			sDesc.OutputFrameFreq	= sDesc.InputSampleFreq;
			if (sDesc.SingleFieldPerSample)
				sDesc.OutputFrameFreq.dwNumerator *= 2;
		} else if ((pVih = asVideoInfo()) != NULL) {
			sDesc.dwSampleWidth		= pVih->bmiHeader.biWidth;
			sDesc.dwSampleHeight	= abs(pVih->bmiHeader.biHeight);
			sDesc.dwFourCC			= pVih->bmiHeader.biCompression;
			ConvertFrequency((DWORD)pVih2->AvgTimePerFrame, sDesc.InputSampleFreq.dwNumerator, sDesc.InputSampleFreq.dwDenominator);
			sDesc.OutputFrameFreq	= sDesc.InputSampleFreq;
		}
	}
#endif


	/// Destructeur
	~InputMediaType() {
		FreeMediaType(*this); }
};

/// Structure d'assistance  l'accs aux proprits de la vido
struct PA_Parms {
	BYTE		nFlags;		//!< Slecteur de la proprits
	BYTE		nCtlOfs;	//!< Offset de la valeur nativ
};

// ====================================================================================
// Implmentations des mthodes de la classe CMediaRenderer
// ====================================================================================

/// Initialisation gnrale
HRESULT CMediaRenderer::Init(IPTvM_GraphBase & iGraph)
{
	myprintf(COLR(LCYAN, TEXT("Initialisation du rendu %s (%s)")) EOL, TypeStr(), pszName);

	// cre le filtre de rendu
	HRESULT hr = iGraph.create_and_add_filter(*this);
	if (FAILED(hr))
		return hr;

	// Rcupration de la broche d'entre (pour le ratio d'aspect)
	hr = cherche_pin(pFilter, PINDIR_INPUT, pInputPin);
	check_hr(hr, "CMediaRenderer::Init / cherche_pin");

	return Configure();
}

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
/// Connexion  une sortie de filtre
HRESULT CMediaRenderer::Connect(IPTvM_GraphBase & iGraph, IBaseFilter * pMediaCodec)
{
	if (bConnected)
		return S_OK;	// Dj connect

	HRESULT hr = E_FAIL;

	if (!pFilter)
		return erreur(TEXT("   %s"), hr, TypeStr());
	hr = iGraph.connect_filters(pMediaCodec, pFilter);
	// Problme rcurrent  identifier ici : erreur 0x80040207
	// = VFW_E_NO_ACCEPTABLE_TYPES ("There is no common media type between these pins")
	if (FAILED(hr)) {
		// Ici, on affiche l'erreur mais on ne quitte pas (a laisse la possibilit d'enregistrer)
		erreur(
			TEXT("    %s\n  %s \"%s\".") EOL
			TEXT("Render     .") EOL
			TEXT("     "), hr, TypeStr(), TypeStr(), pszName);
		affiche_erreurs(NULL);
		hr = S_OK;
	} else
		bConnected = true;
	return hr;
}
#endif // #if BUILD_UNUSED_CODE

#if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
/// Dconnexion
HRESULT CMediaRenderer::Disconnect(IPTvM_GraphBase & iGraph)
{
	HRESULT	hr = S_OK;

	if (bConnected) {	// si pas dj dconnect
		if (FAILED(hr = iGraph.disconnect_filter(pFilter, PINDIR_INPUT)))
			return hr;
		bConnected = false;
	}
	return hr;
}
#endif // #if BUILD_UNUSED_CODE

/// Destructeur
CMediaRenderer::~CMediaRenderer()
{
	myprintf(TEXT("Destruction du rendu (%s)") EOL, pszName);

	if (pFilter) {
		CComPtr<IFilterGraph> pGrph = graph_from_filter();

		if (pGrph) {
			// Retrait du filtre (probablement prfrable car ventualit possible de
			// changement de rendu  chaud)
			HRESULT				hr = pGrph->RemoveFilter(pFilter);

			check_hr(hr, "CMediaRenderer / RemoveFilter");
		}
	}
}

// ====================================================================================
// Implmentations des mthodes de la classe CVideoRenderer
// ====================================================================================

/// Initialisation gnrale
HRESULT CVideoRenderer::Init(IPTvM_GraphBase & iGraph)
{
	HRESULT hr = __super::Init(iGraph);

	if (FAILED(hr))
		return hr;

	hr = PrepareSurface();
	if (FAILED(hr))
		return hr;

#if USE_VMR_DEINTERLACE
	if (use_vmr_deinterlace) {
		if (FAILED(hr = SetMixingPrefs()))
			return hr;
		if (FAILED(hr = InitDeinterlace()))
			return hr;
	}
#endif

	hr = InitControl();
	if (FAILED(hr))
		return hr;

	hr = NonDelegatingSetAspectRatioMode(false);	// Mode "letterbox" la premire fois
	if (FAILED(hr))
		return hr;

	// myprintf(CMyP TEXT("Rendu vido (%s) initialis") EOL, SUCCEEDED(hr), pszName);
	return hr;
}

/// Initialisation des rglages de couleur de la vido
HRESULT CVideoRenderer::InitProcAmp(INT32 sInitialValues[4])
{
	if (isProcAmpAvailable() == paa_MissingFromMode)
		return E_NOTIMPL;

	ProcAmpValues	sProcAmpVals;
	HRESULT			hr = GetProcAmpValues(sProcAmpVals);

	if (FAILED(hr))
		return hr;	// Probablement code 0x80040299 = VFW_E_VMR_NO_PROCAMP_HW

	// Conversion des valeurs :
	for (ProcAmpType eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp)
		sProcAmpVals.aVal[eTyp].SetIntVal(sInitialValues[eTyp]);

	hr = SetProcAmpValues(sProcAmpVals);
	if (hr == S_OK)
		eProcAmpAvail = paa_Available;
	return hr;
}

/** \brief Rcupration des informations sur la vido, depuis la broche d'entre du filtre de rendu
 *
 * \param[out]	sVInf	Informations rcupres
 * \return \p true ou \p false selon que les informations ont pu tre rcupres ou non ; cette valeur
 * peut tre ignore si les valeurs par dfaut suffisent.
 * \note Si les informations rclames ne peuvent tre obtenues, des valeurs par dfaut sont
 * retournes
 **/
bool CVideoRenderer::GetVideoInfo(LocalVideoInfo & sVInf) const
{
	if (pInputPin) {
		InputMediaType sAmt(pInputPin);

		if (sAmt.isValid()) {
			const VIDEOINFOHEADER2 * vih = sAmt.asVideoInfo2();

			if (
				vih &&
				/*vih->rcSource.right > 0 && vih->rcSource.bottom > 0 &&*/
				vih->dwPictAspectRatioX > 0 && vih->dwPictAspectRatioY > 0
			) {
				sVInf.sAspectRatio.cx = vih->dwPictAspectRatioX;
				sVInf.sAspectRatio.cy = vih->dwPictAspectRatioY;
				sVInf.sNativeSize.cx = vih->rcSource.right;
				sVInf.sNativeSize.cy = vih->rcSource.bottom;

/*
				myprintf(TEXT("GetVideoInfo: ar=%i/%i, siz=%i/%i, type=%c%c%c%c") EOL,
					sVInf.sAspectRatio.cx, sVInf.sAspectRatio.cy,
					sVInf.sNativeSize.cx, sVInf.sNativeSize.cy,
					LPCSTR(&sAmt.subtype.Data1)[0], LPCSTR(&sAmt.subtype.Data1)[1],
					LPCSTR(&sAmt.subtype.Data1)[2], LPCSTR(&sAmt.subtype.Data1)[3]);
*/
				return true;
			}
		}
	}

	// Valeurs par dfaut :
	sVInf.sAspectRatio.cx = DEFAULT_AR_X;
	sVInf.sAspectRatio.cy = DEFAULT_AR_Y;
	sVInf.sNativeSize.cx = 0;
	sVInf.sNativeSize.cy = 0;
	return false;
}

/// Rcupration du ratio d'aspect courant
SIZE CVideoRenderer::GetAspectRatio() const
{
	static SIZE		aAr[] =
	{
		{0,0},
		{1,1},
		{5,4},
		{4,3},
		{16,10},
		{16,9}
	};

	if (aspect_ratio != ar_Auto)
		return aAr[aspect_ratio];

	LocalVideoInfo	sVInf;

	GetVideoInfo(sVInf);

	SIZE			sAr	= Simplify(sVInf.sAspectRatio);

	// myprintf(TEXT("Aspect ratio: x=%i, y=%i") EOL, sAr.cx, sAr.cy);

	return sAr;
}

/** \brief Dbut de mise  jour de la couche OSD dans l'affichage
 *
 * Cette mthode prpare la couche OSD (effacement, redimensionnement si ncessaire) afin
 * d'en redessiner le contenu.
 * Elle retourne un pointeur sur l'interface  utiliser pour dessiner les objets dans cette
 * couche OSD.
 * \param[in] hDc	"Handle" sur le "device context" de la fentre. Utilis seulement
 *					si la mise  jour est demande par un message \p WM_PAINT avec certains
 *					rendus vido. Dans tous les autres cas, mettre \p NULL.
 * \return	Un pointeur partag sur l'interface de cette couche
 **/
COsdLayerPtr CVideoRenderer::BeginDraw(HDC hDc)
{
	COsdLayerPtr pOsdLayer = COsdLayerPtr(COsdLayer::CreateWinInstance(hDc));

	if (!pOsdLayer->layer_set())
		pOsdLayer->Create();
	else
		pOsdLayer->Erase();
	return pOsdLayer;
}

void CVideoRenderer::Paint(HDC hDC)
{
	if (pOSD)
		pOSD->Redraw(hDC);
}

void CVideoRenderer::UpdateVideoPos()
{
	DWORD tick		= GetTickCount();
	DWORD elapsed	= tick-last_tick;

	if (elapsed < DUREE_VIDEO_MOVE) {
		// Le dernier appel  cette fonction est vieux de moins que DUREE_VIDEO_MOVE :
		// On ne fait rien, et on programme un timer pour rappeler la mme fonction
		// lorsque ce dlai sera coul.
		SetTimer(hMainWnd, TIMER_VIDEO_MOVE, DUREE_VIDEO_MOVE-elapsed, NULL);
		return;
	}
	last_tick = tick;

	// Calcule le rectangle de la vido :
	RECT sRect = PlaceVideoInRect(CalcCliRect(), GetAspectRatio());

	// myprintf(TEXT("SetVideoPosition([") RECTFMT(",") TEXT("])") EOL, RECTPARM(sRect));
	SetAspectRatioMode(video_pan_mode == pm_Stretch || aspect_ratio != ar_Auto);
	SetVideoPosition(&sRect);
}

// ====================================================================================
// Dclaration de la classe CVideoRendererWithBitmapOSD
// ====================================================================================

/// Classe intermdiaire pour les rendus vido synthtisant l'OSD dans une image bitmap
class CVideoRendererWithBitmapOSD : public CVideoRenderer
{
protected:

	shared_ptr<COsdLayer>			m_pOsdLayer;
	bool							bBmInVideo;		//!<  \p true si un bitmap est charg dans l'affichage

	// static float					ArR(const RECT & r) {
	//									return float(r.right-r.left) / float (r.bottom - r.top); }
private:
	DWORD							dwLastOsdSet;	//!< Temps du dernier changement OSD

	virtual HRESULT		PrepareSurface();

	/// Incrustation de la couche OSD dans l'affichage
	virtual HRESULT		PresentOsdLayer() = 0;

	/// Retrait de la couche OSD de l'affichage
	virtual HRESULT		RemoveOsdLayer() = 0;

	/// Fonction destine  limiter le nombre de modifications par seconde aux
	/// incrustations OSD (pour viter que des changements trop rapides puissent
	/// "casser" la vido).  appeler juste avant l'incrustation de la couche OSD.
	void				LimitOsdRate();

protected:
	CVideoRendererWithBitmapOSD(LPCTSTR pszNam, REFCLSID clsid) :
		CVideoRenderer(pszNam, clsid),
		bBmInVideo(false),
		dwLastOsdSet(0)
	{
		eProcAmpAvail = paa_MissingFromGPU;
	}

	/// Calcul de la taille  attribuer au bitmap  incruster dans la vido
	virtual SIZE		GetBitmapSize() const {
							return RSiz(CalcCliRect()); }	// Retourne la taille du rectangle client si pas surcharg

	/// Calcul de la taille du bitmap si celui-ci doit se limiter  la surface de la vido
	SIZE				GetBitmapInVideoSize() const;

	/// Repositionnement du rectangle du bitmap OSD  l'intrieur de celui de la video
	/// \param[in] sRect	Rectangle client de la fentre
	/// \return				Rectangle calcul
	RECT				PlaceOsdBitmapInVideo(const RECT & sRect);

	/** \brief Dbut de mise  jour de la couche OSD dans l'affichage
	 *
	 * Cette mthode prpare la couche OSD (effacement, redimensionnement si ncessaire) afin
	 * d'en redessiner le contenu.
	 * Elle retourne un pointeur sur l'interface  utiliser pour dessiner les objets dans cette
	 * couche OSD.
	 * \param[in] hDc	"Handle" sur le "device context" de la fentre. Utilis seulement
	 *					si la mise  jour est demande par un message \p WM_PAINT avec certains
	 *					rendus vido. Dans tous les autres cas, mettre \p NULL.
	 * \return	Un pointeur partag sur l'interface de cette couche
	 **/
	virtual COsdLayerPtr
						BeginDraw(HDC hDC = NULL);

	/** \brief Fin de mise  jour (avec validation) de la couche OSD dans l'affichage
	 *
	 * Cette mthode termine la construction de la couche OSD, et assure sa prsentation
	 * effective dans l'affichage.
	 **/
	virtual HRESULT		EndDraw();
};

// ====================================================================================
// Dclaration de la classe CNullVideoRenderer
// ====================================================================================

/// Classe de substitution pour l'absence de rendu vido
class CNullVideoRenderer : public CVideoRenderer
{
	HPEN hPen;

	/// Point d'appel pour redessiner cet objet OSD en rponse  \p WM_PAINT
	/// \param[in]	hDC	"Device context"  utiliser
	virtual void	Paint(HDC hDC);

	/// Appel sur notification de redimensionnement de la fentre
	virtual void	Resize() {
						InvalidateRect(hMainWnd, NULL, TRUE);}

public:
	/// Constructeur
	CNullVideoRenderer();

	/// Destructeur
	virtual ~CNullVideoRenderer();
};

// ====================================================================================
// Dclaration de la classe CVideoMixingRenderer7
// ====================================================================================

/// Classe commune pour les implmentations utilisant le Video Mixing Renderer version 7
class CVideoMixingRenderer7 : public CVideoRenderer
{
	CComPtr<IVMRWindowlessControl> pVMR7Control;

#if USE_VMR_DEINTERLACE
	virtual HRESULT		SetMixingPrefs();
	virtual HRESULT		InitDeinterlace();	//!< \todo Voir \p CVideoRenderer::InitDeinterlace
#endif
	virtual HRESULT		InitControl();

	VMR_ASPECT_RATIO_MODE
						ARMode(bool bEtirer) const {
							return bEtirer ? VMR_ARMODE_NONE : VMR_ARMODE_LETTER_BOX; }

	/// Dfinition de la position d'affichage et du ratio d'aspect.
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT		SetVideoPosition(const LPRECT pDstRect);

	/// Obtenir la rsolution native de la vido
	virtual SIZE		GetNativeSize() const;

	virtual HRESULT		DisplayModeChanged();

	/// Rcuprer l'image affiche (en vue d'une capture statique)
	/// \param[out]	pDib	Pointeur recevant l'image capture
	virtual HRESULT		GetCurrentImage(CComHeapPtr<BYTE> & pDib);

protected:

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT			NonDelegatingSetAspectRatioMode(bool bEtirer);

public:
	// Constructeur
	CVideoMixingRenderer7() :
		CVideoRenderer(TEXT("Video Mixing Renderer 7"), CLSID_VideoMixingRenderer)
	{
	}
};

// ====================================================================================
// Dclaration de la classe CVideoMixingRenderer7_renderless
// ====================================================================================

/// Implmentation utilisant le Video Mixing Renderer version 7 en mode "renderless"
class CVideoMixingRenderer7_renderless : public CVideoMixingRenderer7
{
	virtual HRESULT		Configure();
	virtual HRESULT		PrepareSurface();
public:
	// Constructeur
	CVideoMixingRenderer7_renderless() :
		CVideoMixingRenderer7()
	{
	}

	/// Obtenir le type de VMR courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_VMR7_renderless;}
};

// ====================================================================================
// Dclaration de la classe CVideoMixingRenderer7_windowless
// ====================================================================================

/// Implmentation utilisant le Video Mixing Renderer version 7 en mode "windowless"
class CVideoMixingRenderer7_windowless : public CVideoMixingRenderer7
{
	virtual HRESULT		Configure();
public:
	// Constructeur
	CVideoMixingRenderer7_windowless() :
		CVideoMixingRenderer7()
	{
	}

	/// Obtenir le type de VMR courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_VMR7_windowless;}
};

// ====================================================================================
// Dclaration de la classe CVideoMixingRenderer9
// ====================================================================================

/// Comparaison de rectangles \p VMR9NormalizedRect pour galit
inline bool operator == (const VMR9NormalizedRect & sRct1, const VMR9NormalizedRect & sRct2)
{
	return
		sRct1.left   == sRct2.left  &&
		sRct1.top    == sRct2.top   &&
		sRct1.right  == sRct2.right &&
		sRct1.bottom == sRct2.bottom;
}

/// Implmentation utilisant le Video Mixing Renderer version 9
class CVideoMixingRenderer9 : public CVideoRendererWithBitmapOSD
{
protected:

	CComPtr<IVMRWindowlessControl9> pVMR9Control;	//!< Interface de contrle
	CComPtr<IVMRMixerControl9>		pVMR9Mixer;		//!< Interface du mixer
	CComPtr<IVMRMixerBitmap9>		pBMP;			//!< Interface de transfert du bitmap
	static const VMR9NormalizedRect	rRefNRect;		//!< Rectangle normalis de rfrence {0.0f,0.0f,1.0f,1.0f}
	static const PA_Parms			aParms[];		//!< Paramtres utiliss pour le contrle de la vido

	HRESULT				UpdateOutputRect_internal(const VMR9NormalizedRect & rNewOut);
	HRESULT				UpdateVideoPosition_internal(const LPRECT lpSRCRect, const LPRECT lpDSTRect);
#if BUILD_UNUSED_CODE
	HRESULT				UpdateAlphaBitmapPos_internal(const VMR9NormalizedRect & rNewPos);
#endif

	// static float		ArN(const VMR9NormalizedRect & r) {
	//						return float(r.right-r.left) / float (r.bottom - r.top); }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		NonDelegatingSetAspectRatioMode(bool bEtirer);

protected:

	virtual HRESULT		Configure();
	virtual HRESULT		InitControl();

	/// Obtenir le code de mode de ration d'aspect
	/// \param[in]	bEtirer	\p true pour mode remplissage d'cran
	VMR9AspectRatioMode	ARMode(bool bEtirer) const {
							return bEtirer ? VMR9ARMode_None : VMR9ARMode_LetterBox; }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		SetAspectRatioMode(bool bEtirer) {
							return NonDelegatingSetAspectRatioMode(true); }
#if NEW_VMR9_OUTPUT
	/// Dfinition de la position d'affichage et du ratio d'aspect.
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT		SetVideoPosition(const LPRECT pDstRect);
#endif

	/// Obtenir la rsolution native de la vido
	virtual SIZE		GetNativeSize() const;

	/// Incrustation de la couche OSD dans l'affichage
	virtual HRESULT		PresentOsdLayer() {
							return m_pOsdLayer->Present(pBMP); }

	/// Retrait de la couche OSD de l'affichage
	virtual HRESULT		RemoveOsdLayer();

	virtual HRESULT		DisplayModeChanged();

	/// Rcuprer l'image affiche (en vue d'une capture statique)
	/// \param[out]	pDib	Pointeur recevant l'image capture
	virtual HRESULT		GetCurrentImage(CComHeapPtr<BYTE> & pDib);

	/// Redessiner l'espace de la vido
	virtual void		Paint(HDC hDC);

	virtual HRESULT		GetProcAmpValues(ProcAmpValues & sValues) const;

	virtual HRESULT		SetProcAmpValues(ProcAmpValues & sValues);

public:

	CVideoMixingRenderer9() :
		CVideoRendererWithBitmapOSD(TEXT("Video Mixing Renderer 9"), CLSID_VideoMixingRenderer9)
	{
	}

#if NEW_VMR9_OUTPUT
	/// Obtenir le type de VMR courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_VMR9_windowless_new; }
#endif
};

/// Drivation de la classe \p CVideoMixingRenderer9 pour retrouver le fonctionnement de l'ancienne version
class CVideoMixingRenderer9_Old : public CVideoMixingRenderer9
{
	/// Calcul du rectangle utilis pour la vido ( l'usage de la construction de l'OSD)
	/// \param[in] sRect	Rectangle client de la fentre, calcul par \p CalcCliRect
	/// \return				Rectangle calcul, utilisable pour l'OSD
	virtual RECT			CalcOsdRect(RECT sRect) {
								return PlaceOsdBitmapInVideo(sRect); }

	/// Calcul de la taille  attribuer au bitmap  incruster dans la vido
	virtual SIZE			GetBitmapSize() const {
								return GetBitmapInVideoSize(); }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT			SetAspectRatioMode(bool bEtirer) {
								return NonDelegatingSetAspectRatioMode(bEtirer); }

	/// Dfinition de la position d'affichage et du ratio d'aspect (ancienne version).
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT			SetVideoPosition(const LPRECT pDstRect);

public:

	/// Obtenir le type de VMR courant
	virtual RdrTypes		GetRdrType() const {
								return rdt_VMR9_windowless; }
};

// ====================================================================================
// Dclaration de la classe CEnhancedVideoRenderer
// ====================================================================================

/// Comparaison de rectangles \p MFVideoNormalizedRect pour galit
inline bool operator == (const MFVideoNormalizedRect & sRct1, const MFVideoNormalizedRect & sRct2)
{
	return
		sRct1.left   == sRct2.left  &&
		sRct1.top    == sRct2.top   &&
		sRct1.right  == sRct2.right &&
		sRct1.bottom == sRct2.bottom;
}

/// Implmentation utilisant le Enhanced Video Renderer
class CEnhancedVideoRenderer : public CVideoRendererWithBitmapOSD
{
protected:

	CComPtr<IMFVideoDisplayControl>		pEVRControl;	//!< Interface de contrle
	CComPtr<IMFVideoProcessor>			pEVRVideoProc;	//!< Interface du processeur vido
	CComPtr<IMFVideoMixerControl>		pEVRMixCtrl;	//!< Interface de contrle du mlangeur
	CComPtr<IMFVideoMixerBitmap>		pBMP;			//!< Interface de transfert du bitmap
	static const MFVideoNormalizedRect	rRefNRect;		//!< Rectangle normalis de rfrence {0.0f,0.0f,1.0f,1.0f}
	static const PA_Parms				aParms[];		//!< Paramtres utiliss pour le contrle de la vido

	HRESULT				UpdateOutputRect_internal(const MFVideoNormalizedRect & rNewOut);
	HRESULT				UpdateVideoPosition_internal(const MFVideoNormalizedRect * pnrcSource, const LPRECT prcDest);

	// static float		ArN(const MFVideoNormalizedRect & r) {
	//						return float(r.right-r.left) / float (r.bottom - r.top); }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		NonDelegatingSetAspectRatioMode(bool bEtirer);

private:

	virtual HRESULT		InitControl();

	/// Obtenir le code de mode de ration d'aspect
	/// \param[in]	bEtirer	\p true pour mode remplissage d'cran
	MFVideoAspectRatioMode
						ARMode(bool bEtirer) const {
							return bEtirer ? MFVideoARMode_None : MFVideoARMode_PreservePicture; }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		SetAspectRatioMode(bool bEtirer) {
							return NonDelegatingSetAspectRatioMode(true); }

	/// Dfinition de la position d'affichage et du ratio d'aspect.
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT		SetVideoPosition(const LPRECT pDstRect);

	/// Obtenir la rsolution native de la vido
	virtual SIZE		GetNativeSize() const;

	/// Incrustation de la couche OSD dans l'affichage
	virtual HRESULT		PresentOsdLayer() {
							return m_pOsdLayer->Present(pBMP); }

	/// Retrait de la couche OSD de l'affichage
	virtual HRESULT		RemoveOsdLayer() {
							return pBMP->ClearAlphaBitmap(); }

	/// Rcuprer l'image affiche (en vue d'une capture statique)
	/// \param[out]	pDib	Pointeur recevant l'image capture
	virtual HRESULT		GetCurrentImage(CComHeapPtr<BYTE> & pDib);

	/// Redessiner l'espace de la vido
	virtual void		Paint(HDC hDC);

	virtual HRESULT		GetProcAmpValues(ProcAmpValues & sValues) const;

	virtual HRESULT		SetProcAmpValues(ProcAmpValues & sValues);

public:

	CEnhancedVideoRenderer() :
		CVideoRendererWithBitmapOSD(TEXT("Enhanced Video Renderer"), CLSID_EnhancedVideoRenderer)
	{
	}

	/// Obtenir le type de VMR courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_EVR_new; }
};

/// Drivation de la classe \p CEnhancedVideoRenderer pour retrouver le fonctionnement de l'ancienne version
class CEnhancedVideoRenderer_Old : public CEnhancedVideoRenderer
{
	/// Fonction "callback" utilise pour le rafrachissement retard de la vido
	static VOID CALLBACK	SetVideoPosTimerProc(HWND hWnd, UINT, UINT_PTR idEvent, DWORD);

	/// Calcul du rectangle utilis pour la vido ( l'usage de la construction de l'OSD)
	/// \param[in] sRect	Rectangle client de la fentre, calcul par \p CalcCliRect
	/// \return				Rectangle calcul, utilisable pour l'OSD
	virtual RECT			CalcOsdRect(RECT sRect) {
								return PlaceOsdBitmapInVideo(sRect); }

	virtual SIZE			GetBitmapSize() const {
								return GetBitmapInVideoSize(); }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT			SetAspectRatioMode(bool bEtirer) {
								return NonDelegatingSetAspectRatioMode(bEtirer); }

	/// Dfinition de la position d'affichage et du ratio d'aspect (ancienne version).
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT			SetVideoPosition(const LPRECT pDstRect);

public:

	/// Obtenir le type de VMR courant
	virtual RdrTypes		GetRdrType() const {
								return rdt_EVR; }
};

// ====================================================================================
// Dclaration de la classe CDirectSoundRenderer
// ====================================================================================

class CDirectSoundRenderer : public CAudioRenderer
{
public:
	CDirectSoundRenderer() :
		CAudioRenderer(TEXT("DirectSound"), CLSID_DSoundRender)
	{
	}

	/// Obtenir le type de rendu courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_DirectSound; }
};

// ====================================================================================
// Dclaration de la classe CNullAudioRenderer
// ====================================================================================

class CNullAudioRenderer : public CAudioRenderer
{
public:
	CNullAudioRenderer() :
		CAudioRenderer(TEXT("Null Audio Renderer"), CLSID_NullRenderer)
	{
	}
};

// ====================================================================================
// Implmentations des mthodes de la classe CVideoRendererWithBitmapOSD
// ====================================================================================

HRESULT CVideoRendererWithBitmapOSD::PrepareSurface()
{
	m_pOsdLayer = COsdLayerPtr(COsdLayer::CreateInstance(lt_BMP));
	if (!m_pOsdLayer)
		return E_FAIL;
	if (pOSD)
		pOSD->SetFontAllocator(m_pOsdLayer->CreateFontAllocator());
	return m_pOsdLayer->Initialize();
}

/// Fonction destine  limiter le nombre de modifications par seconde aux
/// incrustations OSD (pour viter que des changements trop rapides puissent
/// "casser" la vido).  appeler juste avant l'incrustation de la couche OSD.
void CVideoRendererWithBitmapOSD::LimitOsdRate()
{
	#define OSD_RATE_LIMIT 200	// 200 millisecondes (5 fois par seconde)

	DWORD	dwTicks = GetTickCount();
	UINT	nDiff	= dwTicks-dwLastOsdSet;

	if (nDiff < OSD_RATE_LIMIT)
		Sleep(OSD_RATE_LIMIT-nDiff);
	dwLastOsdSet = dwTicks;
}

/// Calcul de la taille du bitmap si celui-ci doit se limiter  la surface de la vido
SIZE CVideoRendererWithBitmapOSD::GetBitmapInVideoSize() const
{
	if (video_pan_mode == pm_Stretch)
		return CVideoRendererWithBitmapOSD::GetBitmapSize();

	SIZE	sSize	= GetNativeSize();

	// myprintf(TEXT("GetBitmapSize, sSize: cx=%u/cy=%u") EOL, sSize.cx, sSize.cy);

	return sSize;
}

/// Repositionnement du rectangle du bitmap OSD  l'intrieur de celui de la video
/// \param[in] sRect	Rectangle client de la fentre
/// \return				Rectangle calcul
RECT CVideoRendererWithBitmapOSD::PlaceOsdBitmapInVideo(const RECT & sRect)
{
	if (IsRectEmpty(&sRect))
		return sRect;

	RECT	sIRct;
	RECT	sZRct = PlaceVideoInRect(sRect, GetAspectRatio(), &sIRct);
	SIZE	sVSiz = GetBitmapSize();

	// Ici, "sIRct" contient le rectangle utilisable pour l'OSD dans la fentre.
	// On va transformer a pour le rcuprer dans les coordonnes de la vido native.

	float	fx = 1.0f - float(RWdt(sIRct))/float(RWdt(sZRct));
	float	fy = 1.0f - float(RHgt(sIRct))/float(RHgt(sZRct));

	LONG	dx = LONG(sVSiz.cx * fx)/2;
	LONG	dy = LONG(sVSiz.cy * fy)/2;

	return MkRECT(dx, dy, sVSiz.cx-dx, sVSiz.cy-dy);
}

/** \brief Dbut de mise  jour de la couche OSD dans l'affichage
 *
 * Cette mthode prpare la couche OSD (effacement, redimensionnement si ncessaire) afin
 * d'en redessiner le contenu.
 * Elle retourne un pointeur sur l'interface  utiliser pour dessiner les objets dans cette
 * couche OSD.
 * \param[in] hDc	"Handle" sur le "device context" de la fentre. Utilis seulement
 *					si la mise  jour est demande par un message \p WM_PAINT avec certains
 *					rendus vido. Dans tous les autres cas, mettre \p NULL.
 * \return	Un pointeur partag sur l'interface de cette couche
 **/
COsdLayerPtr CVideoRendererWithBitmapOSD::BeginDraw(HDC)
{	// (paramtre HDC ici inutilis)
	if (m_pOsdLayer) {
		m_pOsdLayer->SetSize(GetBitmapSize());
		if (!m_pOsdLayer->layer_set())
			m_pOsdLayer->Create();
		else if (m_pOsdLayer->notEmpty())
			m_pOsdLayer->Erase();
	}
	return m_pOsdLayer;
}

/** \brief Fin de mise  jour (avec validation) de la couche OSD dans l'affichage
 *
 * Cette mthode termine la construction de la couche OSD, et assure sa prsentation
 * effective dans l'affichage.
 **/
HRESULT CVideoRendererWithBitmapOSD::EndDraw()
{
	HRESULT hr = S_OK;

	if (m_pOsdLayer && m_pOsdLayer->layer_set() && m_pOsdLayer->notEmpty()) {
		LimitOsdRate();			// Limitations du nombre de changements par seconde

		hr = PresentOsdLayer();	// Incrustation de la couche OSD dans l'affichage

		myprintf(CMyP FMT_ERRX("PresentOsdLayer") ERRX(TEXT("(%s)")),
			FAILED(hr), hr, DShowErrorToString(hr).c_str());

		if (SUCCEEDED(hr))
			bBmInVideo = true;
	} else if (bBmInVideo) {
		hr = RemoveOsdLayer();	// Retrait de la couche OSD de l'affichage

		if (SUCCEEDED(hr))
			bBmInVideo = false;
	}
	return hr;
}

// ====================================================================================
// Implmentations des mthodes de la classe CNullVideoRenderer
// ====================================================================================

/// Constructeur
CNullVideoRenderer::CNullVideoRenderer() :
	CVideoRenderer(TEXT("Null Video Renderer"), CLSID_NullRenderer),
	hPen(NULL)
{
	LOGBRUSH sLbrsh = {
		BS_SOLID,
		RGB(0, 0, 192),		// bleu pas trop intense
		0
	};

	hPen = ExtCreatePen(PS_COSMETIC|PS_SOLID, 1, &sLbrsh, 0, NULL);
	PostMessage(hMainWnd, WM_APP_VIDEO_UPDATED, 0, NULL);
}

/// Point d'appel pour redessiner cet objet OSD en rponse  \p WM_PAINT
/// \param[in]	hDC	"Device context"  utiliser
/// \todo Ajouter ventuellement un texte "Sortie vido dsactive" en milieu d'affichage
void CNullVideoRenderer::Paint(HDC hDC)
{

	__super::Paint(hDC);	// Grer l'OSD standard

	RECT				sRect = CalcCliRect();

	// Dessiner une croix en travers de l'affichage
	AutoSelect<HPEN>	cPS(hDC, hPen);

	MoveToEx(hDC, sRect.left, sRect.top, NULL);
	LineTo(hDC, sRect.right, sRect.bottom);
	MoveToEx(hDC, sRect.left, sRect.bottom, NULL);
	LineTo(hDC, sRect.right, sRect.top);
}

/// Destructeur
CNullVideoRenderer::~CNullVideoRenderer()
{
	if (hPen)
		DeleteObject(hPen);
}

// ====================================================================================
// Modle pour la prslection du moniteur
// ====================================================================================

/// Modle pour la prslection du moniteur dans certains rendus vido
template <typename I, typename D>
	class MonitorList : public vector<D>
{
public:
	CComPtr<I>		pMonCfg;

	MonitorList(CComPtr<IBaseFilter> pFilter)
	{
		pFilter.QueryInterface(&pMonCfg);
		if (pMonCfg) {
			DWORD	nNumMon;
			HRESULT	hr = pMonCfg->GetAvailableMonitors(NULL, 0, &nNumMon);

			if (hr == S_OK) {
				resize(nNumMon);
				hr = pMonCfg->GetAvailableMonitors(&(*this)[0], nNumMon, &nNumMon);
			}
		}
	}

	const D * FindBest(HWND hWnd) const
	{
		RECT	sRect;
		UINT	nMaxSurface = 0;
		UINT	nIdx = 0;

		if (GetWindowRect(hWnd, &sRect)) {
			for (ITERATE_CONST_VECTOR(*this, D, it)) {
				RECT	sIRct;

				if (IntersectRect(&sIRct, &it->rcMonitor, &sRect)) {
					UINT nSurface = RWdt(sIRct) * RHgt(sIRct);

					if (nSurface > nMaxSurface) {
						nMaxSurface = nSurface;
						nIdx = (UINT)distance(begin(), it);
					}
				}
			}
		}
		return &(*this)[nIdx];
	}
};

// ====================================================================================
// Implmentations des mthodes de la classe CVideoMixingRenderer7
// ====================================================================================

#if USE_VMR_DEINTERLACE

HRESULT CVideoMixingRenderer7::SetMixingPrefs()
{
	CComQIPtr<IVMRMixerControl> pMixControl(pFilter);

	if (!pMixControl)
		return erreur(
			TEXT(" IVMRMixerControl "),
			E_NOINTERFACE);

	HRESULT hr =
		pMixControl->SetMixingPrefs(
			MixerPref_NoDecimation | MixerPref_BiLinearFiltering | MixerPref_RenderTargetYUV);

	if (FAILED(hr))
		return erreur(TEXT("  SetMixingPrefs"), hr);
	return hr;
}

HRESULT CVideoMixingRenderer7::InitDeinterlace()
{
	CComQIPtr<IVMRDeinterlaceControl>	pDeint(pFilter);

	if (!pDeint)
		return erreur(
			TEXT(" IVMRDeinterlaceControl "),
			E_NOINTERFACE);

	VMRVideoDesc						desc;
	InputMediaType						sAmt(pInputPin);

	sAmt.FillVideoDesc(desc);

	vector<GUID>						vGuids;
	HRESULT								hr = GetDeinterlaceModes(pDeint.p, desc, vGuids);

	if (FAILED(hr))
		return hr;

	for (ITERATE_VECTOR(vGuids, GUID, it)) {
		myprintf(TEXT("Dsentrelacement mode %i = ") GUIDFMT_N EOL,
			distance(vGuids.begin(), it), GUIDPARM_N(*it));
		if (*it == DXVA_DeinterlaceBobDevice) {
			hr = pDeint->SetDeinterlaceMode(0xffffffff, &*it);
			check_hr(hr, "SetDeinterlaceMode");
			break;
		}
	}

	return hr;
}

#endif	// #if USE_VMR_DEINTERLACE

HRESULT CVideoMixingRenderer7::InitControl()
{
	HRESULT hr = pFilter.QueryInterface(&pVMR7Control);

	if (FAILED(hr))
		return erreur(TEXT("   (VMR)"), hr);

	hr = pVMR7Control->SetVideoClippingWindow(hMainWnd);
	check_hr(hr, "SetVideoClippingWindow");

	hr = pVMR7Control->SetColorKey(colorKey);
	check_hr(hr, "SetColorKey");
	return hr;
}

/// Dfinition du mode de ratio d'aspect.
/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
HRESULT CVideoMixingRenderer7::NonDelegatingSetAspectRatioMode(bool bEtirer)
{
	CComQIPtr<IVMRAspectRatioControl> pARC(pFilter);

	if (!pARC)
		return erreur(
			TEXT("   (aspect ratio)"),
			E_NOINTERFACE);

	HRESULT hr = pARC->SetAspectRatioMode(ARMode(bEtirer));

	check_hr(hr, "SetAspectRatioMode");
	return hr;
}

/// Dfinition de la position d'affichage et du ratio d'aspect.
/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
HRESULT CVideoMixingRenderer7::SetVideoPosition(const LPRECT pDstRect)
{
	if (!pVMR7Control || !pDstRect || IsRectEmpty(pDstRect))
		return E_FAIL;

	HRESULT hr = pVMR7Control->SetVideoPosition(NULL, pDstRect);

	check_hr(hr, "SetVideoPosition");
	return hr;
}

HRESULT CVideoMixingRenderer7::DisplayModeChanged()
{
	if (pVMR7Control)
		return pVMR7Control->DisplayModeChanged();
	return E_FAIL;
}

/// Rcuprer l'image affiche (en vue d'une capture statique)
/// \param[out]	pDib	Pointeur recevant l'image capture
HRESULT CVideoMixingRenderer7::GetCurrentImage(CComHeapPtr<BYTE> & pDib)
{
	if (pVMR7Control)
		return pVMR7Control->GetCurrentImage(&pDib);
	return E_FAIL;
}

/// Obtenir la rsolution native de la vido
SIZE CVideoMixingRenderer7::GetNativeSize() const
{
	SIZE	sSz	= {768, 576};	// valeurs par dfaut en cas d'erreur
	HRESULT hr	= pVMR7Control->GetNativeVideoSize(&sSz.cx, &sSz.cy, NULL, NULL);

	check_hr(hr, "GetNativeVideoSize");

	return sSz;
}

// ====================================================================================
// Implmentations des mthodes de la classe CVideoMixingRenderer7_renderless
// ====================================================================================

HRESULT CVideoMixingRenderer7_renderless::Configure()
{
	CComQIPtr<IVMRFilterConfig> pVMRConfig(pFilter);

	if (!pVMRConfig)
		return erreur(
			TEXT("    (VMR)"),
			E_NOINTERFACE);

	HRESULT hr = pVMRConfig->SetRenderingMode(VMRMode_Renderless);

#if USE_VMR_DEINTERLACE
	if (SUCCEEDED(hr) && use_vmr_deinterlace)
		hr = pVMRConfig->SetNumberOfStreams(1);
#endif

	return hr;
}

HRESULT CVideoMixingRenderer7_renderless::PrepareSurface()
{
	CComQIPtr<IVMRSurfaceAllocatorNotify> pNotify(pFilter);

	if (!pNotify)
		return erreur(
			TEXT(" IVMRSurfaceAllocatorNotify "),
			E_NOINTERFACE);

	CComPtr<IVMRSurfaceAllocator> pSurfAlloc;

	HRESULT hr = pSurfAlloc.CoCreateInstance(CLSID_AllocPresenter, NULL, CLSCTX_INPROC_SERVER);

	if (FAILED(hr))
		return erreur(TEXT(" IVMRSurfaceAllocator "), hr);

	{
		CComQIPtr<IVMRImagePresenterConfig> pPresConf(pSurfAlloc);

		if (!pPresConf)
			return erreur(
				TEXT(" IVMRImagePresenterConfig "),
				E_NOINTERFACE);

		hr = pPresConf->SetRenderingPrefs(
				eWinVers>=wv_Vista ?
				RenderPrefs_AllowOffscreen :
				(RenderPrefs_ForceOverlays | RenderPrefs_DoNotRenderColorKeyAndBorder)
			);

		if (FAILED(hr))
			return erreur(TEXT("  SetRenderingPrefs"), hr);
	}

	hr = pNotify->AdviseSurfaceAllocator(VMR_INSTANCE, pSurfAlloc);
	check_hr(hr, "AdviseSurfaceAllocator");
	hr = pSurfAlloc->AdviseNotify(pNotify);
	check_hr(hr, "AdviseNotify");

	return hr;
}

// ====================================================================================
// Implmentations des mthodes de la classe CVideoMixingRenderer7_windowless
// ====================================================================================

HRESULT CVideoMixingRenderer7_windowless::Configure()
{
	CComQIPtr<IVMRFilterConfig> pVMRConfig(pFilter);

	if (!pVMRConfig)
		return erreur(
			TEXT("    (VMR)"),
			E_NOINTERFACE);

	HRESULT hr = pVMRConfig->SetRenderingMode(VMRMode_Windowless);

	if (SUCCEEDED(hr)) {
		hr = pVMRConfig->SetRenderingPrefs(
			eWinVers >= wv_Vista ?
			RenderPrefs_AllowOffscreen :
			(RenderPrefs_ForceOverlays | RenderPrefs_DoNotRenderColorKeyAndBorder)
		);
#if USE_VMR_DEINTERLACE
		if (SUCCEEDED(hr) && use_vmr_deinterlace)
			hr = pVMRConfig->SetNumberOfStreams(1);
#endif
	}

	if (SUCCEEDED(hr)) {
		MonitorList<IVMRMonitorConfig, VMRMONITORINFO>	vMonitors(pFilter);

		if (vMonitors.size() > 0) {
			hr = vMonitors.pMonCfg->SetMonitor(&vMonitors.FindBest(hMainWnd)->guid);
		}
	}
	return hr;
}

// ====================================================================================
// Implmentations des mthodes de la classe CVideoMixingRenderer9
// ====================================================================================

const VMR9NormalizedRect CVideoMixingRenderer9::rRefNRect = {0.0f, 0.0f, 1.0f, 1.0f};

HRESULT CVideoMixingRenderer9::Configure()
{
	CComQIPtr<IVMRFilterConfig9> pVMRConfig(pFilter);

	if (!pVMRConfig)
		return erreur(
			TEXT("    (VMR)"),
			E_NOINTERFACE);

	HRESULT hr = pVMRConfig->SetRenderingMode(VMR9Mode_Windowless);

	// Requis pour pouvoir obtenir l'interface IVMRMixerControl9
	if (SUCCEEDED(hr))
		hr = pVMRConfig->SetNumberOfStreams(1);

	if (SUCCEEDED(hr)) {
		MonitorList<IVMRMonitorConfig9, VMR9MonitorInfo>	vMonitors(pFilter);

		if (vMonitors.size() > 0) {
			hr = vMonitors.pMonCfg->SetMonitor(vMonitors.FindBest(hMainWnd)->uDevID);
		}
	}
	return hr;
}

/// \test Exprimenter avec et sans cette option
#define BILINEAR_FILTERING

HRESULT CVideoMixingRenderer9::InitControl()
{
	HRESULT hr = pFilter.QueryInterface(&pVMR9Control);
	if (FAILED(hr)) {
		erreur(TEXT("    (VMR)"), hr);
		return hr;
	}

	hr = pFilter.QueryInterface(&pVMR9Mixer);
	if (FAILED(hr)) {
		erreur(TEXT("    (mixer VMR)"), hr);
		return hr;
	}

	hr = pVMR9Control->SetVideoClippingWindow(hMainWnd);
	check_hr(hr, "SetVideoClippingWindow");

	hr = pVMR9Control->SetBorderColor(colorKey);
	if (check_hr(hr, "SetBorderColor"))
		return hr;

#ifndef BILINEAR_FILTERING
	// Request point filtering (instead of bilinear filtering)
	// to improve the text quality.  In general, if you are
	// not scaling the app image, you should use point filtering.
	// This is very important if you are doing source color keying.

	DWORD dwPrefs=0;

	hr = pVMR9Mixer->GetMixingPrefs(&dwPrefs);

	if (SUCCEEDED(hr)) {
		dwPrefs |= MixerPref_PointFiltering;
		dwPrefs &= ~(MixerPref_BiLinearFiltering);

		hr = pVMR9Mixer->SetMixingPrefs(dwPrefs);
	}
#endif

	// Get alpha-blended bitmap interface
	hr = pFilter.QueryInterface(&pBMP);

	return hr;
}

/// Dfinition du mode de ratio d'aspect.
/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
HRESULT CVideoMixingRenderer9::NonDelegatingSetAspectRatioMode(bool bEtirer)
{
	CComQIPtr<IVMRAspectRatioControl9> pARC(pFilter);

	if (!pARC)
		return erreur(
			TEXT("    (aspect ratio)"),
			E_NOINTERFACE);

	HRESULT hr = pARC->SetAspectRatioMode(ARMode(bEtirer));

	check_hr(hr, "SetAspectRatioMode");
	return hr;
}

HRESULT CVideoMixingRenderer9::UpdateOutputRect_internal(const VMR9NormalizedRect & rNewOut)
{
	VMR9NormalizedRect	rOldOut;
	HRESULT				hr;

	hr = pVMR9Mixer->GetOutputRect(0, &rOldOut);
	check_hr(hr, "GetOutputRect");
	/*
	myprintf(COLR(LYELLOW, TEXT("GetOutputRect ")
		TEXT("- rct=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(rOldOut), ArN(rOldOut));
	*/

	if (hr != S_OK)
		return hr;
	if (rOldOut == rNewOut)
		return S_FALSE;	// pas d'erreur mais aucun changement

	/*
	myprintf(COLR(HYELLOW, TEXT("SetOutputRect ")
		TEXT("- rct=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(rNewOut), ArN(rNewOut));
	*/
	hr = pVMR9Mixer->SetOutputRect(0, &rNewOut);
	check_hr(hr, "SetOutputRect");
	return hr;
}

HRESULT CVideoMixingRenderer9::UpdateVideoPosition_internal(const LPRECT lpSRCRect, const LPRECT lpDSTRect)
{
	RECT	rOldSrc, rOldDst;
	HRESULT	hr;

	hr = pVMR9Control->GetVideoPosition(&rOldSrc, &rOldDst);
	check_hr(hr, "GetVideoPosition");
	/*
	myprintf(COLR(LYELLOW, TEXT("GetVideoPosition ")
		TEXT(", src=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))
		TEXT(", dst=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))) EOL,
		RECTPARM(rOldSrc), ArR(rOldSrc),
		RECTPARM(rOldDst), ArR(rOldDst));
	*/

	if (hr != S_OK)
		return hr;
	if ((lpSRCRect == NULL || rOldSrc == *lpSRCRect) && (lpDSTRect == NULL || rOldDst == *lpDSTRect))
		return S_FALSE;	// pas d'erreur mais aucun changement

	/*
  #ifdef _DEBUG
	myprint(COLR(HYELLOW, TEXT("SetVideoPosition")));
	if (lpSRCRect)
		myprintf(COLR(HYELLOW, TEXT(", src=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))), RECTPARM(*lpSRCRect), ArR(*lpSRCRect));
	if (lpDSTRect)
		myprintf(COLR(HYELLOW, TEXT(", dst=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))), RECTPARM(*lpDSTRect), ArR(*lpDSTRect));
	myprint(EOL);
  #endif
	*/

	hr = pVMR9Control->SetVideoPosition(lpSRCRect, lpDSTRect);
	check_hr(hr, "SetVideoPosition");
	if (hr != S_OK)
		return hr;
	// S_OK si une partie du fond est mise  dcouvert
	return (rOldDst | *lpDSTRect) != *lpDSTRect ? S_OK : S_FALSE;
}

#if BUILD_UNUSED_CODE
HRESULT CVideoMixingRenderer9::UpdateAlphaBitmapPos_internal(const VMR9NormalizedRect & rNewPos)
{
	VMR9AlphaBitmap	sBmpParms;
	HRESULT			hr;

	hr = pBMP->GetAlphaBitmapParameters(&sBmpParms);
	check_hr(hr, "GetAlphaBitmapParameters");
	/*
	myprintf(COLR(LYELLOW, TEXT("IVMRMixerBitmap9::GetAlphaBitmapParameters ")
		TEXT("- src=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))
		TEXT("- dst=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(sBmpParms.rSrc), ArR(sBmpParms.rSrc),
		RECTPARM(sBmpParms.rDest), ArN(sBmpParms.rDest));
	*/

	if (hr != S_OK)
		return hr;
	if (sBmpParms.rDest == rNewPos)
		return S_FALSE;	// pas d'erreur mais aucun changement

	sBmpParms.rDest = rNewPos;
	/*
	myprintf(COLR(HYELLOW, TEXT("IVMRMixerBitmap9::UpdateAlphaBitmapParameters ")
		TEXT("- src=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))
		TEXT("- dst=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(sBmpParms.rSrc), ArR(sBmpParms.rSrc),
		RECTPARM(sBmpParms.rDest), ArN(sBmpParms.rDest));
	*/
	hr = pBMP->UpdateAlphaBitmapParameters(&sBmpParms);
	check_hr(hr, "UpdateAlphaBitmapParameters");
	return hr;
}
#endif

#if NEW_VMR9_OUTPUT
/// Dfinition de la position d'affichage et du ratio d'aspect.
/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
/// \todo En VMR9 seulement, si on positionne la fentre hors de l'cran par en haut et/ou par
///  gauche, la vido refuse de sortir de l'cran par ces cts-l, se dcalant en consquence
/// par rapport au rectangle d'affichage. Ce n'est que modrment gnant, mais on pourrait
/// quand mme y chercher une solution ventuellement.
HRESULT CVideoMixingRenderer9::SetVideoPosition(const LPRECT pDstRect)
{
	if (!pVMR9Control || !pVMR9Mixer || !pDstRect || IsRectEmpty(pDstRect))
		return E_FAIL;

	// Doc. : Coordinate Mapping in the VMR
	// http://msdn.microsoft.com/en-us/library/dd387932%28v=VS.85%29.aspx

	HRESULT							hr;
	RECT							rNewDst		= CalcCliRect();

	// Construire le rectangle de sortie et le rectangle source de l'affichage
	VMR9NormalizedRect				rNewOut		= TransposeRect(rNewDst, *pDstRect, rRefNRect);

/*
	// Actualiser la position du bitmap
	rNormDstRect = TransposeRect(rTemp, rNewDst, rRefNRect);
	if (bBmInVideo)
		hr = UpdateAlphaBitmapPos_internal(rNormDstRect);
*/

	hr = UpdateOutputRect_internal(rNewOut);
	if (SUCCEEDED(hr))	// S_OK ou S_FALSE
		hr = UpdateVideoPosition_internal(NULL, &rNewDst);

	return hr;
}
#endif

HRESULT CVideoMixingRenderer9::DisplayModeChanged()
{
	if (pVMR9Control)
		return pVMR9Control->DisplayModeChanged();
	return E_FAIL;
}

/// Rcuprer l'image affiche (en vue d'une capture statique)
/// \param[out]	pDib	Pointeur recevant l'image capture
HRESULT CVideoMixingRenderer9::GetCurrentImage(CComHeapPtr<BYTE> & pDib)
{
	if (pVMR9Control)
		return pVMR9Control->GetCurrentImage(&pDib);
	return E_FAIL;
}

void CVideoMixingRenderer9::Paint(HDC hDC)
{
	if (pVMR9Control) {
		// myprint(TEXT("pVMR9Control->RepaintVideo(hMainWnd, hDC)") EOL);
		HRESULT hr = pVMR9Control->RepaintVideo(hMainWnd, hDC);
		check_hr(hr, "RepaintVideo");
	}
}

/// Obtenir la rsolution native de la vido
SIZE CVideoMixingRenderer9::GetNativeSize() const
{
	SIZE	sSz	= {768, 576};	// valeurs par dfaut en cas d'erreur
	HRESULT hr	= pVMR9Control->GetNativeVideoSize(&sSz.cx, &sSz.cy, NULL, NULL);

	check_hr(hr, "GetNativeVideoSize FAILED!");

	return sSz;
}

/// Retrait de la couche OSD de l'affichage
HRESULT CVideoMixingRenderer9::RemoveOsdLayer()
{
	if (!pBMP)
		return E_FAIL;

	// Retrait du bitmap de la vido :
	VMR9AlphaBitmap sBmpParms = {
		VMR9AlphaBitmap_Disable					// dwFlags
		// le reste  zro (implicite)
	};

	return pBMP->SetAlphaBitmap(&sBmpParms);
}

const PA_Parms CVideoMixingRenderer9::aParms[pat_MAX] = {
	{ProcAmpControl9_Brightness,	offsetof(VMR9ProcAmpControl, Brightness)},
	{ProcAmpControl9_Contrast,		offsetof(VMR9ProcAmpControl, Contrast)},
	{ProcAmpControl9_Hue,			offsetof(VMR9ProcAmpControl, Hue)},
	{ProcAmpControl9_Saturation,	offsetof(VMR9ProcAmpControl, Saturation)}
};

HRESULT CVideoMixingRenderer9::GetProcAmpValues(ProcAmpValues & sValues) const
{
	if (!pVMR9Mixer)
		return E_FAIL;

	VMR9ProcAmpControl	sCtl		= {sizeof(sCtl)};
	DWORD				nStreamID	= 0;	// Toujours 0 pour le moment
	HRESULT				hr			= pVMR9Mixer->GetProcAmpControl(nStreamID, &sCtl);

	if (check_hr(hr, "GetProcAmpControl"))
		return hr;

	for (ProcAmpType eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp) {
		PA_Parms		sParms = aParms[eTyp];

		if ((sParms.nFlags & sCtl.dwFlags) == 0)
			continue;	// Le paramtre n'est pas gr

		VMR9ProcAmpControlRange
			sRng = {
				sizeof(sRng),
				(VMR9ProcAmpControlFlags)sParms.nFlags
				// Le reste  0 (implicite)
			};

		hr = pVMR9Mixer->GetProcAmpControlRange(nStreamID, &sRng);
		if (check_hr(hr, "GetProcAmpControlRange"))
			return hr;

		ProcAmpVal &	sVal = sValues.aVal[eTyp];

		sVal.fCurrent	= memberof(sCtl, float, sParms.nCtlOfs);
		sVal.fMax		= sRng.MaxValue;
		sVal.fMin		= sRng.MinValue;
		sVal.fDefault	= sRng.DefaultValue;
		sVal.fStep		= sRng.StepSize;
		sVal.bValid		= true;
	}

	return hr;
}

HRESULT CVideoMixingRenderer9::SetProcAmpValues(ProcAmpValues & sValues)
{
	VMR9ProcAmpControl	sCtl		= {sizeof(sCtl)};
	DWORD				nStreamID	= 0;	// Toujours 0 pour le moment
	HRESULT				hr			= S_OK;
	ProcAmpType			eTyp;
	PA_Parms			sParms;

	for (eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp) {
		ProcAmpVal &	sVal = sValues.aVal[eTyp];

		if (!sVal.bToSet)
			continue;	// Le paramtre n'est pas modifi
		sParms = aParms[eTyp];

		float &			fCur = memberof(sCtl, float, sParms.nCtlOfs);

		sCtl.dwFlags |= sParms.nFlags;
		sVal.CheckLimits();
		fCur = sVal.fCurrent;
	}

	if (sCtl.dwFlags) {
		hr = pVMR9Mixer->SetProcAmpControl(nStreamID, &sCtl);
		check_hr(hr, "SetProcAmpControl (1)");

		// Ce qui suit est un palliatif destin  certains pilotes de cartes vido qui sont affects
		// d'un bug qui les fait ignorer la valeur qui leur est soumise si elle est identique  leur
		// valeur par dfaut :
		// On rcupre les nouvelles valeurs des rglages depuis la carte aprs modification, on les
		// compare  celles qu'elles auraient d avoir, et si une diffrence est trouve, on modifie
		// la valeur est la dcalant d'un pas gal au pas minimal de dcalage, et on rapplique.

		VMR9ProcAmpControl	sVerif		= {sizeof(sVerif)};

		hr = pVMR9Mixer->GetProcAmpControl(nStreamID, &sVerif);	// Rcupration des nouveaux rglages
		check_hr(hr, "GetProcAmpControl (vrification)");
		sCtl.dwFlags = 0;
		for (eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp) {
			ProcAmpVal &	sVal = sValues.aVal[eTyp];

			if (!sVal.bToSet)
				continue;	// Le paramtre n'avait pas t modifi
			sVal.bToSet	= false;
			sParms = aParms[eTyp];

			float &			fCur = memberof(sCtl, float, sParms.nCtlOfs);

			if (memberof(sVerif, float, sParms.nCtlOfs) == fCur)
				continue;
			sCtl.dwFlags |= sParms.nFlags;
			fCur = min<float>(sVal.fMax, sVal.fCurrent + sVal.fStep);
		}
		if (sCtl.dwFlags) {
			hr = pVMR9Mixer->SetProcAmpControl(nStreamID, &sCtl);
			check_hr(hr, "SetProcAmpControl (2)");
		}
	}
	return hr;
}

/// Dfinition de la position d'affichage et du ratio d'aspect (ancienne version).
/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
HRESULT CVideoMixingRenderer9_Old::SetVideoPosition(const LPRECT pDstRect)
{
	if (!pVMR9Control || !pVMR9Mixer || !pDstRect || IsRectEmpty(pDstRect))
		return E_FAIL;

	// Doc. : Coordinate Mapping in the VMR
	// http://msdn.microsoft.com/en-us/library/dd387932%28v=VS.85%29.aspx

	HRESULT	hr = UpdateVideoPosition_internal(NULL, pDstRect);

	// S'assurer que les bordures autour de l'image resteront noires :
	if (hr == S_OK)		// si une partie du fond est mise  dcouvert (sinon S_FALSE)
		InvalidateRect(hMainWnd, NULL, TRUE);

	return hr;
}

// ====================================================================================
// Implmentations des mthodes de la classe CEnhancedVideoRenderer
// ====================================================================================

const MFVideoNormalizedRect CEnhancedVideoRenderer::rRefNRect = {0.0f, 0.0f, 1.0f, 1.0f};

HRESULT CEnhancedVideoRenderer::InitControl()
{
	CComQIPtr<IMFGetService> pMFGetService(pFilter);

	if (!pMFGetService)
		return erreur(TEXT(" IMFGetService "), E_NOINTERFACE);

	HRESULT hr = pMFGetService->GetService(MR_VIDEO_RENDER_SERVICE, IID_IMFVideoDisplayControl, (void **) &pEVRControl);

	if (FAILED(hr))
		return erreur(TEXT("    (EVR)"), hr);

	hr = pMFGetService->GetService(MR_VIDEO_MIXER_SERVICE, IID_IMFVideoProcessor, (void **) &pEVRVideoProc);
	if (FAILED(hr))
		return erreur(TEXT("    (EVR)"), hr);

	hr = pEVRControl->SetVideoWindow(hMainWnd);
	if (check_hr(hr, "SetVideoWindow"))
		return hr;

	hr = pEVRControl->SetBorderColor(colorKey);
	if (check_hr(hr, "SetBorderColor"))
		return hr;

	hr = pMFGetService->GetService(MR_VIDEO_MIXER_SERVICE, IID_IMFVideoMixerControl, (void **) &pEVRMixCtrl);
	if (FAILED(hr))
		return erreur(TEXT(" IMFVideoMixerControl   "), hr);

	hr = pMFGetService->GetService(MR_VIDEO_MIXER_SERVICE, IID_IMFVideoMixerBitmap, (void **) &pBMP);
	if (FAILED(hr))
		return erreur(TEXT(" IMFVideoMixerBitmap "), hr);

	CComPtr<IMFVideoProcessor>	pVidProc;

	hr = pMFGetService->GetService(MR_VIDEO_MIXER_SERVICE, IID_IMFVideoProcessor, (void **) &pVidProc);
	if (FAILED(hr))
		return erreur(TEXT(" IMFVideoProcessor "), hr);

	pVidProc->SetBackgroundColor(colorKey);
	return hr;
}

/// Dfinition du mode de ratio d'aspect.
/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
HRESULT CEnhancedVideoRenderer::NonDelegatingSetAspectRatioMode(bool bEtirer)
{
	if (!pEVRControl)
		return E_FAIL;

	HRESULT hr = pEVRControl->SetAspectRatioMode(ARMode(bEtirer));

	check_hr(hr, "SetAspectRatioMode");
	return hr;
}

HRESULT CEnhancedVideoRenderer::UpdateOutputRect_internal(const MFVideoNormalizedRect & rNewOut)
{
	MFVideoNormalizedRect	rOldOut;
	HRESULT					hr;

	hr = pEVRMixCtrl->GetStreamOutputRect(0, &rOldOut);
	check_hr(hr, "GetStreamOutputRect");
	/*
	myprintf(COLR(LYELLOW, TEXT("GetStreamOutputRect ")
		TEXT("- rct=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(rOldOut), ArN(rOldOut));
	*/

	if (hr != S_OK)
		return hr;
	if (rOldOut == rNewOut)
		return S_FALSE;	// pas d'erreur mais aucun changement

	/*
	myprintf(COLR(HYELLOW, TEXT("SetStreamOutputRect ")
		TEXT("- rct=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))) EOL,
		RECTPARM(rNewOut), ArN(rNewOut));
	*/
	hr = pEVRMixCtrl->SetStreamOutputRect(0, &rNewOut);
	check_hr(hr, "SetStreamOutputRect");
	return hr;
}

HRESULT CEnhancedVideoRenderer::UpdateVideoPosition_internal(const MFVideoNormalizedRect * pnrcSource, const LPRECT prcDest)
{
	MFVideoNormalizedRect	rOldSrc;
	RECT					rOldDst;
	HRESULT					hr;

	hr = pEVRControl->GetVideoPosition(&rOldSrc, &rOldDst);
	check_hr(hr, "GetVideoPosition");
	/*
	myprintf(COLR(LYELLOW, TEXT("GetVideoPosition ")
		TEXT("- src=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))
		TEXT(", dst=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))) EOL,
		RECTPARM(rOldSrc), ArN(rOldSrc),
		RECTPARM(rOldDst), ArR(rOldDst));
	*/

	if (hr != S_OK)
		return hr;
	if ((pnrcSource == NULL || rOldSrc == *pnrcSource) && (prcDest == NULL || rOldDst == *prcDest))
		return S_FALSE;	// pas d'erreur mais aucun changement

	/*
  #ifdef _DEBUG
	myprint(COLR(HYELLOW, TEXT("SetVideoPosition")));
	if (pnrcSource)
		myprintf(COLR(HYELLOW, TEXT(", src=") COLR(HWHITE, TEXT("{t=%.3f,l=%.3f,b=%.3f,r=%.3f} - ar=%.3f"))), RECTPARM(*pnrcSource), ArN(*pnrcSource));
	if (prcDest)
		myprintf(COLR(HYELLOW, TEXT(", dst=") COLR(HWHITE, TEXT("{")RECTFMT(",") TEXT("} - ar=%.3f"))), RECTPARM(*prcDest), ArR(*prcDest));
	myprint(EOL);
  #endif
	*/

	hr = pEVRControl->SetVideoPosition(pnrcSource, prcDest);
	check_hr(hr, "SetVideoPosition");
	return hr;
}

/// Dfinition de la position d'affichage et du ratio d'aspect.
/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
HRESULT CEnhancedVideoRenderer::SetVideoPosition(const LPRECT pDstRect)
{
	if (!pEVRControl || !pEVRMixCtrl || !pDstRect || IsRectEmpty(pDstRect))
		return E_FAIL;

	RECT							rNewDst = CalcCliRect();

	// Dterminer le plus petit rectangle qui encadre  la fois le rectangle client de la
	// fentre et le rectangle d'affichage souhait de la vido
	RECT							rTemp	= rNewDst | *pDstRect;

	// Agrandir ce rectangle pour avoir le mme ratio d'aspect que celui du rectangle client
	MinArRect(rTemp, RSiz(rNewDst));

	// Construire le rectangle de sortie et le rectangle source de l'affichage
	MFVideoNormalizedRect			rNewSrc = TransposeRect(rTemp, rNewDst,  rRefNRect);
	MFVideoNormalizedRect			rNewOut	= TransposeRect(rTemp, *pDstRect, rRefNRect);
	HRESULT							hr		= UpdateOutputRect_internal(rNewOut);

	if (SUCCEEDED(hr))
		hr = UpdateVideoPosition_internal(&rNewSrc, &rNewDst);
	return hr;
}

/// Rcuprer l'image affiche (en vue d'une capture statique)
/// \param[out]	pDib	Pointeur recevant l'image capture
HRESULT CEnhancedVideoRenderer::GetCurrentImage(CComHeapPtr<BYTE> & pDib)
{
	if (pEVRControl) {
		BITMAPINFOHEADER	pBih = {sizeof(BITMAPINFOHEADER)};
		CComHeapPtr<BYTE>	pDibTmp;
		DWORD				pcbDib;
		HRESULT				hr = pEVRControl->GetCurrentImage(&pBih, &pDibTmp, &pcbDib, NULL);

		if (FAILED(hr))
			return hr;

		pDib.AllocateBytes(sizeof(BITMAPINFOHEADER) + pcbDib);
		CopyMemory(pDib, &pBih, sizeof(BITMAPINFOHEADER));
		CopyMemory(pDib + sizeof(BITMAPINFOHEADER), pDibTmp, pcbDib);

		return S_OK;
	}
	return E_FAIL;
}

void CEnhancedVideoRenderer::Paint(HDC hDC)
{
	if (pEVRControl) {
		// myprint(TEXT("pEVRControl->RepaintVideo()") EOL);
		pEVRControl->RepaintVideo();
	}
}

/// Obtenir la rsolution native de la vido
SIZE CEnhancedVideoRenderer::GetNativeSize() const
{
#if 0
	LocalVideoInfo	sVInf;

	GetVideoInfo(sVInf);

	// Ajustement de la taille horizontale, d'une manire similaire  ce qui semble tre fait dans la fonction
	// "GetNativeVideoSize" des autres modes ; passage par un type "double" car les valeurs utilises pour
	// exprimer le ratio d'aspect sont parfois trs leves :
	SIZE sSz = {
		LONG((double(sVInf.sNativeSize.cy)+.5) * sVInf.sAspectRatio.cx / sVInf.sAspectRatio.cy),
		sVInf.sNativeSize.cy
	};
#else
	// Note : On a tent d'utiliser pEVRControl->GetNativeVideoSize, mais il semble qu'en EVR,
	// les valeurs retournes par cette fonction soient trop incohrentes pour tre exploites,
	// contrairement  ce qui se passe dans les modes VMR.

	SIZE	sSz	= {768, 576};					// valeurs par dfaut en cas d'erreur
	SIZE	sAr	= {DEFAULT_AR_X, DEFAULT_AR_Y};	// valeurs par dfaut en cas d'erreur
	HRESULT hr	= pEVRControl->GetNativeVideoSize(&sSz, &sAr);

	check_hr(hr, "GetNativeVideoSize");
	sAr = Simplify(sAr);

	// Ajustement de la taille horizontale, d'une manire similaire  ce qui semble tre fait dans la fonction
	// "GetNativeVideoSize" des autres modes ; passage par un type "double" car les valeurs utilises pour
	// exprimer le ratio d'aspect sont parfois trs leves :
	if (sAr.cx!=0 && sAr.cy!=0)
		sSz.cx = LONG((double(sSz.cy) +.5) * sAr.cx / sAr.cy);
#endif

	return sSz;
}

const PA_Parms CEnhancedVideoRenderer::aParms[pat_MAX] = {
	{DXVA2_ProcAmp_Brightness,	offsetof(DXVA2_ProcAmpValues, Brightness)},
	{DXVA2_ProcAmp_Contrast,	offsetof(DXVA2_ProcAmpValues, Contrast)},
	{DXVA2_ProcAmp_Hue,			offsetof(DXVA2_ProcAmpValues, Hue)},
	{DXVA2_ProcAmp_Saturation,	offsetof(DXVA2_ProcAmpValues, Saturation)}
};

HRESULT CEnhancedVideoRenderer::GetProcAmpValues(ProcAmpValues & sValues) const
{
	DXVA2_ProcAmpValues		sCtl;
	HRESULT					hr	= pEVRVideoProc->GetProcAmpValues(
									DXVA2_ProcAmp_Brightness|DXVA2_ProcAmp_Contrast|
									DXVA2_ProcAmp_Hue|DXVA2_ProcAmp_Saturation, &sCtl);

	if (check_hr(hr, "GetProcAmpValues"))
		return hr;

	for (ProcAmpType eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp) {
		PA_Parms			sParms = aParms[eTyp];
		DXVA2_ValueRange	sRng = {0};

		hr = pEVRVideoProc->GetProcAmpRange(sParms.nFlags, &sRng);
		if (check_hr(hr, "GetProcAmpRange"))
			return hr;

		ProcAmpVal &		sVal = sValues.aVal[eTyp];

		sVal.fCurrent	= DXVA2FixedToFloat(memberof(sCtl, DXVA2_Fixed32, sParms.nCtlOfs));
		sVal.fMax		= DXVA2FixedToFloat(sRng.MaxValue);
		sVal.fMin		= DXVA2FixedToFloat(sRng.MinValue);
		sVal.fDefault	= DXVA2FixedToFloat(sRng.DefaultValue);
		sVal.fStep		= DXVA2FixedToFloat(sRng.StepSize);
		sVal.bValid		= true;
	}

	return hr;
}

HRESULT CEnhancedVideoRenderer::SetProcAmpValues(ProcAmpValues & sValues)
{
	DXVA2_ProcAmpValues		sCtl	= {0};
	HRESULT					hr		= S_OK;
	DWORD					dwFlags	= 0;

	for (ProcAmpType eTyp = pat_Brightness; eTyp < pat_MAX; ++eTyp) {
		PA_Parms		sParms	= aParms[eTyp];
		ProcAmpVal &	sVal	= sValues.aVal[eTyp];

		if (!sVal.bToSet)
			continue;	// Le paramtre n'est pas modifi

		// Limitation des valeurs (le jeu des imprcisions en virgule flottante
		//  pouvant induire un lger dpassement)
		if (sVal.fCurrent <= sVal.fMin)
			sVal.fCurrent = sVal.fMin;
		if (sVal.fCurrent >= sVal.fMax)
			sVal.fCurrent = sVal.fMax;
		dwFlags |= sParms.nFlags;
		memberof(sCtl, DXVA2_Fixed32, sParms.nCtlOfs) = DXVA2FloatToFixed(sVal.fCurrent);
		sVal.bToSet = false;
	}

	if (dwFlags) {
		hr = pEVRVideoProc->SetProcAmpValues(dwFlags, &sCtl);
		check_hr(hr, "SetProcAmpValues");
	}
	return hr;
}

/// Fonction "callback" utilise pour le rafrachissement retard de la vido
VOID CALLBACK CEnhancedVideoRenderer_Old::SetVideoPosTimerProc(HWND hWnd, UINT, UINT_PTR idEvent, DWORD)
{
	KillTimer(hWnd, idEvent);
	InvalidateRect(hWnd, NULL, TRUE);
}

/// Dfinition de la position d'affichage et du ratio d'aspect (ancienne version).
/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
HRESULT	CEnhancedVideoRenderer_Old::SetVideoPosition(const LPRECT pDstRect)
{
	if (!pEVRControl || !pEVRMixCtrl || !pDstRect || IsRectEmpty(pDstRect))
		return E_FAIL;

	HRESULT	hr  = UpdateVideoPosition_internal(&rRefNRect, pDstRect);

	// S'assurer que les bordures autour de l'image resteront noires :
	// Ceci est une tentative, qui ne marche mme pas dans tous les cas, de pallier
	// un problme de positionnement de la vido en cas de zoom ngatif,
	// en retardant l'effacement du pourtour. Lire  ce sujet :
	// http://social.msdn.microsoft.com/Forums/en-US/mediafoundationdevelopment/thread/399460eb-8959-49d4-8360-a99227f77ddc
	if (hr == S_OK)	// Si aucun changement n'a eu lieu, hr = S_FALSE
		SetTimer(hMainWnd, TIMER_EVR_UPDATE, 200, SetVideoPosTimerProc);
	return hr;
}

// ====================================================================================
// Fonctions globales (exportes)
// ====================================================================================

/**
 * Redessiner l'affichage ( appeler en rponse  \p WM_PAINT)
 * \param[in] hDC "Device context" retourn par la fonction systme
 * \p BeginPaint (appele au pralable)
 **/
void do_paint(HDC hDC)
{
	if (pVMR)
		// si le VMR existe, on lui dlgue la tche (mme si on
		// sait qu'il va peut-tre la rassigner  l'OSD)
		pVMR->Paint(hDC);
	else if (pOSD)
		// sinon, on fait juste dans l'OSD
		pOSD->Redraw(hDC);
}

/**
 * Mise  jour des coordonnes de l'cran ( appeler en rponse  tout vnement
 * de nature  modifier les dimensions de la vido)
 **/
void update_coords()
{
	// myprint(TEXT("Excution update_coords()") EOL);
	request_osd_update();

	if (pVMR)
		pVMR->UpdateVideoPos();
}

// ====================================================================================
// Mthodes statiques de cration de rendus mdia
// ====================================================================================

/**
 * Cration d'un rendu vido du type spcifi
 * \param[in]	eRdrMode		Type de rendu vido  crer
 * \return		Pointeur sur l'objet "rendu vido" cr
 **/
CVideoRenderer * CVideoRenderer::NewRenderer(RdrTypes eRdrMode)
{
	switch (eRdrMode) {
	case rdt_VMR7_windowless:		return new CVideoMixingRenderer7_windowless();
	case rdt_VMR7_renderless:		return new CVideoMixingRenderer7_renderless();
	case rdt_VMR9_windowless:		return new CVideoMixingRenderer9_Old();
#if NEW_VMR9_OUTPUT
	case rdt_VMR9_windowless_new:	return new CVideoMixingRenderer9();
#endif
	case rdt_EVR:					return new CEnhancedVideoRenderer_Old();
	case rdt_EVR_new:				return new CEnhancedVideoRenderer();
	}

	// par dfaut
	return new CNullVideoRenderer();
}

/**
 * Cration d'un rendu audio du type spcifi
 * \param[in]	eRdrMode		Type de rendu audio  crer
 * \return		Pointeur sur l'objet "rendu audio" cr
 **/
CAudioRenderer * CAudioRenderer::NewRenderer(RdrTypes eRdrMode)
{
	switch (eRdrMode) {
	case rdt_DirectSound:		return new CDirectSoundRenderer();
	}

	// par dfaut
	return new CNullAudioRenderer();
}

// ====================================================================================
// Variables globales (exportes)
// ====================================================================================

/// Pointeur sur l'instance du VMR cre par lors de l'excution du programme
shared_ptr<CVideoRenderer>	pVMR;

/// Pointeur sur l'instance de l'OSD cre par lors de l'excution du programme
shared_ptr<COsdAllObjects>	pOSD;

// ====================================================================================
