/*
 * $Id: rendering.h 270 2010-12-02 20:57:32Z gingko $
 * Copyright (C) 2008 gingko - http://gingko.homeip.net/
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Dfinitions pour le rendu vido.
 *
 *	Dfinitions des classes et structures ncessaires au rendu vido
 *  (\p VMR7, \p VMR9, \p EVR...).
 **/

#pragma once

#include "filters.h"
#include "osd_objects.h"
#include "osd_layer.h"

// ====================================================================================
// Dfinitions
// ====================================================================================

#define DEFAULT_AR_X		4	//!< Ratio d'aspect horizontal par dfaut
#define DEFAULT_AR_Y		3	//!< Ratio d'aspect vertical par dfaut

// ====================================================================================
// Dclaration de la structure LocalVideoInfo
// ====================================================================================

/// \brief Donnes obtenues depuis le graphe concernant les dimensions de la video
struct LocalVideoInfo
{
	SIZE	sNativeSize;	//!< Dimensions natives de la video
	SIZE	sAspectRatio;	//!< Ratio d'aspect
};

// ====================================================================================
// Dclaration de la structure ProcAmpValues
// ====================================================================================

enum ProcAmpAvailability {
	paa_MissingFromMode,
	paa_MissingFromGPU,
	paa_Available
};

inline ProcAmpType operator ++ (ProcAmpType & eTyp)
{
	return eTyp = ProcAmpType(int(eTyp)+1);
}

struct ProcAmpVal
{
	bool	bValid;
	bool	bToSet;
	float	fCurrent;
	float	fMin;
	float	fMax;
	float	fDefault;
	float	fStep;

	/// Constructeurs (avec valeurs par dfaut pour item inactif)
	ProcAmpVal() :
		bValid(false),
		bToSet(false),
		fCurrent(0.0f),
		fMin(-1.0f),
		fMax(1.0f),
		fDefault(0.0f),
		fStep(0.25f)
	{
	}

	/// Obtenir le nombre de dcimales requises pour l'affichage
	int GetPrecision() const {
		return max<int>(0, int(-log10(fStep))); }

	/// Obtenir la valeur courante en tant qu'entier, pour sauvegarde
	INT32 GetIntVal() const {
		return fCurrent == fDefault ? VIDEO_UNDEF : INT32(fCurrent / fStep);
	}

	/// Dfinit la valeur courante  partir d'un entier, issu de la sauvegarde
	void SetIntVal(INT32 nVal) {
		float	fVal = nVal != VIDEO_UNDEF ? (nVal * fStep) : fDefault;

		if (fCurrent != fVal) {
			fCurrent = fVal;
			bToSet = true;
		}
	}

	/// Vrification et ajustement de la valeur courante par rapport aux limites
	void CheckLimits() {
		if (fCurrent < fMin)
			fCurrent = fMin;
		if (fCurrent > fMax)
			fCurrent = fMax;
	}
};

struct ProcAmpValues
{
	ProcAmpVal	aVal[pat_MAX];
};

class IPTvM_GraphBase;	// Dfini ailleurs

// ====================================================================================
// Dclaration de la classe CMediaRenderer
// ====================================================================================

class CMediaRenderer : public FilterData
{
  #if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	bool				bConnected;
  #endif
protected:
	CComPtr<IPin>		pInputPin;	//!< Interface sur la broche d'entre

	/// Configuration
	virtual HRESULT		Configure() {
							return S_OK; }

public:
	/// Initialisation gnrale
	virtual HRESULT		Init(IPTvM_GraphBase & iGraph);

	/// Constructeur
	CMediaRenderer(LPCTSTR pszNam, REFCLSID clsid) :
  #if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
		FilterData(pszNam, clsid),
		bConnected(false)
  #else
		FilterData(pszNam, clsid)
  #endif
	{
		// myprintf(TEXT("Construction du rendu (%s)") EOL, pszName);
	}

	/// Obtenir le type de VMR courant
	virtual RdrTypes	GetRdrType() const {
							return rdt_null; }

	virtual	LPCTSTR		TypeStr() const = 0;

  #if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/// Connexion  une sortie de filtre
	HRESULT				Connect(IPTvM_GraphBase & iGraph, IBaseFilter * pMediaCodec);
  #endif

  #if BUILD_UNUSED_CODE	// Inutilis, mais rserv pour usage futur
	/// Dconnexion
	HRESULT				Disconnect(IPTvM_GraphBase & iGraph);
  #endif

	virtual	~CMediaRenderer();
};

// ====================================================================================
// Dclaration de la classe CAudioRenderer
// ====================================================================================

/** \brief Audio Renderer, classe de base
 *
 *  Ceci est la classe de base dont drivent toutes les diffrentes implmentations
 *  de rendu audio.
 **/
class CAudioRenderer : public CMediaRenderer
{
protected:

public:

	/// Constructeur
	CAudioRenderer(LPCTSTR pszNam, REFCLSID clsid) :
		CMediaRenderer(pszNam, clsid)
	{
	}

	/**
	 * Cration d'un rendu audio du type spcifi
	 * \param[in]	eRdrMode		Type de rendu audio  crer
	 * \return		Pointeur sur l'objet "rendu audio" cr
	 **/
	static CAudioRenderer *
						NewRenderer(RdrTypes eRdrMode);

	virtual	LPCTSTR		TypeStr() const {
							return TEXT("audio"); }
};

// ====================================================================================
// Dclaration de la classe CVideoRenderer
// ====================================================================================

class COsdObject;		// prdclaration "forward" (car utilise par CVideoRenderer)

/** \brief Video Renderer, classe de base
 *
 *  Ceci est la classe de base dont drivent toutes les diffrentes implmentations
 *  de Video Renderer.
 **/
class CVideoRenderer : public CMediaRenderer
{
	DWORD				last_tick;

protected:

	ProcAmpAvailability	eProcAmpAvail;

	/// Prparation de la surface (utile seulement en mode RenderLess) :
	virtual HRESULT		PrepareSurface() {
							return S_OK; }

#if USE_VMR_DEINTERLACE
	virtual HRESULT		SetMixingPrefs() {
							return S_OK; }

	/// Initialisation du dsentrelacement.
	/// \todo Cette mthode, ici sauvagement recopie depuis l'ancienne initialisation
	/// du VMR, n'a jamais t valide et n'a sans doute jamais fonctionn.
	/// L'tudier soigneusement quand on aura le temps, en vue de dcider si on peut
	/// la faire marcher, ou bien s'il est plus sens de tout supprimer.
	virtual HRESULT		InitDeinterlace() {
							return S_OK; }
#endif

	/// Initialisation et configuration de l'interface de contrle
	virtual HRESULT		InitControl() {
							return S_OK; }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		NonDelegatingSetAspectRatioMode(bool bEtirer) {
							return S_OK; }

	/// Dfinition du mode de ratio d'aspect.
	/// \param[in]	bEtirer		Dformer la vido si ncessaire pour remplir le rectangle
	virtual HRESULT		SetAspectRatioMode(bool bEtirer) {
							return NonDelegatingSetAspectRatioMode(bEtirer); }

	/// Dfinition de la position d'affichage et du ratio d'aspect.
	/// \param[in]	pDstRect	Rectangle de la nouvelle position de l'affichage
	virtual HRESULT		SetVideoPosition(const LPRECT pDstRect) {
							return S_OK; }

	/// Obtenir la rsolution native de la vido
	virtual SIZE		GetNativeSize() const  {
							SIZE sSiz = {0, 0}; return sSiz; }

	/// Constructeur ("protected")
	CVideoRenderer(LPCTSTR pszNam, REFCLSID clsid) :
		CMediaRenderer(pszNam, clsid),
		last_tick(0),
		eProcAmpAvail(paa_MissingFromMode)
	{
	}

public:

	/**
	 * Cration d'un rendu vido du type spcifi
	 * \param[in]	eRdrMode		Type de rendu vido  crer
	 * \return		Pointeur sur l'objet "rendu vido" cr
	 **/
	static CVideoRenderer *
						NewRenderer(RdrTypes eRdrMode);

	virtual	LPCTSTR		TypeStr() const {
							return TEXT("vido"); }

	/// Signifier au VMR que le mode d'affichage a chang
	virtual HRESULT		DisplayModeChanged() {
							return S_OK;}

	/// Rcuprer l'image affiche (en vue d'une capture statique)
	/// \param[out]	pDib	Pointeur recevant l'image capture
	virtual HRESULT		GetCurrentImage(CComHeapPtr<BYTE> & pDib)  {
							return S_OK;}

	virtual class COsdFontAllocator *
						CreateFontAllocator() {
							return COsdFontAllocator::CreateInstance(lt_GDI); }

	/// Calcul du rectangle utilis pour la vido ( l'usage de la construction de l'OSD)
	/// \param[in] sRect	Rectangle client de la fentre, calcul par \p CalcCliRect
	/// \return				Rectangle calcul, utilisable pour l'OSD
	virtual RECT		CalcOsdRect(RECT sRect) {
							return sRect; }	// Retourne le mme rectangle si pas surcharg

	/** \brief Dbut de mise  jour de la couche OSD dans l'affichage
	 *
	 * Cette mthode prpare la couche OSD (effacement, redimensionnement si ncessaire) afin
	 * d'en redessiner le contenu.
	 * Elle retourne un pointeur sur l'interface  utiliser pour dessiner les objets dans cette
	 * couche OSD.
	 * \param[in] hDc	"Handle" sur le "device context" de la fentre. Utilis seulement
	 *					si la mise  jour est demande par un message \p WM_PAINT avec certains
	 *					rendus vido. Dans tous les autres cas, mettre \p NULL.
	 * \return	Un pointeur partag sur l'interface de cette couche
	 **/
	virtual COsdLayerPtr
						BeginDraw(HDC hDc = NULL);

	/** \brief Fin de mise  jour (avec validation) de la couche OSD dans l'affichage
	 *
	 * Cette mthode termine la construction de la couche OSD, et assure sa prsentation
	 * effective dans l'affichage.
	 **/
	virtual HRESULT		EndDraw() {
							return S_OK; }

	/// Initialisation gnrale
	virtual HRESULT		Init(IPTvM_GraphBase & iGraph);

	/** \brief Rcupration des informations sur la vido, depuis la broche d'entre du filtre de rendu
	 *
	 * \param[out]	sVInf	Informations rcupres
	 * \return \p true ou \p false selon que les informations ont pu tre rcupres ou non ; cette valeur
	 * peut tre ignore si les valeurs par dfaut suffisent.
	 * \note Si les informations rclames ne peuvent tre obtenues, des valeurs par dfaut sont
	 * retournes
	 **/
	bool				GetVideoInfo(LocalVideoInfo & sVInf) const;

	/// Rcupration du ratio d'aspect courant
	SIZE				GetAspectRatio() const;

	/// Redessiner l'espace de la vido
	virtual void		Paint(HDC hDC);

	/// Appel sur notification de redimensionnement de la fentre
	virtual void		Resize() {
							}

	/// Repositionnement de la vido aprs changement dans les paramtres,
	/// avec une temporisation pour assurer que le changement n'intervient
	/// pas plus souvent que tous les #DUREE_VIDEO_MOVE.
	void				UpdateVideoPos();

	ProcAmpAvailability isProcAmpAvailable() const {
							return eProcAmpAvail; }

	virtual HRESULT		GetProcAmpValues(ProcAmpValues & sValues) const {
							return E_NOTIMPL; }

	virtual HRESULT		SetProcAmpValues(ProcAmpValues & sValues) {
							return E_NOTIMPL; }

	/// Initialisation des rglages de couleur de la vido
	HRESULT				InitProcAmp(INT32 sInitialValues[4]);
};

// ====================================================================================
// Dclarations des fonctions exportes
// ====================================================================================

/**
 * Positionnement de la vido dans un rectangle de forme et de dimensions quelconque
 * (prsum tre un rectangle client de fentre d'affichage).
 * Les valeurs des variables globales \p #video_pan_mode et \p #zoom_ratio sont prises en
 * compte dans ce positionnement.
 *
 * \param[in]	sSrcRect	Rectangle client de la fenetre d'affichage (normalement avec origine
 *							zro, mais les cas o il en irait autrement sont pris en compte).
 * \param[in]	sAr			Ratio d'aspect  utiliser
 * \param[in]	pVRect		Pointeur optionnel (si non NULL) sur un rectangle destin  recevoir
 *							l'intersection entre le rectangle client et la vido positionne
 * \return					Rectangle de la vido positionne.
 *
 * \note Du fait des dplacements produits par les variables \p #video_pan_mode et
 * \p #zoom_ratio, la taille de la vido peut excder celle de la fentre. Ceci induit, selon
 * les valeurs fournies en entre, la possibilit d'avoir des valeurs ngatives en rsultat,
 * particulirement sur le bords gauche et le bord suprieur.
 **/
RECT PlaceVideoInRect(const RECT & sSrcRect, SIZE sAr, LPRECT pVRect=NULL);

/**
 * Redessiner l'affichage ( appeler en rponse  \p WM_PAINT)
 * \param[in] hDC "Device context" retourn par la fonction systme
 * \p BeginPaint (appele au pralable)
 **/
void do_paint(HDC hDC);

/**
 * Mise  jour des coordonnes de l'cran ( appeler en rponse  tout vnement
 * de nature  modifier les dimensions de la vido)
 **/
void update_coords();

// ====================================================================================
// Dclarations des variables exportes
// ====================================================================================

/// Pointeur sur l'instance du VMR cre lors de l'excution du programme
extern shared_ptr<CVideoRenderer>	pVMR;

/// Pointeur sur l'instance de l'OSD cre lors de l'excution du programme
extern shared_ptr<COsdAllObjects>	pOSD;
