/*
 * $Id$
 * Copyright (C) 2006 Pouchin <pouchinteve@free.fr>
 *
 * This file is part of Pouchin TV Mod, a free DVB-T viewer.
 * See http://www.pouchintv.fr/ for updates.
 *
 * Pouchin TV Mod is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Pouchin TV Mod is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *
 * Most files of this project have been changed from the Pouchin TV
 * project (Copyright (C) 2006 Pouchin <pouchinteve@free.fr>),
 * to use with a modified version of this software.
 * See http://pouchinteve.free.fr/ for the original project.
 */

/** \file
 *	\brief Implmentation de la fonctionnalit de recherche de chanes
 *
 *	Ce fichier implmente les fonctionnalits ncessaires  la fonctionnalit
 *	de recherche de chanes (hors interface utilisateur).
 **/

#include "stdafx.h"
#include "scan.h"
#include "parse.h"
#include "search.h"
#include <time.h>

// ====================================================================================
// Recherche de chanes
// ====================================================================================

/**
 * Fonction gnrant en mme temps le log des stats du signal  cet instant, et
 * vrifiant si la qualit du signal est suffisante pour esprer recevoir quelque chose.
 * \param[in]	nTuneDelay	Temps depuis lequel le tuner est syntonis (mS)
 * \return		\p true si la qualit du signal est suffisante
 **/
ChannelsScanner::SignalState ChannelsScanner::CheckQuality(UINT nTuneDelay) const
{
	SignalStatistics	sSStats;

	sSStats.getStats();

	if (sSStats.hr_1st!=S_OK) {
		log(TEXT("Erreur stats signal, hr=0x%08x, tapes=0x%02x") EOL, sSStats.hr_1st, sSStats.nErrStep);
		// return ss_error;
	}

	TCHAR				szStrength[16];
	TCHAR				szQuality[16];

	// crire les statistiques du signal dans le journal (avec la force du signal
	// toujours sous forme native)
	sSStats.getStrengthStr(false, szStrength, _countof(szStrength), TEXT(""));
	sSStats.getQualityStr(szQuality, _countof(szQuality), TEXT("%"));
	log(TEXT("%5u mS : Signal=%s, verrou=%s, force=%s, qualit=%s") EOL, nTuneDelay,
		sSStats.getPresentStr(), sSStats.getLockedStr(), szStrength, szQuality);

	// Transmettre l'information de qualit du signal au dialogue de contrle
	SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY, sn_info_stats, LPARAM(&sSStats));

	if (!sSStats.isPresenceOk())
		return ss_nosignal;
	if (!sSStats.isLockedOk())
		return ss_unlocked;
	if (!sSStats.isQualityOk())
		return ss_bad_quality;
	return ss_signal_ok;
}

/**
 * Syntonisation du tuner pour un canal, et valuation des caractristiques
 * du signal prsent.
 * \param[in]	nNoCanal	Numro du canal  explorer (en principe de 21  69)
 * \param[out]	khz			Frquence de syntonisation
 * \param[out]	sSStats		Statistiques de qualit du signal
 * \return		Estimation du succs de la syntonisation
 **/
ChannelsScanner::SignalState ChannelsScanner::tune(BYTE nNoCanal, INT32 & khz, SignalStatistics & sSStats)
{
	if (!pGraph)
		return ss_error;
	khz = liste_scan.canal2khz(nNoCanal);

	tstring		strErr;
	bool		bAdd5khzTried = false;
	HRESULT		hr;
	SignalState	eSState;

	for (;;) {	// -> Uniquement pour fournir une tiquette de retour pour dcalage 5 kHz
		hr = change_frequence(khz, DVBT_BANDWIDTH, nNoCanal, NULL, &strErr);

		if (FAILED(hr)) {
			log(TEXT("Erreur(s) au changement de frquence : %s") EOL, strErr.c_str());
			strErr.clear();
			// return ss_error;				// (apparemment certains tuners renvoient le code d'erreur
											//  alors qu'ils ont quand mme chang la frquence)
		}

		Sleep(sTunProps.delai_tuner1);		// Dlai de verrouillage
		eSState = CheckQuality(0);

		if (eSState < ss_bad_quality) {
			// Signal pas prsent ou pas verrouill
			if (sTunProps.try_adding_5khz && ! bAdd5khzTried) {
				// Si option slectionne, on tente de se dcaler de 5 kHz
				// (car certains tuners n'aiment pas tre rgls sur la frquence exacte)
				khz += 5;
				bAdd5khzTried = true;
				log(TEXT("Pas de signal ; essai  +5 kHz = %i kHz") EOL, khz);
				continue;
			}

			if (eSState < ss_unlocked) {
				// On quitte si le signal n'est mme pas prsent
				log(TEXT("Aucun signal dtect.") EOL);
				return eSState;
			}
		}
		break;	// toujours
	}

	log(TEXT("Signal dtect : attente stabilisation.") EOL);

	int nTuneDelay = 0;

	// Attente de stabilisation du tuner (dlai fragment en tranches gales
	//  l'intervalle de lecture des statistiques du signal)
	do {
		if (eScanState!=ss_running)
			// On quitte si l'arrt ou l'annulation a t demand
			break;

		int	nSubDelay = min(sTunProps.delai_tuner2 - nTuneDelay, DUREE_SIGNAL_STATS);

		Sleep(nSubDelay);

		nTuneDelay += nSubDelay;
		eSState = CheckQuality(nTuneDelay);

		//  ce stade, on quitte si le signal n'est pas verrouill
	} while (nTuneDelay < sTunProps.delai_tuner2 && eSState >= ss_bad_quality);

	// Retourner l'tat de dtection du signal
	return eSState;
}

bool ChannelsScanner::parse(BYTE nNoCanal, INT32 khz)
{
	if (!pGraph)
		return false;

	Mpeg2SectionsHandler	handler(pGraph->GetMpeg2Data());
	UINT					i;

	// Note : ici, on ne teste l'tat de 'eScanState' que sur l'annulation, pour
	// ne pas courir le risque de valider des donnes de chanes incompltes

	// Choppe les noms des chanes
	handler.Init(p_pid_SDT, si_SDT, 5000, MAKEFOURCC('S','D','T',0));
	while (handler.GetNextSection()) {
		parse_sdt(*reinterpret_cast<const SI_SDT *>(handler.data), nNoCanal, khz);
		if (eScanState==ss_cancelling)
			return false;
	}
	if (!handler.complete)
		return false;	// Pas reu la SDT : inutile de continuer

	if (prev_size >= Canaux.size()) {
		log(TEXT("  ** Aucune chane (supplmentaire) trouve") EOL);
		return true;
	}

	// Choppe la PAT
	handler.Init(p_pid_PAT, si_PAT, 10000, MAKEFOURCC('P','A','T',0));
	while (handler.GetNextSection()) {
		parse_pat(*reinterpret_cast<const SI_PAT *>(handler.data), nNoCanal);
		if (eScanState==ss_cancelling)
			return false;
	}
	if (!handler.complete)
		return false;	// Pas reu la PAT : inutile de continuer

	// Choppe les PMT
	for (i = prev_size; i < Canaux.size(); i++) {
		// Traiter les chanes ajoutes depuis le dbut de l'excution de cette fonction.
		Chaine & canal = Canaux[i];

		if (canal.pmt_pid==0)
			continue;

		handler.Init((PID)canal.pmt_pid, si_PMT, 5000, MAKEFOURCC('P','M','T',0));
		while (handler.GetNextSection()) {
			parse_pmt(*reinterpret_cast<const SI_PMT *>(handler.data), true, canal);
			if (eScanState==ss_cancelling)
				return false;
		}
	}

	// Logging des chanes trouves, et limination des chanes vides, ou bien dont le contenu n'est
	// pas exploitable (note : ne pas sauver 'Canaux.size()' dans une variable car cette valeur peut
	// diminuer au fil de l'exploration, si des chanes invalides sont dtruites) :
	log(TEXT("  ** Chanes trouves :") EOL);
	for (i = prev_size; i < Canaux.size(); i++) {
		const Chaine & canal = Canaux[i];

		log(TEXT("  -- %") A2t TEXT(" [groupe=%") A2t TEXT("],")
			TEXT(" OnId=%u, TsId=%u, SId=%u, vPId=%u, pPid=%u, pcrPid=%u") EOL,
			canal.nom, canal.groupe,
			canal.ONID, canal.TSID, canal.SID, canal.video_pid, canal.pmt_pid, canal.pcr_pid);

		if (canal.video_pid==0 && canal.video_type == vst_Unknown) {
			if (canal.nbr > 0) {
				log(TEXT("    -- Audio seul") EOL);
			} else {
				//vire cette chane
				log(TEXT("    -- Invalide : destruction") EOL);
				Canaux.erase(Canaux.begin()+(i--));
			}
		} else {
			TCHAR	szWrk[128];

			for (int j=0; j<canal.nbr; ++j) {
				canal.tbl[j].getAsLogString(szWrk, _countof(szWrk));
				log(TEXT("    -- %s : %s") EOL, j < canal.nbr ? TEXT("Audio") : TEXT("Autre") , szWrk);
			}
		}
	}

	// Si on a dtruit tout ce qu'on a trouv ...
	if (prev_size >= Canaux.size()) {
		log(TEXT("  ** Aucune chane valide trouve") EOL);
		return true;
	}

	// NIT, pour l'ordre des chanes.

	// Tenter d'attribuer les numros NIT issue des multiplex prcdents aux chanes de celui-ci
	if (place_numeros()) {	// si au moins une chane n'a pas reu de numro, on recherche
							// la table NIT du flux courant
		log(TEXT("  ** Numro(s) de chanes manquant(s) : lecture de la table NIT") EOL);
		PostMessage(hwndDlg, WM_APP_SCAN_NOTIFY, sn_info_nit, 0);	// Pour dclencher affichage info "NIT"
		handler.Init(p_pid_NIT, si_NIT, 10000, MAKEFOURCC('N','I','T',0));
		while (handler.GetNextSection()) {
			UINT old_size = (UINT)cInfoNIT.size();

			parse_nit(*reinterpret_cast<const SI_NIT *>(handler.data), cInfoNIT);

			if (!cInfoNIT.strNetworkName.empty())
				log(TEXT("    -- Nom du rseau : %s") EOL, cInfoNIT.strNetworkName.c_str());

			// Logging de ce qui a t trouv :
			UINT new_size = (UINT)cInfoNIT.size();

			if (new_size>old_size) {
				log(TEXT("    -- Numros trouvs :") EOL);
				for (i = old_size; i < new_size; i++) {
					const ChaineIdsNIT & sCIds = cInfoNIT[i];

					log(TEXT("      -- %3u pour SId=%4u (TsId=%2u, OnId=%4u)") EOL,
						sCIds.nNumeroNit, sCIds.SID, sCIds.TSID, sCIds.ONID);
				}
			} else
				log(TEXT("    -- Aucun numro trouv") EOL);

			if (eScanState==ss_cancelling)
				return false;
		}

		// Retenter d'attribuer les numros
		if (handler.complete)
			place_numeros();
	}

	return true;
}

/// Rinitialisation de la carte des chanes  vide
void ChannelsScanner::raz_carte()
{
	// Tous les pointeurs de numros de chanes  -1 :
	for (int nInx=0; nInx<_countof(aRefNumeros); ++nInx)
		aRefNumeros[nInx] = -1;
}

void ChannelsScanner::cartographier()
{
	raz_carte();

	SHORT	i;
	SHORT	siz = (SHORT)Canaux.size();

	for (i = 0; i < siz; ++i) {
		Chaine &	canal = Canaux[i];

		if (canal.nNumeroChaine==0)
			continue;	// Numro gal  zro (pas normal, mais on va rparer)
		if (canal.nNumeroChaine<1000) {
			SHORT & nRef =  aRefNumeros[canal.nNumeroChaine];

			if (nRef<0) {
				nRef = i;
				continue;
			}
		}
		// S'il existait un doublon antrieurement, ou bien un numro suprieur  999
		// (gros bug, mais bon, on va grer)
		canal.nNumeroChaine = 0;
	}
}

/// Sortir les informations concernant un filtre
void ChannelsScanner::logDeviceInfo(LPCTSTR pszTitle, LPCTSTR pszDevicePath)
{
	log(EOL TEXT("Pilote (%s) :") EOL, pszTitle);

	RegFilterInfo	sInfo;

	if (CRegControlSet().GetDeviceInfo(pszDevicePath, sInfo)) {
		if (!sInfo.strConnection.empty())
			log(TEXT("  -- Connexion   : %s") EOL, sInfo.strConnection.c_str());
		if (!sInfo.strDriverDesc.empty())
			log(TEXT("  -- Description : %s") EOL, sInfo.strDriverDesc.c_str());
		if (!sInfo.strDriver.empty())
			log(TEXT("  -- Fichier(s)  : %s") EOL, sInfo.strDriver.c_str());
		if (!sInfo.strDriverVersion.empty())
			log(TEXT("  -- Version     : %s") EOL, sInfo.strDriverVersion.c_str());
		if (!sInfo.strDriverDate.empty())
			log(TEXT("  -- Date        : %s") EOL, sInfo.strDriverDate.c_str());
		if (!sInfo.strProviderName.empty())
			log(TEXT("  -- Fabricant   : %s") EOL, sInfo.strProviderName.c_str());
	}
}

DWORD ChannelsScanner::ThreadProc()
{
	setlocale(LC_ALL, "Bulgarian_Bulgaria.1251");		// Pour affichage de la date

	eScanState = ss_running;			// Change l'tat => on lance un scan

	time_t	sTime;
	tm		sTm;

	// Initialiser selon la mthode de recherche :
	switch (methode_recherche) {

	case sm_incremental:	// Recherche incrmentale (ajout de chanes)
		// Cartographier les numros dj allous :
		cartographier();
		break;

	case sm_prefs:			// Recherche en conservant les prfrences
	case sm_full:			// Recherche totalement rinitialise
		// Effacement de la liste des chanes
		Canaux.clear();
	}

	log(EOL TEXT("################################################################") EOL);
	time(&sTime);
	if (localtime_s(&sTm, &sTime)==0) {
		TCHAR		szWrk[64];

		_tcsftime(szWrk, _countof(szWrk), TEXT("%#c"), &sTm);
		log(TEXT("Dbut de la recherche : %s") EOL, szWrk);
	}

	log(TEXT("  [%s version %s]") EOL, szAppName, szAppVersion);
	log(TEXT("  [Windows version %s]") EOL EOL, GetWindowsVersionStr(windows_version).c_str());

	NamesAndDevPaths	sNnP;

	if (CRegControlSet().GetAllNamesAndPaths(sTunProps, sNnP)) {
		if (!sNnP.strTunerName.empty())
			log(TEXT("Tuner        : %s") EOL, sNnP.strTunerName.c_str());
		if (!sNnP.strDemodName.empty())
			log(TEXT("Dmodulateur : %s") EOL, sNnP.strDemodName.c_str());
		if (!sNnP.strCaptureName.empty())
			log(TEXT("Capture      : %s") EOL, sNnP.strCaptureName.c_str());

		switch (sNnP.DevPathCompare()) {

		case ud_t:
			logDeviceInfo(TEXT("tuner"), sNnP.strTunerPath.c_str());
			break;

		case ud_tc:
			logDeviceInfo(TEXT("tuner & capture"), sNnP.strTunerPath.c_str());
			break;

		case ud_tdc:
			logDeviceInfo(TEXT("tuner & dmodulateur & capture"), sNnP.strTunerPath.c_str());
			break;

		case ud_t_plus_c:
			logDeviceInfo(TEXT("tuner"), sNnP.strTunerPath.c_str());
			logDeviceInfo(TEXT("capture"), sNnP.strCapturePath.c_str());
			break;

		case ud_t_plus_dc:
			logDeviceInfo(TEXT("tuner"), sNnP.strTunerPath.c_str());
			logDeviceInfo(TEXT("dmodulateur & capture"), sNnP.strDemodPath.c_str());
			break;

		case ud_t_plus_d_plus_c:
			logDeviceInfo(TEXT("tuner"), sNnP.strTunerPath.c_str());
			logDeviceInfo(TEXT("dmodulateur"), sNnP.strDemodPath.c_str());
			logDeviceInfo(TEXT("capture"), sNnP.strCapturePath.c_str());
			break;

		case ud_td_plus_c:
			logDeviceInfo(TEXT("tuner & dmodulateur"), sNnP.strTunerPath.c_str());
			logDeviceInfo(TEXT("capture"), sNnP.strCapturePath.c_str());
		}

	} else {
		log(TEXT("Erreur en cherchant  identifier les composants") EOL);
	}

	log(EOL TEXT("Options :") EOL
		TEXT("  -- Ignorer prsence du signal = %s") EOL
		TEXT("  -- Ignorer signal verrouill  = %s") EOL
		TEXT("  -- Ignorer qualit du signal  = %s") EOL,
		boolToYesNo(sTunProps.scan_ignore_presence),
		boolToYesNo(sTunProps.scan_ignore_lock),
		boolToYesNo(sTunProps.scan_ignore_quality));
	log(TEXT("  -- Temps de stabilisation 1   = %u mS") EOL
		TEXT("  -- Temps de stabilisation 2   = %u mS") EOL
		TEXT("  -- Correction frquence tuner = %i kHz (non inclus dans valeurs affiches)") EOL
		TEXT("  -- Essayer +5 kHz si chec    = %s") EOL,
		sTunProps.delai_tuner1, sTunProps.delai_tuner2, sTunProps.offset_tuner,
		boolToYesNo(sTunProps.try_adding_5khz));

	log(EOL TEXT("Canaux  explorer : %s") EOL,
		_tcscmp(choixCanaux, NULL_NAME)==0 ? TEXT("Tous") :
		_tcscmp(choixCanaux, LIST_NAME)==0 ? liste_canaux : choixCanaux);

	bool	bSuccess	= false;
	WORD	nRepCount	= 0;

	do {
		SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY, MAKEWPARAM(sn_init, ++nRepCount), liste_scan.nCount);

		for (nIndex=0; eScanState==ss_running && nIndex < liste_scan.nCount; ++nIndex) {
			BYTE	canal_no	= liste_scan.aNoCan[nIndex];
			INT32	khz			= liste_scan.canal2khz(canal_no);

			log(EOL TEXT("Essai du canal %u, frquence = %i kHz") EOL, canal_no, khz);

			prev_size = (UINT)Canaux.size();	// Mmoriser le nombre de chanes dj identifies  cet instant

			SignalStatistics	sSStats;		// Qualit du signal

			SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY,
				MAKEWPARAM(sn_info_freq, canal_no), LPARAM(khz));

			// Note : bien que 'khz' soit dj charg (pour le log), la mthode 'tune' le calcule  nouveau.
			// En effet, cette fonction est susceptible d'oprer des c orrections de frquence mineures.
			bool				bSignalFound = tune(canal_no, khz, sSStats)==ss_signal_ok;

			if (bSignalFound) {
				log(TEXT("  ** %s%sRecherche de canaux") EOL,
					(sTunProps.scan_ignore_presence ? TEXT("") : TEXT("Signal prsent - ")),
					(sTunProps.scan_ignore_lock ? TEXT("") : TEXT("Signal verrouill - ")));
				bSuccess = parse(canal_no, khz);
			}

			if (eScanState==ss_running) {
				// Notifier au dialogue de recherche qu'on a complt une frquence
				SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY,
					MAKEWPARAM(sn_completed, canal_no), LPARAM(khz));
			}

			// Si on a demand la recherche en boucle, on arrte au premier canal qui a permis de
			// trouver quelque chose (peu importe qu'il y ait de nouvelles chanes ou non)
			if (liste_scan.bScanLoop && bSuccess)
				break;
		}

		if (!liste_scan.bScanLoop || bSuccess)
			break;
		Sleep(250);	// Laisser les affichages s'tablir un peu
	} while (eScanState==ss_running && nRepCount<999);

	if (eScanState == ss_cancelling || eScanState == ss_stopping)
		log(EOL TEXT("Interruption par l'utilisateur") EOL);

	// Recherche termine :
	SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY, sn_info_clear, NULL);

	Canaux.tri();				// Trier les chanes par numro, puis par nom

	cartographier();			// Recartographier, le tri ayant tout fauss

	attribue_numeros();			// Attribuer les numros manquants (les canaux ayant t tris, la
								//   fonction devrait attribuer les numros manquant dans l'ordre
								//   alphabtique des noms des chanes correspondantes)
								// Note : en cas d'abandon de la recherche, il peut se produire
								//   que des numros soient attribus ici  des chanes qui en
								//   auraient eu un autrement (cas o cet abandon a lieu alors
								//   que certaines tables ont dj t lues).

	// Restaurer les prfrences de l'utilisateur, s'il y a lieu (mme en cas de recherche
	// incrmentale : en effet, un utilisateur pourrait avoir fait une recherche avec restauration
	// de prfrences qui aurait moins de succs que la recherche antrieure, oui bien qu'il devrait,
	// en raison de sa situation, effectuer en plusieurs tapes, recherche qu'il pourrait alors
	// vouloir complter incrmentalement)
	if ((methode_recherche==sm_prefs || methode_recherche==sm_incremental) && cInfoUser.size()>0) {
		log(EOL TEXT("Restauration des prfrences de l'utilisateur") EOL);
		SendMessage(hwndDlg, WM_APP_SCAN_NOTIFY, sn_restore, NULL);
		restaure_prefs();
	}

	Canaux.tri();				// Trier  nouveau, les numros attribus ayant vraisemblablement
								//   affect l'ordre de classement rsultant

	PostMessage(hwndDlg, WM_APP_SCAN_NOTIFY,
		MAKEWPARAM(sn_scan_end, eScanState), LPARAM(liste_scan.nCount));
	return eScanState;
}

/**
 * Assignation d'un numro  une chane. Si une autre chane avait ce numro, les numros sont
 * changs.
 * \param[in]	ixChaine	Index de la chane dans la table
 * \param[in]	nNumero		Numro  assigner
 * \param[in]	bPref		\p true s'il s'agit d'une prfrence utilisateur (par opposition
 *							 un numro automatiquement calcul)
 **/
void ChannelsScanner::assigne_numero(int ixChaine, UINT16 nNumero, bool bPref)
{
	Chaine &	canal = Canaux[ixChaine];

	if (nNumero!=canal.nNumeroChaine) {
		// il faut quand mme que le nouveau numro soit diffrent ...
		int ixAutreChaine = aRefNumeros[nNumero];

		if (ixAutreChaine>=0) {
			// Numro dj assign ailleurs
			Chaine & canal2 = Canaux[ixAutreChaine];

			canal2.nNumeroChaine = canal.nNumeroChaine;
			if (!bPref)
				canal2.nNumeroAuto = canal2.nNumeroChaine;
			aRefNumeros[canal2.nNumeroChaine] = (SHORT)ixAutreChaine;
		}

		canal.nNumeroChaine = nNumero;
		if (!bPref)
			canal.nNumeroAuto = nNumero;
		aRefNumeros[nNumero] = (SHORT)ixChaine;
	}
}

/// Transfert des numros de chane de la NIT vers les chanes qui viennent d'tre ajoutes
bool ChannelsScanner::place_numeros()
{
	SHORT	i;
	SHORT	siz = (SHORT)Canaux.size();
	bool	bMissing = false;

	for (i = SHORT(prev_size); i < siz; i++) {	// Pour chaque nouvelle chane de la liste
		Chaine &				canal = Canaux[i];
		const ChaineIdsNIT *	pCIds = cInfoNIT.Cherche(canal);

		if (pCIds) {
			if (pCIds->nNumeroNit!=0) {
				UINT16	nNumeroNit	= pCIds->nNumeroNit;
				UINT16	nNumero		= nouveau_numero_decale(nNumeroNit);

				canal.nNumeroNit	= nNumeroNit;
				assigne_numero(i, nNumero, false);
				if (nNumero==nNumeroNit) {
					log(TEXT("Assignation du numro %u  la chane %") A2t TEXT(", d'aprs les informations du rseau") EOL,
						nNumero, canal.nom);
				} else {
					log(TEXT("Le rseau assigne le numro %u  la chane %") A2t TEXT(", mais celui-ci") EOL
						TEXT("  est dj utilis ; attribution du numro %u  la place") EOL,
						nNumeroNit, canal.nom, nNumero);
				}
			} else
				bMissing = true;
		} else
			bMissing = true;
	}
	return bMissing;
}

/**
 * Attribution d'un numro  toutes les chanes qui n'en ont pas reu.
 * Si les chanes ont t tries auparavant, ce numro sera attribu selon l'ordre
 * alphabtique de leur nom.
 * Au passage, marque galement comme "inactives" les chanes qui ne diffusent pas
 * couramment en norme MPEG2.
 **/

void ChannelsScanner::attribue_numeros()
{
	int		nbChaines	= (int)Canaux.size();
	bool	bHaveMpeg2	= !IsNullName(filtreMPEG2);	// peu vraisemblable qu'il n'y ait pas, mas bon ...
	bool	bHaveH264	= !IsNullName(filtreH264);

	for (int i = 0; i<nbChaines; i++) {
		Chaine & canal = Canaux[i];

		if (canal.nNumeroChaine==0) {
			UINT16 nNumero = nouveau_numero_chaine();

			log(TEXT("Attribution du numro %u  la chane %") A2t TEXT(", qui n'en avait pas") EOL,
				nNumero, canal.nom);
			assigne_numero(i, nNumero, false);
		}

		switch (canal.video_type) {

		case vst_MPEG2:
			if (!bHaveMpeg2)
				canal.etat = ec_active;	// plutt que 'ec_preferee'
			break;

		case vst_H264:
			if (!bHaveH264)
				canal.etat = ec_active;	// plutt que 'ec_preferee'
			break;

		case vst_MPEG2_encrypted:
		case vst_H264_encrypted:		// crypt ...
			canal.etat = ec_inactive;
			break;

		default:
			if (canal.nbr > 0)
				canal.etat = ec_active;	// radio ...
			else
				canal.etat = ec_inactive;
		}
	}
}

/**
 * Attribution d'un nouveau numro pour une nouvelle chane,  partir d'un numro propos.
 * Vrifie d'abord la disponibilit du numro propos, et le retourne s'il est disponible.
 * S'il ne l'est pas, alors on cherche  transformer ce numro par dcalage, en ajoutant
 * (ventuellement plusieurs fois) la valeur de la variable globale \p autonum_duplicate.
 * S'il est impossible de trouver un numro selon ce critre (si tous les numros "dcals" possibles
 * entre 1 et 999 sont pris), on retourne 0.
 * \param[in]	nNumeroPropose	Numro propos pour la chane.
 * \retval Numro trouv (intervalle 1  999)
 * \retval 0 si aucun numro dcal disponible
 **/
UINT16 ChannelsScanner::nouveau_numero_decale(UINT16 nNumeroPropose) const
{
	while (nNumeroPropose<_countof(aRefNumeros)) {
		if (aRefNumeros[nNumeroPropose] < 0)
			return nNumeroPropose;
		nNumeroPropose = nNumeroPropose + UINT16(autonum_duplicate);
	}
	return 0;
}

/**
 * Attribution d'un nouveau numro pour une nouvelle chane qui n'a pas de numro propos.
 * Cherche  attribuer ce numro parmi les numros disponibles, dans l'intervalle compris
 * entre la valeur de la variable \p autonum_undef et 999, puis entre 1 et \p autonum_undef.
 * \retval Numro trouv (intervalle 1  999)
 * \retval 0 si plus aucun numro disponible (hypothse assez improbable)
 **/
UINT16 ChannelsScanner::nouveau_numero_chaine() const
{
	UINT16	nStartNum = min<UINT16>(UINT16(autonum_undef), _countof(aRefNumeros));
	UINT16	nNewNum;

	for (nNewNum=nStartNum; nNewNum<_countof(aRefNumeros); ++nNewNum) {
		if (aRefNumeros[nNewNum] < 0)
			return nNewNum;
	}
	for (nNewNum=0; nNewNum<nStartNum; ++nNewNum) {
		if (aRefNumeros[nNewNum] < 0)
			return nNewNum;
	}
	return 0;
}

/**
 * Restauration des prfrences utilisateur sauvegardes
 **/
void ChannelsScanner::restaure_prefs()
{
	for (ITERATE_CONST_VECTOR(cInfoUser, ChaineIdsUser, it)) {
		int ixChaine = Canaux.trouve_par_Ids(*it);

		if (ixChaine>=0) {
			Chaine & canal = Canaux[ixChaine];

			if (canal.nNumeroChaine!=it->nNumeroChaine) {
				assigne_numero(ixChaine, it->nNumeroChaine, true);
				log(TEXT("  -- Restauration prfrence : numro %u pour %") A2t TEXT ("") EOL,
					canal.nNumeroChaine, canal.nom);
			}
			if (canal.etat!=it->etat) {
				canal.etat = it->etat;
				log(TEXT("  -- Restauration prfrence : %") A2t TEXT (" = %") A2t TEXT ("") EOL,
					canal.nom, EnumToString(canal.etat, "Inactive\0Active\0Prfre\0"));
			}
		}
	}
}

/// Constructeur
ChannelsScanner::ChannelsScanner(HWND hDlg, const CInfoUser & cInfoUsr) :
	WorkingThread(false),
	hwndDlg(hDlg),
	fLog(NULL),
	cInfoUser(cInfoUsr),
	prev_size(0),
	eScanState(ss_idle),
	nIndex(0)

{
	// Init carte des chanes
	raz_carte();

	// Dfinition de l'emplacement du fichier de 'log' de la recherche, et ouverture.
	NomFichierConf<>	sLogFileName(TEXT("scan.log"));
	FILE *				fop;

	if (_tfopen_s(&fop, sLogFileName, TEXT("a"))==0)
		fLog = fop;
}

void ChannelsScanner::log(LPCTSTR pszFmt, ...) const
{
	FMT_BUF(szBuf, 512, pszFmt, pszFmt);

	myprintf(TEXT("%s"), szBuf);
	if (fLog)
		_fputts(szBuf, fLog);
}

/// Destructeur
ChannelsScanner::~ChannelsScanner()
{
	if (fLog)
		fclose(fLog);
}
